/* FILE: currtime.c                 Copyright (c), Symmetric Research, 2003-2005

This program checks the PARGPS current time functionality.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pargps.h"
#include "srhelper.h"


#define TITLE_FMT   "\n    Date       Time      Nsat     Lat         Lon       Alt (ENTER to quit) %3s\n"
#define RESULT_FMT "%04d-%02d-%02d %02d:%02d:%02d.%03d  %2d  %10.3lf  %10.3lf  %6.0lf  YMD=%s HMS=%s\n"

char *ValidParGpsNames[] = {
        "SrParGps",
        "SrParGps378",
        "SrParGps278",
        NULL
        };

char* UsageMsg[] = {
    "USAGE: currtime [ParGpsDriver] [ms] [set]                             ",
    "                                                                      ",
    "Where                                                                 ",
    "     ParGpsDriver is the PARGPS device driver name  (default SrParGps)",
    "     ms           is milliseconds between updates   (default 1000)    ",
    "     set          means currtime should set the PC clock on startup   ",
    "                  and once per day thereafter (you must have          ",
    "                  permission to set time for this option to work)     ",
    "                                                                      ",
    "Examples: currtime SrParGps 1000 set                                  ",
    "          currtime 2000                                               ",
    "                                                                      ",
    NULL
    };

char* StartMsg[] = {
    "                                                                    ",
    "Type currtime /? to get a help usage message.                       ",
    "                                                                    ",
    "Notes: This program only opens the PARGPS driver and retrieves saved",
    "current time information.  You must run an active GPS program       ",
    "that opens and starts BOTH the PARGPS and PARxCH drivers in order   ",
    "to save the current time information.  Active GPS programs include  ",
    "PARGPS diag and the PARxCH acquisition programs (simp and scope)    ",
    "with GPS enabled.  Correct permissions are required to set PC time. ",
    "                                                                    ",
    "The YMD= and HMS= columns on the far right indicate the source of   ",
    "the date and time in the left most columns.  Common values and      ",
    "their meanings are:                                                 ",
    "                                                                    ",
    "   ZDA     from GPS NMEA string ZDA                                 ",
    "   GGA     from GPS NMEA string GGA                                 ",
    "   PC      from PC time updated for crossing date boundaries        ",
    "   OLD     from last valid GPS time updated for elapsed time        ",
    "   blank   source unknown                                           ",
    "                                                                    ",
    "Press ENTER to check current time and ENTER again to stop.          ",
    NULL
    };


// This string array indicates the source for the specified quantity.
// Typically, it is a NMEA string, but may also be from the PC, from an
// old value, or unknown. Entries must be in the SRCSTR_ defines order.

char *SrcStr[] = { "ZDA", "GGA", "GLL", "VTG", "GSA", "GSV", "RMC",
                   "PC ", "OLD", "   " };

// This string array is used in the title line to indicate whether or
// not the PC time will be periodically updated to agree with GPS time.

char *SetPcStr[] = { "   ", "SET" };

int MatchDriverName( char *trialname, int *match );
void PrintMsg( char *MsgArray[] );
void Usage( char* SpecificMsgText );

extern unsigned long ParGpsLastDriverError;


int main( int argc, char **argv ) {

        int       i, match, ret, TitleCount, UpdateMs, SetPcTime, ErrorCode;
        int       Yr, Mo, Dy, Hh, Mm, Ss, Msec, Nsat;
        int       SrcYMD, SrcHMS, SrcNsat, SrcPos, LastDay;
        long      Usec;
        double    Lat, Lon, Alt, Time;
        DEVHANDLE ParGpsHandle;
        char     *ParGpsName;


        // Display the leader message.

        printf( "PARGPS currtime [ParGpsDriver] [ms] [set]: (c) SR 2005 : Rev {03/25/05}\n" );


        // Allow keyboard check to return if no key pressed.

        SrKeyboardNonBlock();



        // Get values from command line

        if ( argc > 4 )
                Usage( "Command line must have no more than 3 arguments" );


        ParGpsName  = NULL;
        SetPcTime   = -1;
        UpdateMs    = 0;
        match       = 0;

        for ( i = 1 ; i < argc ; i++ ) {

                if ( strcmp( argv[i], "/?" ) == 0 )
                                Usage( NULL );
                
                else if ( MatchDriverName( argv[i], &match ) ) {
                        if (!ParGpsName)
                                ParGpsName = ValidParGpsNames[match];
                        else
                                Usage( "Two ParGpsDriver names on command line" );
                        }
                else if ( strcmp( argv[i], "set" ) == 0 ) {
                        printf("matched set argument\n");
                        if (SetPcTime < 0)
                                SetPcTime = 1;
                        else
                                Usage( "Two /set values on command line" );
                        }
                else {
                        if (UpdateMs <= 0) {
                                ret = sscanf( argv[i], "%d", &UpdateMs );
                                if ( ret != 1 || UpdateMs <= 0 )
                                        Usage( "Invalid ms value" );
                                }
                        else
                                Usage( "Two ms values on command line" );

                        }
                
                } // end for i


        // Values were not give on command line, so use defaults

        if (SetPcTime < 0)
                SetPcTime = 0;
        if (UpdateMs <= 0)
                UpdateMs = 1000;
        if (!ParGpsName)
                ParGpsName = ValidParGpsNames[0];


        

        // Open PARGPS driver.
        
        printf( "Opening PARGPS driver ...       " );  SrFlushPrint();

        ParGpsHandle = ParGpsOpenForTimeOnly( ParGpsName, &ErrorCode );
        if (ParGpsHandle == BAD_DEVHANDLE ) {
                printf( "ERROR: %s\n", PARGPS_ERROR_MSG[ErrorCode] );
                exit( 1 );
                }
        printf( "Success\n" );



        // Print intro message and wait for a key, allowing Ctrl+C to quit

        PrintMsg( StartMsg );
        if ( SrGetKeyWait() == 0x3 ) {  // Ctrl+C
                ParGpsClose( ParGpsHandle );
                exit( 0 );
                }


        // Loop around checking current time.

        LastDay    = 0;
        TitleCount = 0;
        printf( TITLE_FMT, SetPcStr[SetPcTime] );

        while ( 1 ) {

                ret = ParGpsGetCurrentTime( ParGpsHandle, &Time, &Yr, &Mo, &Dy,
                                            &Hh, &Mm, &Ss, &Usec, &Nsat,
                                            &Lat, &Lon, &Alt,
                                            &SrcYMD, &SrcHMS, &SrcNsat, &SrcPos,
                                            &ErrorCode
                                           );

                Msec = Usec / 1000;

                if (!ret && (ErrorCode != PARGPS_ERROR_NONE)) {
                        printf( "ERROR: %s\n", PARGPS_ERROR_MSG[ErrorCode] );
                        break;
                        }
                else if (!ret) { // No saved time available

                        // Set unfilled values, then get time from PC

                        Msec = 0;  Usec = 0L;  Lat = Lon = Alt = 0.0L;
                        SrcYMD = SrcHMS = SRCSTR_PC;
                        ParGpsGetPcTime( &Time, &Yr, &Mo, &Dy, &Hh, &Mm, &Ss );
                        }
                


                // If setting the PC time is requested, we didn't just
                // get the time from the PC, and it is a new day ...
                // then set the PC time.
                
                if ( SetPcTime == 1  &&  SrcHMS != SRCSTR_PC  &&  LastDay != Dy ) {
                        LastDay = Dy;
                        ret = ParGpsSetPcTime( Time, Yr, Mo, Dy, Hh, Mm, Ss, 1 );
                        if (!ret) {
                                printf( "WARNING: Could not set PC time to " );
                                printf( "%04d-%02d-%02d %02d:%02d:%02d (err=%lu)\n",
                                          Yr, Mo, Dy, Hh, Mm, Ss,
                                          ParGpsLastDriverError );
                                }
                        }
                

                // Display title occasionally

                if (TitleCount == 20) {
                        printf( TITLE_FMT, SetPcStr[SetPcTime] );
                        TitleCount = 0;
                        }
                else
                        TitleCount++;


                // Correct source string index if needed

                if (SrcYMD <= SRCSTR_MIN || SrcYMD >= SRCSTR_MAX )
                        SrcYMD = SRCSTR_UNK;
                if (SrcHMS <= SRCSTR_MIN || SrcHMS >= SRCSTR_MAX)
                        SrcHMS = SRCSTR_UNK;


                // Display results

                printf( RESULT_FMT, Yr, Mo, Dy, Hh, Mm, Ss, Msec, 
                                    Nsat, Lat, Lon, Alt,
                                    SrcStr[SrcYMD], SrcStr[SrcHMS] );
                          


                // Wait for specified interval, check for keypress

                SrSleep( UpdateMs );
                
                if ( SrGetKeyCheck() != EOF )
                        break;

                } // end of while loop




        // Close PARGPS driver.
        
        ParGpsClose( ParGpsHandle );

        exit( 0 );
}
int MatchDriverName( char *trialname, int *match ) {
        
        int i;
        
        i      = 0;
        *match = -1;

        while ( ValidParGpsNames[i] != NULL ) {
                if ( strcmp( trialname, ValidParGpsNames[i] ) == 0 ) {
                        *match = i;
                        break;
                        }
                i++;
                }
        if (*match == -1)
                return( 0 );
        else
                return( 1 );
}
void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );

}
void Usage( char* SpecificMsgText ) {
        int i;

        if ( SpecificMsgText )
                printf( "\nERROR: %s\n", SpecificMsgText );
        else
                printf( "\n" );

        PrintMsg( UsageMsg );
        
        printf( "Valid ParGps driver names are: ");

        i = 0;
        while ( ValidParGpsNames[i] != NULL ) {
                if (i != 0)
                        printf( "," );
                printf( " %s", ValidParGpsNames[i] );
                i++;
                }
        printf( "\n" );
                        
        exit( 1 );
}
