// FILE: indriver.c              Copyright (c), Symmetric Research, 2003
//
// Program to install PARGPS kernel mode driver.
//
// This is done by starting the Service Control Manager, creating the
// driver service, adding driver information to the registry, and then
// starting the service.  Additional registry entries are also added to
// allow PARGPS driver messages to be properly interpreted by the
// system log EventViewer.
//
// Refer to the readme.txt file for information about how to install
// the device driver manually.
//
//


#include <windows.h>
#include <stdio.h>

#include "pargpskd.h"     // for REG_PARM constants




// Include the DriverInfo[] array with allowed driver names:

#include "indriver.h"
DRIVER_INFO_TABLE *Driver;




// Helper function protos:

void PrintMsg( char *MsgArray[] );
void Usage( void );
void Error( char *Msg, int Fatal );



// Global variables shared by main and helper functions:

SC_HANDLE SCManager, SCService;




// Usage and Error message arrays:

char* UsageMsg[] = {

    "                                                                       ",
    "ERROR: Command Line Usage: indriver ParGpsDriverName                   ",
    "                                                                       ",
    "        example: > indriver SrParGps                                   ",
    "                                                                       ",
    "This utility installs the PARGPS device driver.  Use ParGpsDriverName  ",
    "as the device name when using rmdriver, with the ParGpsOpen() function ",
    "and with application programs like simp or scope.                      ",
    "                                                                       ",
    "ParGpsDriverName must be from the following list:                      ",
    "                                                                       ",

    NULL
    };

char* SuccessMsg[] = {

  "                                                                       ",
  "SUCCESS: The device has been installed.                                ",
  "                                                                       ",
  "Now run the PARGPS diagnostic program diag to verify the PARGPS is     ",
  "working correctly.                                                     ",
  "                                                                       ",
  "Note that the exact driver name is needed when accessing the PARGPS in:",
  "      calls to ParGpsOpen                                              ",
  "      the scope gps dialog                                             ",
  "      the simp ini file                                                ",
  "      the rmdriver utility to remove the driver                        ",
  "                                                                       ",
  "You can find the name by clicking the Windows NT Devices Applet from   ",
  "the Control Panel and looking for names that start with Sr.            ",
  "                                                                       ",
  "The name of the installed PARGPS device driver is:                     ",
  "                                                                       ",

  NULL
  };

char* ErrorServiceExistsMsg[] = {

  "  Note: Use rmdriver.exe to clear out the previous version and        ",
  "  try again.                                                          ",
  
  NULL
  };


// Character string areas:

#define MAXSTR  256

char
        ParameterKeyName[MAXSTR],
        EventLogKeyName[MAXSTR],
        DriverPath[MAXSTR],
        EventMessageFile[MAXSTR*2],
        DriverDescBuffer[MAXSTR];





// Main:

void main( int argc, char *argv[] ) {

        HKEY  ParGpsKey;
        BOOL  Ret;
        LONG  Code, Code1, Code2;
        DWORD TagId, Disposition, TypesSupported;
        char *DriverDesc, *LoadGroup;



        
        // Banner message

        printf( "PARGPS Kernel Mode Driver Installation Program : Rev {03/05/03}\n");



        // Process command line arguments for driver name

        if ( argc != 2 ) 
                Usage();

        Driver = DriverInfo;

        while ( Driver->Name )  {
                if (strcmp(Driver->Name, argv[1]) == 0)
                        break;
                Driver++;
                }
        
        if (Driver->Name == NULL)
                Usage();





        // Open Service Control Manager

        SCManager = NULL;
        SCService = NULL;
        
        SCManager = OpenSCManager(
                        NULL,                  // use local machine
                        NULL,                  // use SERVICES_ACTIVE_DATABASE
                        SC_MANAGER_ALL_ACCESS  // full access permission
                        );

        if (!SCManager)
                Error("Could not create service manager", 1);
        



        // Create the device driver service.
        //
        // This involves creating a registry key with several 
        // values and adding the service to the services database.  
        // The registry entries can be made by hand.  But in that
        // case, you must reboot the machine before the service 
        // will be included in the services database.  Once the 
        // service is available, it can be started.  The pertinent 
        // registry entries are as follows (notes in parenthesis are
        // just information and are not included in the registry):
        //
        // HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\SrParGpsXXX
        //  DisplayName  = PARGPS A/D Device Driver 0xXXX (DriverDesc)
        //  ErrorControl = 0x00000001                     (SERVICE_ERROR_NORMAL)
        //  Group        = GPS Timing Devices             (LoadGroup)
        //  ImagePath    = \??\C:\sr\pargps\driver\XXX\pargpskd.sys  (\??\C:DriverPath)
        //  Start        = 0x00000002                     (SERVICE_AUTO_START)
        //  Tag          = 0x00000001                     (TagId)
        //  Type         = 0x00000001                     (SERVICE_KERNEL_DRIVER)
        //

        GetCurrentDirectory( MAXSTR-13, DriverPath );
        lstrcat( DriverPath, "\\pargpskd.sys" );
        
        sprintf( DriverDescBuffer, "%s PARGPS Timing Device Driver", Driver->Name );
        DriverDesc  = DriverDescBuffer;
        LoadGroup   = "GPS Timing Devices";
        TagId       = 0x1L;
        
        SCService = CreateService(
                        SCManager,             // Service Control Manager
                        Driver->Name,          // name of device driver
                        DriverDesc,            // description of driver
                        SERVICE_ALL_ACCESS,    // full access permission
                        SERVICE_KERNEL_DRIVER, // type is a kernel driver
                        SERVICE_AUTO_START,    // starts at end of booting
                        SERVICE_ERROR_NORMAL,  // give warning, but still boot
                        DriverPath,            // location of .sys file
                        LoadGroup,             // loads in custom A/D group
                        &TagId,                // load order within group
                        NULL,                  // depends on no other services
                        NULL,                  // use default ServiceStartName
                        NULL                   // no password
                       );


        if (!SCService)
                Error("Could not create device driver service", 1);




        // Add registry entries for device specific parameters
        
        lstrcpy( ParameterKeyName, "System\\CurrentControlSet\\Services\\");
        lstrcat( ParameterKeyName, Driver->Name );
        lstrcat( ParameterKeyName, "\\Parameters" );

        Code = RegCreateKeyEx(
                       HKEY_LOCAL_MACHINE,      // put under this key
                       ParameterKeyName,        // at this location
                       0,                       // reserved
                       NULL,                    // object type of key
                       REG_OPTION_NON_VOLATILE, // keep value when reboot
                       KEY_ALL_ACCESS,          // full access permission
                       NULL,                    // key handle not inheritable
                       &ParGpsKey,              // handle to key
                       &Disposition             // indicates key new or existing
                      );

        if (Code != ERROR_SUCCESS)
                Error("Failed to create parameter key", 1);


        Code1 = RegSetValueEx(
                    ParGpsKey,                      // use this key
                    REG_PARM_DRIVERNAME,            // value name
                    0,                              // reserved
                    REG_SZ,                         // value data type
                    &(CONST BYTE)Driver->Name[0],   // ptr to value contents
                    strlen(Driver->Name)+1          // contents size
                   );
        
        if ( (Code1 != ERROR_SUCCESS) )
                printf("Failed writing name parameter key values to registry\n");
        
                       
        Code = RegCloseKey( ParGpsKey );
        
        if (Code != ERROR_SUCCESS) 
                Error("Failed to close parameter key", 1);
        

        

        // Add registry entries for interpreting EventViewer entries
        //
        // Note that these entries are NOT deleted when rmdriver.exe
        // is run.  This is so that System Log entries relating to the
        // PARGPS driver can still be interpreted even when the driver
        // is no longer running.  Except for making the System Log
        // entries less informative, deleting these registry entries
        // by hand produces no bad effects.
        //
        
        lstrcpy( EventLogKeyName,
                 "System\\CurrentControlSet\\Services\\EventLog\\System\\");
        lstrcat( EventLogKeyName, Driver->Name );


        TypesSupported = 0x7L;
        lstrcpy( EventMessageFile, DriverPath );
        lstrcat( EventMessageFile, ";%SystemRoot%\\System32\\IoLogMsg.dll");
                         
        Code = RegCreateKeyEx(
                       HKEY_LOCAL_MACHINE,      // put under this key
                       EventLogKeyName,         // at this location
                       0,                       // reserved
                       NULL,                    // object type of key
                       REG_OPTION_NON_VOLATILE, // keep value when reboot
                       KEY_ALL_ACCESS,          // full access permission
                       NULL,                    // key handle not inheritable
                       &ParGpsKey,              // handle to key 
                       &Disposition             // indicates key new or existing
                      );

        if (Code != ERROR_SUCCESS) 
                Error("Failed creating Event Log Key", 1);


        Code1 = RegSetValueEx(
                    ParGpsKey,                      // use this key
                    "EventMessageFile",             // value name
                    0,                              // reserved
                    REG_EXPAND_SZ,                  // value data type
                    (CONST BYTE*)EventMessageFile,  // ptr to value contents
                    strlen(EventMessageFile)        // contents size
                    );
        
                              
        Code2 = RegSetValueEx(
                    ParGpsKey,                      // use this key
                    "TypesSupported",               // value name
                    0,                              // reserved
                    REG_DWORD,                      // value data type
                    &(CONST BYTE)TypesSupported,    // ptr to value contents
                    sizeof(DWORD)                   // contents size
                   );
        
        if ( (Code1 != ERROR_SUCCESS) ||
             (Code2 != ERROR_SUCCESS) )
                printf("Failed writing event log key values to registry\n");
        

        Code = RegCloseKey( ParGpsKey );
        if (Code != ERROR_SUCCESS)
                Error("Failed closing Event Log Key", 1);




        // Indicate install successful
        
        printf( "\n%s driver has been successfully installed.\n", Driver->Name );

        
        
        // Start Service

        Ret = StartService(

                   SCService,     // service identifier
                   0,             // number of arguments
                   NULL           // pointer to arguments

                   );


        if (!Ret)
                Error("Failed to start PARGPS device driver service", 1);
        else
                printf( "%s driver was started.\n", Driver->Name );
        

        // Close handles (leaving service loaded and running)

        CloseServiceHandle( SCService );
        CloseServiceHandle( SCManager );


        PrintMsg( SuccessMsg );
        printf( "\t\t%s\n\n", Driver->Name );
        
}



// Helper functions:

void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );

}

void Usage( void ) {

        int i;
        
        PrintMsg( UsageMsg );
                
        for ( i = 0 ; DriverInfo[i].Name ; i++ )
                printf( "     %s\n", DriverInfo[i].Name );
        
        exit( 1 );
}

void Error( char *Msg, int Fatal ) {

        DWORD  Err;
        LPVOID lpMessageBuffer;


        // Display local message, error number, and system message
        
        Err = GetLastError();

        FormatMessage(
                      FORMAT_MESSAGE_ALLOCATE_BUFFER |
                      FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      Err,
                      MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                      (LPTSTR) &lpMessageBuffer,
                      0,
                      NULL
                     );

        printf( "\n%s.  Error #%d\n\n%s\n", Msg, Err, lpMessageBuffer );

        LocalFree( lpMessageBuffer );


        
        // Give additional explanation for common errors
        
        if (Err == ERROR_SERVICE_EXISTS) // #1073
                PrintMsg( ErrorServiceExistsMsg );
        
        else
                printf( "Sorry, no extra help available for this error.\n" );


        // Clean up and exit if fatal.
        
        if ( Fatal ) {

                if (SCService)  CloseServiceHandle( SCService );
                if (SCManager)  CloseServiceHandle( SCManager );

                exit( 1 );

                }
}
