/* FILE: pargpskd.c                Copyright (c), Symmetric Research, 2003

SR PARGPS device driver functions for all OS's.

For Windows NT/2K and Linux, the device driver must first be installed
by running the provided indriver utility.

Users then interact with the driver functions by calling the pargps.c
library functions.


DRIVER ORGANIZATION OVERVIEW

The following information is provided as an overview for those wishing
to understand how this driver code is organized.  A thorough
understanding of device driver programming is recommended for anyone
wanting to modify this code.  The following resources may help:

   Linux Device Drivers by Alessandro Rubini & Jonathan Corbet
   Writing NT Device Drivers by Tony Mason & Peter Giscarola
   WDM Device Drivers by Walter Oney


CONDITIONAL COMPILE TIME DEFINES

Although there is a great deal of similarity from one OS to the next
in how device drivers work, they are not identical.  This means that
some OS specific code is required.  However, we have tried to share
and reuse as much common code as possible since this eases maintence
and thus increases reliablity.  OS specific sections of code are
selected at compile time by use of #if defined statements that check
the SROS_xxx group of defines where xxx indicates one of the supported
OS's.  Currently these are: SROS_WIN2K, SROS_WINNT, SROS_LINUX.


DEVINFO DATA STRUCTURE

The most important data structure in this driver is the device
information structure (DEVINFO).  It contains all of the driver's data
and is passed to almost all of the functions.  To make it easier to
support multiple OS's, these functions have been divided into three
main groups described below.


FUNCTION GROUPS

The first group contains functions that interact with the PARGPS hardware
and contain the fundamental behavior of the driver common to all the
OS's.  Because they are so important, they have been designated as group
1 have been given names that start with ParGps_.  For example ParGps_DrvOpen
and ParGps_GetCurrentTime.  These common functions are sandwiched between
two layers of OS specific functions.  

The second group contains OS specific helper functions which are used to
insulate the common PARGPS functions from differences between the OS's.
They have been given names that start with Os.  For example,
OsGetSystemTime or OsSetIoctlBuffers.  There is one complete set of
these helper functions for each supported OS.  Although all versions are
included in this source file, the conditional compile time SROS_xxx
defines are used to select which set is actually used.

The third group contains top level functions that are specific to an
OS.  They are typically called from the OS itself and have been given
names prefixed with the OS.  For example, Wdm_PnpStartDevice or
Linux_DrvOpen.


CALLING THE DRIVER

The functions in this driver are typically called in one of three ways;
directly from the OS, from the OS as a result of call from user code or
from the related PARxCH driver as the result of a hardware interrupt.

When the indriver installation utility is run, the OS calls some of
the OS specific top level functions to initialize the driver upon
installation. 

Once the device driver has been installed, user code can use the
provided pargps.c library to initiate calls into this driver.  Typically
these calls will be used to do things like open, start, and read data
from the driver.  Many of these calls from user code to the PARGPS
library and on to the OS and one of the top level OS specific
functions before reaching the shared ParGps_DrvIoctl function which deals
with any data transfer before calling an appropriate ParGps_ function to
actually service the request.  Although this sounds complicated, it is
all relatively straightforward compared to the code path related to
interrupts and the PARxCH driver.


WORKING WITH THE PARxCH

To work with the PARGPS, the PARxCH A/D driver must first be compiled
with PARGPS support.  Then, after both drivers have been installed, the
PARxCH driver must be "attached" to the PARGPS driver using a call
from the parxch.c library.  This results in both drivers having a
pointer to a common kernel data area which they share and use to store
pointers to selected functions which the other driver might need to
call.  Finally, interrupts must be enabled with another call from the
PARxCH library.


HARDWARE BEHAVIOR

The PARGPS board generates a high precision PPS pulse per second
signal indicating the start of each second that is fed into the PARxCH
digital I/O.  When the PARxCH board sees this PPS signal, it generates
a hardware interrupt on the parallel port.  This is referred to as the
PPS interrupt.

Between 200 and 500 milliseconds after the PPS signal, the PARGPS
begins sending NMEA messages containing coarse time and location
information to the serial port at 4800 baud.  It typically takes 100
to 300 more milliseconds for the full message to arrive, depending on
how many characters the message contains.

During all this time, the PARxCH is continuously acquiring data.  If
its control bits are correctly set after a PPS signal comes in, the
PARxCH will place a mark in the data stream and send another parallel
port interrupt when the next data sample is ready.  This is referred
to as the DREADY interrupt.

The task of the driver is to save the PPS and NMEA serial information,
set the PARxCH control bits, save the marked data and keep track of
the relationship between all these items.



INTERRUPT PROCESSING

A relatively short time after the hardware interrupt occurs, the
PARxCH interrupt function is called which in turn calls the PARGPS
interrupt function as listed in the common kernel data area.

Since interrupt routines should be very fast, the PARGPS interrupt
function gets the current system time and saves data.  Then it queues
a DPC deferred procedure call or bottom half that will finish the
needed processing.  The DPC function run as soon as possible, but is
allowed to take a little longer since interrupts are no longer
disabled when it runs.  

While there is only one PARGPS interrupt routine, the hardware discussion
above makes it clear that two different situations may generate the
parallel port hardware interrupt.  The interrupt routine uses the
device info IntrSourceNext variable to distinguish between them.  It
performs slightly different tasks and queues up different DPC
functions depending on whether it is servicing a PPS or a DREADY
interrupt.



PPS INTERRUPT

The PARGPS DPC that runs for the PPS interrupt does the following
three things: request a DREADY interrupt, save some serial data,
signal a thread to get new serial data.  These tasks are explained in
more detail below.

The first task, requesting a DREADY interrupt means calling a PARxCH
driver function to request that the PARxCH hardware send another
interrupt signal when the next piece of acquired data is ready.  The
code refers to this as the DREADY interrupt.  

The second task, saving some serial data, involves saving NMEA serial
data by transferring it from a smaller temporary (2 second) array to a
larger, longer term (5 minute) storage array where the data is held
until user code calls to retrieve it.  The data being transferred is
associated with the PPS signal that occurred 2 seconds before the
current signal.  The delay is required to ensure that the full NMEA
string is available.

The third task, signaling a thread to get new serial data, sends a
signal to a kernel system thread which reads from the serial port.
The data it reads is asociated with the PPS signal that occurred 1
second before the current signal, as the serial data for the current
signal may not arrive for another 200 to 500 milliseconds.


DREADY INTERRUPT PROCESSING

The PARGPS DPC that runs for the DREADY interrupt is much simpler and
just saves the PPS and DREADY counter information.  These 64 bit
counters indicate the time of the PPS signal and the time the next
piece of acquired data came ready.


SERIAL MONITORING THREAD

It would have been nice to actually read the NMEA serial data from
inside a DPC function, but this is not possible.  Although the DPC
runs at a "safer" time than the interrupt routine itself, it is still
not safe enough for serial access.  So, the serial access must be done
by a separate kernel system thread

This kernel system thread is started when the PARGPS start function is
called.  Then it waits in a loop looking for a signal.  After getting
the signal, it reads the NMEA data from the serial port.  Then it goes
back to waiting.  It is stopped when the PARGPS stop function is called.


CURRENT TIME

The PARGPS driver can be opened multiple times from user code.  However,
the user must ensure that the pargps.c library ParGpsOpen function which
sets the serial port and initializes the PARGPS is only used once.  The
handle returned from this full open is the only one which should be used
for attaching the PARxCH and starting the PARGPS.

All other user code should use the pargps.c library ParGpsOpenForTimeOnly
function.  The handle returned from this should be used only for retrieving
the current time by calling the ParGpsGetCurrentTime library function.

Driver support for the ParGpsGetCurrentTime function is provided by
keeping a separate CurrentTimeData structure which contains the latest
PPS counter information and the latest NMEA serial information that
was saved as part of the regular functioning of the PPS and DREADY
interrupt DPCs.  In addition, when the function is called, a 64 bit
counter representing the current system time is also saved to the
CurrentTimeData structure.  The whole structure is then returned to
the library and user code.  Care must be taken when working with data
shared between DPC functions and functions called from user code via
the library to be sure simultaneous accesses are prevented.


SUMMARY

The organization of the PARGPS device driver has been discussed.
Hopefully, this will help anyone who needs to understand the workings
of this code and that further questions can be answered by studying
the code itself.  If you still have questions, please send email to
info@symres.com and we will be happy to discuss them with you.


*/




/* DEBUGGING DEFINE:

This define controls the level of printf type debugging that will come
from the kernel mode driver.  -1 = no printf debugging statements.
SRDBG = (1,2,3,4,5) -> various levels of reporting.  Production
compiles are done with -1.

*/

#define SRDBG  -1




/* LOCAL DEFINE:

This define controls whether the driver functions are visible to
other source files.  Most users should let SRLOCAL be defined as
static to avoid namespace clutter and possible function name
contention.  The only exceptions would be for specialized debugging
and diagnostic programs.

*/

#if !defined( SRLOCAL )
#define SRLOCAL static
#endif
   

////////////////////////////////////////////////////////////////////////
/* OS dependent includes, defines, and typedefs.
 *
 *
 *
 */

#if defined( SROS_WIN98 ) || defined( SROS_WIN2K )
#include <wdm.h>
#include <stdio.h>
#include "pargpslg.h"  // includes error log codes

#define NTCOUNT_TO_SEC 10000000

#define ATOMIC  long 
typedef void KTASK;
typedef long *KWAIT;


#elif defined( SROS_WINNT )
#include <ntddk.h>
#include <stdio.h>
#include "pargpslg.h"  // includes error log codes

#define NTCOUNT_TO_SEC 10000000

#define ATOMIC  long 
typedef long KTASK;
typedef long *KWAIT;

#define PowerDeviceD0      1
#define PowerSystemWorking 1
typedef int DEVICE_POWER_STATE;
typedef int SYSTEM_POWER_STATE;
typedef int DEVICE_CAPABILITIES;
typedef int IO_REMOVE_LOCK;


#elif defined( SROS_LINUX )
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>     // printk
#include <linux/ioport.h>     // check_region, etc
#include <linux/fs.h>         // everything ...
#include <linux/smp_lock.h>   // lock_kernel
#include <linux/sched.h>      // current for file.h, pt_regs for interrupt.h
#include <linux/interrupt.h>  // bottom half queue constant
#include <linux/tqueue.h>     // task queue
#include <linux/wait.h>       // wait queue
#include <linux/slab.h>       // kmem_cache_free needed by putname
#include <linux/file.h>       // get_unused_fd, fget, etc
#include "lnxver.h"           // Account for kernel changes. Includes
                              // linux/poll.h, linux/sched.h, linux/init.h, 
                              // asm/uaccess.h, asm/io.h
                              // and maybe linux/vmalloc.h
#ifdef USE_RDTSC_FOR_TIME
#include <asm/msr.h>          // machine specific registers like rdtsc
#endif // USE_RDTSC_FOR_TIME

#define ATOMIC            atomic_t
#define WCHAR             char
#define PowerDeviceD0      1
#define PowerSystemWorking 1

#define STATUS_SUCCESS                          (0x00000000L)
#define STATUS_CANNOT_DELETE                    ((long)(-ESRCH))
#define STATUS_INVALID_PORT_ATTRIBUTES          ((long)(-EINVAL))
#define STATUS_TOKEN_ALREADY_IN_USE             ((long)(-EBUSY))
#define STATUS_NO_TOKEN                         ((long)(-ENXIO))
#define STATUS_INVALID_PARAMETER                ((long)(-EINVAL))
#define STATUS_INVALID_USER_BUFFER              ((long)(-EFAULT))
#define STATUS_INVALID_DEVICE_REQUEST           ((long)(-EINVAL))
#define STATUS_DEVICE_CONFIGURATION_ERROR       ((long)(-ENODEV))
#define STATUS_CANCELLED                        ((long)(-ECHILD))
#define STATUS_MEMORY_NOT_ALLOCATED             ((long)(-ENOMEM))
#define STATUS_PIPE_EMPTY                       ((long)(-EAGAIN))
#define STATUS_TIMEOUT                          ((long)(-ETIME))
#define STATUS_FILE_INVALID                     ((long)(-EBADF))
//#define STATUS_SERIAL_NO_DEVICE_INITED          ((long)(-EBADF))
#define STATUS_SECTION_NOT_EXTENDED             ((long)(-ENOBUFS))

#define NT_SUCCESS(status) ( status == STATUS_SUCCESS ? 1 : 0 )

#define PARGPS_LOG_INFO 1
#define PARGPS_LOG_WARN 2
#define PARGPS_LOG_ERR  3

typedef int DEVICE_POWER_STATE;
typedef int SYSTEM_POWER_STATE;
typedef int DEVICE_CAPABILITIES;
typedef int IO_REMOVE_LOCK;
typedef long DRIVER_OBJECT;
typedef long DEVICE_OBJECT;
typedef char *UNICODE_STRING;
typedef unsigned long NTSTATUS;
typedef unsigned int *PKINTERRUPT;
typedef void (*PKSERVICE_ROUTINE)( int Irq, void *Dev, struct pt_regs *Regs );
typedef struct tq_struct KDPC;
typedef struct wait_queue *KEVENT;
typedef wait_queue_head_t KWAIT;
typedef int (*PKSTART_ROUTINE)( void *Context );
typedef struct task_struct KTASK;
typedef struct semaphore KMUTEX;
typedef unsigned long KIRQL;
typedef spinlock_t KSPIN_LOCK;
typedef struct file FILE;

#endif  // SROS_xxxxx



////////////////////////////////////////////////////////////////////////
/* Common includes and defines:

These includes and defines are the same across all the OS's.

*/

#include "pargps.h"    // includes PARGPS defines shared with the user & DLL
#include "pargpskd.h"  // includes PARGPS IOCTL codes and device name defines
#include "gpsa2dkd.h"  // includes PARGPS + A/D Common Kernal Area defines
#include "dbgprint.h"  // for debugging

// Define defaults.  Common addr,irq values are: 0x378,7  0x278,5

#define DEFAULT_PARGPS_NAME     PARGPS_0x378
#define DEFAULT_PARGPS_MAJOR    121

#define MAX_STR_LENGTH          256

// Be careful changing this.  It must match a bit position.  eg 4,8,16.
#define MAXGPSDATA 4

#define RUNSTATE int
#define PWRSTATE int

#define RUNSTATE_STOPPED  0
#define RUNSTATE_RUNNING  1

#define PWRSTATE_POWEROFF 0
#define PWRSTATE_POWERON  1

#define IOCTL_STANDARD    0
#define IOCTL_INTERNAL    1

#define MUTEX_LOCKED      0
#define MUTEX_AVAILABLE   1






////////////////////////////////////////////////////////////////////////
/* DEVICE INFORMATION STRUCTURE:

This very important structure contains all the global data which the
driver code needs.  It is passed to most functions.

*/

SRLOCAL struct DeviceInformation {

        int DevInfoTagStart;


        // Name of driver
        
        char DriverNameBuffer[MAX_STR_LENGTH];

        
        // Usage count.

        ATOMIC NumberOpens;


        // Windows Driver Model (WDM) support

        DRIVER_OBJECT *DriverObj;
        DEVICE_OBJECT *ThisFdoDevObj;
        DEVICE_OBJECT *LowerDevObj;
        DEVICE_OBJECT *PdoDevObj;
        UNICODE_STRING LinkName;
        RUNSTATE CurrDevState;
        RUNSTATE PrevDevState;
        DEVICE_POWER_STATE CurrDevPower;
        SYSTEM_POWER_STATE CurrSysPower;
        DEVICE_CAPABILITIES DevCaps;
        IO_REMOVE_LOCK RemoveLock;
        KEVENT   PowerEvent;
        NTSTATUS PowerStatus;

// WCT: Review 
//      POWERSTATE CurrPowerState;     // Oney notes saying don't include
//      UNICODE_STRING InterfaceName;  // this OR link name ?


        // Events and handles

        char      *SerialPortName;
        DEVHANDLE  hSerial;
        
        KEVENT     PpsEvent;
        KEVENT     ThreadEvent;

        int        ThreadTerminate;
        KTASK     *ThreadTask;
        KMUTEX     ThreadMutex;
        KWAIT      ThreadWait;
        int        ThreadTimeOutCount;


        // Sr Kernel Area Common to A/D and PARGPS drivers

        GPSCKA *SrCommonArea;


        // Current time data and its spin lock.

        GPSTIMEDATA CurrentTimeData;
        KSPIN_LOCK  CurrentTimeLock;

        KMUTEX      IoctlMutex;


        // Interrupt source, counts, and DPCs
        
        int     IntrSourceLast;
        int     IntrSourceNext;
        int     IntrCount;
        int     GpsDreadyCount;
        KDPC    CustomDpc[PARGPS_INTTYPE_MAX];


        // Temporary storage for PARGPS PPS and serial data
        // filled by interrupt service routine.
        
        int DataIndexCurr;
        int DataIndexPrev;
        int DataIndexPrev2;

        PPSDATA    PpsDataTemp[MAXGPSDATA];
        SERIALDATA SerialDataTemp[MAXGPSDATA];
        

        
        // Software FIFO to hold PARGPS data until user requests it.

        int         PpsDataRdIndex;
        int         PpsDataWrIndex;
        PPSDATA     PpsDataStore[MAX_PARGPS_STORAGE];

        int         SerialDataRdIndex;
        int         SerialDataWrIndex;
        SERIALDATA  SerialDataStore[MAX_PARGPS_STORAGE];



        // Internal circular buffer to hold serial NMEA data

        char SerialBuffer[MAX_NMEA_BUFF];
        int  SbUsePtr;
        int  SbFillPtr;
        int  SbWrap;
        

        // Error message buffer.
        
        WCHAR MsgBuffer[ MAX_STR_LENGTH ];


        int DevInfoTagEnd;

} ParGpsDeviceInfo;

typedef struct DeviceInformation DEVINFO;




// Global common kernel area shared with A/D

GPSCKA SrCommonKernelArea;

// Global Variable

UNICODE_STRING DriverRegPath;








////////////////////////////////////////////////////////////////////////
/* FUNCTION PROTOTYPES:
 *
 *
 *
 *
 */

// Group 1: PARGPS hardware and behavior functions:
// 
// 
// 
// 
// 

SRLOCAL long ParGps_DrvOpen(   DEVINFO *Dev, IRP *Irp );
SRLOCAL long ParGps_DrvClose(  DEVINFO *Dev, IRP *Irp );
SRLOCAL long ParGps_DrvIoctl(  DEVINFO *Dev, IRP *Irp );
SRLOCAL void ParGps_DrvUnload( DEVINFO *Dev );

SRLOCAL void ParGps_FillDevInfo( DEVINFO *Dev, char *DriverName );
SRLOCAL void ParGps_FillCommonArea( DEVINFO *Dev, int FillHeader );
SRLOCAL void ParGps_ClearCurrentTime( DEVINFO *Dev );
SRLOCAL void ParGps_ClearDataIndex( DEVINFO *Dev );
SRLOCAL void ParGps_Start( DEVINFO *Dev );
SRLOCAL void ParGps_Stop( DEVINFO *Dev );
SRLOCAL void ParGps_Init( DEVINFO *Dev, int *IoBuffer );
SRLOCAL void ParGps_Reset( DEVINFO *Dev );
SRLOCAL void ParGps_SetInterruptMode( DEVINFO *Dev, int Mode );
SRLOCAL void ParGps_SetSerialPort( DEVINFO *Dev, int Port );
SRLOCAL void ParGps_SavePpsData( DEVINFO *Dev );
SRLOCAL void ParGps_SaveSerialData( DEVINFO *Dev );
SRLOCAL unsigned long ParGps_ReadPpsData( DEVINFO *Dev, PPSDATA *OutBuffer, int Nout );
SRLOCAL unsigned long ParGps_ReadSerialData( DEVINFO *Dev, SERIALDATA *OutBuffer, int Nout );
SRLOCAL unsigned long ParGps_GetCurrentTime( DEVINFO *Dev, GPSTIMEDATA *OutBuffer );

SRLOCAL long ParGps_ThreadControl( DEVINFO *Dev, int RequestState );
SRLOCAL unsigned long ParGps_ThreadFunc( DEVINFO *Dev );
SRLOCAL void ParGps_SerialProcessData( DEVINFO *Dev );
SRLOCAL long ParGps_SerialReadNmea( DEVINFO *Dev,
                                    int MaxSize,
                                    int *LastMsg,
                                    char *NmeaMsg,
                                    int *Nchar,
                                    int *Error );
SRLOCAL long ParGps_SerialFillBuffer( DEVINFO *Dev );
SRLOCAL long ParGps_SerBufGetToken( DEVINFO *Dev,
                                    char *Token,
                                    char Sep,
                                    int *chksum,
                                    int *numchar );
SRLOCAL char ParGps_SerBufGetNextChar( DEVINFO *Dev );
SRLOCAL int ParGps_SerBufPeekOne( DEVINFO *Dev, char *c );
SRLOCAL int ParGps_SerBufPutOne( DEVINFO *Dev, char c );
SRLOCAL long ParGps_SerialReadChar( DEVINFO *Dev, char *C, int *Nchar  );
SRLOCAL int ParGps_SerBufGetMsgIndex( char *Token );
SRLOCAL int ParGps_SerAtoix( char *src, int n );

int  ParGps_Interrupt( DEVINFO *Dev );
void ParGps_DpcForPpsRoutine( DEVINFO *Dev );
void ParGps_DpcForDreadyRoutine( DEVINFO *Dev );
void ParGps_DpcForPlainRoutine( DEVINFO *Dev );



                         

// Group 2: OS specific helper functions:
// 
// 
// 
// 
// 

SRLOCAL void OsMessage( DEVINFO *Dev, char *Msg, long Status, int Level );
SRLOCAL unsigned long OsCopyToUser( void *UserBuff, const void *KernelBuff, unsigned long Nbytes );
SRLOCAL unsigned long OsCopyFromUser( void *KernelBuff, const void *UserBuff, unsigned long Nbytes );
SRLOCAL unsigned long OsCopyToDirectIo( void *MidlBuff, const void *KernelBuff, unsigned long Nbytes );
SRLOCAL void OsIrpFinish( IRP *Irp, long Status, unsigned long Info );
SRLOCAL void OsSetIoctlCommand( IRP *Irp, unsigned long *Command );
SRLOCAL long OsSetIoctlBuffers( IRP *Irp,
                                unsigned long Command,
                                int **InBuffer,
                                int **OutBuffer,
                                int *Nin,
                                int *Nout );
SRLOCAL void OsReturnIoctlBuffer( IRP *Irp,
                                  unsigned long Command,
                                  int *OutBuffer,
                                  unsigned long ReturnedBytes );

SRLOCAL long OsSerialOpen( DEVINFO *Dev );
SRLOCAL long OsSerialClose( DEVINFO *Dev );
SRLOCAL long OsSerialFlush( DEVINFO *Dev );
SRLOCAL int  OsSerialReadChar( DEVINFO *Dev,
                               char *Buffer,
                               unsigned long BuffLen,
                               unsigned long *Status );
SRLOCAL long OsSerialSignal( DEVINFO *Dev );
SRLOCAL long OsThreadOpen( DEVINFO *Dev );
SRLOCAL long OsThreadClose( DEVINFO *Dev );
SRLOCAL void OsThreadInit( DEVINFO *Dev );
SRLOCAL void OsThreadLeave( DEVINFO *Dev, unsigned long Status );
SRLOCAL void OsGetSystemTime( SR64BIT *SysTime );
SRLOCAL long OsGetCounterFrequency( void );
SRLOCAL unsigned long OsWaitOnPps( DEVINFO *Dev,
                                   SR64BIT *RequestedTimeout,
                                   int Id );



// Group 3: Top level OS dependent initialization and "dispatch" functions:
// 
// 
// 
// 
// 

#if defined( SROS_WIN98 ) ||  defined( SROS_WIN2K )

NTSTATUS DriverEntry(      DRIVER_OBJECT *DriverObject, UNICODE_STRING *RegistryPath );
void     Wdm_DrvUnload(    DRIVER_OBJECT *DriverObject );
NTSTATUS Wdm_DrvAddDevice( DRIVER_OBJECT *DriverObject,  DEVICE_OBJECT *PDO );

NTSTATUS Wdm_DrvOpen(  DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_DrvClose( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_DrvIoctl( DEVICE_OBJECT *FDO, IRP *Irp );

BOOLEAN Wdm_GpsInterrupt( PKINTERRUPT pInterruptObject, void *Context );

void Wdm_DpcForPpsRoutine(    KDPC *DpcForPps,    void *Context, void *SysArg1, void *DpcData );
void Wdm_DpcForDreadyRoutine( KDPC *DpcForDready, void *Context, void *SysArg1, void *DpcData );
void Wdm_DpcForPlainRoutine(  KDPC *DpcForPlain,  void *Context, void *SysArg1, void *DpcData );
                               
void WdmWinnt_ThreadFunc( void *Context );

NTSTATUS Wdm_DrvPnp(               DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpStartDevice(       DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpStopDevice(        DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpRemoveDevice(      DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryCapabilities( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpCancelStop(        DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpCancelRemove(      DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryStop(         DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryRemove(       DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryDevRelations( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpQueryInterface(    DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpFilterResourceReq( DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PnpSurpriseRemoval(   DEVICE_OBJECT *FDO, IRP *Irp );

SRLOCAL NTSTATUS Wdm_RemoveDevice(  DEVICE_OBJECT *FDO );
SRLOCAL NTSTATUS Wdm_ResourceSet(   DEVICE_OBJECT *FDO, IRP *Irp );
SRLOCAL NTSTATUS Wdm_ResourceClear( DEVICE_OBJECT *FDO );
SRLOCAL NTSTATUS Wdm_StartDevice(   DEVICE_OBJECT *FDO );
SRLOCAL NTSTATUS Wdm_StopDevice(    DEVICE_OBJECT *FDO );


NTSTATUS Wdm_DrvPower(                DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PowerSetSystem(          DEVICE_OBJECT *FDO, IRP *Irp, POWER_STATE PowerState );
NTSTATUS Wdm_SendDeviceSetPower(      DEVICE_OBJECT *FDO, IRP *Irp, POWER_STATE PowerState );
void Wdm_OnCompleteDeviceSetPower(    DEVICE_OBJECT *FDO,
                                         UCHAR MinorFunction,
                                         POWER_STATE PowerState,
                                         VOID *Context,
                                         IO_STATUS_BLOCK *IoStatus );
NTSTATUS Wdm_OnCompletePowerUpSystem( DEVICE_OBJECT *FDO, IRP *Irp, void *Context );
NTSTATUS Wdm_PowerSetDevice(          DEVICE_OBJECT *FDO, IRP *Irp, POWER_STATE PowerState );
NTSTATUS Wdm_OnCompletePowerUpDevice( DEVICE_OBJECT *FDO, IRP *Irp, void *Context );
NTSTATUS Wdm_PowerForward(            DEVICE_OBJECT *FDO, IRP *Irp );
NTSTATUS Wdm_PowerFail( IRP *Irp, NTSTATUS Status );

NTSTATUS Wdm_DrvWmi(   DEVICE_OBJECT *FDO, IRP *Irp ); // Windows Mgmt Interface

SRLOCAL NTSTATUS Wdm_OnRequestComplete( DEVICE_OBJECT *FDO,
                                        IRP *Irp,
                                        KEVENT *Event );
SRLOCAL NTSTATUS Wdm_IrpBuildSendAndWait(
                            ULONG           Internal,
                            ULONG           IoControlCode,
                            PDEVICE_OBJECT  TargetDeviceObject,
                            PVOID           InputBuffer,
                            ULONG           InputBufferLength,
                            PVOID           OutputBuffer,
                            ULONG           OutputBufferLength,
                            PLARGE_INTEGER  RequestedTimeout
                            );

SRLOCAL NTSTATUS Wdm_IrpForwardAndWait( DEVICE_OBJECT *FDO, IRP *Irp );
SRLOCAL NTSTATUS Wdm_IrpForward(        DEVICE_OBJECT *FDO, IRP *Irp );

SRLOCAL NTSTATUS Wdm_GetParameters( DRIVER_OBJECT *DriverObject,
                                    UNICODE_STRING *RegistryPath,
                                    UNICODE_STRING *ParGpsName );

SRLOCAL NTSTATUS WdmWinnt_ErrorLog( void *IoObject,
                                    IRP *Irp,
                                    NTSTATUS CurrStatus,
                                    unsigned long ErrCode,
                                    const WCHAR *Msg );



#elif defined( SROS_WINNT )

NTSTATUS DriverEntry( DRIVER_OBJECT *DriverObject,
                      UNICODE_STRING *RegistryPath );
void     Winnt_DrvUnload( DRIVER_OBJECT *DriverObject );

NTSTATUS Winnt_DrvOpen(  DEVICE_OBJECT *DeviceObject, IRP *Irp );
NTSTATUS Winnt_DrvClose( DEVICE_OBJECT *DeviceObject, IRP *Irp );
NTSTATUS Winnt_DrvIoctl( DEVICE_OBJECT *DeviceObject, IRP *Irp );

BOOLEAN Winnt_ParGpsInterrupt( PKINTERRUPT Interrupt, void *Context );

void Winnt_DpcForPpsRoutine(    KDPC *DpcForPps,    void *Context, void *SysArg1, void *DpcData );
void Winnt_DpcForDreadyRoutine( KDPC *DpcForDready, void *Context, void *SysArg1, void *DpcData );
void Winnt_DpcForPlainRoutine(  KDPC *DpcForPlain,  void *Context, void *SysArg1, void *DpcData );

void WdmWinnt_ThreadFunc( void *Context );

SRLOCAL NTSTATUS Winnt_GetParameters( DRIVER_OBJECT *DriverObject,
                                      UNICODE_STRING *RegistryPath,
                                      UNICODE_STRING *ParGpsName );

SRLOCAL void     Winnt_ResourceFree( DRIVER_OBJECT *DriverObject,
                                     UNICODE_STRING *NtDeviceName,
                                     DEVICE_OBJECT *DeviceObject,
                                     UNICODE_STRING *Win32DeviceName,
                                     int SymLink );

SRLOCAL NTSTATUS WdmWinnt_ErrorLog( void *IoObject,
                                    IRP *Irp,
                                    NTSTATUS CurrStatus,
                                    unsigned long ErrCode,
                                    const WCHAR *Msg );

                         

#elif defined( SROS_LINUX )

int Linux_ThreadFunc( void *Context );

void Linux_ParGpsInterrupt( int irq, void *dev_id, struct pt_regs *regs );
void Linux_DpcForPpsRoutine(    void *DpcData );
void Linux_DpcForDreadyRoutine( void *DpcData );
void Linux_DpcForPlainRoutine(  void *DpcData );

SRLOCAL void Linux_GetParameters( void );
SRLOCAL void Linux_ResourceFree( int ParGpsMajor, char *ParGpsName );

int     Linux_DrvOpen(  struct inode *inode, struct file *filp );
close_t Linux_DrvClose( struct inode *inode, struct file *filp );
int     Linux_DrvIoctl( struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg );

void cleanup_module( void );
int     init_module( void );

#endif  // SROS_xxxxx



////////////////////////////////////////////////////////////////////////
// These are some Group 2 Os specific helper defines.  They should
// logically be listed with the other group functions, but the compiler
// needs to know about them before they are used by the Group 1 functions.
//

// *************************************************************************
#if defined( SROS_WIN98 ) || defined( SROS_WIN2K ) || defined( SROS_WINNT )

#define OsModuleUseCountInc
#define OsModuleUseCountDec

#define OsQueueDPC( Dev, Inttype, DeviceObject ) {               \
                KeInsertQueueDpc( &Dev->CustomDpc[Inttype],      \
                                  DeviceObject->CurrentIrp,      \
                                  DeviceObject->DeviceExtension  \
                                 );                              \
                }

// Atomic operation functions

#define OsAtomicSet( pAtom, Value )  InterlockedExchange( (pAtom), (Value) )
#define OsAtomicInc( pAtom )         InterlockedIncrement( (pAtom) )
#define OsAtomicDec( pAtom )         InterlockedDecrement( (pAtom) )
#define OsAtomicAdd( pAtom, Amount ) InterlockedExchangeAdd( (pAtom), (Amount) )
#define OsAtomicSub( pAtom, Amount ) InterlockedExchangeAdd( (pAtom), -(Amount) )
#define OsAtomicRead( pAtom )        *(pAtom)


// Spinlock and mutex functions

#define OsSpinLockAcquire( pLock, OldIrqlOrFlags ) \
        KeAcquireSpinLock( (pLock), &(OldIrqlOrFlags) );
#define OsSpinLockRelease( pLock, OldIrqlOrFlags ) \
        KeReleaseSpinLock( (pLock), (OldIrqlOrFlags) );

#define OsMutexInit( pMutex, LevelOrValue ) \
        KeInitializeMutex( (pMutex), LevelOrValue );
#define OsMutexAcquire( pMutex ) \
        KeWaitForSingleObject( (pMutex), Executive, KernelMode, FALSE, NULL );
#define OsMutexRelease( pMutex ) \
        KeReleaseMutex( (pMutex), FALSE );



// *************************************************************************
#elif defined( SROS_LINUX )

#define OsModuleUseCountInc  MOD_INC_USE_COUNT
#define OsModuleUseCountDec  MOD_DEC_USE_COUNT

#define OsQueueDPC( Dev, Inttype, DeviceObject ) {                    \
                queue_task( &Dev->CustomDpc[Inttype], &tq_immediate ); \
                mark_bh( IMMEDIATE_BH );                               \
                }


// Atomic operation functions

#define OsAtomicSet( pAtom, Value )  atomic_set( (pAtom), (Value) )
#define OsAtomicInc( pAtom )         atomic_inc( (pAtom) )
#define OsAtomicDec( pAtom )         atomic_dec( (pAtom) )
#define OsAtomicAdd( pAtom, Amount ) atomic_add( (Amount), (pAtom) )
#define OsAtomicSub( pAtom, Amount ) atomic_sub( (Amount), (pAtom) )
#define OsAtomicRead( pAtom )        atomic_read( (pAtom) )


// Spinlock and mutex functions

#define OsSpinLockAcquire( pLock, OldIrqlOrFlags ) \
        spin_lock_irqsave( (pLock), (OldIrqlOrFlags) );
#define OsSpinLockRelease( pLock, OldIrqlOrFlags ) \
        spin_unlock_irqrestore( (pLock), (OldIrqlOrFlags) );

//WCT: may need -ERESTARTSYS return if down_interruptible
#define OsMutexInit( pMutex, LevelOrValue ) \
        sema_init( pMutex, MUTEX_AVAILABLE );
#define OsMutexAcquire( pMutex ) \
        down_interruptible( (pMutex) );
#define OsMutexRelease( pMutex ) \
        up( (pMutex) );

#endif // SROS_xxxxx  End of OS dependent helper defines










////////////////////////////////////////////////////////////////////////
// START OF CODE ...
//
//
//




////////////////////////////////////////////////////////////////////////
/* Group 1: PARGPS HARDWARE AND BEHAVIOR FUNCTIONS:
 *
 *
 *
 *
 *
 *
 */

SRLOCAL long ParGps_DrvOpen( DEVINFO *Dev, IRP *Irp ) {

        dbg1_printk("In ParGps_DrvOpen\n");
        
        // We now allow more than one device open at a time.

        OsAtomicInc( &Dev->NumberOpens );
        dbg1_printk("Number of opens is %d\n",OsAtomicRead(&Dev->NumberOpens));

        OsModuleUseCountInc;

        OsIrpFinish( Irp, STATUS_SUCCESS, 0 );        
        
        return( STATUS_SUCCESS );
}

SRLOCAL long ParGps_DrvClose( DEVINFO *Dev, IRP *Irp ) {

        dbg1_printk("In ParGps_DrvClose\n");

        // If this is the last open PARGPS driver, ensure that
        // the thread and serial port are correctly closed

        if ( (OsAtomicRead( &Dev->NumberOpens ) == 1) && Dev->ThreadTask ) {
                dbg1_printk("Last open, and thread still running at close\n");
                ParGps_Stop( Dev );
                }


        // Indicate we're finished with the device.

        dbg1_printk("Closing with num locks %d\n",OsAtomicRead(&Dev->NumberOpens));
        OsAtomicDec( &Dev->NumberOpens );
        
        OsModuleUseCountDec;
        
        OsIrpFinish( Irp, STATUS_SUCCESS, 0 );        
        
        return( STATUS_SUCCESS );
}

SRLOCAL long ParGps_DrvIoctl( DEVINFO *Dev, IRP *Irp ) {

        long           Status;
        unsigned long  ReturnedBytes, Command;
        int           *InBuffer, *OutBuffer, Nin, Nout;
        int            RequestState;

        dbg3_printk("In ParGps_DrvIoctl\n");
         
        
        // Set and check command
        
        OsSetIoctlCommand( Irp, &Command );

        if ( ( IOCTL_PARGPS_TYPE(Command) != PARGPS_DEVICE_TYPE ) ||
             ( IOCTL_PARGPS_FUNC(Command) >= IOCTL_PARGPS_MAXFUNC ) ) {
                dbg0_printk("In ParGps_DrvIoctl, invalid request\n");
                 OsIrpFinish( Irp, STATUS_INVALID_DEVICE_REQUEST, 0 );
                return( STATUS_INVALID_DEVICE_REQUEST );
                }



        // Set, check, and fill the data buffers
        
        Status = OsSetIoctlBuffers( Irp, Command, &InBuffer, &OutBuffer,
                                    &Nin, &Nout );
        
        if (Status != STATUS_SUCCESS) {
                dbg0_printk("In ParGps_DrvIoctl, could not fill buffers\n");
                OsIrpFinish( Irp, Status, 0 );
                return( Status );
                }
        


        
        // Call specified IOCTL request function
        
        Status        = STATUS_SUCCESS;
        ReturnedBytes = 0;

        

        // Handle Current Time request

        if (Command == IOCTL_PARGPS_GET_CURRENT_TIME) {
                dbg1_printk( "Requested IOCTL_PARGPS_GET_CURRENT_TIME\n" );
                ReturnedBytes = ParGps_GetCurrentTime( Dev, (GPSTIMEDATA*)OutBuffer );
                OsReturnIoctlBuffer( Irp, Command, OutBuffer, ReturnedBytes );
                OsIrpFinish( Irp, Status, ReturnedBytes );
                return( Status );
                }

        // All other IOCTL requests are protected by a mutex so they are
        // only called by one user at a time.  This forces any thread
        // trying to call these functions sleep until the mutex is
        // available.

        else {
                OsMutexAcquire( &Dev->IoctlMutex );
                }




        dbg3_printk("Ready to switch on command 0x%lX\n", Command);
        
        switch (Command) {

                case IOCTL_PARGPS_START:
                        dbg2_printk( "Requested IOCTL_PARGPS_START\n" );
                        ParGps_Start( Dev );
                        break;


                case IOCTL_PARGPS_STOP:
                        dbg2_printk( "Requested IOCTL_PARGPS_STOP\n" );
                        ParGps_Stop( Dev );
                        break;


                case IOCTL_PARGPS_INIT:
                        dbg2_printk( "Requested IOCTL_PARGPS_INIT\n" );
                        ParGps_Init( Dev, InBuffer );
                        ReturnedBytes = 2*sizeof(int);
// WCT: when using alternate way of passing parameters
//                      ReturnedBytes = sizeof(INITPARGPSPARMS);
                        break;


                case IOCTL_PARGPS_RESET:
                        dbg1_printk( "Requested IOCTL_PARGPS_RESET\n");
                        ParGps_Reset( Dev );
                        break;

                        
                case IOCTL_PARGPS_SET_INTR_MODE:
                        dbg2_printk( "Requested IOCTL_PARGPS_SET_INTR_MODE\n" );
                        ParGps_SetInterruptMode( Dev, *InBuffer );
                        break;


                case IOCTL_PARGPS_SET_SERIAL_PORT:
                        dbg2_printk( "Requested IOCTL_PARGPS_SET_SERIAL_PORT\n" );
                        ParGps_SetSerialPort( Dev, *InBuffer );
                        break;


                case IOCTL_PARGPS_GET_INTR_COUNT:
                        dbg2_printk( "Requested IOCTL_PARGPS_GET_INTR_COUNT\n" );
                        dbg2_printk( "IntrCount is %d\n", Dev->IntrCount );                        
                        *OutBuffer     = Dev->IntrCount;
                        ReturnedBytes = sizeof(int);
                        break;


                case IOCTL_PARGPS_GET_KAREA:
                        dbg1_printk( "Requested IOCTL_PARGPS_KAREA\n" );
                        dbg1_printk("Common is 0x%p\n",Dev->SrCommonArea);
                        *OutBuffer    = (int)Dev->SrCommonArea;
                        ReturnedBytes = sizeof(long*);
                        break;
                        
                        
                case IOCTL_PARGPS_GET_COUNTERFREQ:
                        dbg4_printk( "Requested IOCTL_PARGPS_COUNTERFREQ\n" );
                        *OutBuffer    = (int)OsGetCounterFrequency();
                        ReturnedBytes = sizeof(long);
                        break;
                        
                        
                case IOCTL_PARGPS_READ_PPS_DATA:
                        ReturnedBytes = ParGps_ReadPpsData( Dev,
                                                (PPSDATA*)OutBuffer, Nout );
                        break;
                        
                case IOCTL_PARGPS_READ_SERIAL_DATA:
                        ReturnedBytes = ParGps_ReadSerialData( Dev,
                                                (SERIALDATA*)OutBuffer, Nout );
                        break;


                case IOCTL_PARGPS_THREAD_STATE:
                        dbg2_printk( "Requested IOCTL_PARGPS_THREAD_STATE\n" );
                        RequestState = *InBuffer;
                        dbg0_printk("Requested thread state %d\n",RequestState);
                        ParGps_ThreadControl( Dev, RequestState );
                        break;

                        
                default:
                    dbg1_printk( "Unknown IOCTL requested\n" );
                    ReturnedBytes = 0;
                    Status        = STATUS_INVALID_DEVICE_REQUEST;
                }


        
        // Release IOCTL one caller mutex

        OsMutexRelease( &Dev->IoctlMutex );


        // Return OutBuffer data to user space
        
        OsReturnIoctlBuffer( Irp, Command, OutBuffer, ReturnedBytes );


        
        OsIrpFinish( Irp, Status, ReturnedBytes );

        return( Status );
}

SRLOCAL void ParGps_DrvUnload( DEVINFO *Dev ) {

        PKFN_RELEASE ReleaseFn;

        dbg0_printk("In ParGps_DrvUnload\n");


        // Ensure that any A/D using the common kernel area has released it
        
        ReleaseFn = (PKFN_RELEASE)Dev->SrCommonArea->AtodFuncRelease;
        if ((void*)ReleaseFn != NULL) {
                ReleaseFn( (void*)Dev->SrCommonArea->AtodDev );
                }
        Dev->SrCommonArea = NULL;
        
}





SRLOCAL void ParGps_FillDevInfo( DEVINFO *Dev, char *DriverName ) {

        dbg1_printk("In ParGps_FillDevInfo\n");

        Dev->DevInfoTagStart = Dev->DevInfoTagEnd = 'RSRS';


        strncpy( Dev->DriverNameBuffer, DriverName, MAX_STR_LENGTH-1 );
        Dev->DriverNameBuffer[MAX_STR_LENGTH-1] = '\0';

        Dev->IntrSourceLast   = PARGPS_INTTYPE_BAD;
        Dev->IntrSourceNext   = PARGPS_INTTYPE_PLAIN;
        Dev->IntrCount        = 0;
        Dev->GpsDreadyCount   = 0;
        
        Dev->hSerial          = BAD_DEVHANDLE;
        Dev->SerialPortName   = NULL;
        
        Dev->ThreadTerminate  = 1;
        Dev->ThreadTask       = NULL;
        Dev->ThreadTimeOutCount = 0;

        OsAtomicSet( &Dev->NumberOpens, 0 );



        // WDM items
        
        Dev->DriverObj        = NULL;
        Dev->ThisFdoDevObj    = NULL;
        Dev->LowerDevObj      = NULL;
        Dev->PdoDevObj        = NULL;
        Dev->PrevDevState     = RUNSTATE_STOPPED;
        Dev->CurrDevState     = RUNSTATE_STOPPED;
        Dev->CurrDevPower     = PowerDeviceD0;
        Dev->CurrSysPower     = PowerSystemWorking;
//        Dev->DevCaps;
//        Dev->RemoveLock;
//        Dev->PowerEvent;
//        Dev->PowerStatus;
        

        OsMutexInit( &Dev->IoctlMutex, MUTEX_AVAILABLE );

        
        ParGps_FillCommonArea( Dev, 1 );
        ParGps_ClearDataIndex( Dev );
        ParGps_ClearCurrentTime( Dev );
}

SRLOCAL void ParGps_FillCommonArea( DEVINFO *Dev, int FillHeader ) {
        
        int   i, j;
        SR64BIT LiZero;
        
        dbg2_printk("In GpsFillCommonArea, SrCKA is 0x%p\n",&SrCommonKernelArea);

        
        // Initialize common area pointer

        LiZero.u.LowPart = LiZero.u.HighPart = 0L;

        Dev->SrCommonArea                           = &SrCommonKernelArea;

        if (FillHeader) {
                Dev->SrCommonArea->CkaRev           = CKA_REV;
                Dev->SrCommonArea->GpsDev           = (void*)Dev;
                Dev->SrCommonArea->AtodDev          = NULL;
                Dev->SrCommonArea->AtodFuncControl  = NULL;
                Dev->SrCommonArea->AtodFuncRelease  = NULL;
                Dev->SrCommonArea->AtodFuncResetGps = NULL;
                Dev->SrCommonArea->GpsFuncInterrupt = (PKFN_INTERRUPT)ParGps_Interrupt;
                }
        
        Dev->DataIndexCurr    = 0;
        Dev->DataIndexPrev    = MAXGPSDATA-1;
        Dev->DataIndexPrev2   = MAXGPSDATA-2;

        for ( i = 0 ; i < MAXGPSDATA ; i++ ) {
        
                Dev->PpsDataTemp[i].ValidFields    = VALID_NONE;
                Dev->PpsDataTemp[i].PpsEventNum    = 0;
                Dev->PpsDataTemp[i].CountAtPps     = LiZero;
                Dev->PpsDataTemp[i].CountAtDready  = LiZero;
                Dev->SerialDataTemp[i].ValidFields = VALID_NONE;
                Dev->SerialDataTemp[i].PpsEventNum = 0;
                Dev->SerialDataTemp[i].NmeaCount   = 0;
                for ( j = 0 ; j < MAX_NMEA_TYPE ; j++ )
                        Dev->SerialDataTemp[i].NmeaMsg[j*MAX_NMEA_SIZE] = '\0';

                }
}

SRLOCAL void ParGps_ClearDataIndex( DEVINFO *Dev ) {

        dbg0_printk("In ParGps_ClearDataIndex\n" );

        // Reset GPS data area

        Dev->PpsDataRdIndex    = 0;
        Dev->PpsDataWrIndex    = 0;
        Dev->SerialDataRdIndex = 0;
        Dev->SerialDataWrIndex = 0;
}

SRLOCAL void ParGps_ClearCurrentTime( DEVINFO *Dev ) {
        
        KIRQL   OldIrqlOrFlags;
        SR64BIT LiZero;
        int     i;

        LiZero.u.LowPart = LiZero.u.HighPart = 0L;

        OsSpinLockAcquire( &Dev->CurrentTimeLock, OldIrqlOrFlags );
        Dev->CurrentTimeData.ValidFields    = VALID_NONE;
        Dev->CurrentTimeData.EventNumP      = 0;
        Dev->CurrentTimeData.EventNumS      = 0;
        Dev->CurrentTimeData.CountAtPps     = LiZero;
        Dev->CurrentTimeData.CountAtNow     = LiZero;
        Dev->CurrentTimeData.NmeaCount      = 0;
        for ( i = 0 ; i < MAX_NMEA_TYPE ; i++ )
                Dev->CurrentTimeData.NmeaMsg[i*MAX_NMEA_SIZE] = '\0';
        OsSpinLockRelease( &Dev->CurrentTimeLock, OldIrqlOrFlags );
}

SRLOCAL void ParGps_Start( DEVINFO *Dev ) {

        dbg0_printk("In ParGps_Start\n" );

        // Reset PARGPS variables

        ParGps_Reset( Dev );


        // Start serial monitoring thread

        ParGps_ThreadControl( Dev, THREAD_STARTING );
}

SRLOCAL void ParGps_Stop( DEVINFO *Dev ) {

        dbg0_printk("In ParGps_Stop\n" );

        // Stop serial monitoring thread

        ParGps_ThreadControl( Dev, THREAD_STOPPING );


        // Reset PARGPS data area and current time.  We do this on stop so
        // old values will not be inadvertantly used by a second process
        // that just has the driver open to get current time.

        ParGps_ClearDataIndex( Dev );
        ParGps_ClearCurrentTime( Dev );
}

SRLOCAL void ParGps_Init( DEVINFO *Dev, int *IoBuffer ) {

        int Mode, Port;
        
        dbg0_printk("In ParGps_Init\n" );

// WCT: alternate way of passing parameters
//        INITPARGPSPARMS *Parm;
//        Parm = (PINITPARGPSPARMS)IOBuffer;
//        Mode = Parm->Mode;
//        Port = Parm->Port;

        Port = IoBuffer[0];
        Mode = IoBuffer[1];
        
        dbg0_printk("Mode is %d, Port is %d\n", Mode, Port );

        ParGps_SetInterruptMode( Dev, Mode );
        ParGps_SetSerialPort( Dev, Port );
}
SRLOCAL void ParGps_Reset( DEVINFO *Dev ) {

        PKFN_RESETGPS  ResetGpsFn;

        dbg0_printk("In ParGps_Reset\n" );

        // Reset GPS variables in this driver

        Dev->IntrCount      = 0;
        Dev->GpsDreadyCount = 0;
        ParGps_FillCommonArea( Dev, 0 );
        ParGps_ClearDataIndex( Dev );
        ParGps_ClearCurrentTime( Dev );


        // Reset GPS variables in A/D driver too

        if (Dev->SrCommonArea != NULL) {
                ResetGpsFn = (PKFN_RESETGPS)Dev->SrCommonArea->AtodFuncResetGps;
                if ((void*)ResetGpsFn != NULL) {
                        ResetGpsFn( (void*)Dev->SrCommonArea->AtodDev );
                        }
                }
}

SRLOCAL void ParGps_SetInterruptMode( DEVINFO *Dev, int Mode ) {

        dbg1_printk("In ParGps_SetInterruptMode\n");

        Dev->IntrSourceLast = PARGPS_INTTYPE_BAD;
        Dev->IntrSourceNext = Mode;
}
SRLOCAL void ParGps_SetSerialPort( DEVINFO *Dev, int Port ) {

        dbg1_printk("In ParGps_SetSerialPort\n");

        if (Port == 1)
                Dev->SerialPortName = KERNEL_COM1;
        else if (Port == 2)
                Dev->SerialPortName = KERNEL_COM2;
        else {
                dbg_printk("Invalid port number\n");
                Dev->SerialPortName = NULL;
                }
}
SRLOCAL void ParGps_SavePpsData( DEVINFO *Dev ) {

        KIRQL OldIrqlOrFlags;
        int i, k;

        dbg4_printk("In ParGps_SavePpsData\n");
        

        // Set up indexes for software FIFO and common area
        // (Note: no checks are done to prevent or identify overflow)
        
        i = Dev->PpsDataWrIndex;
        k = Dev->DataIndexCurr;

        // Don't save empty point
        
        if (Dev->PpsDataTemp[k].PpsEventNum == 0) {
                dbg0_printk("Ignoring PARGPS save at index %d\n", k);
                return;
                }



        // Save PPS data from common area
        
        Dev->PpsDataStore[i].ValidFields   = Dev->PpsDataTemp[k].ValidFields;
        Dev->PpsDataStore[i].PpsEventNum   = Dev->PpsDataTemp[k].PpsEventNum;
        Dev->PpsDataStore[i].CountAtPps    = Dev->PpsDataTemp[k].CountAtPps;
        Dev->PpsDataStore[i].CountAtDready = Dev->PpsDataTemp[k].CountAtDready;


        // Update the PPS part of the current time

        OsSpinLockAcquire( &Dev->CurrentTimeLock, OldIrqlOrFlags );
        Dev->CurrentTimeData.ValidFields = VALID_PPS;
        Dev->CurrentTimeData.EventNumP   = Dev->PpsDataStore[i].PpsEventNum;
        Dev->CurrentTimeData.CountAtPps  = Dev->PpsDataStore[i].CountAtPps;
        OsSpinLockRelease( &Dev->CurrentTimeLock, OldIrqlOrFlags );



        // Increment circular buffer PpsDataWrIndex

        Dev->PpsDataWrIndex++;
        if (Dev->PpsDataWrIndex >= MAX_PARGPS_STORAGE)
                Dev->PpsDataWrIndex = 0;

        dbg4_printk("End ParGps_SavePpsData\n");
}


SRLOCAL void ParGps_SaveSerialData( DEVINFO *Dev ) {

        KIRQL OldIrqlOrFlags;
        int i, j, k;

        dbg4_printk("In ParGps_SaveSerialData\n");
        

        // Set up indexes for software FIFO and common area
        // (Note: no checks are done to prevent or identify overflow)
        
        i = Dev->SerialDataWrIndex;
        k = Dev->DataIndexPrev2;



        // Don't save empty point
        
        if (Dev->SerialDataTemp[k].PpsEventNum == 0) {
                dbg0_printk("Ignoring serial save at index %d\n", k);
                return;
                }

        //dbg0_printk("SaveSerial indexes: wr %d, data %d\n", i, k);

        
        // Save serial data from common area
        
        Dev->SerialDataStore[i].ValidFields = Dev->SerialDataTemp[k].ValidFields;
        Dev->SerialDataStore[i].PpsEventNum = Dev->SerialDataTemp[k].PpsEventNum;
        Dev->SerialDataStore[i].NmeaCount   = Dev->SerialDataTemp[k].NmeaCount;
        for ( j = 0 ; j < MAX_NMEA_BUFF ; j++ )
                Dev->SerialDataStore[i].NmeaMsg[j] = Dev->SerialDataTemp[k].NmeaMsg[j];



        // Update the serial part of current time (PPS part already saved)

        // dbg0_printk("i %d: PPS count %d, Ser Count %d\n",
        // i,Dev->PpsDataStore[i].PpsEventNum,Dev->SerialDataStore[i].PpsEventNum);
        
        OsSpinLockAcquire( &Dev->CurrentTimeLock, OldIrqlOrFlags );
        Dev->CurrentTimeData.ValidFields |= VALID_NMEA;
        Dev->CurrentTimeData.EventNumS    = Dev->SerialDataStore[i].PpsEventNum;
        Dev->CurrentTimeData.NmeaCount    = Dev->SerialDataStore[i].NmeaCount;
        for ( j = 0 ; j < MAX_NMEA_BUFF ; j++ )
                Dev->CurrentTimeData.NmeaMsg[j] =
                        Dev->SerialDataStore[i].NmeaMsg[j];
        OsSpinLockRelease( &Dev->CurrentTimeLock, OldIrqlOrFlags );



        // Increment circular buffer SerialDataWrIndex

        Dev->SerialDataWrIndex++;
        if (Dev->SerialDataWrIndex >= MAX_PARGPS_STORAGE)
                Dev->SerialDataWrIndex = 0;


        dbg4_printk("End ParGps_SaveSerialData\n");
}


SRLOCAL unsigned long ParGps_ReadPpsData( DEVINFO *Dev, PPSDATA *OutBuffer, int Nout ) {

        int k, PpsReady, Nvals;
        unsigned long ReturnedBytes;

        dbg4_printk("ParGps_ReadPpsData\n");

        
        // Determine number of GPS data available in circular buffer
        
        PpsReady = Dev->PpsDataWrIndex - Dev->PpsDataRdIndex;
        if (PpsReady < 0)
                PpsReady += MAX_PARGPS_STORAGE;

        
        // Take minimum of number requested and number available

        Nvals = Nout / sizeof(PPSDATA);
        if (Nvals > PpsReady)
                Nvals = PpsReady;



        // Copy PPS data to user buffer and increment PpsDataRdIndex,
        // index of the next PPS item acquired but not yet read by the user

        for ( k = 0 ; k < Nvals ; k++ ) {

                // Copy PPS data to user buffer

                OsCopyToDirectIo( &OutBuffer[k], 
                                  &Dev->PpsDataStore[Dev->PpsDataRdIndex],
                                  sizeof(PPSDATA) 
                                  );


                // Clear out location now that it has been read
                
                Dev->PpsDataStore[Dev->PpsDataRdIndex].ValidFields = VALID_NONE;
                Dev->PpsDataStore[Dev->PpsDataRdIndex].PpsEventNum = 0;


                // Increment read index

                Dev->PpsDataRdIndex++;
                if (Dev->PpsDataRdIndex >= MAX_PARGPS_STORAGE)
                        Dev->PpsDataRdIndex = 0;
                }

        

        ReturnedBytes = Nvals * sizeof(PPSDATA);
        
        return( ReturnedBytes );
}



SRLOCAL unsigned long ParGps_ReadSerialData( DEVINFO *Dev, SERIALDATA *OutBuffer, int Nout ) {

        int k, SerialReady, Nvals;
        unsigned long ReturnedBytes;

        dbg4_printk("ParGps_ReadSerialData\n");


        // Determine number of Serial data available in circular buffer
        
        SerialReady = Dev->SerialDataWrIndex - Dev->SerialDataRdIndex;
        if (SerialReady < 0)
                SerialReady += MAX_PARGPS_STORAGE;

        
        // Take minimum of number requested and number available
        
        Nvals = Nout / sizeof(SERIALDATA);
        if (Nvals > SerialReady)
                Nvals = SerialReady;


        // Copy Serial data to user buffer and increment SerialDataRdIndex,
        // index of the next Serial item acquired but not yet read by the user

        for ( k = 0 ; k < Nvals ; k++ ) {

                // Copy Serial data to user buffer

                OsCopyToDirectIo( &OutBuffer[k], 
                                  &Dev->SerialDataStore[Dev->SerialDataRdIndex],
                                  sizeof(SERIALDATA) 
                                  );


                // Clear out location now that it has been read

                Dev->SerialDataStore[Dev->SerialDataRdIndex].ValidFields = VALID_NONE;
                Dev->SerialDataStore[Dev->SerialDataRdIndex].PpsEventNum = 0;


                // Increment read index

                Dev->SerialDataRdIndex++;
                if (Dev->SerialDataRdIndex >= MAX_PARGPS_STORAGE)
                        Dev->SerialDataRdIndex = 0;
                }

        

        ReturnedBytes = Nvals * sizeof(SERIALDATA);
        
        return( ReturnedBytes );
}


SRLOCAL unsigned long ParGps_GetCurrentTime( DEVINFO *Dev, GPSTIMEDATA *OutBuffer ) {

        KIRQL   OldIrqlOrFlags;
        SR64BIT tv;

        // Get current system time and save it.

        OsGetSystemTime( &tv );

        dbg4_printk("ParGps_GetCurrentTime START\n");
        
        OsSpinLockAcquire( &Dev->CurrentTimeLock, OldIrqlOrFlags );

        *OutBuffer = Dev->CurrentTimeData;
        OutBuffer->ValidFields |= VALID_DREADY;
        OutBuffer->CountAtNow   = tv;

        OsSpinLockRelease( &Dev->CurrentTimeLock, OldIrqlOrFlags );

        
        dbg4_printk("ParGps_GetCurrentTime END\n");
        return( sizeof(GPSTIMEDATA) );
}











// Serial Monitoring Thread & Helper Functions
//
//


#define ParGps_SerBufFlush() ( Dev->SbUsePtr = Dev->SbFillPtr = Dev->SbWrap = 0 )
#define ParGps_SerBufEmpty() ( (Dev->SbUsePtr == Dev->SbFillPtr && Dev->SbWrap <= 0) ? 1 : 0 )
#define ParGps_SerBufFull()  ( (Dev->SbUsePtr == Dev->SbFillPtr && Dev->SbWrap >  0) ? 1 : 0 )

#define ParGps_SerBufIncUsePtr( ) {                 \
        Dev->SbUsePtr++;                        \
        if (Dev->SbUsePtr >= MAX_NMEA_BUFF) {   \
                Dev->SbUsePtr = 0;              \
                Dev->SbWrap -= MAX_NMEA_BUFF;   \
                }                               \
        }                                        

#define ParGps_SerBufIncFillPtr( ) {                \
        Dev->SbFillPtr++;                       \
        if (Dev->SbFillPtr >= MAX_NMEA_BUFF) {  \
                Dev->SbFillPtr = 0;             \
                Dev->SbWrap += MAX_NMEA_BUFF;   \
                }                               \
        }                                       


SRLOCAL long ParGps_ThreadControl( DEVINFO *Dev, int RequestState ) {

        long Status;

        Status  = STATUS_SUCCESS;
        
        switch (RequestState) {
                
            case THREAD_STARTING:
                Dev->ThreadTerminate = 0;
                Status = OsThreadOpen( Dev );
                if (Status != STATUS_SUCCESS) {
                        dbg_printk("Failed to create system thread\n");
                        return( STATUS_CANCELLED );
                        }
                break;

            case THREAD_RUNNING:
                break;
                
            case THREAD_STOPPING:
                OsThreadClose( Dev );
                break;
                
            default:
                dbg_printk("Unknown thread state requested\n");
                
            } // end switch

        return( Status );
}



SRLOCAL unsigned long ParGps_ThreadFunc( DEVINFO *Dev ) {

        NTSTATUS Status;
        int      iev, nc;
        char     ErrMsg[MAX_STR_LENGTH];


        OsThreadInit( Dev );
        Dev->ThreadTimeOutCount = 0;

        
	// Open serial port (init done at user level)

        Status = OsSerialOpen( Dev );

        
        if ( !NT_SUCCESS( Status ) ) {
                dbg_printk("Failed to open serial port %s. Status = 0x%lX\n", Dev->SerialPortName, Status);
		OsThreadLeave( Dev, Status );
                return( (unsigned long)Status );
                }


        OsSerialFlush( Dev );
        ParGps_SerBufFlush();


        // Automatically start running

        // Loop while no termination request getting PPS events

        iev = 0;

        while (Dev->ThreadTerminate == 0) {

                Status = OsWaitOnPps( Dev, NULL, iev );

                if (Dev->ThreadTerminate == 1) // Check for termination request
                        break;

                
                // Display warning message if wait was not ended by the arrival
                // of a PPS signal, but then continue anyway.  This serial
                // serial monitoring thread needs to continue until it is
                // specifically asked to terminate since PPS signals may
                // start coming in again even after they'be been missing for
                // awhile.

                if ( Status == STATUS_SUCCESS ) {
                        // Print notice and flush serial when we first
                        // recover from timeouts so things start off right
                        // with the next PPS signal
                        if (Dev->ThreadTimeOutCount > 0) {
                                sprintf(ErrMsg,
                                        "Recovered after %d timeouts waiting for PPS\n",
                                        Dev->ThreadTimeOutCount);
                                OsMessage( Dev, ErrMsg, Status, PARGPS_LOG_WARN );
                                OsSerialFlush( Dev );
                                ParGps_SerBufFlush();
                                Dev->ThreadTimeOutCount = 0; // reset timeout count
                                }
                        }

                else if ( Status == STATUS_TIMEOUT ) {
                        // Only show timeout message once
                        if (Dev->ThreadTimeOutCount == 0) {
                                OsMessage(Dev, 
                                          "Timeout occurred waiting for PPS signal\n",
                                          Status, 
                                          PARGPS_LOG_WARN );
                                }
                        OsSerialFlush( Dev );
                        ParGps_SerBufFlush();
                        Dev->ThreadTimeOutCount++;
                        }


                else {
                         OsMessage( Dev, 
                                   "Failed waiting for PPS signal\n",
                                    Status, 
                                    PARGPS_LOG_WARN );
                        }


                if (Dev->ThreadTerminate == 1) // Check for termination request
                        break;


                // Fill internal circular buffer with everything
                // that is in the serial stream

                nc = ParGps_SerialFillBuffer( Dev );
                            

                if (Dev->ThreadTerminate == 1) // Check for termination request
                        break;


                
                // Process the NMEA messages from the internal buffer

                ParGps_SerialProcessData( Dev );

                iev++;

                } // end while loop



        // Close Serial Port and end thread

        OsSerialClose( Dev );
        
        dbg2_printk("Normal ThreadFunc termination\n");

        OsThreadLeave( Dev, STATUS_SUCCESS );

        // Under Linux, this value is never actually returned since
        // the thread itself has exited with the OsThreadLeave call
        // above.  Win2000 also doesn't ever return from one of the
        // system thread termination functions it calls in OsThreadLeave.

        return( (unsigned long)STATUS_SUCCESS );
}


#define NMEA_ERR_NO_DOLLAR      0
#define NMEA_ERR_NO_MSGID       1
#define NMEA_ERR_NO_SEQUENCE    2
#define NMEA_ERR_NO_DATA        3
#define NMEA_ERR_NO_CR          4
#define NMEA_ERR_NO_CHECKSUM    5
#define NMEA_ERR_NO_LF          6

SRLOCAL void ParGps_SerialProcessData( DEVINFO *Dev ) {

        int     inmea, istart, nc, kp, NmeaError, LastMsg;
        long    Status;


        dbg4_printk("ParGps_SerialProcessData\n");

        kp = Dev->DataIndexPrev;

        
        // Zero out previous strings

        Dev->SerialDataTemp[kp].NmeaCount = 0;
        for ( inmea = 0 ; inmea < MAX_NMEA_TYPE ; inmea++ ) {
                istart = inmea*MAX_NMEA_SIZE;
                Dev->SerialDataTemp[kp].NmeaMsg[istart] = '\0';
                }


        // Read new strings

        LastMsg = -1;
        for ( inmea = 0 ; inmea < MAX_NMEA_TYPE ; inmea++ ) {
                istart = inmea*MAX_NMEA_SIZE;
                Status = ParGps_SerialReadNmea(
                                Dev,
                                MAX_NMEA_SIZE,
                                &LastMsg,
                                &Dev->SerialDataTemp[kp].NmeaMsg[istart],
                                &nc,
                                &NmeaError
                                );

                
                // Error or end of messages
                
                if (Status != STATUS_SUCCESS) {
                        
                        // Note: We expect to get one NO_DOLLAR "error" each
                        // time after all NMEA messages have been read
                        
                        if (NmeaError != NMEA_ERR_NO_DOLLAR) {
                                if (NmeaError != NMEA_ERR_NO_SEQUENCE) {
                                        dbg_printk("ParGps_SerialReadNmea didn't get expected NMEA characters\n");
                                        dbg_printk("PpsEventNum %d, Status 0x%lX, ErrCode %d\n", Dev->SerialDataTemp[kp].PpsEventNum,Status,NmeaError);
                                        }
                                else {
                                        dbg0_printk("ParGps_SerialReadNmea found out of sequence\n");
                                        dbg0_printk("PpsEventNum %d, Status 0x%lX, ErrCode %d\n", Dev->SerialDataTemp[kp].PpsEventNum,Status,NmeaError);
                                        }
                                }
                        break;
                        }

                
                // No characters
                
                if (nc == 0)
                        break;

                }

        
         Dev->SerialDataTemp[kp].NmeaCount = inmea;
         Dev->SerialDataTemp[kp].ValidFields |= VALID_NMEA;

}
                        
SRLOCAL long ParGps_SerialReadNmea( DEVINFO *Dev,
                                int MaxSize,
                                int *LastMsg,
                                char *NmeaMsg,
                                int *Nchar,
                                int *NmeaError ) {

        int  i, j, icount, checksum, readsum, nc, saveuse, CurrMsg;
        char Token[2*MAX_NMEA_SIZE];
        long Status;

        dbg5_printk("ParGps_SerialReadNmea\n");


        // Initialize variables

        i        = 0;
        checksum = 0;
        *Nchar   = 0;
        
        
        // Discard characters until matching a $

        Status = ParGps_SerBufGetToken( Dev, Token, '$', NULL, &nc );
        icount = nc;

        
        if (Status != STATUS_SUCCESS) {
                // We'll get this at the end of each NMEA set
                // dbg3_printk("Failed character read while looking for $\n" );
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_DOLLAR;
                return( Status );
                }

        saveuse = Dev->SbUsePtr;                      // Save current location
        NmeaMsg[i++] = ParGps_SerBufGetNextChar( Dev );   // Get matched $


        // Get MSGID, adding to checksum, by matching until ,
        
        Status = ParGps_SerBufGetToken( Dev, Token, ',', &checksum, &nc );
        if (Status != STATUS_SUCCESS) {
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_MSGID;
                return( Status );
                }


        // Test for correct msgid sequence and size.  If fail, leave
        // those characters in the buffer by restoring the use pointer
        // and reset the last message index.
        
        CurrMsg = ParGps_SerBufGetMsgIndex( Token );
        if ( (CurrMsg < *LastMsg) || (nc != 5) ) {
                Dev->SbUsePtr = saveuse;                  // Restore saved location
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_SEQUENCE;
                dbg0_printk("Bad Sequence Token %s\n", Token);
                return( STATUS_NO_TOKEN );
                }
        *LastMsg = CurrMsg;


        for ( j = 0 ; j < nc ; j++ )            // Save MSGID
                NmeaMsg[i++] = Token[j];


        // Get msg data, adding to checksum, by matching to *

        Status = ParGps_SerBufGetToken( Dev, Token, '*', &checksum, &nc );
        if (Status != STATUS_SUCCESS) {
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_DATA;
                NmeaMsg[i] = '\0';
                dbg_printk("Message   %s\n",NmeaMsg);
                dbg_printk("Bad Token %s\n", Token);
                return( Status );
                }

        for ( j = 0 ; j < nc ; j++ )                 // Save msg data
                NmeaMsg[i++] = Token[j];

        NmeaMsg[i++] = ParGps_SerBufGetNextChar( Dev );  // Get matched *
        


        // Get checksum by matching until CR

        Status = ParGps_SerBufGetToken( Dev, Token, '\r', NULL, &nc );
        if (Status != STATUS_SUCCESS) {
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_CR;
                return( Status );
                }
        if ( nc != 2 ) {
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_CR;
                return( STATUS_NO_TOKEN );
                }

        for ( j = 0 ; j < nc ; j++ )                 // Save read check sum
                NmeaMsg[i++] = Token[j];

        NmeaMsg[i++] = ParGps_SerBufGetNextChar( Dev );  // Get matched CR



        // Test the read value of the checksum against the computed
        // value which is derived by taking the XOR of all characters
        // between but not including the first $ and the last *

        readsum = ParGps_SerAtoix( Token, 2 );
        if (checksum != readsum) {
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_CHECKSUM;
                return( STATUS_INVALID_PARAMETER );
                }


        // Match and get LF

        Status = ParGps_SerBufGetToken( Dev, Token, '\n', NULL, &nc );
        if (Status != STATUS_SUCCESS) {
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_LF;
                return( Status );
                }
        if ( nc != 0 ) {
                if (NmeaError)  *NmeaError = NMEA_ERR_NO_LF;
                return( STATUS_NO_TOKEN );
                }

        NmeaMsg[i++] = ParGps_SerBufGetNextChar( Dev );  // Get matched LF



        // Null terminate

        NmeaMsg[i] = '\0';
        *Nchar = i;
        

        dbg0_printk("(CHK 0x%X %d) %03d: %s", checksum, icount, i, NmeaMsg );

        return( STATUS_SUCCESS );
}


SRLOCAL long ParGps_SerialFillBuffer( DEVINFO *Dev ) {

        long Status;
        int  nc, nctotal, full;
        char c;

        dbg5_printk("ParGps_SerialFillBuffer\n");

        nctotal = 0;

        Status = ParGps_SerialReadChar( Dev, &c, &nc );

        while ( (Status == STATUS_SUCCESS) && (nc == 1) ) {

                nctotal += nc;
                
                full = ParGps_SerBufPutOne( Dev, c );
                if (full) {
                        dbg5_printk("ParGps_SerialFillBuffer buffer full.\n");
                        break;
                        }
                
                Status = ParGps_SerialReadChar( Dev, &c, &nc );
                }

        
        return( (long)nctotal );
}


SRLOCAL long ParGps_SerBufGetToken( DEVINFO *Dev,
                               char *Token,
                               char Sep,
                               int *chksum,
                               int *numchar ) {
        int  nc, exist;
        char c;

        dbg5_printk("ParGps_SerBufGetToken\n");

        // Move character from SerialData buffer to the Token buffer
        // up to but NOT including the designated Separator character.

        nc = 0;
        exist = ParGps_SerBufPeekOne( Dev, &c );

        while (exist && c != Sep) {
                if (chksum)
                        *chksum ^= c;
                Token[nc] = c;
                nc++;
                ParGps_SerBufIncUsePtr( );
                exist = ParGps_SerBufPeekOne( Dev, &c );
                }

        Token[nc] = '\0';
        if (numchar)
                *numchar = nc;

        dbg5_printk("ParGps_SerBufGetToken exist %d, nc %d\n", exist, nc);

        if (exist)
                return( STATUS_SUCCESS );
        else
                return( STATUS_PIPE_EMPTY );

}

SRLOCAL char ParGps_SerBufGetNextChar( DEVINFO *Dev ) {

        char c;

        c = ' ';

        // Peek at next character, if it exists then change
        // the peek to a get by incrementing the use pointer
        
        if ( ParGps_SerBufPeekOne( Dev, &c ) )
                ParGps_SerBufIncUsePtr( );

        return( c );
}

SRLOCAL int ParGps_SerBufPeekOne( DEVINFO *Dev, char *c ) {

        // Check for data

        if (ParGps_SerBufEmpty())
                return( 0 );


        // Get next character

        *c = Dev->SerialBuffer[Dev->SbUsePtr];
        return( 1 );
}

SRLOCAL int ParGps_SerBufPutOne( DEVINFO *Dev, char c ) {

        // Save character into next free position in buffer

        Dev->SerialBuffer[Dev->SbFillPtr] = c;
        ParGps_SerBufIncFillPtr( );


        // Check for fullness

        return( ParGps_SerBufFull() );
}
                
SRLOCAL long ParGps_SerialReadChar( DEVINFO *Dev, char *C, int *Nchar  ) {

        long Status;

        
        // Read 1 character

        Status = STATUS_SUCCESS;
        
        *Nchar = OsSerialReadChar( Dev, C, 1L, &Status );

        if (*Nchar < 0)
                Status = -(*Nchar);

        if (*Nchar == 0)
                Status = STATUS_PIPE_EMPTY;
        
        return( Status );
}

char *NmeaMsgIdList[] = { "GPZDA",
                          "GPGGA",
                          "GPGLL",
                          "GPVTG",
                          "GPGSA",
                          "GPGSV",
                          "GPRMC",
                          NULL
                         };

SRLOCAL int ParGps_SerBufGetMsgIndex( char *Token ) {

        char **m;
        int i, j, match;

        m = NmeaMsgIdList;
        i = 0;

        while (*m) {
                match = 1;
                for ( j = 0 ; j < 5 ; j++ ) {
                        
                        if ( *m[i] != Token[i] ) {
                                match = 0;
                                break;
                                }
                        }
                
                if (match)
                        return( i );
                m++;
                i++;
                }

        return( -1 );
}

SRLOCAL int ParGps_SerAtoix( char *src, int n ) {

        char *hexdigits, *str, c;
        int   i, result, digit;
        
        hexdigits = "0123456789ABCDEF";
        digit     = -1;
        result    = 0;
        str       = src;

        while ( n-- ) {
                
                c = *str++;
                
                for ( i = 0 ; i < 16 ; i++ ) {
                        if (hexdigits[i] == c) {
                                digit = i;
                                break;
                                }
                        }
                
                result = (result * 0x10) + digit;
                }
        
        return( result );
}





// Interrupt and Deferred Procedure Call (bottom half) functions
//
//
//

int ParGps_Interrupt( DEVINFO *Dev ) {

        SR64BIT tv;
        int     k;
        DEVICE_OBJECT *DeviceObject;


        // SR64BIT =
        //   { DWORD LowPart, LONG HighPart } or  { LONGLONG QuadPart }
        //     unsigned long, long            or    hyper

        
        // Get system time

        OsGetSystemTime( &tv );

        
        dbg5_printk( "ParGps_Interrupt\n" );


        DeviceObject = (DEVICE_OBJECT *)Dev->ThisFdoDevObj;


        
        // Process PPS interrupt
        //
        //

        if (Dev->IntrSourceNext == PARGPS_INTTYPE_PPS ) {

                // A new PPS has come in, so update circular index
                
                Dev->DataIndexCurr++;
                Dev->DataIndexPrev++;
                Dev->DataIndexPrev2++;
                Dev->DataIndexCurr  &= (MAXGPSDATA-1);
                Dev->DataIndexPrev  &= (MAXGPSDATA-1);
                Dev->DataIndexPrev2 &= (MAXGPSDATA-1);

                k = Dev->DataIndexCurr;
                Dev->PpsDataTemp[k].ValidFields    = VALID_NONE;
                Dev->SerialDataTemp[k].ValidFields = VALID_NONE;
                
                
                // Save PPS time info
                
                Dev->PpsDataTemp[k].ValidFields |= VALID_PPS;
                Dev->PpsDataTemp[k].CountAtPps = tv;

        
                // Error check alternation and toggle interrupt type 
                
                if (Dev->IntrSourceLast == Dev->IntrSourceNext )
                        Dev->IntrSourceNext = PARGPS_INTTYPE_BAD;
                else {
                        Dev->IntrSourceLast = Dev->IntrSourceNext;
                        Dev->IntrSourceNext = PARGPS_INTTYPE_DREADY;
                        }

                Dev->IntrCount++;


                
                // Call DPC
                //   Signals serial monitoring event
                //   Saves previous Serial data structure

                OsQueueDPC( Dev, PARGPS_INTTYPE_PPS, DeviceObject );
                }



        
        // Process DREADY interrupt
        //
        //
        
        else if (Dev->IntrSourceNext == PARGPS_INTTYPE_DREADY ) {

                // Save Dready time and count
                
                k = Dev->DataIndexCurr;
                Dev->GpsDreadyCount++;
                Dev->PpsDataTemp[k].ValidFields |= VALID_DREADY;
                Dev->PpsDataTemp[k].CountAtDready = tv;
                Dev->PpsDataTemp[k].PpsEventNum = Dev->GpsDreadyCount;
                Dev->SerialDataTemp[k].PpsEventNum = Dev->GpsDreadyCount;


                // Don't need to call ControlFn to request the PARxCH pass
                // on a PPS interrupt from the PARGPS board since the PARxCH
                // automatically switches back to PPS after one DREADY
                
                
                // Error check alternation and toggle interrupt type 
                
                if (Dev->IntrSourceLast == Dev->IntrSourceNext )
                        Dev->IntrSourceNext = PARGPS_INTTYPE_BAD;
                else {
                        Dev->IntrSourceLast = Dev->IntrSourceNext;
                        Dev->IntrSourceNext = PARGPS_INTTYPE_PPS;
                        }
                
                Dev->IntrCount++;



                // Call DPC
                //   Saves PPS data
                
                OsQueueDPC( Dev, PARGPS_INTTYPE_DREADY, DeviceObject );
                }



        
        // Process Plain interrupt
        //
        //
        
        else { // Dev->IntrSourceNext == PARGPS_INTTYPE_PLAIN or unknown

                // Save Plain PPS time
                
                k = Dev->DataIndexCurr;
                Dev->PpsDataTemp[k].ValidFields |= VALID_PPS;
                Dev->PpsDataTemp[k].CountAtPps = tv;

                
                // Increment count
                
                Dev->IntrSourceLast = Dev->IntrSourceNext;
                Dev->IntrCount++;


                OsQueueDPC( Dev, PARGPS_INTTYPE_PLAIN, DeviceObject );
                }
        
    
        return( 1 );
}

void ParGps_DpcForPpsRoutine( DEVINFO *Dev ) {

        long         PrevState;
        PKFN_CONTROL ControlFn;

        dbg5_printk( "ParGps_DpcForPpsRoutine\n" );
        
        // Process PPS DPC


        // Error check for alternating types

        if (Dev->IntrSourceNext == PARGPS_INTTYPE_BAD) {
                dbg_printk("Two %d interrupts in a row !!\n", Dev->IntrSourceLast);
                }


        
        // Request PARxCH send DREADY interrupt

        ControlFn = (PKFN_CONTROL)Dev->SrCommonArea->AtodFuncControl;

        if ((void*)ControlFn != NULL)
                ControlFn( (void*)Dev->SrCommonArea->AtodDev );
        else {
                dbg_printk("OsMessage with Null ControlFn\n");
                OsMessage( Dev, "Null ControlFn", STATUS_SUCCESS, PARGPS_LOG_ERR );
                }



        // Save Serial data from 2nd previous PPS
        
        ParGps_SaveSerialData( Dev );



        // Signal serial monitoring event to read data from 1st previous PPS

        PrevState = OsSerialSignal( Dev );
        if (PrevState != 0) {
                dbg_printk("In DPC, Event is still set from last PPS time %d\n",Dev->IntrCount);
                }
}


void ParGps_DpcForDreadyRoutine( DEVINFO *Dev ) {

        // Process DREADY DPC

        dbg5_printk( "ParGps_DpcForDreadyRoutine\n" );
        
        
        // Error check for alternating types

        if (Dev->IntrSourceNext == PARGPS_INTTYPE_BAD) {
                dbg_printk("Two %d interrupts in a row !!\n", Dev->IntrSourceLast);
                }

        
        
        // Save GPS data from current PPS

        ParGps_SavePpsData( Dev );
}

void ParGps_DpcForPlainRoutine( DEVINFO *Dev ) {
        
        int  k;
        long PrevState;

        dbg5_printk( "ParGps_DpcForPlainRoutine\n" );
        

        // Process Plain DPC
        
        k = Dev->DataIndexCurr;
        Dev->PpsDataTemp[k].PpsEventNum = Dev->IntrCount;


        
        // Error check interrupt source

        if (Dev->IntrSourceNext == PARGPS_INTTYPE_BAD) {
                dbg_printk("Bad interrupt, last was %d!!\n", Dev->IntrSourceLast);
                }



        // Save GPS data from current PPS

        ParGps_SavePpsData( Dev );



        
        // Signal serial monitoring event

        PrevState = OsSerialSignal( Dev );
        if (PrevState != 0) {
                dbg_printk("In DPC, Event is still set from last PPS time %d\n",Dev->IntrCount);
                }

}








////////////////////////////////////////////////////////////////////////
/* Group 2: OS SPECIFIC HELPER FUNCTIONS AND MACROS:
 *
 * These functions are repeated several times, so you MUST give a
 * valid SROS_xxx compile time define to select which set of these
 * functions is appropriate for your OS.
 *
 */


// *************************************************************************

#if defined( SROS_WIN98 ) || defined( SROS_WIN2K ) || defined( SROS_WINNT )

SRLOCAL void OsMessage( DEVINFO *Dev, char *Msg, long Status, int Level ) {

        int  i;
        char *m;

        i = 0;
        m = Msg;  

        while ( (*m != '\0') && (i < MAX_STR_LENGTH - 2) ) {
                swprintf( &Dev->MsgBuffer[i], L"%c", *m++ );
                i++;
                }
        swprintf( &Dev->MsgBuffer[i], L"\0" );

                
        dbg_printk( "Msg is %ls", Dev->MsgBuffer );

        WdmWinnt_ErrorLog( Dev->ThisFdoDevObj,
                           NULL,
                           (NTSTATUS)Status,
                           Level,
                           Dev->MsgBuffer );
        
        if (Level == PARGPS_LOG_ERR) {
                // Die ?
                }
        else if (Level == PARGPS_LOG_WARN) {
                // Continue ?
                }
        else {
                // Continue ?
                }
                

        return;
}
SRLOCAL unsigned long OsCopyToUser( void *UserBuff, const void *KernelBuff, unsigned long Nbytes ) {
        RtlCopyMemory( UserBuff, KernelBuff, Nbytes );
        return( Nbytes );
}
SRLOCAL unsigned long CopyFromUser( void *KernelBuff, const void *UserBuff, unsigned long Nbytes ) {
        RtlCopyMemory( KernelBuff, UserBuff, Nbytes );
        return( Nbytes );
}
SRLOCAL unsigned long OsCopyToDirectIo( void *MidlBuff, const void *KernelBuff, unsigned long Nbytes ) {
        unsigned long i;
        unsigned char *midlptr;
        unsigned char *kernelptr;
        
        midlptr   = (unsigned char *)MidlBuff;
        kernelptr = (unsigned char *)KernelBuff;
        
        for ( i = 0 ; i < Nbytes ; i++ )
                *midlptr++ = *kernelptr++;
        
        return( Nbytes );
}

SRLOCAL void OsIrpFinish( IRP *Irp, long Status, unsigned long Info ) {
        dbg3_printk("In OsIrpFinish info %lu (0x%X), status 0x%X\n", Info, Info, Status );
        
        Irp->IoStatus.Status      = (NTSTATUS)Status;
        Irp->IoStatus.Information = Info;
        IoCompleteRequest( Irp, IO_NO_INCREMENT );
}

SRLOCAL void OsSetIoctlCommand( IRP *Irp, unsigned long *Command ) {
        
        IO_STACK_LOCATION *IrpStack;
        
        IrpStack  = IoGetCurrentIrpStackLocation( Irp );
        *Command  = IrpStack->Parameters.DeviceIoControl.IoControlCode;
}

SRLOCAL long OsSetIoctlBuffers( IRP *Irp,
                                unsigned long Command,
                                int **InBuffer,
                                int **OutBuffer,
                                int *Nin,
                                int *Nout ) {

        IO_STACK_LOCATION *IrpStack;

        IrpStack  = IoGetCurrentIrpStackLocation( Irp );


        // Assign buffer sizes

        *Nin  = IrpStack->Parameters.DeviceIoControl.InputBufferLength;
        *Nout = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;


        // Assign buffers

        *InBuffer  = (int *)Irp->AssociatedIrp.SystemBuffer;
        *OutBuffer = (int *)Irp->AssociatedIrp.SystemBuffer;

        if ((IOCTL_PARGPS_METH(Command) == METHOD_IN_DIRECT) ||
            (IOCTL_PARGPS_METH(Command) == METHOD_OUT_DIRECT)) {

                if (Irp->MdlAddress)
#if defined( SROS_WIN2K )
                        *OutBuffer = (int*)MmGetSystemAddressForMdlSafe(Irp->MdlAddress,NormalPagePriority);
#else // defined( SROS_WIN98 ) || defined( SROS_WINNT )
                        *OutBuffer = (int*)MmGetSystemAddressForMdl(Irp->MdlAddress);
#endif
                else
                        *OutBuffer = NULL;
                }


        // Check that buffer exists since all commands pass (or get) data
        
        if ( !(*InBuffer) && !(*OutBuffer) )
                return( STATUS_INVALID_USER_BUFFER );


        // No filling needed for InBuffer since OS functions do that
                    
        return( STATUS_SUCCESS );
}

SRLOCAL void OsReturnIoctlBuffer( IRP *Irp,
                                  unsigned long Command,
                                  int *OutBuffer,
                                  unsigned long ReturnedBytes ) {
        
        // No filling needed for OutBuffer since OS functions do that

        return;
}

SRLOCAL long OsSerialOpen( DEVINFO *Dev ) {

        NTSTATUS          Status;
        OBJECT_ATTRIBUTES ObjAttr;
        UNICODE_STRING    ObjName;
        IO_STATUS_BLOCK   IoStatusBlock;
        ULONG             CreateOpts;
        HANDLE            SerialHandle;

        dbg2_printk("Calling OsSerialOpen\n");

        // Check if already open

        if ( Dev->hSerial != BAD_DEVHANDLE )
                return( STATUS_TOKEN_ALREADY_IN_USE );


        // Prepare serial port name and attributes

        swprintf( Dev->MsgBuffer, L"%S", Dev->SerialPortName );
        RtlInitUnicodeString( &ObjName, Dev->MsgBuffer );

        InitializeObjectAttributes(
                                   &ObjAttr,      // OUT POBJECT_ATTRIBUTES  InitializedAttributes,
                                   &ObjName,      // IN PUNICODE_STRING  ObjectName,
                                   OBJ_INHERIT,   // or OBJ_KERNEL_HANDLE
                                   NULL,          // IN HANDLE  RootDirectory,
                                   NULL           // IN PSECURITY_DESCRIPTOR  SecurityDescriptor
                                  );

        CreateOpts = FILE_SYNCHRONOUS_IO_ALERT | FILE_NON_DIRECTORY_FILE;


        // Open serial port

        Status = ZwCreateFile(
                              &SerialHandle,
                              GENERIC_READ | SYNCHRONIZE,
                              &ObjAttr,
                              &IoStatusBlock,
                              NULL,                 // PLARGE_INTEGER AllocationSize
                              FILE_ATTRIBUTE_NORMAL,
                              FILE_SHARE_READ,      // or 0 = no share
                              FILE_OPEN,            // fail if not exist
                              CreateOpts,           // create options?
                              NULL,
                              0
                             );

        if (Status == STATUS_SUCCESS)
                Dev->hSerial = (DEVHANDLE)SerialHandle;

        dbg2_printk("Leaving OsSerialOpen with hSerial 0x%p\n",Dev->hSerial);
        return( Status );
}

SRLOCAL long OsSerialClose( DEVINFO *Dev ) {

        NTSTATUS Status;

        dbg2_printk("Calling OsSerialClose\n");
        
        // Early exit if already closed

        if ( Dev->hSerial == BAD_DEVHANDLE )
                return( STATUS_SUCCESS );


        // Close serial handle

        Status = ZwClose( Dev->hSerial );
        Dev->hSerial = BAD_DEVHANDLE;
        dbg2_printk("OsSerialClose ZwClose close status 0x%lX\n", Status);

        return( STATUS_SUCCESS );
}
SRLOCAL long OsSerialFlush( DEVINFO *Dev ) {

        long Status;
        char c;
        int  nchar;

        dbg2_printk("Calling OsSerialFlush\n");
        
        Status = STATUS_SUCCESS;

        while (Status == STATUS_SUCCESS) {
                Status = ParGps_SerialReadChar( Dev, &c, &nchar );
                }

        return( Status );
}

SRLOCAL int OsSerialReadChar( DEVINFO *Dev,
                              char *Buffer,
                              unsigned long BuffLen,
                              unsigned long *ErrorCode ) {
        int             Nchar;
        NTSTATUS        Status;
        LARGE_INTEGER   ByteOffset;
        IO_STATUS_BLOCK IoStatusBlock;

        dbg5_printk("Calling OsSerialReadChar\n");
        
        ByteOffset.u.LowPart  = 0; // request current pos
        ByteOffset.u.HighPart = 0;


        Status = ZwReadFile(
                            Dev->hSerial,
                            NULL,            // Event
                            NULL,            // ApcRoutine
                            NULL,            // ApcContext
                            &IoStatusBlock,  
                            Buffer,          // data area
                            BuffLen,         // length
                            &ByteOffset,     // read from current pos
                            NULL
                           );

        if ( !NT_SUCCESS(Status)) {
                dbg_printk("OsSerialReadChar failed ZwRead with Status = 0x%X, IOB.status = 0x%X\n",
                           Status, IoStatusBlock.Status);
        }

        if (ErrorCode)
                *ErrorCode = Status;

        Nchar = IoStatusBlock.Information;

        return( Nchar );
}

SRLOCAL long OsSerialSignal( DEVINFO *Dev ) {
        
        long PrevState;
        
        PrevState = KeSetEvent( &Dev->PpsEvent, // IN PKEVENT Event,
                                0,              // IN KPRIORITY Increment,
                                FALSE           // IN don't wait
                              );
        return( PrevState );
}


SRLOCAL long OsThreadOpen( DEVINFO *Dev ) {

        long      Status;
        DEVHANDLE ThreadHandle;

        dbg2_printk("Calling OsThreadOpen\n");

        // Check to be sure it isn't already open

        if (Dev->ThreadTask) {
                dbg_printk( "Thread already open\n" );
                return( STATUS_TOKEN_ALREADY_IN_USE );
                }

        // Create serial monitoring thread

        Status = PsCreateSystemThread(
                                      &ThreadHandle,
                                      THREAD_ALL_ACCESS, // DesiredAccess
                                      NULL,              // POBJECT_ATTRIBUTES
                                      (HANDLE)0L,        // run in system context, ProcessHandle
                                      NULL,              // ClientId
                                      WdmWinnt_ThreadFunc,  // PKSTART_ROUTINE
                                      Dev                // Context
                                     );
        if ( !NT_SUCCESS( Status ) ) {
                dbg_printk( "Failed to create serial monitoring thread\n" );
                return( Status );
                }



        // Get a reference to the thread so we can wait on its demise

        Status = ObReferenceObjectByHandle(
                                          ThreadHandle,       // Handle
                                          THREAD_ALL_ACCESS,  // Desired access
                                          NULL,               // Obj type
                                          KernelMode,         // Processor mode
                                          &Dev->ThreadTask,   // Object
                                          NULL                // Obj handle info
                                        );
        
        if ( !NT_SUCCESS( Status ) ) {
                dbg_printk( "Failed to get object handle for serial thread\n" );
                // Send a wakeup to force termination
                Dev->ThreadTerminate = 1;
                KeSetEvent( &Dev->PpsEvent, // IN PKEVENT Event,
                            0,              // IN KPRIORITY Increment,
                            FALSE           // IN don't wait
                          );
                return( Status );
                }


        // We don't need to save the original handle now that we have the ref.

        ZwClose( ThreadHandle );


        dbg2_printk("Leaving OsThreadOpen\n");
        return( Status );
}
SRLOCAL long OsThreadClose( DEVINFO *Dev ) {

        NTSTATUS      Status;

        dbg2_printk("Calling OsThreadClose\n");
        
        // Early exit if thread already closed

        if (Dev->ThreadTask == NULL) {
                dbg_printk( "Thread is already closed\n" );
                return( STATUS_CANNOT_DELETE );
                }


        
        // Request the thread to kill itself

        Dev->ThreadTerminate = 1;


        // Force a wake up by sending a pretend PPS event
        
        KeSetEvent( &Dev->PpsEvent, // IN PKEVENT Event,
                    0,              // IN KPRIORITY Increment,
                    FALSE           // IN don't wait
                  );


        // Wait on thread object itself for the thread to die

        Status = KeWaitForSingleObject(
                                       Dev->ThreadTask,
                                       Executive,     // WaitReason
                                       KernelMode,    // WaitMode
                                       FALSE,         // not Alertable
                                       NULL           // wait forever
                                      );        


        // Clean up reference to the thread object now

        ObDereferenceObject( Dev->ThreadTask );
        Dev->ThreadTask = NULL;



        // This is here just in case, it should have been done in the thread

        if (Dev->hSerial != BAD_DEVHANDLE) {
                dbg_printk("Calling serial close from thread close\n");
                OsSerialClose( Dev );
                }

        dbg2_printk("Leaving OsThreadClose\n");
        
        return( STATUS_SUCCESS );
}

SRLOCAL void OsThreadInit( DEVINFO *Dev ) {

        // This function called from inside the thread when it starts

        KPRIORITY oldPriority, DesiredPriority;

        dbg2_printk("Calling OsThreadInit\n");
        
        // Set thread priority to lowest realtime level.

        DesiredPriority = LOW_REALTIME_PRIORITY;
        oldPriority = KeSetPriorityThread( KeGetCurrentThread(),
                                           LOW_REALTIME_PRIORITY );
        dbg1_printk("Thread old priority was %d, new is %d\n",
                    oldPriority, LOW_REALTIME_PRIORITY );

}
SRLOCAL void OsThreadLeave( DEVINFO *Dev, unsigned long Status ) {

        // This function called from inside the thread when it exits

        dbg2_printk("Calling OsThreadLeave\n");
        
        Dev->ThreadTerminate = 1;

        PsTerminateSystemThread( Status );  // This call never really returns
}

SRLOCAL void OsGetSystemTime( SR64BIT *SysTime ) {

        LARGE_INTEGER PerfCounter;

        // The following two are only accurate to 10 ms
        // KeQuerySystemTime( &tv );
        // KeQueryTickCount( &tv );

#ifdef USE_RDTSC_FOR_TIME
        
        // Do the rdtsc instruction
        __asm {
                 push eax      ; Save registers we will overwrite (eax, ebx, edx).
                 push ebx
                 push edx
                 _emit 0x0F    ; The RDTSC instruction consists of these two bytes.
                 _emit 0x31
                 mov ebx, SysTime ; Put the address where the timestamp goes in ebx.
                 mov [ebx],eax   ; Save low 4 bytes of timestamp there.
                 mov [ebx+4],edx ; Save high 4 bytes of timestamp next.
                 pop edx         ; Restore overwritten registers.
                 pop ebx
                 pop eax
                 }
#else
        PerfCounter = KeQueryPerformanceCounter( NULL );
        SysTime->QuadPart = PerfCounter.QuadPart;
        
#endif // USE_RDTSC_FOR_TIME

}

SRLOCAL long OsGetCounterFrequency( void ) {
        
        LARGE_INTEGER PerfCounter, PrefFreq;
        
        PerfCounter = KeQueryPerformanceCounter( &PrefFreq );

        if (PrefFreq.HighPart != 0x00000000) {
                dbg_printk("Performance counter frequency exceeds 32 bits!\n");
                return( 0L );
                }

        return( PrefFreq.LowPart );
}

SRLOCAL unsigned long OsWaitOnPps( DEVINFO *Dev,
                                  SR64BIT *RequestedTimeout,
                                  int Id ) {

        NTSTATUS      Status;
        LARGE_INTEGER Timeout;

        // Set timeout by request or default
        // - means relative, + means absolute
        
        if (RequestedTimeout)
                Timeout.QuadPart = RequestedTimeout->QuadPart;
        else
                Timeout.QuadPart = -(2 * NTCOUNT_TO_SEC);




        // Wait for PPS event to signal
        
        Status = KeWaitForSingleObject(
                                       &Dev->PpsEvent,
                                       Executive,     // WaitReason
                                       KernelMode,    // WaitMode
                                       FALSE,         // not Alertable
                                       &Timeout       // PLARGE_INTEGER Timeout
                                      );        

        if (Status == STATUS_SUCCESS) {
                dbg5_printk("Received PPS event %d in thread\n", Id);
                KeClearEvent( &Dev->PpsEvent );
                }
        else if (Status == STATUS_TIMEOUT) {
                dbg1_printk("Timed out on PPS event %d in thread\n", Id);
                }
        else {
                dbg1_printk("Status of 0x%X on PPS event %d in thread\n", Status, Id);
                }

        return( (unsigned long)Status );
}








// *************************************************************************

#elif defined( SROS_LINUX )

SRLOCAL void OsMessage( DEVINFO *Dev, char *Msg, long Status, int Level ) {

        dbg_printk( "%s (Status 0x%lX)\n", Msg, Status );
        
        if (Level == PARGPS_LOG_ERR) {
                // Die ?
                }
        else if (Level == PARGPS_LOG_WARN) {
                // Continue ?
                }
        else {
                // Continue ?
                }
                

        return;
}
SRLOCAL unsigned long OsCopyToUser( void *UserBuff, const void *KernelBuff, unsigned long Nbytes ) {
        unsigned long result;
        
        result = __copy_to_user( UserBuff, KernelBuff, Nbytes );
        
        return( result );
}        
SRLOCAL unsigned long OsCopyFromUser( void *KernelBuff, const void *UserBuff, unsigned long Nbytes ) {
        unsigned long result;
        
        result = __copy_from_user( KernelBuff, UserBuff, Nbytes );
        
        return( result );
}        
// This is identical to OsCopyToUser since Linux has no Midl
SRLOCAL unsigned long OsCopyToDirectIo( void *MidlBuff, const void *KernelBuff, unsigned long Nbytes ) {
        unsigned long result;
	
        result = __copy_to_user( MidlBuff, KernelBuff, Nbytes );
        
        return( result );
}        

SRLOCAL void OsIrpFinish( IRP *Irp, long Status, unsigned long Info ) {
        if (Irp)
                Irp->ReturnedBytes = Info;
}

SRLOCAL void OsSetIoctlCommand( IRP *Irp, unsigned long *Command ) {
        *Command  = Irp->Command;
}

SRLOCAL long OsSetIoctlBuffers( IRP *Irp,
                                unsigned long Command,
                                int **InBuffer,
                                int **OutBuffer,
                                int *Nin,
                                int *Nout ) {

        // BufferArea is defined at maximum size needed for any command

        static unsigned char BufferArea[MAX_PARGPS_STORAGE*sizeof(SERIALDATA)];
        int ReadCopy, WriteCopy, Direct, MaxBuffArea, Err;
        

        /* _IOC type is user-oriented, verify_area is kernel-oriented
         * so read/write are reversed.  VERIFY_WRITE handles R/W.
         * eg _IOC_READ means read from the device, so driver must write
         * to user space.
         */
        
        // Check if driver must write to user space
        
        if ( _IOC_DIR(Command) & _IOC_READ )
                WriteCopy = 1;
        else
                WriteCopy = 0;

        
        // Check if driver must read from user space
        
        if ( _IOC_DIR(Command) & _IOC_WRITE )
                ReadCopy  = 1;
        else
                ReadCopy  = 0;



        // Verify that IoctlIrp structure is valid for read/write

        Err = access_ok( VERIFY_WRITE,
                         (void*)Irp,
                         _IOC_SIZE(Command) );
        if (!Err)
                return( STATUS_INVALID_USER_BUFFER );



        // Unpack IoctlIrp structure

        if ( ( Irp->DataMethod == METHOD_IN_DIRECT ) ||
             ( Irp->DataMethod == METHOD_OUT_DIRECT ) )
                Direct  = 1;
        else
                Direct  = 0;


        *Nin  = (int)Irp->InSize;
        *Nout = (int)Irp->OutSize;


        if (Direct) {
                *InBuffer  = (int*)Irp->InBuffer;
                *OutBuffer = (int*)Irp->OutBuffer;
                }
        else {
                MaxBuffArea = sizeof(BufferArea);
                if ( (*Nin > MaxBuffArea) || (*Nout > MaxBuffArea) )
                        return( STATUS_SECTION_NOT_EXTENDED );
                *InBuffer = *OutBuffer = (int *)BufferArea;
                }

        
        // Check input buffer is ok to read

        if ( ReadCopy ) {
                Err = access_ok( VERIFY_READ,
                                 (void*)Irp->InBuffer,
                                 *Nin );
                if (!Err)
                        return( STATUS_INVALID_USER_BUFFER );
                }


        // Check output buffer is ok to read/write

        if ( WriteCopy ) {
                Err = access_ok( VERIFY_WRITE,
                                 (void*)Irp->OutBuffer,
                                 *Nout );
                if (!Err)
                        return( STATUS_INVALID_USER_BUFFER );
                }


        // Fill Ioctl buffer from user space if not using direct IO

        if (ReadCopy && !Direct)
                OsCopyFromUser( *InBuffer, Irp->InBuffer, *Nin );


        return( STATUS_SUCCESS );
}

SRLOCAL void OsReturnIoctlBuffer( IRP *Irp,
                                  unsigned long Command,
                                  int *OutBuffer,
                                  unsigned long ReturnedBytes ) {


        // Save number of bytes being returned

        Irp->ReturnedBytes = ReturnedBytes;


        // Copy has been done else where for this command

        if ((Irp->DataMethod == METHOD_IN_DIRECT) ||
            (Irp->DataMethod == METHOD_OUT_DIRECT) ) {
                return;
                }


        // Check if driver must write to user space (by READing from kernel)

        if ( _IOC_DIR(Command) & _IOC_READ )
                OsCopyToUser( Irp->OutBuffer, OutBuffer, Irp->OutSize );

}

SRLOCAL long OsSerialOpen( DEVINFO *Dev ) {

        long Status;
        char *tmp;
        int fd;
        struct file *file;
        mm_segment_t fs, ds;

        dbg2_printk("Calling OsSerialOpen\n");


        // Check if already open

        if ( Dev->hSerial != BAD_DEVHANDLE )
                return( STATUS_TOKEN_ALREADY_IN_USE );



        fs = get_fs();
        ds = get_ds();
        set_fs( ds );

        tmp = getname( Dev->SerialPortName );
        fd  = PTR_ERR(tmp);
        if (!IS_ERR(tmp)) {
                lock_kernel();
                fd = get_unused_fd();
                if (fd >= 0) {
                        file = filp_open( tmp, O_RDONLY | O_NONBLOCK, 0444 );
                        if (!IS_ERR(file)) {
                                fd_install( fd, file );
                                }
                        else {
//                                put_unused_fd(fd);  // unknown symbol
                                fd = PTR_ERR(file);
                                }
                        }
                unlock_kernel();
                putname(tmp);
                }       

        set_fs( fs );

        Dev->hSerial = fd;

        if ( Dev->hSerial < 0 ) 
                Status = STATUS_INVALID_PORT_ATTRIBUTES;
        else
                Status = STATUS_SUCCESS;

        return( Status );
}

SRLOCAL long OsSerialClose( DEVINFO *Dev ) {

        dbg2_printk("Calling OsSerialClose\n");

        // Early exit if already closed

        if ( Dev->hSerial == BAD_DEVHANDLE )
                return( STATUS_SUCCESS );


        sys_close( Dev->hSerial );
        Dev->hSerial = BAD_DEVHANDLE;
        
        return( STATUS_SUCCESS );
}

SRLOCAL long OsSerialFlush( DEVINFO *Dev ) {

        long           Status;
        int            Nret;
        unsigned int   cmd;
        struct file   *file;
//      struct inode  *inode;
//      struct dentry *dentry;
        mm_segment_t   fs, ds;

        dbg2_printk("Calling OsSerialFlush\n");

        Status = STATUS_SUCCESS;

        cmd = TCFLSH;

        fs = get_fs();
        ds = get_ds();
        set_fs( ds );

        lock_kernel();
        file = fget( Dev->hSerial );

        if (file && 
            file->f_dentry &&
            file->f_dentry->d_inode &&
            file->f_op && 
            file->f_op->ioctl) {
                Nret = file->f_op->ioctl( file->f_dentry->d_inode,
                                          file, 
                                          cmd, 
                                          TCIOFLUSH);
                }
        else
                Status = STATUS_FILE_INVALID;

        if (file) fput( file );
        unlock_kernel();
        set_fs( fs );

        return( Status );
}

SRLOCAL int OsSerialReadChar( DEVINFO *Dev,
                              char *Buffer,
                              unsigned long BuffLen,
                              unsigned long *Status ) {
        int Nchar;
        struct file *file;
        mm_segment_t fs, ds;

        dbg5_printk("Calling OsSerialReadChar\n");

        *Status = STATUS_SUCCESS;
        Nchar   = 0;

        fs = get_fs();
        ds = get_ds();
        set_fs( ds );


        lock_kernel();
        file = fget( Dev->hSerial );

        // Consider adding a check of locks_verify_area like in 
        // sys_read in fs/read_write.c
        
        if (file && file->f_op && file->f_op->read)
                Nchar = file->f_op->read( file, Buffer, BuffLen, &file->f_pos);
        else {
                dbg_printk("Failed to perform serial read ...\n");
                *Status = STATUS_FILE_INVALID;
                }

        if (file) fput( file );
        unlock_kernel();
        set_fs( fs );

        if (*Status != STATUS_SUCCESS)
                return( 0 );

        if (Nchar == 0)
                *Status = STATUS_PIPE_EMPTY;
        else if (Nchar < 0) {
                *Status = -Nchar;
                Nchar   = 0;
                }

        return( Nchar );
}

SRLOCAL long OsSerialSignal( DEVINFO *Dev ) {
        
        long PrevState;
        
	wake_up_interruptible( &Dev->ThreadWait );
	dbg2_printk("wake_up serial monitor thread\n" );

        // The DPC later checks this to see if the event was processed.
        // Winnt gets PrevState when calling KeSetEvent, we just hardwire it.
        
        PrevState = 0;
        
        return( PrevState );
}

SRLOCAL long OsThreadOpen( DEVINFO *Dev ) {

        dbg2_printk("OsThreadOpen Enter\n");

        // Check for thread already open
        
        if (Dev->ThreadTask) {
                dbg_printk( "Thread already open\n" );
                return( STATUS_TOKEN_ALREADY_IN_USE );
                }

        
	init_waitqueue_head( &Dev->ThreadWait );

/*

Initialize the semaphore:

We start with the semaphore locked. The new kernel thread will setup
its stuff and unlock it. This control flow (the one that creates the
thread) blocks in the down operation below until the thread has
reached the up() operation.  

*/

        sema_init( &Dev->ThreadMutex, MUTEX_LOCKED );


        // Create the new thread

        kernel_thread( Linux_ThreadFunc, (void *)Dev, 0 );



        // Wait till it has reached the OsThreadInit routine

        down( &Dev->ThreadMutex );

        return( STATUS_SUCCESS );
}

SRLOCAL long OsThreadClose( DEVINFO *Dev ) {

        dbg2_printk("OsThreadClose Enter\n");

        if (Dev->ThreadTask == NULL) {
                dbg_printk("thread_drv: killing non existing thread!\n");
                return( STATUS_CANNOT_DELETE );
                }

        // This function needs to be protected with the big
        // kernel lock (lock_kernel()). The lock must be
        // grabbed before changing the terminate
        // flag and released after the down() call.

        lock_kernel();

        // Initialize the semaphore. We lock it here, the
        // leave_thread call of the thread to be terminated
        // will unlock it. As soon as we see the semaphore
        // unlocked, we know that the thread has exited.

        sema_init( &Dev->ThreadMutex, MUTEX_LOCKED );


        // We need to do a memory barrier here to be sure that
        // the flags are visible on all CPUs. 

        mb();


        // Set flag to request thread termination

        Dev->ThreadTerminate = 1;

        // We need to do a memory barrier here to be sure that
        // the flags are visible on all CPUs. 

        mb();
        kill_proc( Dev->ThreadTask->pid, SIGKILL, 1 );


        // Block till thread terminated

        down( &Dev->ThreadMutex );


        // release the big kernel lock

        unlock_kernel();

        return( STATUS_SUCCESS );
}

SRLOCAL void OsThreadInit( DEVINFO *Dev ) {

        dbg2_printk("OsThreadInit Enter\n");


        // Lock the kernel. A new kernel thread starts without
        // the big kernel lock, regardless of the lock state
        // of the creator (the lock level is *not* inheritated)

        lock_kernel();


        // Put ourself in background, release all user space connection.

        Linux_daemonize();


        // Set name of this process (max 15 chars + 0 !)

        sprintf( current->comm, "pargpsthread" );


        // Fill in thread task structure

        Dev->ThreadTask = current;



        // Set signal mask to what we want to respond to

        siginitsetinv( &current->blocked, 
                       sigmask(SIGKILL)|sigmask(SIGINT)|sigmask(SIGTERM));


        // Initialize wait queue

        init_waitqueue_head( &Dev->ThreadWait );

        

        // Let others run

        unlock_kernel();


        // Tell the creator that we are ready and let him run

        up( &Dev->ThreadMutex );
}

SRLOCAL void OsThreadLeave( DEVINFO *Dev, unsigned long Status ) {

        dbg2_printk("OsThreadLeave Enter\n");

        
        // Lock the kernel, the exit will unlock it

        lock_kernel();

        Dev->ThreadTask = NULL;

        mb();  // Memory barrier


        // Notify the OsThreadClose routine that we are terminating.

        up( &Dev->ThreadMutex );

/* 

The kernel_thread that called clone() does a do_exit here.

There is no race here between unloading the module and real
termination of the thread (race window between up and do_exit), since
both the thread and the unload function are running with the kernel
lock hold.  The kernel lock will be freed after the thread exited, so
the code is really not executed anymore as soon as the unload
functions gets the kernel lock back.  The init process may not have
made the cleanup of the process here, but the cleanup can be done
safely with the module unloaded.  

*/

}

SRLOCAL void OsGetSystemTime( SR64BIT *SysTime ) {
        
#ifdef USE_RDTSC_FOR_TIME
        
        register unsigned long eax asm("ax");
        register unsigned long edx asm("dx");
        
        if (current_cpu_data.x86_capability & X86_FEATURE_TSC) {
                rdtsc( eax, edx );
                SysTime->u.HighPart = edx;
                SysTime->u.LowPart  = eax;
                }
        else {
                SysTime->u.HighPart = 0L;
                SysTime->u.LowPart  = 0L;
                }
                
#else        
        struct timeval tv;

        do_gettimeofday( &tv );
        SysTime->u.HighPart = tv.tv_sec;
        SysTime->u.LowPart  = tv.tv_usec;

#endif // USE_RDTSC_FOR_TIME

}

SRLOCAL long OsGetCounterFrequency( void ) {

        // Linux timeval is always in microseconds
        
        return( 1000000L );
}

SRLOCAL unsigned long OsWaitOnPps( DEVINFO *Dev,
                                   SR64BIT *RequestedTimeout,
                                   int Id ) {

        long Status, Remain, Timeout;


        Status = STATUS_SUCCESS;

        // Set timeout by request or default
        // (always must be positive and relative)

        if (RequestedTimeout)
                Timeout = RequestedTimeout->u.HighPart;
        else
                Timeout = (4000*HZ)/1000;




        // Wait for PPS event to signal



        // Fall asleep for requested timeout

        Remain = interruptible_sleep_on_timeout( &Dev->ThreadWait, Timeout );


        // We need to do a memory barrier here to be sure that
        // the flags are visible on all CPUs. 

        mb();



        // Here we are back from sleep, either due to the timeout
        // or receiving a signal, or being woken up by PPS DPC.

        if (Remain > 0L)
                Status = STATUS_SUCCESS;
        else
                Status = STATUS_TIMEOUT;


        if (Status == STATUS_SUCCESS) {
                dbg0_printk("Received PPS event %d in thread (remain %ld)\n", Id, Remain);
                
                // Winnt uses KeClearEvent here to indicate that the
                // Pps event was received
        }
        else if (Status == STATUS_TIMEOUT) {
                dbg0_printk("Timed out PPS event %d in thread (remain %ld)\n",Id, Remain);
        }
        else {
                dbg0_printk("Status of 0x%lX on PPS event %d in thread\n", Status, Id);
        }

        return( (unsigned long)Status );
}



#endif // SROS_xxxxx  End of OS dependent helper functions section






////////////////////////////////////////////////////////////////////////
/* Group 3: OS DEPENDENT INITIALIZATION AND DISPATCH FUNCTIONS:

The functions below are extremely system dependent as they involve very
OS specific initialization code.

For programmers interested in the low level mechanics of how it all
works, a brief overview follows:

When the indriver utility is run, the OS calls a driver
initialization function with a known name.  This function gets any
user supplied information, sets up pointers to the "dispatch"
functions for read and I/O control, and fills the Device Information
structure which holds global data for the driver.

When the user calls a pargps.c library function, it calls an OS
function that in turn calls a driver function from pargpskd.c using
the previously set "dispatch" pointers.

Most of the library function calls eventually get sent through
ParGps_DrvIoctl which has a big switch statement to select from the
many available functions.


A short comparison between some driver items for Win2K/NT and Linux:

item description     Win2K/NT          Linux                 
----------------     -----------       -----------           
init known name      DriverEntry       init_module           
src of user info     registry          insmod command line   
OS read function     ReadFile          read                  
OS I/O ctrl func     DeviceIoControl   ioctl                 

*/


// *************************************************************************

#if defined( SROS_WIN98 ) || defined( SROS_WIN2K )

NTSTATUS DriverEntry( DRIVER_OBJECT *DriverObject,
                      UNICODE_STRING *RegistryPath ) {
        
        dbg0_printk("In PARGPS Wdm DriverEntry\n");

   
        // STEP 1: Set up dispatch table

        DriverObject->DriverExtension->AddDevice           = Wdm_DrvAddDevice;
        DriverObject->MajorFunction[IRP_MJ_PNP]            = Wdm_DrvPnp;
        DriverObject->MajorFunction[IRP_MJ_POWER]          = Wdm_DrvPower;
        DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = Wdm_DrvWmi;
        DriverObject->MajorFunction[IRP_MJ_CREATE]         = Wdm_DrvOpen;
        DriverObject->MajorFunction[IRP_MJ_CLOSE]          = Wdm_DrvClose;
        DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = Wdm_DrvIoctl;
        DriverObject->DriverUnload                         = Wdm_DrvUnload;

        
        // STEP 2: Save registry path to service key
        

        //WCT: Oney did PagedPool, is that sufficient ?

        DriverRegPath.MaximumLength = RegistryPath->Length + sizeof(WCHAR);
        DriverRegPath.Buffer = (PWSTR)ExAllocatePool( 
                                    PagedPool, DriverRegPath.MaximumLength );
        if (!DriverRegPath.Buffer)
                return( STATUS_INSUFFICIENT_RESOURCES );
        RtlCopyUnicodeString( &DriverRegPath, RegistryPath );
        

        return( STATUS_SUCCESS );
}

void Wdm_DrvUnload( DRIVER_OBJECT *DriverObject ) {

        dbg0_printk("In Wdm_DrvUnload\n");



        // This function undoes everything done in DriverEntry

        RtlFreeUnicodeString( &DriverRegPath );

}

#define DEVICE_NAME_FORMAT_DOS  "\\??\\%s"
#define DEVICE_NAME_FORMAT_NT   "\\Device\\%s"
#define DEVICE_NAME_FORMAT_LINK "\\DosDevices\\%s"

#define DEVICE_NAME_DOS         "\\??\\SrParGps"
#define DEVICE_NAME_NT         L"\\Device\\SrParGps"
#define DEVICE_NAME_LINK       L"\\DosDevices\\SrParGps"

NTSTATUS Wdm_DrvAddDevice( DRIVER_OBJECT *DriverObject, DEVICE_OBJECT *PDO ) {

        NTSTATUS         Status;
        WCHAR            ParGpsNameBuffer[MAX_STR_LENGTH];
        unsigned char    AnsiDriverBuffer[MAX_STR_LENGTH];
        char             DeviceNameBuffer[MAX_STR_LENGTH];
        ANSI_STRING      AnsiDriverString;
        UNICODE_STRING   NtDeviceName, LinkDeviceName, ParGpsName;
        DEVICE_OBJECT   *FDO;
        DEVINFO         *Dev;
        POWER_STATE      InitialPowerState;
        char            *DeviceName;

        
        dbg0_printk("In Wdm_DrvAddDevice\n");

        // DriverObject is same one as passed to DriverEntry
        // 
        // PDO is address of the physical device object at the bottom
        // of the device stack (even though some lower filter driver 
        // may already be on top of it).



        // Get driver name from the registry
        
        ParGpsName.Length        = 0;
        ParGpsName.MaximumLength = MAX_STR_LENGTH * sizeof(WCHAR);
        ParGpsName.Buffer        = ParGpsNameBuffer;

        
        Status = Wdm_GetParameters( DriverObject, &DriverRegPath, &ParGpsName );
        
        if ( !NT_SUCCESS( Status ) ) {
                DbgPrint("Wdm_GetParameters failed with status 0x%X\n", Status);
                return( Status );
                }
                                    
        AnsiDriverString.Length        = 0;
        AnsiDriverString.MaximumLength = MAX_STR_LENGTH * sizeof(char);
        AnsiDriverString.Buffer        = DeviceNameBuffer;
        
        DeviceName = DeviceNameBuffer;
        
        RtlUnicodeStringToAnsiString( &AnsiDriverString, &ParGpsName, FALSE );



        

        // STEP 1a: Create device (and device extension)
        
        // Note: type, characteristics and exclusive can be overridden 
        // by values in the hardware key

        // Consider adding FILE_DEVICE_SECURE_OPEN to characteristics.pg56


        AnsiDriverString.Length        = 0;
        AnsiDriverString.MaximumLength = MAX_STR_LENGTH * sizeof(char);
        AnsiDriverString.Buffer        = AnsiDriverBuffer;
        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_NT, DeviceName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &NtDeviceName, &AnsiDriverString, TRUE);
        
        Status = IoCreateDevice(
                                DriverObject,
                                sizeof(DEVINFO),
                                &NtDeviceName,
                                FILE_DEVICE_UNKNOWN,
                                0,                  // Characteristics
                                FALSE,              // NOT Exclusive (> 1 ok)
                                &FDO
                               );
        
        if ( !NT_SUCCESS( Status ) ) {
                DbgPrint("IoCreateDevice failed with status 0x%X\n", Status);
                RtlFreeUnicodeString( &NtDeviceName );
                return( Status );
                }


        // STEP 2: Register device interfaces OR
        //         Create name and symbolic link

        
        // Create symbolic link

        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_LINK, DeviceName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &LinkDeviceName, &AnsiDriverString, TRUE );
        
        Status = IoCreateSymbolicLink( &LinkDeviceName, &NtDeviceName );

        RtlFreeUnicodeString( &NtDeviceName );
        RtlFreeUnicodeString( &LinkDeviceName );

        if ( !NT_SUCCESS( Status ) ) {
                DbgPrint("IoCreateSymbolicLink failed with status 0x%X\n", Status);
                Wdm_RemoveDevice( FDO );
                return( Status );
                }


        
        // STEP 3: Initialize device extension
        
        Dev = (DEVINFO *)FDO->DeviceExtension;
        
        RtlZeroMemory( Dev, sizeof(DEVINFO) );
        ParGps_FillDevInfo( Dev, DeviceName );
        
        Dev->DriverObj     = DriverObject;
        Dev->ThisFdoDevObj = FDO;
        Dev->PdoDevObj     = PDO;


        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_LINK, DeviceName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &Dev->LinkName, &AnsiDriverString, TRUE );



        // Hook DPC for ISR

        // WCT: Review
        //        IoInitializeDpcRequest( FDO, DpcForIsr );

        
        // Set up custom DPCs for PPS, Dready, and Plain types

        KeInitializeDpc(
                        &Dev->CustomDpc[PARGPS_INTTYPE_PPS], // PRKDPC
                        Wdm_DpcForPpsRoutine,                // PKDEFERRED_ROUTINE
                        FDO                                  // PVOID deferred context
                       );

        KeInitializeDpc(
                        &Dev->CustomDpc[PARGPS_INTTYPE_DREADY], // PRKDPC
                        Wdm_DpcForDreadyRoutine,                // PKDEFERRED_ROUTINE
                        FDO                                     // PVOID deferred context
                       );
        KeInitializeDpc(
                        &Dev->CustomDpc[PARGPS_INTTYPE_PLAIN], // PRKDPC
                        Wdm_DpcForPlainRoutine,                // PKDEFERRED_ROUTINE
                        FDO                                    // PVOID deferred context
                       );



        // Initialize Pps and Thread events to not signaled

        KeInitializeEvent( &Dev->PpsEvent,    NotificationEvent, FALSE );
        KeInitializeEvent( &Dev->ThreadEvent, NotificationEvent, FALSE );
        
        

        // Init remove lock

        IoInitializeRemoveLock( &Dev->RemoveLock, '_MYS', 0, 999 );
        
        // WCT: Consider adding interface code
        //        IoRegisterDeviceInterface( ..., &Dev->InterfaceName );


        // Inform power manager of initial power state
        
        InitialPowerState.DeviceState = Dev->CurrDevPower;
        PoSetPowerState( FDO, DevicePowerState, InitialPowerState );



        
        // STEP 4: Put device object into the stack
        
        Dev->LowerDevObj = IoAttachDeviceToDeviceStack( FDO, PDO );

        if (Dev->LowerDevObj == NULL) {
                Wdm_RemoveDevice( FDO );
                return( STATUS_NO_SUCH_DEVICE );
                }



        
        // Set device options for Read/Write data alignment and passing
        // and indicate done initializing
        
        FDO->AlignmentRequirement = FILE_BYTE_ALIGNMENT;
        FDO->Flags               |= DO_DIRECT_IO;
        FDO->Flags               &= ~DO_DEVICE_INITIALIZING;

        return( Status );
}

SRLOCAL NTSTATUS Wdm_RemoveDevice( DEVICE_OBJECT *FDO ) {

        DEVINFO *Dev;
        
        dbg0_printk("In Wdm_RemoveDevice\n");

        // This function undoes everything done by DrvAddDevice


        if (!FDO)
                return( STATUS_UNSUCCESSFUL );

        Dev = (DEVINFO *)FDO->DeviceExtension;

       
        // Detach the current device from the driver stack
 
        if ( Dev->LowerDevObj ) {
                IoDetachDevice( Dev->LowerDevObj );
                Dev->LowerDevObj = NULL;
                }


        // Remove symbolic link name

        if ( Dev->LinkName.Length > 0 ) {
                IoDeleteSymbolicLink( &Dev->LinkName );
                RtlFreeUnicodeString( &Dev->LinkName );
                Dev->LinkName.Length = 0;
                }


        // Delete the current device
        // Careful, in Win98 IoDeleteDevice calls DriverUnload !!

        IoDeleteDevice( FDO );


        return( STATUS_SUCCESS );
        
}

BOOLEAN Wdm_DrvInterrupt( PKINTERRUPT pInterruptObject, void *Context ) {

        DEVINFO *Dev;
        
        // If this device not interrupting, return FALSE

        if (Context == NULL)
                return( FALSE );
        
        Dev = (DEVINFO *)Context;

        return( (BOOLEAN)ParGps_Interrupt( Dev ) );
}



NTSTATUS Wdm_DrvOpen( DEVICE_OBJECT *FDO, IRP *Irp ) {

        return( (NTSTATUS)ParGps_DrvOpen(
                          (DEVINFO *)FDO->DeviceExtension, Irp ) );
}

NTSTATUS Wdm_DrvClose( DEVICE_OBJECT *FDO, IRP *Irp ) {

        return( (NTSTATUS)ParGps_DrvClose(
                          (DEVINFO *)FDO->DeviceExtension, Irp ) );
}

NTSTATUS Wdm_DrvIoctl( DEVICE_OBJECT *FDO, IRP *Irp ) {
        return( (NTSTATUS)ParGps_DrvIoctl(
                          (DEVINFO *)FDO->DeviceExtension, Irp ) );
}

void Wdm_DpcForPpsRoutine( KDPC *DpcForPps,
                             void *Context, // Device_Object
                             void *SysArg1, // Irp
                             void *DpcData  // Device Extension
                           ) {
        
        ParGps_DpcForPpsRoutine( (DEVINFO *)DpcData );
}

void Wdm_DpcForDreadyRoutine( KDPC *DpcForDready,
                                void *Context, // Device_Object
                                void *SysArg1, // Irp
                                void *DpcData  // Device Extension
                              ) {
        ParGps_DpcForDreadyRoutine( (DEVINFO *)DpcData );
}

void Wdm_DpcForPlainRoutine( KDPC *DpcForPlain,
                               void *Context, // Device_Object
                               void *SysArg1, // Irp
                               void *DpcData  // Device Extension
                             ) {
        ParGps_DpcForPlainRoutine( (DEVINFO *)DpcData );
}

void WdmWinnt_ThreadFunc( void *Context ) {

        dbg2_printk("In Wdm_ThreadFunc with Context 0x%p\n", Context );

        ParGps_ThreadFunc( (DEVINFO*)Context );
}



// Plug and Play functions

NTSTATUS Wdm_DrvPnp( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;
        IO_STACK_LOCATION *IrpStack;
        ULONG MinorFn;

        dbg1_printk("In Wdm_DrvPnp\n");
        
        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        MinorFn  = IrpStack->MinorFunction;

        
        switch( MinorFn ) {
                case IRP_MN_START_DEVICE:
                        Status = Wdm_PnpStartDevice( FDO, Irp );
                        break;
                case IRP_MN_STOP_DEVICE:
                        Status = Wdm_PnpStopDevice( FDO, Irp );
                        break;
                case IRP_MN_REMOVE_DEVICE:
                        Status = Wdm_PnpRemoveDevice( FDO, Irp );
                        break;
                case IRP_MN_QUERY_CAPABILITIES:
                        Status = Wdm_PnpQueryCapabilities( FDO, Irp );
                        break;
                case IRP_MN_CANCEL_STOP_DEVICE:
                        Status = Wdm_PnpCancelStop( FDO, Irp );
                        break;
                case IRP_MN_CANCEL_REMOVE_DEVICE:
                        Status = Wdm_PnpCancelRemove( FDO, Irp );
                        break;
                case IRP_MN_QUERY_STOP_DEVICE:
                        Status = Wdm_PnpQueryStop( FDO, Irp );
                        break;
                case IRP_MN_QUERY_REMOVE_DEVICE:
                        Status = Wdm_PnpQueryRemove( FDO, Irp );
                        break;
                case IRP_MN_QUERY_DEVICE_RELATIONS:
                        Status = Wdm_PnpQueryDevRelations( FDO, Irp );
                        break;
                case IRP_MN_QUERY_INTERFACE:
                        Status = Wdm_PnpQueryInterface( FDO, Irp );
                        break;
                case IRP_MN_FILTER_RESOURCE_REQUIREMENTS:
                        Status = Wdm_PnpFilterResourceReq( FDO, Irp );
                        break;
                case IRP_MN_SURPRISE_REMOVAL:
                        Status = Wdm_PnpSurpriseRemoval( FDO, Irp );
                        break;
                default:
                        dbg2_printk("Pnp MinorFn 0x%X\n", MinorFn);
                        Status = Wdm_IrpForward( FDO, Irp );
                }

        return( Status );
}

NTSTATUS Wdm_PnpStartDevice( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;

        dbg0_printk("In Wdm_PnpStartDevice\n");
        
        
        // Let lower drivers do their starting stuff first.

        Irp->IoStatus.Status = STATUS_SUCCESS;
        Status = Wdm_IrpForwardAndWait( FDO, Irp );
        if ( !NT_SUCCESS(Status )) {
                OsIrpFinish( Irp, Status, Irp->IoStatus.Information );
                return( Status );
                }


        // We have now been assigned resources by the PnP Manager.  So
        // now do our starting work of allocating them and starting the
        // device.

        Wdm_ResourceSet( FDO, Irp );


        
        // Do our starting work
        
        Wdm_StartDevice( FDO );
        
        

        // Done

        OsIrpFinish( Irp, Status, Irp->IoStatus.Information );
        return( Status );
}
SRLOCAL NTSTATUS Wdm_ResourceSet( DEVICE_OBJECT *FDO, IRP *Irp ) {

        dbg0_printk("In Wdm_ResourceSet - no resources \n");

        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Wdm_StartDevice( DEVICE_OBJECT *FDO ) {

        DEVINFO *Dev;
        POWER_STATE powerstate;
        
        dbg0_printk("In Wdm_StartDevice\n");

        Dev = (DEVINFO *)FDO->DeviceExtension;


        // Turn on device and inform the power manager.

        Dev->CurrDevState = RUNSTATE_RUNNING;
        powerstate.DeviceState = PowerDeviceD0;
        
        PoSetPowerState( FDO, DevicePowerState, powerstate );


//        IoSetDeviceInterfaceState( &Dev->InterfaceName, TRUE ); //enable

        
        return( STATUS_SUCCESS );
}

NTSTATUS Wdm_PnpStopDevice( DEVICE_OBJECT *FDO, IRP *Irp ) {

        dbg0_printk("In Wdm_PnpStopDevice\n");

        // Handle queueing and cancellation.


        // Do our stopping work.
        
        Wdm_ResourceClear( FDO );
        Wdm_StopDevice( FDO );
        

        // Let lower drivers do their stopping work.
        
        Irp->IoStatus.Status = STATUS_SUCCESS;
        return( Wdm_IrpForward( FDO, Irp ) );
}

SRLOCAL NTSTATUS Wdm_ResourceClear( DEVICE_OBJECT *FDO ) {
        
        dbg0_printk("In Wdm_ResourceClear - no resources\n");
        
        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Wdm_StopDevice( DEVICE_OBJECT *FDO ) {

        DEVINFO *Dev;
        POWER_STATE powerstate;
        
        dbg0_printk("In Wdm_StopDevice\n");


        Dev = (DEVINFO *)FDO->DeviceExtension;


        // This function undoes everything done in StartDevice

        
        // Turn off device and inform the power manager.

        Dev->CurrDevState = RUNSTATE_RUNNING;
        powerstate.DeviceState = PowerDeviceD0;
        
        PoSetPowerState( FDO, DevicePowerState, powerstate );


//        IoSetDeviceInterfaceState( &Dev->InterfaceName, FALSE ); //disable

        return( STATUS_SUCCESS );
}


NTSTATUS Wdm_PnpRemoveDevice( DEVICE_OBJECT *FDO, IRP *Irp ) {
        
        NTSTATUS Status;
        DEVINFO *Dev;
        
        dbg0_printk("In Wdm_PnpRemoveDevice\n");
        
        Dev = (DEVINFO *)FDO->DeviceExtension;


        // Deregister all interfaces.


        // Do our stopping work.

        Wdm_ResourceClear( FDO );
        Wdm_StopDevice( FDO );


        // Pass down to lower devices.
        
        Irp->IoStatus.Status = STATUS_SUCCESS;
        Status = Wdm_IrpForward( FDO, Irp );


        // RemoveDevice undoes things done in AddDevice
        
        Wdm_RemoveDevice( FDO );
        
        return( Status );
}
NTSTATUS Wdm_PnpQueryCapabilities( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryCapabilities\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpCancelStop( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpCancelStop\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpCancelRemove( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpCancelRemove\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryStop( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryStop\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryRemove( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryRemove\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryDevRelations( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryDevRelations\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpQueryInterface( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpQueryInterface\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpFilterResourceReq( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpFilterResourceReq\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}
NTSTATUS Wdm_PnpSurpriseRemoval( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg0_printk("In Wdm_PnpSurpriseRemoval\n");
        return( Wdm_IrpForward( FDO, Irp ) );
}



// Power Functions

NTSTATUS Wdm_DrvPower( DEVICE_OBJECT *FDO, IRP *Irp ) {

        NTSTATUS Status;
        DEVINFO *Dev;
        IO_STACK_LOCATION *IrpStack;
        ULONG MinorFn;
        POWER_STATE_TYPE PowerType;
        POWER_STATE PowerState;

        dbg0_printk("In Wdm_DrvPower\n");
        
        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        MinorFn    = IrpStack->MinorFunction;
        PowerType  = IrpStack->Parameters.Power.Type;
        PowerState = IrpStack->Parameters.Power.State;


//        Status = IoAcquireRemoveLock();
//        if ( !NT_SUCCESS( Status ) ) {
//            Wdm_PowerFail( Irp, Status );
//            return( Status );
//            }



        switch( MinorFn ) {

            case IRP_MN_SET_POWER:

                // WCT: Do more testing here
                if (PowerType == SystemPowerState)
                        Status = Wdm_PowerSetSystem( FDO, Irp, PowerState );

                else if (PowerType == DevicePowerState)
                        Status = Wdm_PowerSetDevice( FDO, Irp, PowerState );

                else
                        Status = Wdm_PowerForward( FDO, Irp );

                break;

            case IRP_MN_QUERY_POWER:
                DbgPrint("Power IRP_MN_QUERY_POWER always forward\n");
                Status = Wdm_PowerForward( FDO, Irp );
                break;
                
            case IRP_MN_WAIT_WAKE:
            case IRP_MN_POWER_SEQUENCE:
            default:
                if (MinorFn == IRP_MN_WAIT_WAKE)
                        DbgPrint("Power IRP_MN_WAIT_WAKE\n");
                else if (MinorFn == IRP_MN_POWER_SEQUENCE)
                        DbgPrint("Power IRP_MN_POWER_SEQUENCE\n");
                else
                        DbgPrint("Power minor fn 0x%X\n", MinorFn);
                Status = Wdm_PowerForward( FDO, Irp );
            }

//        IoReleaseRemoveLock();

        
        return( Status );
}

NTSTATUS Wdm_PowerSetSystem( DEVICE_OBJECT *FDO,
                               IRP *Irp,
                               POWER_STATE PowerState ) {

        NTSTATUS Status;
        POWER_STATE DevPowerState;
        DEVINFO *Dev;
        
        dbg0_printk("In Wdm_PowerSetSystem\n");
        
        IoMarkIrpPending( Irp );

        Dev = (DEVINFO *)FDO->DeviceExtension;
        

        // Determine desired device state from requested system state
        // Device ON for full system ON, otherwise device OFF

        if (PowerState.SystemState == PowerSystemWorking)
                DevPowerState.DeviceState = PowerDeviceD0;
        else
                DevPowerState.DeviceState = PowerDeviceD3;


        // Update saved system power state

        Dev->CurrSysPower = PowerState.SystemState;


        
        // Powering down (current state < new state )
        
        if (Dev->CurrDevPower < DevPowerState.DeviceState ) {
                
                dbg3_printk("Powering down system ... device from %d to %d\n",Dev->CurrDevPower, DevPowerState.DeviceState);

                Status = Wdm_SendDeviceSetPower( FDO, Irp, DevPowerState );

                if (!NT_SUCCESS( Status )) {
                        Wdm_PowerFail( Irp, Status );
                        return( Status );
                        }

                Status = Wdm_PowerForward( FDO, Irp );
                
                }

        

        // Powering up
        
        else if (Dev->CurrDevPower > DevPowerState.DeviceState ) {

                dbg3_printk("Powering up system ... device from %d to %d\n",Dev->CurrDevPower, DevPowerState.DeviceState);

                IoCopyCurrentIrpStackLocationToNext( Irp );
                IoSetCompletionRoutine( Irp,
                                        Wdm_OnCompletePowerUpSystem,
                                        Dev,
                                        TRUE,
                                        TRUE,
                                        TRUE
                                      );
                Status = PoCallDriver( Dev->LowerDevObj, Irp );

                // WCT: MSDN suggests Status = STATUS_PENDING since we 
                // must wait for device irp ?

                }

        else 
                Status = Wdm_PowerForward( FDO, Irp );
                

        return( Status );
}

NTSTATUS Wdm_SendDeviceSetPower( DEVICE_OBJECT *FDO,
                                 IRP *Irp,
                                 POWER_STATE PowerState ) {

        NTSTATUS Status;
        IO_STACK_LOCATION *IrpStack;
        DEVINFO *Dev;
        
        dbg0_printk("In Wdm_SendDeviceSetPower\n");


        // WCT: Consider sending in desired device power state
        
        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );

        KeInitializeEvent( &Dev->PowerEvent, NotificationEvent, FALSE );

        Status = PoRequestPowerIrp( Dev->PdoDevObj,
                                    IrpStack->MinorFunction,
                                    PowerState,
                                    Wdm_OnCompleteDeviceSetPower,
                                    Dev,
                                    NULL
                                  );

        if (Status == STATUS_PENDING) {
                
                KeWaitForSingleObject( &Dev->PowerEvent,
                                       Executive,
                                       KernelMode,
                                       FALSE,
                                       NULL
                                     );
                Status = Dev->PowerStatus;
                }


        // WCT: Cope with W98 not passing power irp

        return( Status );
}

void Wdm_OnCompleteDeviceSetPower( DEVICE_OBJECT *FDO,
                                     UCHAR MinorFunction,
                                     POWER_STATE PowerState,
                                     VOID *Context,
                                     IO_STATUS_BLOCK *IoStatus ) {
        DEVINFO *Dev;

        dbg0_printk("In Wdm_OnCompleteDeviceSetPower\n");
        
        Dev = (DEVINFO *)Context;
        Dev->PowerStatus = IoStatus->Status;
        KeSetEvent( &Dev->PowerEvent, 0, FALSE );
}

NTSTATUS Wdm_OnCompletePowerUpSystem( DEVICE_OBJECT *FDO, IRP *Irp, void *Context ) {

        DEVINFO *Dev;
        NTSTATUS Status;
        IO_STACK_LOCATION *IrpStack;
        POWER_STATE DevPowerState;

        dbg0_printk("In Wdm_OnCompletePowerUpSystem\n");

        // Remember if a lower driver pended the Irp

        if (Irp->PendingReturned)
                IoMarkIrpPending( Irp );


        // Check if a lower driver failed
        
        Status = Irp->IoStatus.Status;
        if ( !NT_SUCCESS( Status ) ) {
                PoStartNextPowerIrp( Irp );
                return( Status );
                }
        
        

        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );



        // Reset required device power state
        
        if (Dev->CurrSysPower == PowerSystemWorking)
                DevPowerState.DeviceState = PowerDeviceD0;
        else
                DevPowerState.DeviceState = PowerDeviceD3;

        
        // And retest new device power state, setting if needed

        if (Dev->CurrDevPower > DevPowerState.DeviceState)
                Status = Wdm_SendDeviceSetPower( FDO, Irp, DevPowerState );
        
        PoStartNextPowerIrp( Irp );


        // Note: If this function had queued a work item to run at 
        // passive level, the return would have had to be
        // STATUS_MORE_PROCESSING_REQUIRED.  On Win98, power Irp's 
        // are completed at passive level.
        
        return( Status );
}

NTSTATUS Wdm_PowerSetDevice( DEVICE_OBJECT *FDO,
                               IRP *Irp,
                               POWER_STATE PowerState ) {

        NTSTATUS Status;
        DEVINFO *Dev;
        
        dbg0_printk("In Wdm_PowerSetDevice\n");
        
        // Note: HIGHER state numbers mean LOWER power

        Dev = (DEVINFO *)FDO->DeviceExtension;

        
        // Powering down (current state < new state )
        
        if (Dev->CurrDevPower < PowerState.DeviceState ) {

                dbg3_printk("Powering down device from %d to %d\n",Dev->CurrDevPower, PowerState.DeviceState);

                // WCT: Actually do hardware power down here
                
                Dev->CurrDevPower = PowerState.DeviceState;
                PoSetPowerState( Dev->ThisFdoDevObj, DevicePowerState, PowerState );
                
                Status = Wdm_PowerForward( FDO, Irp );
                }

        // Powering up
        
        else if (Dev->CurrDevPower > PowerState.DeviceState ) {

                        dbg3_printk("Powering up device from %d to %d\n",Dev->CurrDevPower, PowerState.DeviceState);

                        IoCopyCurrentIrpStackLocationToNext( Irp );
                        IoSetCompletionRoutine( Irp,
                                                Wdm_OnCompletePowerUpDevice,
                                                Dev,
                                                TRUE,
                                                TRUE,
                                                TRUE
                                              );
                        Status = PoCallDriver( Dev->LowerDevObj, Irp );

                        // WCT: MSDN suggests calling IoMarkIrpPending
                        //      and returning STATUS_PENDING

                }

        else // No change
                Status = Wdm_PowerForward( FDO, Irp );


        return( Status );
}

NTSTATUS Wdm_OnCompletePowerUpDevice( DEVICE_OBJECT *FDO, IRP *Irp, void *Context ) {

        NTSTATUS Status;
        IO_STACK_LOCATION *IrpStack;
        POWER_STATE PowerState;
        DEVINFO *Dev;

        dbg0_printk("In Wdm_OnCompletePowerUpDevice\n");

        // Remember if a lower driver pended the Irp

        if (Irp->PendingReturned)
                IoMarkIrpPending( Irp );


        // Check if a lower driver failed
        
        Status = Irp->IoStatus.Status;
        if ( !NT_SUCCESS( Status ) ) {
                PoStartNextPowerIrp( Irp );
                return( Status );
                }
        

        
        Dev = (DEVINFO *)FDO->DeviceExtension;
        IrpStack = IoGetCurrentIrpStackLocation( Irp );
        PowerState = IrpStack->Parameters.Power.State;

        
        // Finish powering up

        // WCT: Actually do hardware power on here
        
        Dev->CurrDevPower = PowerState.DeviceState;
        PoSetPowerState( Dev->ThisFdoDevObj, DevicePowerState, PowerState );


        PoStartNextPowerIrp( Irp );


        // Note: If this function had queued a work item to run at 
        // passive level, the return would have had to be
        // STATUS_MORE_PROCESSING_REQUIRED.  On Win98, power Irp's are
        // completed at passive level.
        
        return( Status );
}

NTSTATUS Wdm_PowerForward( DEVICE_OBJECT *FDO, IRP *Irp ) {

        DEVINFO *Dev;

        dbg0_printk("In Wdm_PowerForward\n");
        
        Dev = (DEVINFO *)FDO->DeviceExtension;
        PoStartNextPowerIrp( Irp );
        IoSkipCurrentIrpStackLocation( Irp );
        return( PoCallDriver( Dev->LowerDevObj, Irp ) );
}

NTSTATUS Wdm_PowerFail( IRP *Irp, NTSTATUS Status ) {

        dbg0_printk("In Wdm_PowerFail\n");
        
        PoStartNextPowerIrp( Irp );
        OsIrpFinish( Irp, Status, 0 );
        return( Status );
}




// Windows Management Instrumentation (WMI) functions

// WCT: Support WMI here

NTSTATUS Wdm_DrvWmi( DEVICE_OBJECT *FDO, IRP *Irp ) {
        dbg1_printk("In Wdm_DrvWmi\n");
        return( STATUS_SUCCESS );
}




SRLOCAL NTSTATUS Wdm_OnRequestComplete( DEVICE_OBJECT *FDO,
                                        IRP *Irp,
                                        KEVENT *Event ) {
        
        dbg3_printk("In Wdm_OnRequestComplete\n");
        
        KeSetEvent( Event, 0, FALSE );
        return( STATUS_MORE_PROCESSING_REQUIRED );
}

SRLOCAL NTSTATUS Wdm_IrpBuildSendAndWait(
                        ULONG           Internal,
                        ULONG           IoControlCode,
                        PDEVICE_OBJECT  TargetDeviceObject,
                        PVOID           InputBuffer,
                        ULONG           InputBufferLength,
                        PVOID           OutputBuffer,
                        ULONG           OutputBufferLength,
                        PLARGE_INTEGER  RequestedTimeout
                        ) {

/*
    This routine builds and sends an Internal IOCTL to the TargetDeviceObject, waits
    for the IOCTL to complete, and returns status to the caller.

    *** This function does not support Input and Output in the same IOCTL

    Internal            - flag indicating if the IOCTL is internal or not
    IoControlCode       - the IOCTL to send
    TargetDeviceObject  - who to send the IOCTL to
    InputBuffer         - pointer to input buffer, if any
    InputBufferLength,  - length of input buffer
    OutputBuffer        - pointer to output buffer, if any
    OutputBufferLength, - length of output buffer
    Timeout             - how long to wait for request to complete,
                          NULL==use driver global AcquirePortTimeout

    All parameters are inputs except for OutputBuffer.
    InputBuffer, OutputBuffer, and RequestedTimeout are OPTIONAL.

    Returns    Status
*/
        
        KEVENT             Event;
        NTSTATUS           Status;
        PIRP               Irp;
        LARGE_INTEGER      Timeout;
        PIO_STACK_LOCATION IrpSp;
        BOOLEAN            NeedToCopyOutputBuffer = FALSE;
        ULONG              i;
        char              *pBuffIn, *pBuffOut;


//        PAGED_CODE();
        
        dbg3_printk("In Wdm_IrpBuildSendAndWait IOCTL 0x%X\n", IoControlCode);




        // Allocate and initialize IRP

        Irp = IoAllocateIrp( (CCHAR)(TargetDeviceObject->StackSize + 1), FALSE );
        if ( !Irp )
                return( STATUS_INSUFFICIENT_RESOURCES );

        IrpSp = IoGetNextIrpStackLocation( Irp );

        if (Internal)
                IrpSp->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
        else
                IrpSp->MajorFunction = IRP_MJ_DEVICE_CONTROL;

        IrpSp->Parameters.DeviceIoControl.OutputBufferLength = OutputBufferLength;
        IrpSp->Parameters.DeviceIoControl.InputBufferLength  = InputBufferLength;
        IrpSp->Parameters.DeviceIoControl.IoControlCode      = IoControlCode;



        // Set the system buffer to be the larger of the two buffers
        // If not using the input buffer, then copy the input data.

        if ( InputBufferLength >= OutputBufferLength ) {
                Irp->AssociatedIrp.SystemBuffer = InputBuffer;
                }
        else {
                Irp->AssociatedIrp.SystemBuffer = OutputBuffer;
                pBuffIn  = (char*)InputBuffer;
                pBuffOut = (char*)OutputBuffer;
                for ( i = 0 ; i < InputBufferLength ; i++ )
                        *pBuffOut++ = *pBuffIn++;
                }

        
        // Set completion routine and send IRP
        
        KeInitializeEvent( &Event, NotificationEvent, FALSE );
                
        IoSetCompletionRoutine( Irp,
                                (PIO_COMPLETION_ROUTINE)Wdm_OnRequestComplete,
                                (PVOID) &Event,
                                TRUE,
                                TRUE,
                                TRUE
                              );

        
        Status = IoCallDriver( TargetDeviceObject, Irp );
        if ( !NT_SUCCESS( Status ) ) {
                IoFreeIrp( Irp );
                return( Status );
                }



        // Set timeout in 100ns units and wait

        if (RequestedTimeout != NULL)
                Timeout = *RequestedTimeout;
        else
                Timeout.QuadPart = -((LONGLONG) 10*1000*500); // default 500 ms
        
        Status = KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, &Timeout );


        // If we timed out, cancel the IRP

        if ( Status == STATUS_TIMEOUT ) {
                IoCancelIrp( Irp );
                KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, NULL );
                }


        // Else the IRP is complete (or canceled).
        
        Status = Irp->IoStatus.Status;
        IoFreeIrp( Irp );


        // If using the input buffer as the system buffer, 
        // then copy the output data.

        if ( InputBufferLength >= OutputBufferLength ) {
                
                pBuffIn  = (char*)InputBuffer;
                pBuffOut = (char*)OutputBuffer;
                for ( i = 0 ; i < OutputBufferLength ; i++ )
                        *pBuffOut++ = *pBuffIn++ ;
                }
        
        
        return( Status );
}

SRLOCAL NTSTATUS Wdm_IrpForwardAndWait( DEVICE_OBJECT *FDO, IRP *Irp ) {

        DEVINFO *Dev;
        KEVENT   Event;

        dbg3_printk("In Wdm_IrpForwardAndWait\n");
        
        Dev = (DEVINFO *)FDO->DeviceExtension;

        // Setup event, Irp stack, and completion routine
        
        KeInitializeEvent( &Event, NotificationEvent, FALSE );
                
        IoCopyCurrentIrpStackLocationToNext( Irp );
        IoSetCompletionRoutine( Irp,
                                (PIO_COMPLETION_ROUTINE)Wdm_OnRequestComplete,
                                (PVOID) &Event,
                                TRUE,
                                TRUE,
                                TRUE
                              );

        
        // Pass Irp down and wait until lower object is done
        
        IoCallDriver( Dev->LowerDevObj, Irp );
        KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, NULL );
        
        return( Irp->IoStatus.Status );
}

SRLOCAL NTSTATUS Wdm_IrpForward( DEVICE_OBJECT *FDO, IRP *Irp ) {

        DEVINFO *Dev;

        dbg3_printk("In Wdm_IrpForward\n");
        
        Dev = (DEVINFO *)FDO->DeviceExtension;
        IoSkipCurrentIrpStackLocation( Irp );
        return( IoCallDriver( Dev->LowerDevObj, Irp ) );
}


SRLOCAL NTSTATUS WdmWinnt_ErrorLog( void *IoObject,
                                    IRP *Irp,
                                    NTSTATUS CurrStatus, 
                                    unsigned long ErrCode,
                                    const WCHAR *Msg ) {

        IO_ERROR_LOG_PACKET *errorLogEntry;
        IO_STACK_LOCATION   *IrpStack;
        unsigned long        CtrlCode;
        unsigned char        MajorFunction, len1, len2;

        dbg2_printk("In WdmWinnt_ErrorLog\n");
        dbg2_printk("Message is ls %ls\n", Msg);

        
        // Initialize entry variables
        
        CtrlCode      = 0;
        MajorFunction = 0;

        if (Irp) {
                IrpStack      = IoGetCurrentIrpStackLocation( Irp );
                MajorFunction = IrpStack->MajorFunction;
                if (MajorFunction == IRP_MJ_DEVICE_CONTROL)
                        CtrlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
                }



        // Calculate packet length including a NULL terminator for the string
        
        len1 = (unsigned char)( (wcslen( Msg ) + 1) * sizeof(WCHAR) );
        len2 = (unsigned char)( sizeof(IO_ERROR_LOG_PACKET) + len1 );
                       
        if (len2 > ERROR_LOG_MAXIMUM_SIZE) {
                len2 = ERROR_LOG_MAXIMUM_SIZE;
                len1 = len2 - sizeof(IO_ERROR_LOG_PACKET);
                if (len1 <= 0)
                        return( STATUS_NO_MEMORY );
                }



        
        // Allocate the packet
        
        errorLogEntry = IoAllocateErrorLogEntry( IoObject, len2 );

        if (!errorLogEntry)
                return( STATUS_NO_MEMORY );

        
        
        // Fill in the packet

        errorLogEntry->ErrorCode         = ErrCode;
        errorLogEntry->NumberOfStrings   = 1 ;
        errorLogEntry->StringOffset      = FIELD_OFFSET(
                                           IO_ERROR_LOG_PACKET, DumpData );
        errorLogEntry->MajorFunctionCode = MajorFunction;
        errorLogEntry->IoControlCode     = CtrlCode;
        errorLogEntry->UniqueErrorValue  = 0;
        errorLogEntry->SequenceNumber    = 0;
        errorLogEntry->FinalStatus       = CurrStatus;


        
        
        // Copy the message string to the packet
        
        RtlCopyMemory( &errorLogEntry->DumpData[0], Msg, len1 );



        // Write the packet to the event log

        IoWriteErrorLogEntry( errorLogEntry );


        
        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Wdm_GetParameters( DRIVER_OBJECT  *DriverObject,
                                    UNICODE_STRING *RegistryPath,
                                    UNICODE_STRING *pParGpsName ) {

        
        // Get PARGPS driver name parameter from the registry.
        
        NTSTATUS                 Status;
        WCHAR                   *queryPath;
        RTL_QUERY_REGISTRY_TABLE queryTable[2];  // Leave room for all 0's
        UNICODE_STRING           defaultName;


        dbg1_printk("In Wdm_GetParameters\n");


        RtlInitUnicodeString( &defaultName, L"SrParGps378" );


        // Generate wide character string having the \Parameters 
        // subkey registry path to query
        
        queryPath = ExAllocatePool( PagedPool, MAX_STR_LENGTH );
        
        if (queryPath == NULL)
                return( STATUS_NO_MEMORY );
        
        RtlZeroMemory( queryPath, MAX_STR_LENGTH );
        RtlMoveMemory( queryPath, RegistryPath->Buffer, RegistryPath->Length );
        wcscat( queryPath, L"\\Parameters" );

        

        // Set up table of parameters to query
        
        RtlZeroMemory( &queryTable[0], sizeof(queryTable) );

        queryTable[0].QueryRoutine  = NULL;
        queryTable[0].Flags         = RTL_QUERY_REGISTRY_DIRECT;
        queryTable[0].Name          = LREG_PARM_DRIVERNAME;
        queryTable[0].EntryContext  = pParGpsName;
        queryTable[0].DefaultType   = REG_SZ;
        queryTable[0].DefaultData   = &defaultName;
        queryTable[0].DefaultLength = 0; // compute value based on default

        queryTable[1].QueryRoutine  = NULL;  // marks the end of the table
        queryTable[1].Name          = NULL;


        // Do the query
        
        Status = RtlQueryRegistryValues(
                      RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
                      queryPath, &queryTable[0], NULL, NULL );


        // Set defaults if the query failed
        
        if ( !NT_SUCCESS(Status) ) {
                WdmWinnt_ErrorLog( DriverObject->DeviceObject, NULL, Status, 
                                   PARGPS_LOG_INFO, L"using default driver name parameter" );
                }
        
        ExFreePool( queryPath );

        
        return( STATUS_SUCCESS );
}


// *************************************************************************

#elif defined( SROS_WINNT )

#define SYMLINK_INVALID 0
#define SYMLINK_VALID   1

#define DEVICE_NAME_FORMAT_NT   "\\Device\\%s"
#define DEVICE_NAME_FORMAT_DOS  "\\??\\%s"


NTSTATUS DriverEntry( DRIVER_OBJECT *DriverObject,
                      UNICODE_STRING *RegistryPath ) {
        
        NTSTATUS          Status;
        char              DriverNameBuffer[MAX_STR_LENGTH];
        unsigned char     AnsiDriverBuffer[MAX_STR_LENGTH];
        WCHAR             ParGpsNameBuffer[MAX_STR_LENGTH];
        char             *DriverName;
        ANSI_STRING       AnsiDriverString;
        UNICODE_STRING    ParGpsName;
        UNICODE_STRING    NtDeviceName, Win32DeviceName;
        DEVICE_OBJECT    *DeviceObject;
        DEVINFO          *Dev;  // Device extension ie device information

        
        dbg0_printk("In PARGPS DriverEntry\n");

   
        // Set up dispatch table

        DriverObject->MajorFunction[IRP_MJ_CREATE]          = Winnt_DrvOpen;
        DriverObject->MajorFunction[IRP_MJ_CLOSE]           = Winnt_DrvClose;
        DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = Winnt_DrvIoctl;
        DriverObject->DriverUnload                          = Winnt_DrvUnload;

        
        // Prepare default name and get parameters from registry 

        swprintf( ParGpsNameBuffer, L"SrParGps378" );
        RtlInitUnicodeString( &ParGpsName, ParGpsNameBuffer );
        ParGpsName.MaximumLength = sizeof(ParGpsNameBuffer);
        
        Status = Winnt_GetParameters( DriverObject, RegistryPath, &ParGpsName );

        if ( !NT_SUCCESS( Status ) ) {
                DbgPrint("Winnt_GetParameters failed with status 0x%X\n", Status);
                return( Status );
                }


        // Generate a simple character string for the driver name
        //   1. Point to the driver name buffer
        //   2. Set up an empty counted ansi string using this buffer
        //   3. Fill it from the just gotten counted unicode string
        //      (ansi string still uses old ansi buffer)
        
        DriverName = DriverNameBuffer;
        
        AnsiDriverBuffer[0] = '\0';
        RtlInitAnsiString( &AnsiDriverString, DriverNameBuffer );
        AnsiDriverString.MaximumLength = sizeof(DriverNameBuffer);
        
        RtlUnicodeStringToAnsiString( &AnsiDriverString, &ParGpsName, FALSE);
        
        dbg0_printk("DriverName is %s\n", DriverName );


        
        // Create device

        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_NT, DriverName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &NtDeviceName, &AnsiDriverString, TRUE);
        
        Status = IoCreateDevice(
                                DriverObject,
                                sizeof(DEVINFO),
                                &NtDeviceName,
                                PARGPS_DEVICE_TYPE,
                                0,                 // Characteristics
                                FALSE,             // NOT Exclusive (> 1 ok)
                                &DeviceObject
                               );
        
        if ( !NT_SUCCESS( Status ) ) {
                DbgPrint("IoCreateDevice failed with status 0x%X\n", Status);
                Winnt_ResourceFree( DriverObject, &NtDeviceName,
                                    NULL, NULL, SYMLINK_INVALID );
                return( Status );
                }


        
        // Create symbolic link

        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_DOS, DriverName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &Win32DeviceName, &AnsiDriverString, TRUE );
        
        Status = IoCreateSymbolicLink( &Win32DeviceName, &NtDeviceName );
        
        if ( !NT_SUCCESS( Status ) ) {
                DbgPrint("IoCreateSymbolicLink failed with status 0x%X\n", Status);
                Winnt_ResourceFree( DriverObject, &NtDeviceName,
                                    DeviceObject, &Win32DeviceName,
                                    SYMLINK_INVALID );
                return( Status );
                }


        
        // Set device options for Read/Write data passing and alignment
        
        DeviceObject->Flags |= DO_DIRECT_IO;
        DeviceObject->AlignmentRequirement = FILE_BYTE_ALIGNMENT;


        
        // Initialize device information
        
        Dev = (DEVINFO *)DeviceObject->DeviceExtension;
        
        RtlZeroMemory( Dev, sizeof(DEVINFO) );
        ParGps_FillDevInfo( Dev, DriverName );

        Dev->DriverObj     = DriverObject;
        Dev->ThisFdoDevObj = DeviceObject;


        // Set up custom DPCs for PPS, Dready, and Plain types

        KeInitializeDpc(
                        &Dev->CustomDpc[PARGPS_INTTYPE_PPS], // PRKDPC
                        Winnt_DpcForPpsRoutine,              // PKDEFERRED_ROUTINE
                        DeviceObject                         // PVOID deferred context
                       );

        KeInitializeDpc(
                        &Dev->CustomDpc[PARGPS_INTTYPE_DREADY], // PRKDPC
                        Winnt_DpcForDreadyRoutine,              // PKDEFERRED_ROUTINE
                        DeviceObject                            // PVOID deferred context
                       );
        KeInitializeDpc(
                        &Dev->CustomDpc[PARGPS_INTTYPE_PLAIN], // PRKDPC
                        Winnt_DpcForPlainRoutine,              // PKDEFERRED_ROUTINE
                        DeviceObject                           // PVOID deferred context
                       );





        // Initialize Pps and Thread events to not signaled

        KeInitializeEvent( &Dev->PpsEvent,    NotificationEvent, FALSE );
        KeInitializeEvent( &Dev->ThreadEvent, NotificationEvent, FALSE );

        


        // Announce successful installation
        
        swprintf( Dev->MsgBuffer, L"no interrupts" );
        WdmWinnt_ErrorLog( DeviceObject, NULL, STATUS_SUCCESS, 
                           PARGPS_LOG_INFO_LOAD, Dev->MsgBuffer );

        

        // Free temporary unicode strings
        
        Winnt_ResourceFree( NULL, &NtDeviceName,
                            NULL, &Win32DeviceName,
                            SYMLINK_INVALID );


        return( Status );
}

void Winnt_DrvUnload( DRIVER_OBJECT *DriverObject ) {

        DEVINFO        *Dev;
        char           *DriverName;
        unsigned char   AnsiDriverBuffer[256];
        ANSI_STRING     AnsiDriverString;
        UNICODE_STRING  Win32DeviceName;

        dbg0_printk("In Winnt_DrvUnload\n");

        Dev = (DEVINFO *)DriverObject->DeviceObject->DeviceExtension;

        
        // Announce unloading driver

        swprintf( Dev->MsgBuffer, L"no interrupt" );
        WdmWinnt_ErrorLog( DriverObject->DeviceObject, NULL, STATUS_SUCCESS,
                           PARGPS_LOG_INFO_UNLOAD, Dev->MsgBuffer );




        // Prepare SymLink name string
        
        DriverName = Dev->DriverNameBuffer;

        sprintf( AnsiDriverBuffer, DEVICE_NAME_FORMAT_DOS, DriverName );
        RtlInitAnsiString( &AnsiDriverString, AnsiDriverBuffer );
        RtlAnsiStringToUnicodeString( &Win32DeviceName, &AnsiDriverString, TRUE );



        // Free device information area
        
        ParGps_DrvUnload( Dev );


        
        // Clean up symlink, device, name

        Winnt_ResourceFree( DriverObject, NULL,
                            DriverObject->DeviceObject, &Win32DeviceName,
                            SYMLINK_VALID );
        
}

NTSTATUS Winnt_DrvOpen( DEVICE_OBJECT *DeviceObject, IRP *Irp ) {

        return( (NTSTATUS)ParGps_DrvOpen(
                          (DEVINFO *)DeviceObject->DeviceExtension, Irp ) );
}

NTSTATUS Winnt_DrvClose( DEVICE_OBJECT *DeviceObject, IRP *Irp ) {

        return( (NTSTATUS)ParGps_DrvClose(
                          (DEVINFO *)DeviceObject->DeviceExtension, Irp ) );
}

NTSTATUS Winnt_DrvIoctl( DEVICE_OBJECT *DeviceObject, IRP *Irp ) {
        return( (NTSTATUS)ParGps_DrvIoctl(
                          (DEVINFO *)DeviceObject->DeviceExtension, Irp ) );
}

BOOLEAN Winnt_ParGpsInterrupt( PKINTERRUPT pInterrupt, void *Context ) {

        if (Context == NULL)
                return( FALSE );
        
        return( (BOOLEAN)ParGps_Interrupt( (DEVINFO *)Context ) );
}

void Winnt_DpcForPpsRoutine( KDPC *DpcForPps,
                             void *Context, // Device_Object
                             void *SysArg1, // Irp
                             void *DpcData  // Device Extension
                           ) {
        
        ParGps_DpcForPpsRoutine( (DEVINFO *)DpcData );
}

void Winnt_DpcForDreadyRoutine( KDPC *DpcForDready,
                                void *Context, // Device_Object
                                void *SysArg1, // Irp
                                void *DpcData  // Device Extension
                              ) {
        ParGps_DpcForDreadyRoutine( (DEVINFO *)DpcData );
}

void Winnt_DpcForPlainRoutine( KDPC *DpcForPlain,
                               void *Context, // Device_Object
                               void *SysArg1, // Irp
                               void *DpcData  // Device Extension
                             ) {
        ParGps_DpcForPlainRoutine( (DEVINFO *)DpcData );
}

void WdmWinnt_ThreadFunc( void *Context ) {

        dbg2_printk("In Winnt_ThreadFunc with Context 0x%p\n", Context );
        ParGps_ThreadFunc( (DEVINFO*)Context );
}


SRLOCAL NTSTATUS Winnt_GetParameters( DRIVER_OBJECT  *DriverObject,
                                      UNICODE_STRING *RegistryPath,
                                      UNICODE_STRING *pParGpsName ) {

        
        // Get PARGPS driver name parameters from the registry.
        
        NTSTATUS                 Status;
        WCHAR                   *queryPath;
        RTL_QUERY_REGISTRY_TABLE queryTable[2];  // Leave room for all 0's
        UNICODE_STRING           defaultName;

        dbg1_printk("In Winnt_GetParameters\n");


        RtlInitUnicodeString( &defaultName, L"SrParGps378" );


        // Generate wide character string having the \Parameters 
        // subkey registry path to query
        
        queryPath = ExAllocatePool( PagedPool, MAX_STR_LENGTH );
        
        if (queryPath == NULL)
                return( STATUS_NO_MEMORY );
        
        RtlZeroMemory( queryPath, MAX_STR_LENGTH );
        RtlMoveMemory( queryPath, RegistryPath->Buffer, RegistryPath->Length );
        wcscat( queryPath, L"\\Parameters" );

        

        // Set up table of parameters to query
        
        RtlZeroMemory( &queryTable[0], sizeof(queryTable) );

        queryTable[0].QueryRoutine  = NULL;
        queryTable[0].Flags         = RTL_QUERY_REGISTRY_DIRECT;
        queryTable[0].Name          = LREG_PARM_DRIVERNAME;
        queryTable[0].EntryContext  = pParGpsName;
        queryTable[0].DefaultType   = REG_SZ;
        queryTable[0].DefaultData   = &defaultName;
        queryTable[0].DefaultLength = 0; // compute value based on default

        queryTable[1].QueryRoutine  = NULL;  // marks the end of the table
        queryTable[1].Name          = NULL;


        // Do the query
        
        Status = RtlQueryRegistryValues(
                      RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
                      queryPath, &queryTable[0], NULL, NULL );


        // Set defaults if the query failed
        
        if ( !NT_SUCCESS(Status) ) {
                WdmWinnt_ErrorLog( DriverObject->DeviceObject, NULL, Status, 
                                   PARGPS_LOG_INFO, L"using default driver name parameter" );
                }
        
        ExFreePool( queryPath );

        
        return( STATUS_SUCCESS );
}

SRLOCAL NTSTATUS Winnt_SynchCompletionRoutine( DEVICE_OBJECT *DeviceObject,
                                               IRP *Irp,
                                               KEVENT *Event
                                             )
{

//    This routine is for use with synchronous IRP processing.
//    All it does is signal an event, so the driver knows it
//    can continue.
//
//    DriverObject - Pointer to driver object created by system.
//    Irp          - Irp that just completed
//    Event        - Event we'll signal to say Irp is done

    
    KeSetEvent( Event, 0, FALSE );
    return( STATUS_MORE_PROCESSING_REQUIRED );
}

SRLOCAL NTSTATUS Winnt_SendIoctl( unsigned long   IoControlCode,
                                  DEVICE_OBJECT  *TargetDeviceObject,
                                  void           *InputBuffer,
                                  unsigned long   InputBufferLength,
                                  void           *OutputBuffer,
                                  unsigned long   OutputBufferLength,
                                  LARGE_INTEGER  *RequestedTimeout
                                )
{

// This routine builds and sends an IOCTL to the
// TargetDeviceObject, waits for the IOCTL to complete, and
// returns status to the caller.

// ** Note: Does not support Input and Output in the same IOCTL.

//    IoControlCode       - the IOCTL to send
//    TargetDeviceObject  - who to send the IOCTL to
//    InputBuffer         - pointer to input buffer, if any
//    InputBufferLength,  - length of input buffer
//    OutputBuffer        - pointer to output buffer, if any
//    OutputBufferLength, - length of output buffer
//    Timeout             - how long to wait for request to complete,
//                          NULL==use driver global AcquirePortTimeout

        NTSTATUS           Status;
        PIRP               irp;
        LARGE_INTEGER      timeout;
        KEVENT             event;
        PIO_STACK_LOCATION irpSp;
        BOOLEAN            needToCopyOutputBuffer = FALSE;
        LARGE_INTEGER      AcquirePortTimeout;

//        PAGED_CODE();

//        DbgPrint("Winnt_SendInternalIoctl: Enter 0x%X\n", IoControlCode);


        AcquirePortTimeout.QuadPart = -((LONGLONG) 10*1000*500); // 500 ms (timeouts in 100ns units)

        
        //
        // Current limitation is that this function does not handle a request 
        //   with both InputBufferLength and OutputBufferLength > 0
        //
        if ( InputBufferLength != 0 && OutputBufferLength != 0 ) {
                return( STATUS_INVALID_DEVICE_REQUEST );
                }


        //
        // Allocate and initialize IRP
        //
        irp = IoAllocateIrp( (CCHAR)(TargetDeviceObject->StackSize + 1), FALSE );
        if ( !irp ) {
                return( STATUS_INSUFFICIENT_RESOURCES );
                }

        
        irpSp = IoGetNextIrpStackLocation( irp );

        irpSp->MajorFunction = IRP_MJ_DEVICE_CONTROL;

        irpSp->Parameters.DeviceIoControl.OutputBufferLength = OutputBufferLength;
        irpSp->Parameters.DeviceIoControl.InputBufferLength  = InputBufferLength;
        irpSp->Parameters.DeviceIoControl.IoControlCode      = IoControlCode;


        if ( InputBufferLength != 0 ) {
                irp->AssociatedIrp.SystemBuffer = InputBuffer;
                }
        else if ( OutputBufferLength != 0 ) {
                irp->AssociatedIrp.SystemBuffer = OutputBuffer;
                }


        //
        // Set completion routine and send IRP
        //
        KeInitializeEvent( &event, NotificationEvent, FALSE );
        IoSetCompletionRoutine( irp, Winnt_SynchCompletionRoutine, &event, TRUE, TRUE, TRUE );

        Status = IoCallDriver( TargetDeviceObject, irp );

        if ( !NT_SUCCESS( Status ) ) {
                DbgPrint("Winnt_SendInternalIoctl: IoCallDriver FAILED - status=%x\n",Status);
                IoFreeIrp( irp );
                return( Status );
                }

        //
        // Set timeout and wait
        //
        //                                      user specified   : default
        if (RequestedTimeout)
                timeout = *RequestedTimeout;
        else
                timeout = AcquirePortTimeout;
        Status = KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, &timeout);

        //
        // Did we timeout or did the IRP complete?
        //
        if ( Status == STATUS_TIMEOUT ) {
                // we timed out - cancel the IRP
                IoCancelIrp( irp );
                KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
                }

        //
        // Irp is complete, grab the status and free the irp
        //
        Status = irp->IoStatus.Status;
        IoFreeIrp( irp );

        return( Status );
}




SRLOCAL NTSTATUS WdmWinnt_ErrorLog( void *IoObject,
                                    IRP *Irp,
                                    NTSTATUS CurrStatus, 
                                    unsigned long ErrCode,
                                    const WCHAR *Msg ) {

        IO_ERROR_LOG_PACKET *errorLogEntry;
        IO_STACK_LOCATION   *IrpStack;
        unsigned long        CtrlCode;
        unsigned char        MajorFunction, len1, len2;

        dbg2_printk("In WdmWinnt_ErrorLog\n");
        dbg2_printk("Message is ls %ls\n", Msg);

        
        // Initialize entry variables
        
        CtrlCode      = 0;
        MajorFunction = 0;

        if (Irp) {
                IrpStack      = IoGetCurrentIrpStackLocation( Irp );
                MajorFunction = IrpStack->MajorFunction;
                if (MajorFunction == IRP_MJ_DEVICE_CONTROL)
                        CtrlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
                }



        // Calculate packet length including a NULL terminator for the string
        
        len1 = (unsigned char)( (wcslen( Msg ) + 1) * sizeof(WCHAR) );
        len2 = (unsigned char)( sizeof(IO_ERROR_LOG_PACKET) + len1 );
                       
        if (len2 > ERROR_LOG_MAXIMUM_SIZE) {
                len2 = ERROR_LOG_MAXIMUM_SIZE;
                len1 = len2 - sizeof(IO_ERROR_LOG_PACKET);
                if (len1 <= 0)
                        return( STATUS_NO_MEMORY );
                }



        
        // Allocate the packet
        
        errorLogEntry = IoAllocateErrorLogEntry( IoObject, len2 );

        if (!errorLogEntry)
                return( STATUS_NO_MEMORY );

        
        
        // Fill in the packet

        errorLogEntry->ErrorCode         = ErrCode;
        errorLogEntry->NumberOfStrings   = 1 ;
        errorLogEntry->StringOffset      = FIELD_OFFSET(
                                           IO_ERROR_LOG_PACKET, DumpData );
        errorLogEntry->MajorFunctionCode = MajorFunction;
        errorLogEntry->IoControlCode     = CtrlCode;
        errorLogEntry->UniqueErrorValue  = 0;
        errorLogEntry->SequenceNumber    = 0;
        errorLogEntry->FinalStatus       = CurrStatus;


        
        
        // Copy the message string to the packet
        
        RtlCopyMemory( &errorLogEntry->DumpData[0], Msg, len1 );



        // Write the packet to the event log

        IoWriteErrorLogEntry( errorLogEntry );


        
        return( STATUS_SUCCESS );
}


SRLOCAL void Winnt_ResourceFree( DRIVER_OBJECT *DriverObject,
                                 UNICODE_STRING *NtDeviceName,
                                 DEVICE_OBJECT *DeviceObject,
                                 UNICODE_STRING *Win32DeviceName,
                                 int SymLink
                               ) {

        dbg0_printk("In Winnt_ResourceFree\n");

        if (SymLink == SYMLINK_VALID)
                IoDeleteSymbolicLink( Win32DeviceName );
                
        if (Win32DeviceName)
                RtlFreeUnicodeString( Win32DeviceName );

        if (DeviceObject)
                IoDeleteDevice( DeviceObject );

        if (NtDeviceName)
                RtlFreeUnicodeString( NtDeviceName );

}




// *************************************************************************

#elif defined( SROS_LINUX )


// Allow these variables to be overridden from the insmod command line.

MODULE_DESCRIPTION("Symmetric Research PARGPS Linux Device Driver");
MODULE_LICENSE("GPL"); // Proprietary is other option
MODULE_PARM( pargps_name,  "s" );
MODULE_PARM( pargps_major, "i" );

SRLOCAL char *pargps_name  = NULL;
SRLOCAL int   pargps_major = -1;

// Dispatch function table

SRLOCAL struct file_operations MajorFunction = {
        open:    Linux_DrvOpen,
        release: Linux_DrvClose,
        ioctl:   Linux_DrvIoctl,
        };

int Linux_ThreadFunc( void *Context ) {

        dbg2_printk("In Linux_ThreadFunc with Context 0x%p\n", Context );
        
        return( (int)ParGps_ThreadFunc( (DEVINFO*)Context ) );
}

SRLOCAL void Linux_GetParameters( void ) {
        
        // Set global parameters from defaults if not already
        // set from command line values

        if (pargps_name == NULL)
                pargps_name = DEFAULT_PARGPS_NAME;

        if (pargps_major == -1)
                pargps_major = DEFAULT_PARGPS_MAJOR;

}


SRLOCAL void Linux_ResourceFree( int ParGpsMajor, char *ParGpsName ) {

        dbg0_printk( "Linux_ResourceFree\n" );
        
        
        // Unregister the character device

        if (ParGpsName) {
                unregister_chrdev( ParGpsMajor, ParGpsName );
                dbg0_printk( "Unregistering char device = %d\n", ParGpsMajor );
                ParGpsName = NULL;
                ParGpsMajor = -1;
                }

}

int Linux_DrvOpen( struct inode *inode, struct file *filp ) {

        filp->private_data = &ParGpsDeviceInfo;
        return( ParGps_DrvOpen( filp->private_data, NULL ) );
}

close_t Linux_DrvClose( struct inode *inode, struct file *filp ) {

        ParGps_DrvClose( filp->private_data, NULL );

        close_return( 0 );
}

int Linux_DrvIoctl( struct inode *inode, struct file *filp,
                    unsigned int cmd, unsigned long arg ) {

        return( ParGps_DrvIoctl( filp->private_data, (IRP*)arg ) );
}


void cleanup_module( void ) {

        dbg0_printk( "cleanup_module\n" );

        
        // Free device information area

        ParGps_DrvUnload( &ParGpsDeviceInfo );


        // Clean up device
        
        Linux_ResourceFree( pargps_major, pargps_name );
}

int init_module( void ) {

        int Status;
        DEVINFO *Dev;

        dbg0_printk( "init_module\n" );

        EXPORT_NO_SYMBOLS;


        // Set up dispatch table
        // This is done above by initializing the MajorFunction structure.



        // Get parameters from command line values or defaults
        
        Linux_GetParameters( );




        
        // Register character device

        Status = register_chrdev( pargps_major, pargps_name, &MajorFunction );

        if (Status < 0) {
                dbg0_printk( "PARGPS: Can't get major %d\n", pargps_major );
                return( STATUS_DEVICE_CONFIGURATION_ERROR );
                }

        if (pargps_major == 0)             // Allow dynamic allocation
                pargps_major = Status;
        
        dbg0_printk( "Registered as char device %d\n", pargps_major );


        // Fill device information

        ParGps_FillDevInfo( &ParGpsDeviceInfo, pargps_name );
        Dev = (DEVINFO *)&ParGpsDeviceInfo;



        // Set up custom DPCs for PPS, Dready, and Plain types
        
        INIT_TQUEUE( (&Dev->CustomDpc[PARGPS_INTTYPE_PPS]),
                                                Linux_DpcForPpsRoutine,    Dev );
        INIT_TQUEUE( (&Dev->CustomDpc[PARGPS_INTTYPE_DREADY]), 
                                                Linux_DpcForDreadyRoutine, Dev );
        INIT_TQUEUE( (&Dev->CustomDpc[PARGPS_INTTYPE_PLAIN]),  
                                                Linux_DpcForPlainRoutine,  Dev );

        return( 0 ); // success
}


// The interrupt handler.

void Linux_ParGpsInterrupt( int irq, void *dev_id, struct pt_regs *regs ) {

        ParGps_Interrupt( (DEVINFO *)dev_id );
}
        
void Linux_DpcForPpsRoutine( void *DpcData ) {

        ParGps_DpcForPpsRoutine( (DEVINFO *)DpcData );
}        

void Linux_DpcForDreadyRoutine( void *DpcData ) {
        
        ParGps_DpcForDreadyRoutine( (DEVINFO *)DpcData );
}        

void Linux_DpcForPlainRoutine( void *DpcData ) {
        
        ParGps_DpcForPlainRoutine( (DEVINFO *)DpcData );
}        

#endif // SROS_xxxxx  End of top level OS specific functions
