// FILE: rmdriver.c                Copyright (c), Symmetric Research, 2003
//
// Program to remove PARGPS kernel mode driver.
//
// This is done by starting the Service Control Manager, opening,
// stopping, and deleting the driver service.  On deletion, driver
// information that was added to the registry under the driver's key is
// also deleted.  Additional registry entries, added to allow PARGPS driver
// messages to be properly interpreted by the system log EventViewer,
// are not deleted so that existing system log messages can still be
// deciphered.
//
// Refer to the readme.txt file for information about how to remove the
// device driver manually.
//
//
//

#include <windows.h>
#include <stdio.h>




// Include the DriverInfo[] array with allowed driver names:

#include "indriver.h"
DRIVER_INFO_TABLE *Driver;




// Helper function protos:

void RemoveEventLogRegistryEntries( void );

void PrintMsg( char *MsgArray[] );
void Usage( void );
void Error( char *Msg, int Fatal );



// Global variables shared by main and helper functions:

SC_HANDLE SCManager, SCService;





// Usage and Error message arrays:

char* UsageMsg[] = {

  "                                                                      ",
  "ERROR: Command Line Usage: rmdriver ParGpsDriverName                  ",
  "                                                                      ",
  "This utility removes the PARGPS driver.  To see what drivers are      ",
  "currently installed, click the Windows NT Devices Applet from the     ",
  "Control Panel or under Linux run /sbin/lsmod.                         ",
  "                                                                      ",
  "ParGpsDriverName must be a name from the following list:              ",
  "                                                                      ",
  
  NULL
  };

char *ErrorInvalidServiceControlMsg[] = {

   "This error most likely indicates that the PARGPS device driver       ",
   "could not be stopped because it was not currently running.  Removal  ",
   "continues.                                                           ",

   NULL
   };
        
char *ErrorServiceDoesNotExistMsg[] = {

  "                                                                     ",
  "The PARGPS device driver must be installed before it can be removed. ",
  "Use indriver.exe to install it.                                      ",

  NULL
  };
                


// Main:

void main( int argc, char *argv[] ) {

        BOOL           Ret;
        SERVICE_STATUS ServiceStatus;


        
        // Banner message

        printf( "PARGPS Kernel Mode Driver Removal Program : Rev {09/30/03}\n");


        
        // Process command line arguments for driver name
        
        if ( argc != 2 ) 
                Usage();

        Driver = DriverInfo;

        while ( Driver->Name )  {
                if (strcmp(Driver->Name, argv[1]) == 0)
                        break;
                Driver++;
                }
        
        if (Driver->Name == NULL)
                Usage();




        // Open Service Control Manager

        SCManager = NULL;
        SCService = NULL;

        SCManager = OpenSCManager(
                        NULL,                  // use local machine
                        NULL,                  // use SERVICES_ACTIVE_DATABASE
                        SC_MANAGER_ALL_ACCESS  // full access permission
                        );

        if (!SCManager)
                Error("Could not create service manager", 1);
        



        // Open the device driver service
        
        SCService = OpenService(
                        SCManager,             // Service Control Manager
                        Driver->Name,          // name of device driver
                        SERVICE_ALL_ACCESS     // full access permission
                        );

        if (!SCService)
                Error("Could not open device driver service", 1);




        // Stop Service if not already stopped

        Ret = QueryServiceStatus(
                                 SCService,          // device driver service
                                 &ServiceStatus      // status
                                );
        
        if (!Ret)
                Error("Could not query device driver status", 0);
        
        else if (ServiceStatus.dwCurrentState != SERVICE_STOPPED) {

                Ret = ControlService(
                             SCService,              // device driver service
                             SERVICE_CONTROL_STOP,   // request STOP
                             &ServiceStatus          // status
                            );

                if (!Ret)
                        Error("Could not stop device driver", 0);
                else
                        printf( "\n%s driver has been stopped.\n", Driver->Name );
                }

        
        
        // Delete Service
        //
        // This function both removes the device driver from the
        // services database and removes the driver's registry key.
        // Because the entire key is deleted, both the registry entries
        // automatically generated by the CreateService call AND the
        // explicitly created Parameters subkey registry entries are
        // deleted.

        Ret = DeleteService( SCService );

        if (!Ret)
                Error("Could not delete device driver", 1);
        else
                printf( "%s driver has been successfully removed.\n",
                          Driver->Name );


        
        // Close handles

        CloseServiceHandle( SCService );
        CloseServiceHandle( SCManager );


        // By default, the code below is compiled.  However, if you wish to
        // have rmdriver.exe not automatically delete the EventLog registry
        // entries, you can compile with the KEEP_EVENTLOG_REGISTRY_ENTRIES
        // macro definition.


#if ! defined( KEEP_EVENTLOG_REGISTRY_ENTRIES )
        
        RemoveEventLogRegistryEntries();
#endif        


}





// Helper functions:
        
void RemoveEventLogRegistryEntries( void ) {

        // Remove associated registry entries
        //
        // Do not delete EventLog registry entries explicitly
        // created with indriver.exe.  This is so that System Log 
        // entries relating to the PARGPS driver can still be 
        // interpreted even when the driver is no longer running.  
        // Except for making the System Log entries less informative,
        // deleting these registry entries by hand produces no bad effects.
        //
        // HKEY_LOCAL_MACHINE\System\CurrentControlSet\Services\EventLog\
        //                                                     System\SrParGpsXX
        //   EventMessageFile = C:\SR\PARGPS\DRIVER\pargpskd.sys;SystemRoot%\System32\IoLogMsg.dll
        //   TypesSupported   = 0x7L;
        //
        
        LONG    Code;
        HKEY    ParentKey;
        LPCTSTR ParentKeyName;

        ParentKeyName = "System\\CurrentControlSet\\Services\\EventLog\\System";
                               
        Code = RegOpenKeyEx(
                            HKEY_LOCAL_MACHINE,      // top level key
                            ParentKeyName,           // event log parent key
                            0,                       // reserved
                            KEY_ALL_ACCESS,          // full access permission
                            &ParentKey               // handle to key
                           );

        if (Code != ERROR_SUCCESS)
                Error("Could not open EventLog parent key", 1);


        Code = RegDeleteKey(
                            ParentKey,    // EventLog system file key
                            Driver->Name  // driver subkey to delete
                           );
        
        if (Code != ERROR_SUCCESS)
                Error("Could not delete SrParGps EventLog subkey", 0);

        
        Code = RegCloseKey( ParentKey );
        
        if (Code != ERROR_SUCCESS)
                Error("Could not close EventLog parent key", 1);
}

void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );

}

void Usage( void ) {

        int i;
        
        PrintMsg( UsageMsg );
                
        for ( i = 0 ; DriverInfo[i].Name ; i++ )
                printf( "     %s\n", DriverInfo[i].Name );
        
        exit( 1 );

}

void Error( char *Msg, int Fatal ) {

        DWORD Err;
        LPVOID lpMessageBuffer;
        
        // Display local message, error number, and system message
        
        Err = GetLastError();

        FormatMessage(
                      FORMAT_MESSAGE_ALLOCATE_BUFFER |
                      FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      Err,
                      MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                      (LPTSTR) &lpMessageBuffer,
                      0,
                      NULL
                     );

        printf( "\n%s.  Error #%d\n\n%s\n", Msg, Err, lpMessageBuffer );

        LocalFree( lpMessageBuffer );



        
        // Give additional explanation for common errors
        
        if (Err == ERROR_INVALID_SERVICE_CONTROL) // #1052
                PrintMsg( ErrorInvalidServiceControlMsg );

        else if (Err == ERROR_SERVICE_DOES_NOT_EXIST) // #1060
                PrintMsg( ErrorServiceDoesNotExistMsg );
        else
                printf( "Sorry, no extra help available for this error.\n" );



        // Clean up and exit if fatal.

        if ( Fatal ) {

                if (SCService)  CloseServiceHandle( SCService );
                if (SCManager)  CloseServiceHandle( SCManager );

                exit( 1 );

                }
}
