/* FILE: pargps.c             Copyright (c), Symmetric Research, 2003-2004
 *
 * PARGPS user library functions.
 * 
 */

// OS dependent includes ...

#if defined( SROS_WIN2K ) || defined( SROS_WINNT )
#include <windows.h>     // windef.h -> winnt.h -> defines GENERIC_READ etc
#include <winioctl.h>    // device driver I/O control macros
#include <process.h>     // for thread info
#include <errno.h>       // for errno values


#elif defined( SROS_LINUX )
#include <errno.h>       // errno global
#include <fcntl.h>       // open  function, fcntl, F_ and O_ defines
#include <signal.h>      // for signal processing
#include <string.h>      // for strncopy, strncmp, strcmp
#include <termios.h>     // for terminal io structure and defines
#include <unistd.h>      // close, access, fsync fns and STDIN_FILENO define
#include <sys/ioctl.h>   // ioctl function and macros and TIOCM defines

typedef int BOOLEAN;
#define FALSE 0
#define TRUE  1

#endif  // SROS_xxxxx


// Common includes

#include <stdio.h>
#include <time.h>

#include "pargps.h"
#include "pargpskd.h"




// This define controls whether the PARGPS functions are static or visible
// to other modules.  The latter case is typically only required for
// debugging or with programs like PARGPS diag that require special access.

#if !defined( SRLOCAL )
#define SRLOCAL static
#endif



// Global variables

unsigned long ParGpsLastDriverError = 0;
char ParGpsLastErrorMessage[1024];

#define MAXCIRC 100
TIMESTAMP CircBuff[MAXCIRC];
int CbNumValid;



// These signal length defines indicate the number of seconds between
// the initial rising edge and the final falling edge of the signal in
// question.  They are used when calculating the GPS time of the first
// data sample after the PPS signal arrives.  They account for the fact
// that, due to the way the PARxCH circuitry and firmware is setup, the
// PC is interrupted on the falling (not rising) edge of the PPS and
// Dready signals which are of different length.

#define SIGNAL_LENGTH_PPS         0.000010L  // 10 microseconds
#define SIGNAL_LENGTH_DREADY      0.000052L  // 52 microseconds
#define SIGNAL_LENGTH_CORRECTION (SIGNAL_LENGTH_DREADY - SIGNAL_LENGTH_PPS)


// Use for comparing doubles

#define TOLERANCE 0.00001



// Note: The NMEA message IDs are listed in the order in which they come out.
// Be sure these match up with the NMEA_MSGID defines in pargps.h.

VARTYPE( char ) *NmeaMsgId[] = { "ZDA","GGA","GLL","VTG","GSA","GSV","RMC" };



// Actual allocation of PARGPS error string array, see pargps.h ...

VARTYPE( char ) *PARGPS_ERROR_MSG[] = {

        PARGPS_ERROR_MSG_NONE,
        PARGPS_ERROR_MSG_NOT_AVAILABLE,
        PARGPS_ERROR_MSG_INTERRUPT_NUMBER,
        PARGPS_ERROR_MSG_INTERRUPT_MODE,
        PARGPS_ERROR_MSG_SERIAL_PORT_NUMBER,
        PARGPS_ERROR_MSG_SERIAL_PORT_NOT_OPEN,
        PARGPS_ERROR_MSG_DRIVER_NOT_OPEN,
        PARGPS_ERROR_MSG_DRIVER_REQUEST_FAILED,
        PARGPS_ERROR_MSG_KERNEL_AREA,
        PARGPS_ERROR_MSG_KEYPRESS,
        PARGPS_ERROR_MSG_BAD_PARAMETER,
        PARGPS_ERROR_MSG_MAX

        };


#if defined( SROS_LINUX )
DEVHANDLE Signal_hSerial;     // for signal handler
DEVHANDLE save_serial_handle; // for closing serial
#endif




// Prototypes for private PARGPS helper and OS dependent functions
// Most users should only use the standard PARGPS functions which are
// declared in pargps.h.

FUNCTYPE( int ) ParGpsIoctl( DEVHANDLE ParGpsHandle, unsigned long IoCtlCode,
                          void *pValueIn, unsigned long InSize,
                          void *pValueOut, unsigned long OutSize,
                          unsigned long *ReturnSize );
FUNCTYPE( int ) ParGpsInit( DEVHANDLE ParGpsHandle, int SerialPort, int InterruptMode );
FUNCTYPE( int ) ParGpsSerialSetup( DEVHANDLE ParGpsHandle, int PortNumber );
FUNCTYPE( int ) ParGpsSetInterruptMode( DEVHANDLE ParGpsHandle, int Mode );
FUNCTYPE( int ) ParGpsReset( DEVHANDLE ParGpsHandle );
FUNCTYPE( int ) ParGpsGetInterruptCount( DEVHANDLE ParGpsHandle, int *Count );
FUNCTYPE( int ) ParGpsExtractNsat( NMEAPARSE *ParseData, int *Nsat, int *Source );
FUNCTYPE( int ) ParGpsExtractHMS( NMEAPARSE *ParseData,
                               int *Hour, int *Minute, int *Secone,
                               long *MicroSecond, int *Source );
FUNCTYPE( int ) ParGpsExtractYMD( NMEAPARSE *ParseData, 
                               int *Year, int *Mon, int *Day, int *Source );
FUNCTYPE( int ) ParGpsExtractLatLonAlt( NMEAPARSE *ParseData,
                                     double *Latitude, double *Longitude,
                                     double *Altitude, int *Source );
        
SRLOCAL int ParGpsInitCircBuff( void );
SRLOCAL int ParGpsSkipCircBuff( void );
SRLOCAL int ParGpsReadCircBuff( TIMESTAMP *Buffer );
SRLOCAL int ParGpsAnalogWriteCircBuff( unsigned int PpsEvent, long Sample, int Obc );
SRLOCAL int ParGpsPpsWriteCircBuff( unsigned int PpsEvent, SR64BIT CountAtPps,
                                    SR64BIT CountAtDready );
SRLOCAL int ParGpsSerialWriteCircBuff( unsigned int PpsEvent, double SecSince1970,
                                       int YMDSource, int HMSSource, int Nsat, int Erase );
SRLOCAL int ParGpsIncCbPtr( int *PtrIndex );
SRLOCAL int ParGpsUpdateCbPtrForSkip( int *PtrIndex );
SRLOCAL int ParGpsCheckEventNum( unsigned int PpsEvent, int *wrptrAPS, char *StrAPS );
SRLOCAL int ParGpsSrcToSource( int YmdSrc, int HmsSrc, int *YmdSource, int *HmsSource );
SRLOCAL int ParGpsEquation2pt( double x1, double y1,
                               double x2, double y2, 
                               double xk, double *yk );
SRLOCAL int ParGpsEquationSlope( double m,
                                 double x1, double y1,
                                 double xk, double *yk );


SRLOCAL DEVHANDLE ParGpsOsDriverOpen( char *DriverName );
SRLOCAL int ParGpsOsDriverClose( DEVHANDLE *ParGpsHandle );
SRLOCAL int ParGpsOsDriverIoctl( DEVHANDLE ParGpsHandle, unsigned long IoCtlCode,
                           void *pValueIn, unsigned long InSize,
                           void *pValueOut, unsigned long OutSize,
                           unsigned long *pBytesReturned );
SRLOCAL long ParGpsOsGetLastError( void );
SRLOCAL DEVHANDLE ParGpsOsSerialOpen( char *SerialPortName );
SRLOCAL int ParGpsOsSerialInit( DEVHANDLE hSerial );
SRLOCAL void ParGpsOsSerialClose( DEVHANDLE hSerial );
SRLOCAL int ParGpsOsSetPcTime( double Time,
                               int Year, int Month, int Day, 
                               int Hour, int Minute, int Second,
                               int FromYMD );
SRLOCAL long ParGpsOsLargeIntDiff( SR64BIT A, SR64BIT B );





/* RELEASE REV INFO:

This function allows the user to determine the library rev they are
working with.

Short rev history:

  PARGPS_REV  100  ( 07/25/2002 )  < first beta release
  PARGPS_REV  101  ( 08/25/2002 )  < next  beta release
  PARGPS_REV  102  ( 09/05/2002 )  < added ChangeDate to Inc/Dec time funcs
  PARGPS_REV  103  ( 04/15/2003 )  < reorg + added CurrentTime function
  PARGPS_REV  104  ( 06/15/2003 )  < added NMEA ZDA string YMD support
  PARGPS_REV  105  ( 08/05/2003 )  < changed NMEA reading + wait timeout in driver
  PARGPS_REV  106  ( 09/10/2003 )  < Added Gps prefix to Os functions
  PARGPS_REV  107  ( 09/30/2003 )  < Have GetRev return value
  PARGPS_REV  108  ( 11/15/2003 )  < Added function to get PC time
  PARGPS_REV  109  ( 02/01/2004 )  < Added error checking for bad serial open
  PARGPS_REV  110  ( 03/20/2004 )  < Added PARPGS LabView support
  PARGPS_REV  111  ( 05/30/2004 )  < Added SetPcTime + chg SecTime split/combine
  PARGPS_REV  112  ( 10/31/2004 )  < Added timestamp functions
  PARGPS_REV  113  ( 11/15/2004 )  < Included timestamp init function

*/

#define PARGPS_REV      113

FUNCTYPE( int ) ParGpsGetRev( int *Rev ) {

        *Rev = PARGPS_REV;
        return( *Rev );
}







/*************************************************************************** 
 *
 * Main Library Functions:
 *
 * These are listed in pargps.h and can be called by all users to control
 * the PARGPS board.  See pargps.h for a description of the typical calling
 * sequence.  Study the PARxCH simp.c acquisition program and the PARGPS
 * Visual Basic demo program for examples of differnt ways in which 
 * these functions can be used.
 *
 *
 */

#define ERROR_RETURN( err ) {   if ( Error )              \
                                        *Error = err;     \
                                return( BAD_DEVHANDLE );  \
                                }

FUNCTYPE( DEVHANDLE ) ParGpsOpen( char *ParGpsName, int SerialPortNumber, int *Error ) {

        DEVHANDLE ParGpsHandle;
        int       ret;

        ParGpsLastErrorMessage[0] = '\0';


        // Error check the inputs

        if ( (SerialPortNumber < 1) || (SerialPortNumber > 8) )
                ERROR_RETURN( PARGPS_ERROR_SERIAL_PORT_NUMBER );



        // Open device driver

        ParGpsHandle = ParGpsOsDriverOpen( ParGpsName );

        if ( ParGpsHandle == BAD_DEVHANDLE ) {
                ParGpsLastDriverError = ParGpsOsGetLastError();
                ERROR_RETURN( PARGPS_ERROR_DRIVER_NOT_OPEN );
                }

        ParGpsLastDriverError = 0;



        // Setup serial port baud rate, etc

        ret = ParGpsSerialSetup( ParGpsHandle, SerialPortNumber );
        if (!ret) {
                ParGpsOsDriverClose( &ParGpsHandle );
                ERROR_RETURN( PARGPS_ERROR_SERIAL_PORT_NOT_OPEN );
                }


        // Initialize driver by seting 
        //    serial port, typically 1 or 2, and
        //    interrupt mode to INTERRUPT_ALTERNATING

        ret = ParGpsInit( ParGpsHandle, SerialPortNumber, INTERRUPT_ALTERNATING );
        if (!ret) {
                sprintf( ParGpsLastErrorMessage,
                         "Could not init serial port,interrupt mode\n");
                ParGpsOsDriverClose( &ParGpsHandle );
                ERROR_RETURN( PARGPS_ERROR_SERIAL_PORT_NOT_OPEN );
                }


        // Initialize circular time stamp buffer
        
        ParGpsInitCircBuff();


        return( ParGpsHandle );
}

FUNCTYPE( DEVHANDLE ) ParGpsOpenForTimeOnly( char *ParGpsName, int *Error ) {

        DEVHANDLE ParGpsHandle;


        // Open device driver

        ParGpsHandle = ParGpsOsDriverOpen( ParGpsName );

        if ( ParGpsHandle == BAD_DEVHANDLE ) {
                ParGpsLastDriverError = ParGpsOsGetLastError();
                ERROR_RETURN( PARGPS_ERROR_DRIVER_NOT_OPEN );
                }

        ParGpsLastDriverError = 0;


        return( ParGpsHandle );
}

FUNCTYPE( int ) ParGpsClose( DEVHANDLE ParGpsHandle ) {

        // Only close open drivers

        if (ParGpsHandle == BAD_DEVHANDLE) {
                ParGpsLastDriverError = ERROR_SERVICE_DISABLED;
                return( 0 );
                }


#if defined( SROS_WIN2K ) || defined( SROS_WINNT )
        // Already closed earlier
#elif defined( SROS_LINUX )
        ParGpsOsSerialClose( save_serial_handle );
#endif


        // Close driver

        if ( !ParGpsOsDriverClose( &ParGpsHandle ) )
                return( 0 );

        ParGpsLastDriverError = 0;

        return( 1 );
}

FUNCTYPE( int ) ParGpsStart( DEVHANDLE ParGpsHandle ) {

        int ret, dummy;
        ret = ParGpsIoctl( ParGpsHandle,
                        IOCTL_PARGPS_START,
                        &dummy,
                        sizeof(int),
                        NULL,
                        0,
                        NULL
                      );
        
        return( ret );
}

FUNCTYPE( int ) ParGpsStop( DEVHANDLE ParGpsHandle ) {

        int ret, dummy;
        ret = ParGpsIoctl( ParGpsHandle,
                        IOCTL_PARGPS_STOP,
                        &dummy,
                        sizeof(int),
                        NULL,
                        0,
                        NULL
                      );

        return( ret );
}


FUNCTYPE( int ) ParGpsGetKernelArea( DEVHANDLE ParGpsHandle, void **KernelArea ) {

        int ret;

        ret = ParGpsIoctl( ParGpsHandle,
                        IOCTL_PARGPS_GET_KAREA,
                        NULL,
                        0,
                        KernelArea,
                        sizeof(void*),
                        NULL
                      );

        return( ret );
}


FUNCTYPE( int ) ParGpsGetCounterFrequency( DEVHANDLE ParGpsHandle, long *CounterFreq ) {

        int ret;

        ret = ParGpsIoctl( ParGpsHandle,
                        IOCTL_PARGPS_GET_COUNTERFREQ,
                        NULL,
                        0,
                        CounterFreq,
                        sizeof(long),
                        NULL
                      );

        return( ret );
}


FUNCTYPE( unsigned int ) ParGpsReadPpsData( DEVHANDLE     ParGpsHandle,
                                         void         *Values,
                                         unsigned int  Nvalues,
                                         int          *Error
                                         ) {
        
        int ret, Nbytes, NvaluesReturned;
        unsigned long NbytesReturned;

        // Quick exit if not asking for any data

        if (Nvalues == 0) {
                if (Error)
                        *Error = PARGPS_ERROR_NONE;
                return( 0 );
                }

        
        // Don't ask for more data than MAX_PARGPS_STORAGE since that is
        // the total amount that can be stored in the PARGPS device driver
        // software FIFO without losing data.
        
        if (Nvalues < MAX_PARGPS_STORAGE)
                Nbytes = Nvalues * sizeof(PPSDATA);
        else
                Nbytes = MAX_PARGPS_STORAGE * sizeof(PPSDATA);


        ret = ParGpsIoctl(
                       ParGpsHandle,
                       IOCTL_PARGPS_READ_PPS_DATA,
                       NULL,
                       0,
                       Values,
                       Nbytes,
                       &NbytesReturned
                      );

        NvaluesReturned = NbytesReturned / sizeof(PPSDATA);

        if (Error) {
                if (!ret)
                        *Error = PARGPS_ERROR_DRIVER_REQUEST_FAILED;
                else
                        *Error = PARGPS_ERROR_NONE;
                }
                
        return( NvaluesReturned );
}

FUNCTYPE( unsigned int ) ParGpsReadSerialData( DEVHANDLE     ParGpsHandle,
                                            void         *Values,
                                            unsigned int  Nvalues,
                                            int          *Error
                                           ) {
        
        int ret, Nbytes, NvaluesReturned;
        unsigned long NbytesReturned;

        
        // Quick exit if not asking for any data

        if (Nvalues == 0) {
                if (Error)
                        *Error = PARGPS_ERROR_NONE;
                return( 0 );
                }

        
        // Don't ask for more data than MAX_PARGPS_STORAGE since that is
        // the total amount that can be stored in the PARGPS device driver
        // software FIFO without losing data.
        
        if (Nvalues < MAX_PARGPS_STORAGE)
                Nbytes = Nvalues * sizeof(SERIALDATA);
        else
                Nbytes = MAX_PARGPS_STORAGE * sizeof(SERIALDATA);


        ret = ParGpsIoctl(
                       ParGpsHandle,
                       IOCTL_PARGPS_READ_SERIAL_DATA,
                       NULL,
                       0,
                       Values,
                       Nbytes,
                       &NbytesReturned
                      );

        NvaluesReturned = NbytesReturned / sizeof(SERIALDATA);

        if (Error) {
                if (!ret)
                        *Error = PARGPS_ERROR_DRIVER_REQUEST_FAILED;
                else
                        *Error = PARGPS_ERROR_NONE;
                }
                
        return( NvaluesReturned );
}

FUNCTYPE( int ) ParGpsGetCurrentTime( DEVHANDLE ParGpsHandle,
                                   double *TimeAsSecsSince1970,
                                   int *Year,
                                   int *Month,
                                   int *Day,
                                   int *Hour,
                                   int *Minute,
                                   int *Second,
                                   long *MicroSecond,
                                   int *NumSatellites,
                                   double *Latitude,
                                   double *Longitude,
                                   double *Altitude,
                                   int *SourceYMD,
                                   int *SourceHMS,
                                   int *SourceNsat,
                                   int *SourcePos,
                                   int *Error ) {

static unsigned int ValidPpsEventNum = 0;
static double       ValidTime        = 0.0;

        int         i, ret, Amount, FoundNsat, FoundPos, FoundHMS, FoundYMD;
        int         Nsat, Hh, Mm, Ss, Yr, Mo, Dy;
        int         SrcYMD, SrcHMS, SrcNsat, SrcPos, ValidNmea;
        long        DiffCount, CounterFreq, Us;
        double      Lat, Lon, Alt, Tm, DiffTm, AbsTm, DelayTm;
        char        *NmeaMsg;
        GPSTIMEDATA CurrentTimeData;
        NMEAPARSE   ParseData;
        time_t      ttNow;



        // Initialize variables.

        if (Error)               *Error               = PARGPS_ERROR_NONE;
        if (TimeAsSecsSince1970) *TimeAsSecsSince1970 = 0.0;
        if (Year)                *Year                = 0;
        if (Month)               *Month               = 0;
        if (Day)                 *Day                 = 0;
        if (Hour)                *Hour                = 0;
        if (Minute)              *Minute              = 0;
        if (Second)              *Second              = 0;
        if (MicroSecond)         *MicroSecond         = 0L;
        if (NumSatellites)       *NumSatellites       = 0;
        if (Latitude)            *Latitude            = 0;
        if (Longitude)           *Longitude           = 0;
        if (Altitude)            *Altitude            = 0;
        if (SourceYMD)           *SourceYMD           = SRCSTR_MIN;
        if (SourceHMS)           *SourceHMS           = SRCSTR_MIN;
        if (SourceNsat)          *SourceNsat          = SRCSTR_MIN;
        if (SourcePos)           *SourcePos           = SRCSTR_MIN;



        // Call kernel function.

        ret = ParGpsIoctl( ParGpsHandle,
                        IOCTL_PARGPS_GET_CURRENT_TIME,
                        NULL,
                        0,
                        &CurrentTimeData,
                        sizeof(GPSTIMEDATA),
                        NULL
                      );

        if (!ret) {
                if (Error)  *Error = PARGPS_ERROR_DRIVER_REQUEST_FAILED;
                return( ret );
                }




        // Verify returned kernel time info.

        if ( !(CurrentTimeData.ValidFields & VALID_PPS)    ||
             !(CurrentTimeData.ValidFields & VALID_DREADY) ||
              (CurrentTimeData.EventNumP == 0)             ||
              (CurrentTimeData.EventNumS == 0)             ) {
                if (Error)  *Error = PARGPS_ERROR_NONE;
                return( 0 );
                }


        // The PPS and NMEA info will be from Events separated by
        // 1 second or 2 seconds depending on when the CurrentTimeData
        // was requested.  If the request comes after a PPS but before
        // the next DREADY then the VALID_NMEA field is set and the
        // PPS to NMEA serial separation is 2.

        DelayTm = CurrentTimeData.EventNumP - CurrentTimeData.EventNumS;

        ValidNmea = CurrentTimeData.ValidFields & VALID_NMEA;
        if ( (ValidNmea && (DelayTm != 1)) ||
            (!ValidNmea && (DelayTm != 2)) ) {
                        if (Error)  *Error = PARGPS_ERROR_BAD_PARAMETER;
                        return( 0 );
                        }


        // Get counter frequency for interpreting PPS and DREADY counts

        ParGpsGetCounterFrequency( ParGpsHandle, &CounterFreq );
        if (CounterFreq == 0L) {
                if (Error)  *Error = PARGPS_ERROR_NOT_AVAILABLE;
                return( 0 );
                }
                        

        
        // Loop through all NMEA messages looking for number of
        // satellites, time and the lat,lon,alt position.

        SrcYMD = SrcHMS = SrcNsat = SrcPos = SRCSTR_MIN;
        FoundYMD = FoundHMS = FoundNsat = FoundPos = 0;  
        Yr = Mo = Dy = Hh = Mm = Ss = 0; Us = 0L;
        Nsat = 0; Lat = Lon = Alt = 0.0L;

        for ( i = 0 ; i < CurrentTimeData.NmeaCount ; i++ ) {

                NmeaMsg = &CurrentTimeData.NmeaMsg[i*MAX_NMEA_SIZE];
                
                if ( ParGpsParseNmeaData( NmeaMsg, &ParseData ) ) {

                        if (!FoundHMS)
                                FoundHMS = ParGpsExtractHMS( &ParseData,
                                                  &Hh, &Mm, &Ss, &Us, &SrcHMS );
                        if (!FoundYMD)
                                FoundYMD = ParGpsExtractYMD( &ParseData,
                                                       &Yr, &Mo, &Dy, &SrcYMD );
                        if (!FoundNsat)
                                FoundNsat = ParGpsExtractNsat( &ParseData, 
                                                            &Nsat, &SrcNsat );
                        if (!FoundPos)
                                FoundPos  = ParGpsExtractLatLonAlt( &ParseData,
                                                    &Lat, &Lon, &Alt, &SrcPos );
                        }

                if (FoundHMS && FoundYMD && FoundNsat && FoundPos)
                        break;
                }




        // Process time taken from NMEA strings


        // Clock and calendar parts are good so use it and save it as valid
        
        if (FoundYMD && FoundHMS) {

                ParGpsSecTimeCombine( Yr, Mo, Dy, Hh, Mm, Ss, Us, &Tm );
                
                ValidPpsEventNum = CurrentTimeData.EventNumP;
                Tm              += DelayTm;
                ValidTime        = Tm;
                }


        // Clock part is good, so get calendar part from PC.
        // Use and save as valid.

        else if (FoundHMS) {

                SrcYMD = SRCSTR_PC; // indicate YMD from PC

                ParGpsGetPcTime( NULL, &Yr, &Mo, &Dy, NULL, NULL, NULL );

                ParGpsSecTimeCombine( Yr, Mo, Dy, Hh, Mm, Ss, Us, &Tm );

                // Force combined time to be within 1 hour of PC time
                // to ensure no date skip occurred.

                time( &ttNow );
                DiffTm = Tm - (double)ttNow;
                if (DiffTm > 0)
                        AbsTm = DiffTm;
                else
                        AbsTm = -DiffTm;
                if (AbsTm > SECPERHR) { // Time off by > 1 hr, so correct date
                        if (DiffTm > 0)
                                Tm -= SECPERDAY;
                        else
                                Tm += SECPERDAY;
                        ParGpsSecTimeSplit( Tm, &Yr, &Mo, &Dy, NULL,NULL,NULL,NULL );
                        }
                
                ValidPpsEventNum = CurrentTimeData.EventNumP;
                Tm              += DelayTm;
                ValidTime        = Tm;
                }


        // Neither clock nor calendar is good, compute as diff from prev value
        
        else if (ValidPpsEventNum) {

                SrcYMD = SRCSTR_OLD; // indicate YMD from OLD value
                SrcHMS = SRCSTR_OLD; // indicate HMS from OLD value

                Amount = CurrentTimeData.EventNumP - ValidPpsEventNum;
                Tm     = ValidTime + Amount;
                }


        // No NMEA time and no previous, so just send back Nsat info

        else {
                if (Error)         *Error         = PARGPS_ERROR_NONE;
                if (NumSatellites) *NumSatellites = Nsat;
                return( 0 );
                }


        
        
        // Adjust the time for the difference since the last PPS

        
        DiffCount = ParGpsOsLargeIntDiff( CurrentTimeData.CountAtNow,
                                    CurrentTimeData.CountAtPps );
        DiffTm    = (double)DiffCount/(double)CounterFreq;
        Tm       += DiffTm;

        ParGpsSecTimeSplit( Tm, &Yr, &Mo, &Dy, &Hh, &Mm, &Ss, &Us );


        // Fill user parameters.
        
        if (TimeAsSecsSince1970) *TimeAsSecsSince1970 = Tm;
        if (Year)                *Year                = Yr;
        if (Month)               *Month               = Mo;
        if (Day)                 *Day                 = Dy;
        if (Hour)                *Hour                = Hh;
        if (Minute)              *Minute              = Mm;
        if (Second)              *Second              = Ss;
        if (MicroSecond)         *MicroSecond         = Us;
        if (NumSatellites)       *NumSatellites       = Nsat;
        if (Latitude)            *Latitude            = Lat;
        if (Longitude)           *Longitude           = Lon;
        if (Altitude)            *Altitude            = Alt;
        if (SourceYMD)           *SourceYMD           = SrcYMD;
        if (SourceHMS)           *SourceHMS           = SrcHMS;
        if (SourceNsat)          *SourceNsat          = SrcNsat;
        if (SourcePos)           *SourcePos           = SrcPos;


        return( 1 );
}


/*************************************************************************** 
 *
 * Alternate and Structure Helper Functions:
 *
 * Some of the previous functions pass structures or have large numbers
 * of arguments which makes them difficult to call from other languages.
 * So these alternate and helper functions have been provided.  If you
 * have a choice, it is better to use an original function than an
 * alternate since the alternate just calls the original and then
 * repackages the results.
 *
 */

FUNCTYPE( int ) ParGpsGetCurrentTimeAlt( DEVHANDLE ParGpsHandle,
                                         int *IntReturns,
                                         long *LongReturns,
                                         double *DoubleReturns,
                                         int *Error ) {

        int Year, Month, Day, Hour, Minute, Second, NumSatellites, RetStd;
        int SourceYMD, SourceHMS, SourceNsat, SourcePos, ErrorStd;
        double Latitude, Longitude, Altitude, TimeAsSecsSince1970;
        long MicroSecond;
        

        // Call standard function
        
        RetStd = ParGpsGetCurrentTime( ParGpsHandle,
                                       &TimeAsSecsSince1970,
                                       &Year, &Month, &Day,
                                       &Hour, &Minute, &Second,
                                       &MicroSecond,
                                       &NumSatellites,
                                       &Latitude, &Longitude, &Altitude,
                                       &SourceYMD, &SourceHMS,
                                       &SourceNsat, &SourcePos,
                                       &ErrorStd );


        // Repackage values in alternate configuration
        
        if (IntReturns) {
                IntReturns[0]  = Year;
                IntReturns[1]  = Month;
                IntReturns[2]  = Day;
                IntReturns[3]  = Hour;
                IntReturns[4]  = Minute;
                IntReturns[5]  = Second;
                IntReturns[6]  = NumSatellites;
                IntReturns[7]  = SourceYMD;
                IntReturns[8]  = SourceHMS;
                IntReturns[9]  = SourceNsat;
                IntReturns[10] = SourcePos;
                }

        if (LongReturns) {
                LongReturns[0] = MicroSecond;
                }
        
        if (DoubleReturns) {
                DoubleReturns[0] = TimeAsSecsSince1970;
                DoubleReturns[1] = Latitude;
                DoubleReturns[2] = Longitude;
                DoubleReturns[3] = Altitude;
                }

        if (Error)
                *Error = ErrorStd;

        return( RetStd );
}

FUNCTYPE( int ) ParGpsSplitCombinePpsData( int           SplitCombine,
                                           unsigned int  Nvalues,
                                           void         *PpsData,
                                           int          *ValidFields,
                                           int          *PpsEventNum,
                                           int          *CountAtPpsHigh,
                                           unsigned int *CountAtPpsLow,
                                           int          *CountAtDreadyHigh,
                                           unsigned int *CountAtDreadyLow,
                                           int          *Error
                                            ) {
        
        unsigned int i;
        PPSDATA     *PpsDataStructure;

        // Quick exit if not working with any data

        if (Nvalues == 0) {
                if (Error)
                        *Error = PARGPS_ERROR_NONE;
                return( 1 );
                }


        // Error exit if no structure

        if (!PpsData) {
                if (Error)
                        *Error = PARGPS_ERROR_BAD_PARAMETER;
                return( 0 );
                }
        
        PpsDataStructure = (PPSDATA *)PpsData;


        // Split the structure into its pieces

        if (SplitCombine == 0) {

                for ( i = 0 ; i < Nvalues ; i++ ) {
                
                        if (ValidFields)
                                ValidFields[i] = PpsDataStructure[i].ValidFields;
            
                        if (PpsEventNum)
                                PpsEventNum[i] = PpsDataStructure[i].PpsEventNum;

                        if (CountAtPpsHigh)
                                CountAtPpsHigh[i] = PpsDataStructure[i].CountAtPps.u.HighPart;
            
                        if (CountAtPpsLow)
                                CountAtPpsLow[i] = PpsDataStructure[i].CountAtPps.u.LowPart;
            
                        if (CountAtDreadyHigh)
                                CountAtDreadyHigh[i] = PpsDataStructure[i].CountAtDready.u.HighPart;
            
                        if (CountAtDreadyLow)
                                CountAtDreadyLow[i] = PpsDataStructure[i].CountAtDready.u.LowPart;
                        
                        } // end for i
                
                } // end split


        // Combine the pieces into the structure

        else { //  (SplitCombine == 1)

                // Must have all parts when combining
                
                if (!ValidFields || !PpsEventNum ||
                    !CountAtPpsHigh || !CountAtPpsLow ||
                    !CountAtDreadyHigh || !CountAtDreadyLow ) {
                        if (Error)
                                *Error = PARGPS_ERROR_BAD_PARAMETER;
                        return( 0 );
                        }

                
                for ( i = 0 ; i < Nvalues ; i++ ) {
                
                        PpsDataStructure[i].ValidFields = ValidFields[i];
                        PpsDataStructure[i].PpsEventNum = PpsEventNum[i];
                        PpsDataStructure[i].CountAtPps.u.HighPart = CountAtPpsHigh[i];
                        PpsDataStructure[i].CountAtPps.u.LowPart = CountAtPpsLow[i];
                        PpsDataStructure[i].CountAtDready.u.HighPart = CountAtDreadyHigh[i];
                        PpsDataStructure[i].CountAtDready.u.LowPart = CountAtDreadyLow[i];
                        
                        } // end for i
                
                } // end combine


        return( 1 );
}

FUNCTYPE( int ) ParGpsSplitCombineSerialData( int           SplitCombine,
                                              unsigned int  Nvalues,
                                              void         *SerialData,
                                              int          *ValidFields,
                                              int          *PpsEventNum,
                                              int          *NmeaCount,
                                              char         *NmeaMsg,
                                              int          *Error
                                              ) {

        unsigned int i;
        int          istart;
        SERIALDATA  *SerialDataStructure;

        // Quick exit if not working with any data

        if (Nvalues == 0) {
                if (Error)
                        *Error = PARGPS_ERROR_NONE;
                return( 0 );
                }


        // Error exit if no structure

        if (!SerialData) {
                if (Error)
                        *Error = PARGPS_ERROR_BAD_PARAMETER;
                return( 0 );
                }
        
        SerialDataStructure = (SERIALDATA *)SerialData;
        istart = 0;

        

        // Split the structure into its pieces

        if (SplitCombine == 0) {

                for ( i = 0 ; i < Nvalues ; i++ ) {

                        if (ValidFields)
                                ValidFields[i] = SerialDataStructure[i].ValidFields;

                        if (PpsEventNum)
                                PpsEventNum[i] = SerialDataStructure[i].PpsEventNum;

                        if (NmeaCount)
                                NmeaCount[i] = SerialDataStructure[i].NmeaCount;

                        if (NmeaMsg)
                                memcpy( &NmeaMsg[istart], SerialDataStructure[i].NmeaMsg, MAX_NMEA_BUFF );


                        istart += MAX_NMEA_BUFF;
                        
                        } // end for i
            
                } // end split


        // Combine the pieces into the structure

        else { //  (SplitCombine == 1)

                // Must have all parts when combining
                
                if (!ValidFields || !PpsEventNum ||
                    !NmeaCount || !NmeaMsg ) {
                        if (Error)
                                *Error = PARGPS_ERROR_BAD_PARAMETER;
                        return( 0 );
                        }

                
                for ( i = 0 ; i < Nvalues ; i++ ) {
                        
                        SerialDataStructure[i].ValidFields = ValidFields[i];
                        SerialDataStructure[i].PpsEventNum = PpsEventNum[i];
                        SerialDataStructure[i].NmeaCount = NmeaCount[i];
                        memcpy( SerialDataStructure[i].NmeaMsg, &NmeaMsg[istart], MAX_NMEA_BUFF );
                        istart += MAX_NMEA_BUFF;
                        
                        } // end for i
                
                } // end combine

        return( 1 );
}

FUNCTYPE( int ) ParGpsGetStructureSize( int *PpsDataSize, int *SerialDataSize ) {
                                        
        if (PpsDataSize)
                *PpsDataSize = sizeof( PPSDATA );

        if (SerialDataSize)
                *SerialDataSize = sizeof( SERIALDATA );

        return( 1 );
}

FUNCTYPE( int ) ParGpsScaleGpsMark( int   MarkIndex,
                                    int   Nchannels,
                                    int   Nsamples,
                                    long  ScaleValue,
                                    long *Data,
                                    int  *Error
                                   ) {
        
        int i, j;

        // Quick exit if not working with any data

        if (Nchannels == 0 || Nsamples == 0) {
                if (Error)
                        *Error = PARGPS_ERROR_NONE;
                return( 1 );
                }


        // Error exit if no data array or bad mark index

        if (!Data || MarkIndex < 0 || MarkIndex > Nchannels) {
                if (Error)
                        *Error = PARGPS_ERROR_BAD_PARAMETER;
                return( 0 );
                }
        


        // Jump through multiplexed data array processing GPS Mark samples

        j = MarkIndex;
        
        for ( i = 0 ; i < Nsamples ; i++ ) {

                if ( Data[j] != 0L )
                    Data[j] = ScaleValue;

                j += Nchannels;
                }
                

        return( 1 );
}



/*************************************************************************** 
 *
 * NMEA Helper Functions
 *
 *
 */

FUNCTYPE( int ) ParGpsGetNmeaSatellites( char *NmeaMsg, int *NumSat, int *Source ) {

        NMEAPARSE ParseData;
        int Src, Nsat;


        // Initialize value

        if (NumSat)  *NumSat = 0;
        if (Source)  *Source = SRCSTR_MIN;

        
        // Parse NMEA string into data tokens
        
        if ( !ParGpsParseNmeaData( NmeaMsg, &ParseData ) )
                return( 0 );

        
        
        // Extract number of satellites from tokens
                
        if ( !ParGpsExtractNsat( &ParseData, &Nsat, &Src ) )
                return( 0 );

        if (NumSat)  *NumSat = Nsat;
        if (Source)  *Source = Src;
        
        return( 1 );
}


FUNCTYPE( int ) ParGpsGetNmeaTime( char *NmeaMsg, double *Time,
                                int *SourceYMD, int *SourceHMS ) {

        NMEAPARSE ParseData;
        double    Tm;
        int       hh, mm, ss, yr, mo, dy, SrcYMD, SrcHMS;
        long      us;
        int       ValidTime, ValidDate;

        ValidTime = ValidDate = 0;

        // Initialize arguments

        if (Time)       *Time      = 0.0;
        if (SourceYMD)  *SourceYMD = 0;
        if (SourceHMS)  *SourceHMS = 0;

        
        // Parse NMEA string into data tokens
        
        if ( !ParGpsParseNmeaData( NmeaMsg, &ParseData ) )
                return( 0 );

        
        
        // Extract clock and date part of time from NMEA string tokens
                
        ValidDate = ParGpsExtractYMD( &ParseData, &yr, &mo, &dy, &SrcYMD );
        ValidTime = ParGpsExtractHMS( &ParseData, &hh, &mm, &ss, &us, &SrcHMS );
        if (SourceYMD)  *SourceYMD = SrcYMD;
        if (SourceHMS)  *SourceHMS = SrcHMS;
        if ( !ValidDate || !ValidTime )
                return( 0 );


        // Combine the parts into the total time

        ValidTime = ParGpsSecTimeCombine( yr, mo, dy, hh, mm, ss, us, &Tm );
        
        if (Time)  *Time = Tm;
        
        return( ValidTime );
}


FUNCTYPE( int ) ParGpsGetNmeaHMS( char *NmeaMsg, double *HMS,
                               int *Hour, int *Minute, int *Second,
                               long *MicroSecond, int *Source ) {

        NMEAPARSE ParseData;
        double    hmstime;
        int       hh, mm, ss, SrcHMS;
        long      us;
        
        // Fill user arguments with defaults

        if (HMS)         *HMS         = 0.0;
        if (Hour)        *Hour        = 0;
        if (Minute)      *Minute      = 0;
        if (Second)      *Second      = 0;
        if (MicroSecond) *MicroSecond = 0L;
        if (Source)      *Source      = SRCSTR_MIN;
        
        
        // Parse NMEA string into data tokens
        
        if ( !ParGpsParseNmeaData( NmeaMsg, &ParseData ) )
                return( 0 );

        
        
        // Extract clock part of time from the NMEA tokens
                
        if ( !ParGpsExtractHMS( &ParseData, &hh, &mm, &ss, &us, &SrcHMS  ) )
                return( 0 );


        // Combine the parts into the total time

        if ( ! ParGpsSecTimeCombine( 0, 0, 0, hh, mm, ss, us, &hmstime ) )
                return( 0 );
        

        // Fill user arguments for return

        if (HMS)         *HMS         = hmstime;
        if (Hour)        *Hour        = hh;
        if (Minute)      *Minute      = mm;
        if (Second)      *Second      = ss;
        if (MicroSecond) *MicroSecond = us;
        if (Source)      *Source      = SrcHMS;
        
        return( 1 );
}


FUNCTYPE( int ) ParGpsGetNmeaYMD( char *NmeaMsg, double *YMD, 
                               int *Year, int *Month, int *Day, int *Source ) {

        NMEAPARSE ParseData;
        double    ymdtime;
        int       yr, mo, dy, SrcYMD;


        // Fill user arguments with defaults

        if (YMD)    *YMD    = 0.0;
        if (Year)   *Year   = 0;
        if (Month)  *Month  = 0;
        if (Day)    *Day    = 0;
        if (Source) *Source = SRCSTR_MIN;

        
        // Parse NMEA string into data tokens
        
        if ( !ParGpsParseNmeaData( NmeaMsg, &ParseData ) )
                return( 0 );

        
        
        // Extract date part of time from NMEA string tokens
                
        if ( !ParGpsExtractYMD( &ParseData, &yr, &mo, &dy, &SrcYMD ) )
                return( 0 );


        // Combine the parts into the total date

        if ( ! ParGpsSecTimeCombine( yr, mo, dy, 0, 0, 0, 0, &ymdtime ) )
                return( 0 );
        

        
        // Fill user arguments for return

        if (YMD)    *YMD    = ymdtime;
        if (Year)   *Year   = yr;
        if (Month)  *Month  = mo;
        if (Day)    *Day    = dy;
        if (Source) *Source = SrcYMD;

        return( 1 );
}


FUNCTYPE( int ) ParGpsGetNmeaPosition( char *NmeaMsg, 
                                    double *Latitude, double *Longitude,
                                    double *Altitude, int *Source ) {

        NMEAPARSE ParseData;
        int       SrcPos;
        double    Lat, Lon, Alt;


        // Fill user arguments with defaults

        if (Latitude)  *Latitude  = 0.0L;
        if (Longitude) *Longitude = 0.0L;
        if (Altitude)  *Altitude  = 0.0L;
        if (Source)    *Source    = SRCSTR_MIN;


        
        // Parse NMEA string into data tokens
        
        if ( !ParGpsParseNmeaData( NmeaMsg, &ParseData ) )
                return( 0 );

        
        // Extract position information from tokens
                
        if ( !ParGpsExtractLatLonAlt( &ParseData, &Lat, &Lon, &Alt, &SrcPos ) )
                return( 0 );


        // Fill arguments

        if (Latitude)  *Latitude  = Lat;
        if (Longitude) *Longitude = Lon;
        if (Altitude)  *Altitude  = Alt;
        if (Source)    *Source    = SrcPos;


        return( 1 );
}


FUNCTYPE( int ) ParGpsParseNmeaData( char *NmeaMsg, NMEAPARSE *ParseData ) {

        char *field, c;
        int i, done, ifield, checksum, tmpsum;


        // Error check Nmea string

        if ( (!NmeaMsg) || (*NmeaMsg++ != '$') )
                return( 0 );


        // Initialze ParseData structure

        ParseData->Nfields  = 0;
        ParseData->CheckSum = 0;
        for ( i = 0 ; i < MAX_NMEA_FIELDS ; i++ )
                ParseData->Field[i][0] = '\0';


        // Initialize local variables

        done     = 0;
        checksum = 0;
        tmpsum   = 0;
        ifield   = 0;
        field    = ParseData->Field[ifield];


        // Parsing loop

        while ( !done ) {

                c = *NmeaMsg++;

                switch ( c ) {

                // Found unexpected start of next string

                        case '$':
                                return( 0 );
                                break;


                // Found end of current field

                        case ',':
                        case '*':
                        case '\r':
                                *field = '\0';
                                ifield++;
                                if (ifield < MAX_NMEA_FIELDS)
                                        field = ParseData->Field[ifield];
                                else 
                                        return( 0 );

                                if (c == '*')
                                        checksum = tmpsum;
                                else
                                        tmpsum ^= c;
                                break;


                // Finished with string

                        case '\n':
                        case '\0':
                                done = 1;
                                break;


                // Add current character to field string and checksum

                        default:
                                *field++ = c;
                                tmpsum ^= c;

                        }

                } // end while



        // Set number of fields and checksum

        ParseData->Nfields = ifield;
        field = ParseData->Field[ifield-1];
        sscanf( field, "%X", &ParseData->CheckSum );

        if (ParseData->CheckSum != checksum)
                return( 0 );

        return( 1 );
}


/*************************************************************************** 
 *
 * Time Helper Functions
 *
 * Packed digit time is a double where two digits have been allocated to
 * each of hours, minutes, and seconds (or year, month, day).  For
 * example 112307 represents 23 minutes and 7 seconds past 11 o'clock.
 * See pargps.h for more details.
 *
 */



#define SHIFTDIG4   10000
#define SHIFTDIG2     100

FUNCTYPE( int ) ParGpsUnpack3( int Xxx, int *Top, int *Middle, int *Bottom ){

        int tt, mm, bb;
        
        // Unpack 3 pieces (ttmmbb) from one int

        tt = Xxx / SHIFTDIG4;
        mm = (Xxx-tt*SHIFTDIG4) / SHIFTDIG2;
        bb = (Xxx-tt*SHIFTDIG4-mm*SHIFTDIG2);

        
        // Save pieces
        
        *Top    = tt;
        *Middle = mm;
        *Bottom = bb;
        
        return( 1 );
}

FUNCTYPE( int ) ParGpsPack3( int Top, int Middle, int Bottom, int *Xxx  ){

        // Pack 3 pieces (ttmmbb) into one int

        *Xxx = (Top * SHIFTDIG4) + (Middle * SHIFTDIG2) + Bottom;
        
        return( 1 );
}

FUNCTYPE( int ) ParGpsGetPcTime( double *Time,
                                 int *Year, int *Month, int *Day, 
                                 int *Hour, int *Minute, int *Second ) {
        
        time_t    ttNow;
        struct tm tmNow;

        // Return time as YMD HMS for UTC (not local)

        time( &ttNow );
        tmNow   = *gmtime( &ttNow );

        if (Time)    *Time   = (double)ttNow;
        if (Year)    *Year   = tmNow.tm_year + 1900;
        if (Month)   *Month  = tmNow.tm_mon + 1;
        if (Day)     *Day    = tmNow.tm_mday;
        if (Hour)    *Hour   = tmNow.tm_hour;
        if (Minute)  *Minute = tmNow.tm_min;
        if (Second)  *Second = tmNow.tm_sec;

        return( 1 );
}

FUNCTYPE( int ) ParGpsSetPcTime( double Time,
                                 int Year, int Month, int Day, 
                                 int Hour, int Minute, int Second,
                                 int FromYMD ) {

        // This function returns 1 for success, 0 for failure.
        // FromYMD is 1 when the time to set is taken from the Year, Month, etc
        // values and and 0 when it is taken from the Time in seconds since
        // 1970.  The underlying Windows function uses YMD, while the underlying
        // Linux function uses Time.  But both work with either setting.

        // NOTE: You MUST have the proper permissions to successfully set
        //       the PC time.  Typically this means being an Administrator
        //       or Power User under Windows and being root or running
        //       "set uid" root under Linux.
        

        return( ParGpsOsSetPcTime( Time, Year, Month, Day, Hour, Minute, Second,
                                   FromYMD )  );
}

FUNCTYPE( int ) ParGpsCalcTimeZoneCorr( long StartTime ) {
        
        time_t    TargetTime, NewTime;
        struct tm tm;
        int       ZoneCorrection;

        // This function does two conversions between time stored as a structure
        // and time as seconds since the epoch (1970-01-01).  They should cancel
        // out, but the first is done using GMT time and the second is done using
        // local time.  The difference gives us the amount of the local correction.
        // Subtract this value from a local time to get a GMT time.
        
        
        // Convert epoch seconds to time structure for gmt
        
        TargetTime = (time_t)StartTime;
        tm = *gmtime( &TargetTime );

        
        // Convert time structure for local to epoch seconds (force DST unknown)
        
        tm.tm_isdst = -1;
        NewTime = mktime( &tm );


        // Difference is local correction

        ZoneCorrection = NewTime - TargetTime;
        return( ZoneCorrection );
}

FUNCTYPE( int ) ParGpsSecTimeSplit( double Time,
                                    int *Year, int *Month, int *Day,
                                    int *Hour, int *Minute, int *Second, 
                                    long *Microsecond ) {
        time_t    ttTime;
        struct tm tmTime;
        double    usec;


        // Split time into pieces (for UTC)

        ttTime = (time_t)Time;
        tmTime = *gmtime( &ttTime );
        usec   = Time - (double)ttTime + 0.0000005;
        
        // Save pieces
        
        if (Year)         *Year        = tmTime.tm_year + 1900;
        if (Month)        *Month       = tmTime.tm_mon + 1;
        if (Day)          *Day         = tmTime.tm_mday;
        if (Hour)         *Hour        = tmTime.tm_hour;
        if (Minute)       *Minute      = tmTime.tm_min;
        if (Second)       *Second      = tmTime.tm_sec;
        if (Microsecond)  *Microsecond = (long)(usec * USPERSEC);

        return( 1 );
}

FUNCTYPE( int ) ParGpsSecTimeCombine( int Year, int Month, int Day,
                                   int Hour, int Minute, int Second, 
                                   long MicroSecond, double *Time ) {

        time_t    ttTime;
        struct tm tmTime;

        // Set default YMD values when 0's are passed in

        if (Year == 0)   Year  = 1970;
        if (Month == 0)  Month = 1;
        if (Day == 0)    Day   = 1;


        // Fill pieces
        
        tmTime.tm_year  = Year - 1900;
        tmTime.tm_mon   = Month - 1;
        tmTime.tm_mday  = Day;
        tmTime.tm_hour  = Hour;
        tmTime.tm_min   = Minute;
        tmTime.tm_sec   = Second;
        tmTime.tm_wday  = 0;
        tmTime.tm_yday  = 0;
        tmTime.tm_isdst = -1;


        // Combine time from GMT pieces to epoch seconds using local correction,
        // then use the zone offset to remove the local correction

        ttTime = mktime( &tmTime );
        if (ttTime == (time_t)-1)
                return( 0 );
        ttTime -= ParGpsCalcTimeZoneCorr( ttTime );
        if (Time)  *Time  = (double)ttTime + (double)MicroSecond / (double)USPERSEC;
        
        return( 1 );
}







/*************************************************************************** 
 *
 * Private Library Helper Functions:
 *
 * Most users should NOT use these functions, but only the main ones
 * above.  These are intended for use by the main library functions and
 * programs like diag.c with special access. 
 *
 *
 */

FUNCTYPE( int ) ParGpsIoctl( DEVHANDLE     ParGpsHandle,
                             unsigned long IoCtlCode,
                             void         *pValueIn,
                             unsigned long InSize,
                             void         *pValueOut, 
                             unsigned long OutSize,
                             unsigned long *ReturnSize
                           ) {

        int           Result;
        unsigned long BytesReturned;


        // Device driver must be opened first.

        if (ParGpsHandle == BAD_DEVHANDLE) {
                ParGpsLastDriverError = ERROR_SERVICE_DISABLED;
                return( 0 );
                }



        // Request specialized services from device driver.

        Result = ParGpsOsDriverIoctl(
                               ParGpsHandle,     // Handle to device
                               IoCtlCode,        // IO Control code
                               pValueIn,         // Input data to driver
                               InSize,           // Length of in buffer in bytes
                               pValueOut,        // Output data from driver
                               OutSize,          // Length of out buffer in bytes
                               &BytesReturned    // Bytes placed in output buffer
                              );

        if (ReturnSize)
                *ReturnSize = BytesReturned;


        
        // Failure output.

        if (!Result) {
                ParGpsLastDriverError = ParGpsOsGetLastError();
                return( 0 );
                }

        ParGpsLastDriverError = 0;
        return( 1 );
}

FUNCTYPE( int ) ParGpsReset( DEVHANDLE ParGpsHandle ) {

        int dummy, ret;

        ret = ParGpsIoctl( ParGpsHandle,
                           IOCTL_PARGPS_RESET,
                           &dummy,
                           sizeof(int),
                           NULL,
                           0,
                           NULL
                         );

        return( ret );
}

FUNCTYPE( int ) ParGpsGetInterruptCount( DEVHANDLE ParGpsHandle, int *count ) {

        int ret;

        ret = ParGpsIoctl( ParGpsHandle, 
                           IOCTL_PARGPS_GET_INTR_COUNT,
                           NULL,
                           0,
                           count,
                           sizeof(int),
                           NULL
                         );

        return( ret );
}

FUNCTYPE( int ) ParGpsSetInterruptMode( DEVHANDLE ParGpsHandle, int Mode ) {

        int IntrMode, ret;

        // Change user define to kernel define

        if (Mode == INTERRUPT_ALTERNATING)
                IntrMode = PARGPS_INTTYPE_PPS;
        else
                IntrMode = PARGPS_INTTYPE_PLAIN;



        ret = ParGpsIoctl( ParGpsHandle,
                           IOCTL_PARGPS_SET_INTR_MODE,
                           &IntrMode,
                           sizeof(int),
                           NULL,
                           0,
                           NULL
                         );

        return( ret );
}

FUNCTYPE( int ) ParGpsSerialSetup( DEVHANDLE ParGpsHandle, int PortNumber ) {

        DEVHANDLE  SerialHandle;
        char      *SerialPortName;
        int        ret;


        // Translate port number to a port name

        if (PortNumber == 1)
                SerialPortName = COM1;
        else if (PortNumber == 2)
                SerialPortName = COM2;
        else
                return( 0 );


        // Open serial port in user space

        SerialHandle = ParGpsOsSerialOpen( SerialPortName );
        if (SerialHandle == BAD_DEVHANDLE)
                return( 0 );


        // Set serial port baud rate, etc

        ret = ParGpsOsSerialInit( SerialHandle );
        if (!ret) {
                ParGpsOsSerialClose( SerialHandle );
                return( ret );
                }



#if defined( SROS_WIN2K ) || defined( SROS_WINNT )
        // Close serial port in user space, kernel will re-open
        ParGpsOsSerialClose( SerialHandle );
#elif defined( SROS_LINUX )
        // Don't do close so kernel serial access continues to work
        save_serial_handle = SerialHandle;
#endif

        return( ret );
}

FUNCTYPE( int ) ParGpsInit( DEVHANDLE ParGpsHandle,
                            int SerialPort,
                            int InterruptMode ) {

        int ret, ParGpsInitParms[2];


        // Error check input values

        if ( (SerialPort != SERIAL_PORT_1) &&
             (SerialPort != SERIAL_PORT_2) )
                return( 0 );
        
        if ( (InterruptMode != INTERRUPT_PPS_ONLY)    &&
             (InterruptMode != INTERRUPT_ALTERNATING) )
                return( 0 );


        // Set up parameters and call kernel driver

        ParGpsInitParms[0] = SerialPort;
        ParGpsInitParms[1] = InterruptMode;
        
        ret = ParGpsIoctl( ParGpsHandle,
                           IOCTL_PARGPS_INIT,
                           ParGpsInitParms,
                           sizeof(ParGpsInitParms),
                           ParGpsInitParms,
                           sizeof(ParGpsInitParms),
                           NULL
                         );

        return( ret );
}


// This structure contains information about which fields in the
// various NMEA messages contain certain common types of information
// such as time and position.

struct NMEA_field_process_table {
        int     NmeaMsgIndex;
        char    *GPMsgId;
        BOOLEAN HasTimeField;
        int     TimeField;
        BOOLEAN HasValidField;
        int     ValidField;
        char    ValidValue;
        BOOLEAN HasDateField;
        int     DateField;
        int     DateLength;
        BOOLEAN HasLatLonField;
        int     LatLonField;
        BOOLEAN HasAltitudeField;
        int     AltitudeField;
        BOOLEAN HasNsatField;
        int     NsatField;
        } *t, NmeaTable[] = {

NMEA_MSGID_ZDA, "GPZDA",  TRUE, 1, FALSE, 0, '\0',  TRUE, 2, 3, FALSE, 0, FALSE, 0, FALSE, 0,
NMEA_MSGID_RMC, "GPRMC",  TRUE, 1,  TRUE, 2,  'A',  TRUE, 9, 1,  TRUE, 3, FALSE, 0, FALSE, 0,
NMEA_MSGID_GGA, "GPGGA",  TRUE, 1,  TRUE, 6,  '1', FALSE, 0, 0,  TRUE, 2,  TRUE, 9,  TRUE, 7,
NMEA_MSGID_GLL, "GPGLL",  TRUE, 5,  TRUE, 6,  'A', FALSE, 0, 0,  TRUE, 1, FALSE, 0, FALSE, 0,
NMEA_MSGID_VTG, "GPVTG", FALSE, 0, FALSE, 0, '\0', FALSE, 0, 0, FALSE, 0, FALSE, 0, FALSE, 0,
NMEA_MSGID_GSA, "GPGSA", FALSE, 0, FALSE, 0, '\0', FALSE, 0, 0, FALSE, 0, FALSE, 0, FALSE, 0,
NMEA_MSGID_GSV, "GPGSV", FALSE, 0, FALSE, 0, '\0', FALSE, 0, 0, FALSE, 0, FALSE, 0,  TRUE, 3,
NMEA_MSGID_MIN, NULL,    FALSE, 0, FALSE, 0, '\0', FALSE, 0, 0, FALSE, 0, FALSE, 0, FALSE, 0
        };


FUNCTYPE( int ) ParGpsExtractNsat( NMEAPARSE *ParseData, int *Nsat, int *Source ) {

        int numsat;


        *Nsat   = 0;
        *Source = SRCSTR_MIN;

        // Find the NMEA table entry matching the parsed data id

        t = NmeaTable;
        while ( t->GPMsgId && strcmp( ParseData->Field[0], t->GPMsgId ) )
                t++;


        // Ensure that it is valid and has a number of satellites field

        if ( (t->GPMsgId == NULL) || (!t->HasNsatField) )
                return( 0 );



        // Extract number of satellites

        if (sscanf( ParseData->Field[t->NsatField], "%d", &numsat ) != 1)
                return( 0 );

        *Nsat   = numsat;
        *Source = t->NmeaMsgIndex; // Assumes SRCSTR_ defines match NMEA_MSGID_

        return( 1 );
}

FUNCTYPE( int ) ParGpsExtractHMS( NMEAPARSE *ParseData,
                                  int *Hour,
                                  int *Minute,
                                  int *Second,
                                  long *MicroSecond,
                                  int *Source ) {

        char MsgChar;
        int  hhmmss;


        // This function is always called from other library
        // routines and not directly from the user, so we assume
        // that the passed arguments all exist.

        *Hour        = 0;
        *Minute      = 0;
        *Second      = 0;
        *MicroSecond = 0L;
        *Source      = SRCSTR_MIN;


        // Find table entry that matches parse data MsgId

        t = NmeaTable;
        while ( t->GPMsgId && strcmp( ParseData->Field[0], t->GPMsgId ) )
                t++;


        // Error check

        if ( (t->GPMsgId == NULL) || (!t->HasTimeField) )
                return( 0 );


        // Extract just the integer portion of the time since
        // the PPS mark is exactly on the second mark

        if (sscanf( ParseData->Field[t->TimeField], "%d", &hhmmss ) != 1)
                return( 0 );


        // Unpack clock part of time into HMS

        ParGpsUnpack3( hhmmss, Hour, Minute, Second );
        *MicroSecond = 0L;

        *Source = t->NmeaMsgIndex; // Assumes SRCSTR_ defines match NMEA_MSGID_
        
        
        // Check validity, assume valid by default

        MsgChar = '\0';

        if (t->HasValidField) {
                sscanf( ParseData->Field[t->ValidField], "%c", &MsgChar );
                if (MsgChar != t->ValidValue)
                        return( 0 );
                }


        return( 1 );
}

FUNCTYPE( int ) ParGpsExtractYMD( NMEAPARSE *ParseData,
                                  int *Year,
                                  int *Month,
                                  int *Day,
                                  int *Source ) {
        
        int  ddmmyy, dd, mm, yy;
        char MsgChar;


        // This function is always called from other library
        // routines and not directly from the user, so we assume
        // that the passed arguments all exist.
        
        *Year   = 0;
        *Month  = 0;
        *Day    = 0;
        *Source = SRCSTR_MIN;


        // Find table entry that matches parse data MsgId

        t = NmeaTable;
        while ( t->GPMsgId && strcmp( ParseData->Field[0], t->GPMsgId ) )
                t++;


        // Error check

        if ( (t->GPMsgId == NULL) || (!t->HasDateField) )
                return( 0 );


        // Extract YMD date

        if ( t->DateLength == 1 ) {
                if (sscanf( ParseData->Field[t->DateField], "%d", &ddmmyy ) != 1)
                        return( 0 );

                ParGpsUnpack3( ddmmyy, &yy, &mm, &dd );
                
                *Day   = dd;
                *Month = mm;
                *Year  = yy + 2000;
                }

        else if( t->DateLength == 3 ) {
                if (sscanf( ParseData->Field[t->DateField+0], "%d", Day ) != 1)
                        return( 0 );
                if (sscanf( ParseData->Field[t->DateField+1], "%d", Month ) != 1)
                        return( 0 );
                if (sscanf( ParseData->Field[t->DateField+2], "%d", Year ) != 1)
                        return( 0 );
                }
        else {
                return( 0 );
                }


        *Source = t->NmeaMsgIndex; // Assumes SRCSTR_ defines match NMEA_MSGID_


        
        // Check validity, assume valid by default

        MsgChar = '\0';

        if (t->HasValidField) {
                sscanf( ParseData->Field[t->ValidField], "%c", &MsgChar );
                if (MsgChar != t->ValidValue)
                        return( 0 );
                }


        return( 1 );
}

FUNCTYPE( int ) ParGpsExtractLatLonAlt( NMEAPARSE *ParseData, double *Latitude, 
                                        double *Longitude, double *Altitude,
                                        int *Source ) {

        char LatQuad, LonQuad, AltUnits;

        
        // This function is always called from other library
        // routines and not directly from the user, so we assume
        // that the passed arguments all exist.
        
         LatQuad  =  LonQuad   =  AltUnits = 'X';
        *Latitude = *Longitude = *Altitude = 0.0L;
        *Source   = SRCSTR_MIN;


        // Find table entry that matches parse data MsgId

        t = NmeaTable;
        while ( t->GPMsgId && strcmp( ParseData->Field[0], t->GPMsgId ) )
                t++;


        // Error check

        if ( (t->GPMsgId == NULL) || (!t->HasLatLonField) )
                return( 0 );


        // Extract lat/lon and verify quadrants

        sscanf( ParseData->Field[t->LatLonField],   "%lf", Latitude );
        sscanf( ParseData->Field[t->LatLonField+1], "%c",  &LatQuad );
        sscanf( ParseData->Field[t->LatLonField+2], "%lf", Longitude );
        sscanf( ParseData->Field[t->LatLonField+3], "%c",  &LonQuad );
        
        if (LatQuad == 'S')
                *Latitude = -(*Latitude);
        else if (LatQuad != 'N')
                return( 0 );

        if (LonQuad == 'W')
                *Longitude = -(*Longitude);
        else if (LonQuad != 'E')
                return( 0 );


        
        // Check altitude, accept any units for right now

        if (t->HasAltitudeField) {
                sscanf( ParseData->Field[t->AltitudeField],   "%lf", Altitude );
                sscanf( ParseData->Field[t->AltitudeField+1], "%c", &AltUnits );
                
                if (AltUnits != 'M')
                        ;
                }


        *Source = t->NmeaMsgIndex; // Assumes SRCSTR_ defines match NMEA_MSGID_

        return( 1 );
}





/* SR64BIT LARGE INTEGER HELPER FUNCTIONS:
 * 
 * Functions for subtracting two SR64BIT values and for adding to a
 * SR64BIT value.  The computation varies depending on what type of
 * data is stored in the the SR64BIT variable.
 *
 */

long ParGpsLargeIntDiffType( SR64BIT A, SR64BIT B, int Type ) {

        long Answer, seconds, usec;

        // Windows style large ints
        
        if (Type == PC_COUNTER_TYPE_INT64) {
                Answer = (long)(A.QuadPart - B.QuadPart);
                }
        
        // Linux style large ints
        
        else { // Type == PC_COUNTER_TYPE_HIGH32LOW32
                seconds = (long)(A.u.HighPart - B.u.HighPart);
                usec    = (long)(A.u.LowPart  - B.u.LowPart);
                Answer  = (seconds * 1000000) + usec;
                }
        
        return( Answer );
}

SR64BIT ParGpsLargeIntAddType( SR64BIT A, long More, int Type ) {

        SR64BIT C;

        // Windows style large ints
        
        if (Type == PC_COUNTER_TYPE_INT64) {
                C.QuadPart = A.QuadPart + More;
                }
        
        else { // Type == PC_COUNTER_TYPE_HIGH32LOW32
                C.u.HighPart = A.u.HighPart;         // seconds part

                if ( More >= 0 ) {
                        while ( More > 1000000 ) {
                                C.u.HighPart++;
                                More -= 1000000;
                                }
                        C.u.LowPart  = A.u.LowPart + More;   // usec part
                        while ( C.u.LowPart >= 1000000) {
                                C.u.HighPart++;
                                C.u.LowPart -= 1000000;
                                }
                        }
                else { // ( More < 0 )
                        while ( More < -1000000 ) {
                                C.u.HighPart--;
                                More += 1000000;
                                }
                        while ( (long)A.u.LowPart < -More ) {
                                C.u.HighPart--;
                                A.u.LowPart += 1000000;
                                }
                        C.u.LowPart  = A.u.LowPart + More;   // usec part
                        }
                }
        
        return( C );
}



/* COLLECT ANALOG/PPS/SERIAL DATA INTO TIMESTAMPS FUNCTIONS:
 * 
 * For each PpsEvent, three types of data may be available - analog,
 * pps, and serial.  The following functions collect the
 * data in what ever order it arrives into timestamp structures.
 * See pargps.h for more details.
 *
 */

FUNCTYPE( int ) ParGpsTsInit( void ) {
        
        // Initialize time stamp circular buffer

        return( ParGpsInitCircBuff() );
}

FUNCTYPE( int ) ParGpsTsProcessAnalog( long *DataValues,
                                       long NumSamples,
                                       long NumChannels,
                                       long FirstSample,
                                       long MarkChannel,
                                       long ObcChannel,
                                       double ObcCountsPerSample,
                                       unsigned int *PpsEvent,
                                       int  *Nvalid
                                     ) {
        long         i;
        unsigned int EventNum;
        int          Obc;
        static long  LastObc   = MARKCHANNEL_NONE;
        static long  LastObcNp = -1;

        long *DataGpsPtr, *DataObcPtr;

        // Search through analog data checking the "mark" channel and
        // recording the mark information when that channel is not 0.
        //
        // MarkChannel  says which channel number has the mark info
        // FirstSample  says which number is the starting time point 
        
        // The algorithm for processing with OBC is described here:
        // Look for a valid OBC (ie non 0) and then for a valid PPS 
        // EventNum (Gps Mark).  In most cases, these occur on the same 
        // sample point.  However, when the next sample comes very soon 
        // after PPS, the OBC will catch it while the Mark will not be 
        // seen until the following sample.
        //
        // While this is a complication, it does not result in an error 
        // since the Mark information (the Dready count saved for that 
        // EventNum) does include the time for the extra sample.  The 
        // following code increments the OBC count, if needed, so it is 
        // correct for the Marked sample. The OBC and its point index are
        // saved in static variables to handle the case where the OBC
        // comes on the last sample of one record and the Mark comes on
        // the first sample of the next.
        

        ParGpsLastErrorMessage[0] = '\0';


        // Initialize optional return parameters

        if ( PpsEvent )  *PpsEvent = 0;
        if ( Nvalid )    *Nvalid   = 0;


        // Error check incoming values
        
        if ( !DataValues )
                return( 0 );


        // Early exit if no points or no GPS mark information is included
        
        if ( NumSamples <= 0  ||  MarkChannel == MARKCHANNEL_NONE )
                return( 0 );


        EventNum = 0;

        // Processing without on board counter
        
        if (ObcChannel == MARKCHANNEL_NONE) { 

                DataGpsPtr = DataValues + MarkChannel;
                
                for ( i = 0 ; i < NumSamples ; i++ ) {
                        if ( *DataGpsPtr !=  0L ) {
                                EventNum = (int)*DataGpsPtr;
                                ParGpsAnalogWriteCircBuff( EventNum,
                                                     FirstSample+i,
                                                     ObcChannel );
                                }
                        DataGpsPtr += NumChannels;
                        
                        }  // end for i < NumSamples
                
                } // end if ObcChannel = NONE

        // Processing with on board counter (OBC).

        else {
                
                DataGpsPtr = DataValues + MarkChannel;
                DataObcPtr = DataValues + ObcChannel;

                for ( i = 0 ; i < NumSamples ; i++ ) {
                        
                        if ( *DataObcPtr != 0L ) {
                                LastObc   = *DataObcPtr;
                                LastObcNp = i;
                                }
                        
                        if ( *DataGpsPtr != 0L ) {
                                EventNum = (int)*DataGpsPtr;
                                
                                if (LastObcNp < 0)
                                        Obc  = MARKCHANNEL_NONE;
                                else {
                                        Obc = (int)(LastObc +
                                                    (i - LastObcNp) *
                                                    ObcCountsPerSample);
                                        LastObcNp = -1;
                                        }
                                
                                ParGpsAnalogWriteCircBuff( EventNum,
                                                     FirstSample+i,
                                                     Obc );
                                }
                        DataGpsPtr += NumChannels;
                        DataObcPtr += NumChannels;
                        
                        } // end for i < Numsamples
                
                } // end else ObcChannel


        // Fill optional return parameters

        if ( PpsEvent )  *PpsEvent = EventNum;
        if ( Nvalid )    *Nvalid   = CbNumValid;

        return( 1 );
}

FUNCTYPE( int ) ParGpsTsProcessPps( PPSDATA *PpsValues, long NumPps,
                                    unsigned int *PpsEvent, int *Nvalid ) {
        long         i;
        int          ValidMask;
        unsigned int EventNum;

        // Save the GPS PPS information in a timestamp structure.  The
        // values saved are the PPS event number (MarkNum), the number of
        // 64 bit PC counter counts at a PPS (CountAtPps) and the next data
        // ready (CountAtDready).

                 
        ParGpsLastErrorMessage[0] = '\0';
        
        // Initialize optional return parameters

        if ( PpsEvent )  *PpsEvent = 0;
        if ( Nvalid )    *Nvalid   = 0;


        // Error check

        if ( !PpsValues  ||  NumPps <= 0 )
                return( 0 );

        

        // Loop through recording the PARGPS PPS data.

        ValidMask = VALID_PPS | VALID_DREADY;
        EventNum  = 0;

        for ( i = 0 ; i < NumPps ; i++ ) {

                if ( (PpsValues[i].ValidFields & ValidMask) == ValidMask) {
                        ParGpsPpsWriteCircBuff( PpsValues[i].PpsEventNum, 
                                          PpsValues[i].CountAtPps, 
                                          PpsValues[i].CountAtDready
                                          );
                        EventNum = PpsValues[i].PpsEventNum;
                        }
                else {
                        sprintf( ParGpsLastErrorMessage,
                                 "ERROR: Pps data for event %d is invalid\n",
                                 PpsValues[i].PpsEventNum );
                        }
                }

        // Fill optional return parameters

        if ( PpsEvent )  *PpsEvent = EventNum;
        if ( Nvalid )    *Nvalid   = CbNumValid;

        return( 1 );
}

FUNCTYPE( int ) ParGpsTsProcessSerial( SERIALDATA *SerialValues, long NumSerial,
                                       int SerialDelay,
                                       unsigned int *PrevPpsEvent, double *PrevSecSince1970,
                                       unsigned int *PpsEvent, int *Nvalid ) {
        long         i;
        unsigned int EventNum, Ncorrect;
        int          ValidTime, ValidHMS, ValidYMD, ValidNsat;
        int          Year, Month, Day, Hour, Minute, Second;
        int          PrevYear, PrevMonth, PrevDay, NumNmea, Nsat;
        int          inmea, SrcYMD, SrcYMDt, SrcHMS, SrcHMSt, SrcNsat;
        int          YmdSource, HmsSource;
        long         MicroSecond;
        double       SecSince1970, YMD, HMS, DiffTime, AbsTime;
        char        *NmeaMsg;

        ParGpsLastErrorMessage[0] = '\0';
        
        // Initialize optional return parameters

        if ( PpsEvent )  *PpsEvent = 0;
        if ( Nvalid )    *Nvalid   = 0;


        // Error check
        
        if ( !SerialValues  ||  NumSerial <= 0 )
                return( 0 );

        EventNum = 0;
        Nsat     = 0;
        
        // Loop through extracting the GPS time from the NMEA strings
        // and recording the GPS serial data.  If not enough satellites
        // have been seen to produce a valid GPS time, then just
        // increment a previous value since it should come every second.

        for ( i = 0 ; i < NumSerial ; i++ ) {
                
                ValidTime = ValidHMS = ValidYMD = ValidNsat = 0;
                EventNum  = SerialValues[i].PpsEventNum + SerialDelay;
                NumNmea   = SerialValues[i].NmeaCount;
                NmeaMsg   = &SerialValues[i].NmeaMsg[0];

                for ( inmea = 0 ; inmea < NumNmea ; inmea++ ) {

                        if (!ValidTime) {
                                ValidTime = ParGpsGetNmeaTime( NmeaMsg, 
                                            &SecSince1970, &SrcYMDt, &SrcHMSt );
                                }

                        if (!ValidYMD) {
                                ValidYMD = ParGpsGetNmeaYMD( NmeaMsg, 
                                           &YMD, &Year, &Month, &Day, &SrcYMD );
                                }

                        if (!ValidHMS) {
                                ValidHMS = ParGpsGetNmeaHMS( NmeaMsg, 
                                           &HMS, &Hour, &Minute, &Second,
                                           &MicroSecond, &SrcHMS );
                                }

                        if (!ValidNsat)
                                ValidNsat = ParGpsGetNmeaSatellites( NmeaMsg, 
                                                   &Nsat, &SrcNsat );
                        
                        if (ValidTime && ValidYMD && ValidHMS && ValidNsat)
                                break; // exit from for inmea loop

                        NmeaMsg += MAX_NMEA_SIZE;  // next message
                        
                        } // end of inmea loop



                // NMEA time is valid, so use it
                
                if (ValidTime) {

                        ParGpsSrcToSource( SrcYMDt, SrcHMSt, &YmdSource, &HmsSource );
        
                        SecSince1970 += SerialDelay;
                        ParGpsSerialWriteCircBuff( EventNum, SecSince1970,
                                             YmdSource, HmsSource,
                                             Nsat, 0 );

                        *PrevPpsEvent     = EventNum;
                        *PrevSecSince1970 = SecSince1970;
                        }
                
                // The HMS NMEA time is valid, so combine it with YMD from prev
                
                else if (ValidHMS) {

                        ParGpsSecTimeSplit( *PrevSecSince1970,
                                            &PrevYear, &PrevMonth, &PrevDay,
                                            NULL, NULL, NULL, NULL );

                        ParGpsSecTimeCombine( PrevYear, PrevMonth, PrevDay,
                                              Hour, Minute, Second, 
                                              MicroSecond, &SecSince1970 );
                        SecSince1970 += SerialDelay;

                        // Force combined time to be within 1 hour of prev 
                        // time to ensure no date skip occurred.

                        DiffTime = SecSince1970 - *PrevSecSince1970;
                        if (DiffTime > 0)
                                AbsTime = DiffTime;
                        else
                                AbsTime = -DiffTime;
                        if (AbsTime > SECPERHR) { // Time off by > 1 hr, so correct date
                                if (DiffTime > 0)
                                        SecSince1970 -= SECPERDAY;
                                else
                                        SecSince1970 += SECPERDAY;
                                ParGpsSecTimeSplit( SecSince1970,
                                                 &Year, &Month, &Day,
                                                 NULL, NULL, NULL, NULL );
                                }

                                ParGpsSrcToSource( SrcYMD, SrcHMS, NULL, &HmsSource );

                                ParGpsSerialWriteCircBuff( EventNum, SecSince1970,
                                                     TIME_SOURCE_CALC, HmsSource,
                                                     Nsat, 0 );

                                *PrevPpsEvent     = EventNum;
                                *PrevSecSince1970 = SecSince1970;
                                }


                // A previous NMEA time is valid, so correct it to be current
                
                else if (*PrevPpsEvent != INVALID_PPSEVENT) {

                                Ncorrect     = EventNum - *PrevPpsEvent;
                                SecSince1970 = *PrevSecSince1970 + Ncorrect;
                                ParGpsSerialWriteCircBuff( EventNum, SecSince1970,
                                                     TIME_SOURCE_CALC, TIME_SOURCE_CALC,
                                                     Nsat, 0 );

                                *PrevPpsEvent     = EventNum;
                                *PrevSecSince1970 = SecSince1970;
                                }


                // No valid NMEA time ever, so erase this entire buffer entry

                else {

                        SecSince1970 = *PrevSecSince1970 + 1;
                        ParGpsSerialWriteCircBuff( EventNum, SecSince1970,
                                             TIME_SOURCE_NONE, TIME_SOURCE_NONE,
                                             Nsat, 1 );
                        }


                } // end for i < NumSerial


        // Fill optional return parameters

        if ( PpsEvent )  *PpsEvent = EventNum;
        if ( Nvalid )    *Nvalid   = CbNumValid;
        
        return( 1 );
}


/* GET VALID FROM LIBRARY AND READ/WRITE TO FILE TIMESTAMP FUNCTIONS:
 * 
 * Get valid timestamps from the library's circular buffer.  Read
 * and write them from/to an open ascii file.
 *
 */

FUNCTYPE( int ) ParGpsTsReadValid( TIMESTAMP *TS ) {

        // Error check input
        if ( !TS )
                return( 0 );


        if ( !ParGpsReadCircBuff( TS ) )
                return( 0 );               // none ready (valid)

        return( 1 );
}


// This is used by both printf and scanf, so don't give any precisions
// in the format for reading/writing time stamp from/to an open ascii file

#define TSTITLE "Valid  PpsEvent      Sample    CountObc         CountAtPps      CountAtDready   SecSince1970     Nsat Ysrc Hsrc\n"
#define TSFMT   "0x%1X  %10ld  %10ld  %10ld  %08lX %08lX  %08lX %08lX  %18lf  %2d  %2d  %2d\n"
   
FUNCTYPE( int ) ParGpsTsFileRead(  FILE *tsfile, TIMESTAMP *TS, int skiptitle ){

        int c, nval;

        if ( !tsfile || !TS )
                return( 0 );

        if ( skiptitle ) {
                c = fgetc( tsfile );
                while ( c != EOF && c != '\n' )
                        c = fgetc( tsfile );
                }


        nval = fscanf( tsfile, TSFMT,
                 &TS->Valid,
                 &TS->PpsEvent,
                 &TS->Sample,
                 &TS->CountObc,
                 &TS->CountAtPps.u.HighPart,
                 &TS->CountAtPps.u.LowPart,
                 &TS->CountAtDready.u.HighPart,
                 &TS->CountAtDready.u.LowPart,
                 &TS->SecSince1970,
                 &TS->NumSat,
                 &TS->YmdSource,
                 &TS->HmsSource
                 );

        if (nval == EOF)
                nval = 0;

        return( nval );
}

FUNCTYPE( int ) ParGpsTsFileWrite( FILE *tsfile, TIMESTAMP *TS, int writetitle ) {

        if ( !tsfile || !TS )
                return( 0 );

        if ( writetitle )
                fprintf( tsfile, TSTITLE );

        fprintf( tsfile, TSFMT,
                 TS->Valid,
                 TS->PpsEvent,
                 TS->Sample,
                 TS->CountObc,
                 TS->CountAtPps.u.HighPart,
                 TS->CountAtPps.u.LowPart,
                 TS->CountAtDready.u.HighPart,
                 TS->CountAtDready.u.LowPart,
                 TS->SecSince1970,
                 TS->NumSat,
                 TS->YmdSource,
                 TS->HmsSource
               );
        
        return( 1 );
}




// Check two timestamps to see if their CountAt values are normal or if
// one is an outlier.  An outlier is defined as one which is more than LimitXxx
// counts away from the StandardXxx.
//
// The number of 64 bit PC counts between successive PPS signals
// and between successive dready mark points should remain fairly
// constant.  We use this fact to check for and discard outlier
// points.  Outlier points typically result when some other
// program does lots of higher priority interrupts and so delays
// the servicing of the PARxCH/PARGPS interrupt routine.
//
// The number of sample points between successive dready marks is
// expected to be NSps.  But periodically it will be one point
// more or less because the sampling rate with respect to true
// time is not exactly an integer value (ie it is 100.001 not 100).

FUNCTYPE( int ) ParGpsTsIsOutlier( TIMESTAMP *TS0, TIMESTAMP *TS1,
                                   long Nsps, int PcCounterType,
                                   double StandardCountsPps,    double LimitPps,    
                                   double StandardCountsDready, double LimitDready, 
                                   double *ActualCountsPps,     double *DiffPps,
                                   double *ActualCountsDready,  double *DiffDready ) {

        int    IsOutlier, Nsec;
        double Pcounts, Dcounts, Pdiff, Ddiff, Correct;
        long   CurrNsamples, StdNsamples;
    

        // Error check to prevent divide by zero

        if (TS0->PpsEvent == TS1->PpsEvent)
                return( -1 );              // This will show as true, meaning outlier



        // Compute PPS counts and difference from standard
        
        Nsec = TS1->PpsEvent - TS0->PpsEvent;
        Pcounts = (double)ParGpsLargeIntDiffType( TS1->CountAtPps,
                                                  TS0->CountAtPps,
                                                  PcCounterType );
        Pdiff = (Pcounts - StandardCountsPps) / (double)Nsec;


        
        // Compute Dready counts and difference from standard, 
        // correcting for number of samples between the two Dreadys

        Dcounts = (double)ParGpsLargeIntDiffType( TS1->CountAtDready,
                                                  TS0->CountAtDready,
                                                  PcCounterType );
    
        CurrNsamples = TS1->Sample - TS0->Sample;
        StdNsamples  = Nsps * Nsec;
        if ( CurrNsamples == StdNsamples )
                Correct = 0.0L;
        else
                Correct = (double)(StdNsamples - CurrNsamples) * 
                          StandardCountsDready / (double)StdNsamples;
        Dcounts += Correct;
    
        Ddiff = Dcounts - StandardCountsDready;


        // Check to see if differences exceed +/- limits

        if ( Pdiff > 0 && Pdiff > LimitPps )
                IsOutlier = 1;
        else if ( Pdiff < 0 && Pdiff < -LimitPps )
                IsOutlier = 1;
        else if ( Ddiff > 0.0L && Ddiff > LimitDready )
                IsOutlier = 1;
        else if ( Ddiff < 0.0L && Ddiff < -LimitDready )
                IsOutlier = 1;
        else
                IsOutlier = 0;

    
        // Fill optional return variables

        if (ActualCountsPps)
                *ActualCountsPps = Pcounts;
        if (DiffPps)
                *DiffPps = Pdiff;
        if (ActualCountsDready)
                *ActualCountsDready = Dcounts;
        if (DiffDready)
                *DiffDready = Ddiff;

        
    return( IsOutlier );
}


// Correct the CountAt values of timestamp TSx so it matches the value
// that would be expected if it were in line with TS0 and TS1

FUNCTYPE( int ) ParGpsTsOutlierAdjust( TIMESTAMP *TS0, TIMESTAMP *TS1, TIMESTAMP *TSx,
                                       long Nsps, int PcCounterType,
                                       double StandardCountsPps, double StandardCountsDready ) {

        int    ValidFlag, Nsec;
        long   NormalCounts, CurrNsamples, StdNsamples;
        double Correct;
    

        // Set validity flag
        
        ValidFlag = 0x0;
        if ( TS1->Valid == TS_VALID_ALL )
                ValidFlag |= 0x1;
        if ( TS0->Valid == TS_VALID_ALL )
                ValidFlag |= 0x2;

        if (ValidFlag == 0x0) {
                sprintf( ParGpsLastErrorMessage,
                         "No timestamps valid, can't compute outlier replacement\n" );
                return( 0 );
                }
    
    
        Nsec = TSx->PpsEvent - TS1->PpsEvent;


        // Compute new Pps counts

        if (ValidFlag & 0x3) { // both time stamps valid

                NormalCounts = ParGpsLargeIntDiffType( TS1->CountAtPps,
                                                       TS0->CountAtPps,
                                                       PcCounterType );
                }
        
        else { // don't have both time stamps valid
                NormalCounts = (long)StandardCountsPps;
                }


        NormalCounts *= Nsec;
        TSx->CountAtPps = ParGpsLargeIntAddType( TS1->CountAtPps,
                                                 NormalCounts,
                                                 PcCounterType );



        // Compute new Dready counts

        CurrNsamples = TSx->Sample - TS1->Sample;
        StdNsamples  = Nsps * Nsec;

        if ( CurrNsamples == StdNsamples )
                Correct = 0.0L;
        else
                Correct = (double)(StdNsamples - CurrNsamples) *
                          StandardCountsDready / (double)StdNsamples;

        if (ValidFlag & 0x3) { // both time stamps valid

                NormalCounts = ParGpsLargeIntDiffType( TS1->CountAtDready,
                                                       TS0->CountAtDready,
                                                       PcCounterType );
                }
        
        else { // don't have both time stamps valid
                NormalCounts = (long)StandardCountsDready;
                }
        

        NormalCounts = (NormalCounts * Nsec) - (long)Correct;
        TSx->CountAtDready = ParGpsLargeIntAddType( TS1->CountAtDready,
                                                    NormalCounts,
                                                    PcCounterType );

    return( 1 );
}



/* COMPUTE TIME USING TIMESTAMP FUNCTIONS:
 * 
 * Given 2 timestamps and some additional data, interpolate to determine
 * the time in seconds since 1970 for the specified data sample.
 *
 * The SPS method just multiplies the number of samples time the sampling
 * rate.  The PPS and OBC methods are somewhat more complex.
 *
 * In those cases, we use one straight line equation to convert from DAQ sample
 * number to 64 bit PC counts:
 *
 *            c4 - c2
 *      cx = --------- * (nx - n2) + c2            (Eq 2)
 *            n4 - n2
 *
 * We use a slightly different version of that equation to convert from
 * DAQ sample number to OBC counts if they are available:
 *
 *      cx = m * (nx - n2) + c2                    (Eq 5)
 *
 * The second step in either case is to use the straight line equation to
 * convert from number of counts (either type as appropriate) to true time:
 *
 *            t3 - t1
 *      tx = --------- * (cx - c1) + t1            (Eq 1)
 *            c3 - c1
 *
 */

FUNCTYPE( int ) ParGpsTsComputeTimeSps( TIMESTAMP *TS0, TIMESTAMP *TS1,
                                        double Sps, long Sample, double *Time ) {

        // Error check for
        //   1. Time stamps exist and are valid
        //   2. No place to put answer
        //   3. Bad sampling rate
        //   4. Sample index ok
        
        if ( !TS0 || TS0->Valid != TS_VALID_ALL  ||
             TS0->YmdSource == TIME_SOURCE_NONE  ||
             TS0->HmsSource == TIME_SOURCE_NONE  ||
             !Time                               ||
             Sps < 0.000001                      ||
             Sample < 0                          ) {
                sprintf( ParGpsLastErrorMessage, "Invalid parameter\n" );
                return( 0 );
                }

        ParGpsLastErrorMessage[0] = '\0';


        *Time = TS0->SecSince1970 + (double)(Sample - TS0->Sample) / Sps;

        return( 1 );
}

FUNCTYPE( int ) ParGpsTsComputeTimePps( TIMESTAMP *TS0, TIMESTAMP *TS1,
                                        double CountsBetweenPps, int CounterType,
                                        long Sample, double *Time ) {

        double  c1, c2, c3, c4, cx, t1, t3, tx;

        
        // Compute Time at Sample using PC counter from PARGPS PPS area
        
        // Error check our incoming data

        if ( !TS0 || TS0->Valid != TS_VALID_ALL         ||
             !TS1 || TS1->Valid != TS_VALID_ALL         ||
             TS0->YmdSource == TIME_SOURCE_NONE         ||
             TS0->HmsSource == TIME_SOURCE_NONE         ||
             TS1->YmdSource == TIME_SOURCE_NONE         ||
             TS1->HmsSource == TIME_SOURCE_NONE         ||
             CountsBetweenPps < 0.00001                 ||
             Sample < 0                                 ||
             CounterType < PC_COUNTER_TYPE_INT64        ||
             CounterType > PC_COUNTER_TYPE_HIGH32LOW32  ||
             !Time                                      ) {
                sprintf( ParGpsLastErrorMessage, "Invalid parameter\n" );
                return( 0 );
                }

        ParGpsLastErrorMessage[0] = '\0';


        
        // Step 0: Perform a translation.
        //
        // This is done to simplify further computations.  Subtract
        // TS0->CountAtPps (ie c1) from all count values and
        // TS0->PpsEvent (ie t1) from all t values.  When translating t
        // values, we could subtract either PpsEvent values or
        // SecSince1970 values.  It doesn't matter which since the
        // adjacent PPS signals always differ by both 1 second and 1
        // PpsEvent.  At the end, we will reverse the -t1 translation by
        // adding SecSince1970 back to the final tx result.
        
        c1 = 0;
        c2 = ParGpsLargeIntDiffType( TS0->CountAtDready,
                                     TS0->CountAtPps,
                                     CounterType );
        c3 = ParGpsLargeIntDiffType( TS1->CountAtPps,
                                     TS0->CountAtPps,
                                     CounterType );
        c4 = ParGpsLargeIntDiffType( TS1->CountAtDready,
                                     TS0->CountAtPps,
                                     CounterType );
        
        t1 = 0;
        t3 = TS1->PpsEvent - TS0->PpsEvent;



        // Step 1: Solve Eq 2 for cx, a 64 bit PC 
        // count for the DAQ sample of interest

        if ( !ParGpsEquation2pt( TS0->Sample, c2,     // (n2,c2)
                                 TS1->Sample, c4,     // (n4,c4)
                                      Sample, &cx ) ) // (nx,cx)
                return( 0 );

        
        // Correct for difference in length of PPS and DREADY signals

        cx = cx - (SIGNAL_LENGTH_CORRECTION * CountsBetweenPps);


        
        // Step 2: Solve Eq 1 for tx, the true
        // time for the DAQ sample of interest

        if ( !ParGpsEquation2pt( c1, t1,   c3, t3,   cx, &tx ) )
                return( 0 );
        

        
        // Finish calculation

        *Time = tx + TS0->SecSince1970;  // Add back in time at t1
        
        
        return( 1 );
}

FUNCTYPE( int ) ParGpsTsComputeTimeObc( TIMESTAMP *TS0, TIMESTAMP *TS1,
                                        double ObcCountsPerSample,
                                        long Sample, double *Time ) {

        double  c1, c2, c3, c4, cx, m, t1, t3, tx;
   
        // Compute time at specified sample point using the
        // PAR8CH 800 ns On Board Counter information from two
        // time stamp structures.

        // Make sure our raw data is ok

        if ( !TS0 || TS0->Valid != TS_VALID_ALL  ||
             !TS1 || TS1->Valid != TS_VALID_ALL  ||
             TS0->YmdSource == TIME_SOURCE_NONE  ||
             TS0->HmsSource == TIME_SOURCE_NONE  ||
             TS1->YmdSource == TIME_SOURCE_NONE  ||
             TS1->HmsSource == TIME_SOURCE_NONE  ||
             ObcCountsPerSample < 0.00001        ||
             Sample < 0                          ||
             !Time                               ) {
                sprintf( ParGpsLastErrorMessage, "Invalid parameter\n" );
                return( 0 );
                }

        ParGpsLastErrorMessage[0] = '\0';


        // Step 0: Initialize some variables
        //
        // We translate the counts variable so C2 is the C origin
        // and the time variable so T1 is the T origin.  For T we
        // can work with either PpsEvent or SecSince1970 since we
        // know there is exactly 1 second per PpsEvent.  We will
        // reverse the tranlation to T at the end by adding back
        // the SecSince1970 at t1 to the final result tx.

        m  = ObcCountsPerSample;
        c2 = 0.0L;
        t1 = 0.0L;
        t3 = TS1->PpsEvent - TS0->PpsEvent;


        // Step 1a: Compute the OBC count at D2 from Eq 5

        if ( !ParGpsEquationSlope( m,
                                   TS0->Sample, c2,     // (n2,c2)
                                   TS1->Sample, &c4 ) ) // (n4,c4)
                return( 0 );


        // Step 1b: Compute the OBC count at Sx from Eq 5

        if ( !ParGpsEquationSlope( m,
                                   TS0->Sample, c2,     // (n2,c2)
                                   Sample,      &cx ) ) // (nx,cx)
                return( 0 );


        // Step 1c: Compute the OBC counts at P1 and P2
        // from the known OBC counts relative to D1 and D1

        c1 = c2 - TS0->CountObc;                        // c2 - o21
        c3 = c4 - TS1->CountObc;                        // c4 - o43

        
        // Step 2: Solve Eq 1 for tx, the true 
        // time for the DAQ sample of interest

        if ( !ParGpsEquation2pt( c1, t1,   c3, t3,   cx, &tx ) )
                return( 0 );

        
        // Finish calculation

        *Time = tx + TS0->SecSince1970;  // Add back in time at t1
   

        return( 1 );
}



// ****************************************************************************
// These functions are helpers and are NOT available for users to call directly
//
//
//


/* CIRCULAR TIMESTAMP BUFFER HELPER FUNCTIONS:
 * 
 * Set up a circular buffer that has 1 read pointer and 3 write
 * pointers.  There is one write pointer for each type of data: analog,
 * GPS PPS, and GPS serial.  The read pointer is used to read "valid"
 * buffers; that is, ones which have had all of their three types of
 * data filled in.
 *
 */


static int rdptr  = 0;
static int rdskip = 0;

SRLOCAL int ParGpsInitCircBuff( void ) {
        
        int i;

        // Initialize time stamp buffer

        for ( i = 0 ; i < MAXCIRC ; i++ ) {
                CircBuff[i].Valid    = TS_VALID_NONE;
                CircBuff[i].PpsEvent = 0;
                }

        CbNumValid = 0;

        return( 1 );
}

SRLOCAL int ParGpsSkipCircBuff( void ) {

        // In special cases, we may need to skip ahead one entry
        // to account for uneven startup conditions (eg where PPS data 
        // for a given analog mark has been read in earlier file that
        // we don't have access to).

        CircBuff[rdptr].Valid = TS_VALID_NONE;
        ParGpsIncCbPtr( &rdptr );
        rdskip = 1;

        return( rdptr );
}

SRLOCAL int ParGpsReadCircBuff( TIMESTAMP *Buffer ) {

        // Check for empty or not ready

        if (CircBuff[rdptr].Valid != TS_VALID_ALL)
                return( 0 );



        // Read data

        *Buffer = CircBuff[rdptr];



        // Clear valid flag and update read pointer

        CircBuff[rdptr].Valid = TS_VALID_NONE;
        CbNumValid--;

        ParGpsIncCbPtr( &rdptr );


        // Indicate we have successfully gotten past startup skips

        rdskip = 0;

        return( 1 );
}

SRLOCAL int CheckValid( int wrptr ) {

        if ( CircBuff[wrptr].Valid == TS_VALID_ALL ) {
                CbNumValid++;
                return( CbNumValid );
                }
        else
                return( 0 );
}

SRLOCAL int ParGpsAnalogWriteCircBuff( unsigned int PpsEvent, long Sample, int Obc ) {

        static int wrptr = 0;


        // Ignore buffer entries already skipped

        ParGpsUpdateCbPtrForSkip( &wrptr );


        // Check for overflow

        if ( (CircBuff[wrptr].Valid & TS_VALID_ANALOG) != TS_VALID_NONE) {
                sprintf( ParGpsLastErrorMessage,
                         "ERROR: CircBuff overflow on Analog write\n");
                return( 0 );
                }


        // Check or write PpsEvent

        if ( !ParGpsCheckEventNum( PpsEvent, &wrptr, "Analog" ) )
                return( 0 );

        
        // Write analog data to buffer

        CircBuff[wrptr].Valid   |= (TS_VALID_ANALOG | TS_VALID_OBC); 
        CircBuff[wrptr].Sample   = Sample;
        CircBuff[wrptr].CountObc = Obc;


        // See if CB element is now completely valid

        CheckValid( wrptr );


        // Update analog write pointer

        ParGpsIncCbPtr( &wrptr );


        return( 1 );
}

SRLOCAL int ParGpsPpsWriteCircBuff( unsigned int PpsEvent,
                                    SR64BIT CountAtPps, SR64BIT CountAtDready ) {

        static int wrptr = 0;


        // Ignore buffer entries already skipped

        ParGpsUpdateCbPtrForSkip( &wrptr );


        // Check for overflow

        if ( (CircBuff[wrptr].Valid & TS_VALID_PPS) != TS_VALID_NONE) {
                sprintf( ParGpsLastErrorMessage,
                         "ERROR: CircBuff overflow on pps write\n");
                return( 0 );
                }


        // Check or write PpsEvent

        if ( !ParGpsCheckEventNum( PpsEvent, &wrptr, "Pps" ) )
                return( 0 );



        // Write Pps data to buffer

        CircBuff[wrptr].Valid        |= TS_VALID_PPS;
        CircBuff[wrptr].CountAtPps    = CountAtPps;
        CircBuff[wrptr].CountAtDready = CountAtDready;


        // See if CB element is now completely valid

        CheckValid( wrptr );


        // Update pps write pointer

        ParGpsIncCbPtr( &wrptr );


        return( 1 );
}

SRLOCAL int ParGpsSerialWriteCircBuff( unsigned int PpsEvent, double SecSince1970,
                                       int YMDSource, int HMSSource, int NumSat,
                                       int Erase ) {

        static int wrptr = 0;



        // Ignore buffer entries already skipped

        ParGpsUpdateCbPtrForSkip( &wrptr );

        
        // Check for overflow

        if ( (CircBuff[wrptr].Valid & TS_VALID_SERIAL) != TS_VALID_NONE) {
                sprintf( ParGpsLastErrorMessage,
                         "ERROR: CircBuff overflow on Serial write\n");
                return( 0 );
                }


        // Check or write PpsEvent

        if ( !ParGpsCheckEventNum( PpsEvent, &wrptr, "Serial" ) )
                return( 0 );



        // Write Serial data to buffer

        CircBuff[wrptr].Valid       |= TS_VALID_SERIAL;
        CircBuff[wrptr].SecSince1970 = SecSince1970;
        CircBuff[wrptr].YmdSource    = YMDSource;
        CircBuff[wrptr].HmsSource    = HMSSource;
        CircBuff[wrptr].NumSat       = NumSat;
//        CircBuff[wrptr].Erase        = Erase;
        

        // See if CB element is now completely valid

        CheckValid( wrptr );


        // Update Serial write pointer
        
        ParGpsIncCbPtr( &wrptr );

        return( 1 );
}

SRLOCAL int ParGpsIncCbPtr( int *PtrIndex ) {
        
        *PtrIndex += 1;
        if (*PtrIndex >= MAXCIRC)
                *PtrIndex = 0;

        return( *PtrIndex );
}

SRLOCAL int ParGpsUpdateCbPtrForSkip( int *PtrIndex ) {
        
        if ( rdskip && (*PtrIndex < rdptr ))
                *PtrIndex = rdptr;
        
        return( *PtrIndex );
}

SRLOCAL int ParGpsCheckEventNum( unsigned int PpsEvent, int *wrptrAPS, char *StrAPS ) {

        int          wrptr;
        unsigned int i, DiffEvent;

        wrptr = *wrptrAPS;
        
        // Check or write PpsEvent

        if ( CircBuff[wrptr].Valid != TS_VALID_NONE) {
                
                if (PpsEvent < CircBuff[wrptr].PpsEvent ) {
                        
                        // This code discards data for an earlier event
                        // since we know we will never get all three
                        // required types of data (analog, pps, serial)
                        // for that earlier event.
                        
                        sprintf( ParGpsLastErrorMessage,
                                 "Ignoring %s PpsEvent %u < existing PpsEvent %u\n",
                                 StrAPS, PpsEvent, CircBuff[wrptr].PpsEvent );
                        return( 0 );
                        }

                else if (PpsEvent > CircBuff[wrptr].PpsEvent ) {
                        
                        // This code corrects for start up conditions
                        // where one of the three types of data (analog,
                        // pps, serial) needed to determine the time of
                        // an event was stored in an earlier file that
                        // we know nothing about.  In this case, we just
                        // throw away the current buffer entry and start
                        // with the next event.
                        
                        DiffEvent = PpsEvent - CircBuff[wrptr].PpsEvent;
                        for ( i = 0 ; i < DiffEvent ; i++ ) {
                                sprintf( ParGpsLastErrorMessage,
                                         "Discarding PpsEvent %u < %s PpsEvent %u\n",
                                         CircBuff[*wrptrAPS].PpsEvent, StrAPS, PpsEvent );
                                CircBuff[*wrptrAPS].Valid = TS_VALID_NONE;
                                ParGpsIncCbPtr( wrptrAPS );
                                ParGpsSkipCircBuff( );
                                }
                        }
                }
        else {
                // Assign the PpsEvent to this buffer entry since it
                // doesn't have one yet.

                CircBuff[wrptr].PpsEvent = PpsEvent;
                }

        return( 1 );
}


SRLOCAL int ParGpsSrcToSource( int YmdSrc, int HmsSrc, int *YmdSource, int *HmsSource ) {

        int YmdTemp, HmsTemp;

        // Translate between specific NMEA string source constants
        // and more general source constants

        // For Year Month Day
        
        if ( YmdSrc == SRCSTR_PC )
                YmdTemp = TIME_SOURCE_PC;
        
        else if ( YmdSrc == SRCSTR_UNK )
                YmdTemp = TIME_SOURCE_NONE;
        
        else if ( YmdSrc > SRCSTR_MIN  &&  YmdSrc < SRCSTR_MAX )
                YmdTemp = TIME_SOURCE_GPS;

        else
                YmdTemp = TIME_SOURCE_NONE;


        // For Hour Minute Second

        if ( HmsSrc == SRCSTR_PC )
                HmsTemp = TIME_SOURCE_PC;
        
        else if ( HmsSrc == SRCSTR_UNK )
                HmsTemp = TIME_SOURCE_NONE;
        
        else if ( HmsSrc > SRCSTR_MIN  &&  HmsSrc < SRCSTR_MAX )
                HmsTemp = TIME_SOURCE_GPS;

        else
                HmsTemp = TIME_SOURCE_NONE;


        // Fill optional return parameters

        if ( YmdSource )
                *YmdSource = YmdTemp;
        if ( HmsSource )
                *HmsSource = HmsTemp;
        
        return( 1 );
}
        




/* INTERPOLATION HELPER FUNCTIONS:
 * 
 * Given the x value for one point and some other data, compute the y
 * value for that point.  The other data may be two additional x,y points
 * or one x,y point and a slope.  Linear interpolation is used in either
 * case.
 *
 */

SRLOCAL int ParGpsEquation2pt( double x1, double y1,
                               double x2, double y2, 
                               double xk, double *yk ) {

        double diff;

        // Error check for
        //   1. No place to put answer
        //   2. Points are identical (this would lead to divide by zero)
        

        if ( x2 > x1 )
                diff = x2 - x1;
        else
                diff = x1 - x2;

        if ( !yk  || diff < TOLERANCE )
                return( 0 );
    


        // Solve the equation
        
        *yk = ((y2 - y1) * (xk - x1) / (x2 - x1)) + y1;


        return( 1 );
}

SRLOCAL int ParGpsEquationSlope( double m,
                                 double x1, double y1,
                                 double xk, double *yk ) {

        // Error check for no place to put answer
        
        if ( !yk ) 
                return( 0 );


        // Solve the equation

        *yk = m * (xk - x1) + y1;

        
        return( 1 );
}



/*************************************************************************** 
 *
 * OS DEPENDENT HELPER FUNCTIONS:
 *
 * These functions exist to hide operating system differences from the
 * main library functions above.
 *
 *
 *
 */

#if defined( SROS_WIN2K ) || defined( SROS_WINNT )

SRLOCAL DEVHANDLE ParGpsOsDriverOpen( char *DriverName ) {

        DEVHANDLE SrHandle;
        HANDLE    NtHandle;
        char      DeviceNameBuffer[256];

        sprintf( DeviceNameBuffer, "\\\\.\\%s", DriverName );


        NtHandle = CreateFile(
                        DeviceNameBuffer,                   // "file" name
                        GENERIC_READ    | GENERIC_WRITE,    // access mode
                        FILE_SHARE_READ | FILE_SHARE_WRITE, // share mode
                        NULL,                               // security
                        OPEN_EXISTING,                      // create mode
                        0,                                  // file attrib
                        NULL                                // copy from
                        );

        if (NtHandle == INVALID_HANDLE_VALUE)
                SrHandle = BAD_DEVHANDLE;
        else
                SrHandle = (DEVHANDLE)NtHandle;
        
        return( SrHandle );
}

SRLOCAL int ParGpsOsDriverClose( DEVHANDLE *ParGpsHandle ) {

        int ret;

        ret = CloseHandle( *ParGpsHandle );
        *ParGpsHandle = BAD_DEVHANDLE;

        return( ret );
}

SRLOCAL int ParGpsOsDriverIoctl( DEVHANDLE      ParGpsHandle,
                                 unsigned long  IoCtlCode,
                                 void          *pValueIn,
                                 unsigned long  InSize,
                                 void          *pValueOut,
                                 unsigned long  OutSize,
                                 unsigned long *pBytesReturned
                               ) {

        int ret;

                ret = DeviceIoControl(
                       ParGpsHandle,     // Handle to device
                       IoCtlCode,        // IO Control code
                       pValueIn,         // Input data to driver
                       InSize,           // Length of in data in bytes
                       pValueOut,        // Output data from driver
                       OutSize,          // Length of out data in bytes
                       pBytesReturned,   // Bytes placed in output buffer
                       NULL              // NULL = wait till I/O completes
                       );

        return( ret );
}

SRLOCAL long ParGpsOsGetLastError( void ) {

        return( GetLastError() );  // Provided by the OS
}

SRLOCAL DEVHANDLE ParGpsOsSerialOpen( char *SerialPortName ) {

        DEVHANDLE hSerial;        
        

        // Open serial port device as non-blocking.

        hSerial = CreateFile( SerialPortName,               // port name
                              GENERIC_READ | GENERIC_WRITE, // access
                              0,                            // no share
                              NULL,                         // no security attrib
                              OPEN_EXISTING,                // create type
                              0,                            // non-overlapped I/O
                              NULL                          // no template file
                            );

//                            FILE_FLAG_OVERLAPPED,         // overlapped I/O


        if ( hSerial == INVALID_HANDLE_VALUE )
                hSerial = BAD_DEVHANDLE;

        return( hSerial );
}

SRLOCAL int ParGpsOsSerialInit( DEVHANDLE hSerial ) {

        DCB NewSettings;
        DWORD EventMask;
        COMMTIMEOUTS NewTimeouts;


        // Can't do init for a bad serial handle

        if (hSerial == BAD_DEVHANDLE)
                return( 0 );
        



        // Prepare new port settings.

        /* For a complete discussion of these flags see the MSDN
         * documentation for the DCB structure.
         *
         */

        NewSettings.DCBlength = sizeof(DCB);
        if (GetCommState( hSerial, &NewSettings ) == 0) {
                sprintf( ParGpsLastErrorMessage, "Failed to get state of COM port\n");
                return( 0 );
                }


        NewSettings.BaudRate        = CBR_4800;     // baud rate 4800
        NewSettings.ByteSize        = 8;            // 8 data bits
        NewSettings.Parity          = NOPARITY;     // no parity
        NewSettings.StopBits        = ONESTOPBIT;   // 1 stop bit
        
        NewSettings.fDsrSensitivity = FALSE;        // local, not modem
        NewSettings.fNull           = FALSE;        // no discard of nulls
        NewSettings.fOutX           = FALSE;        // no outgoing XON/XOFF
        NewSettings.fInX            = FALSE;        // no incoming XON/XOFF
        NewSettings.fOutxCtsFlow    = FALSE;        // no CTS flow control 
        NewSettings.fOutxDsrFlow    = FALSE;        // no DSR flow control 
        NewSettings.fDtrControl     = DTR_CONTROL_DISABLE; // no DTR flow control 
        NewSettings.fRtsControl     = RTS_CONTROL_DISABLE; // no RTS flow control 
        NewSettings.fTXContinueOnXoff = TRUE;       // XOFF continues Tx 
        NewSettings.fErrorChar      = FALSE;        // disable error 
        NewSettings.fAbortOnError   = FALSE;        // no r/w error abort
                                      
        NewSettings.EofChar         = '\n';         // Carriage Return is both
        NewSettings.EvtChar         = '\n';         // end of input and event


        NewTimeouts.ReadIntervalTimeout         = MAXDWORD; // Allow non-blocking
        NewTimeouts.ReadTotalTimeoutMultiplier  = 0;
        NewTimeouts.ReadTotalTimeoutConstant    = 0;
        NewTimeouts.WriteTotalTimeoutMultiplier = 0;        // Don't use
        NewTimeouts.WriteTotalTimeoutConstant   = 0;



//      Linux version registers cleanup function for Ctrl-C interrupt here
//      signal( SIGINT, ParGpsOsSerialSignalHandler );



        // Clear line and activate settings.

        if (SetCommTimeouts( hSerial, &NewTimeouts ) == 0) {
                sprintf( ParGpsLastErrorMessage, "Failed to set Timeouts for COM port\n");
                return( 0 );
                }

        if (SetCommState( hSerial, &NewSettings ) == 0) {
                sprintf( ParGpsLastErrorMessage, "Failed to set State for COM port\n");
                return( 0 );
                }


        
        // Set RTS low to provide negative power for SR PARGPS board

        EscapeCommFunction( hSerial, CLRRTS );

        
        // Set events

        EventMask = EV_RXFLAG | EV_RING;
        SetCommMask( hSerial, EventMask );


        
        // Clear all comm buffers
        
        PurgeComm( hSerial, PURGE_TXABORT | PURGE_TXCLEAR |
                            PURGE_RXABORT | PURGE_RXCLEAR );


        return( 1 );
}

SRLOCAL void ParGpsOsSerialClose( DEVHANDLE hSerial ) {

        
        // Close the serial port.

        CloseHandle( hSerial );
}

SRLOCAL int ParGpsOsSetPcTime( double Time,
                               int Year, int Month, int Day, 
                               int Hour, int Minute, int Second,
                               int FromYMD ) {

        SYSTEMTIME Systim;
        int        Yr, Mo, Dy, Hh, Mm, Ss;
        long       Microsecond;

        // Prepare time in SYSTEMTIME format Windows prefers

        if (FromYMD == 1) {
                Systim.wYear   = Year;
                Systim.wMonth  = Month;
                Systim.wDay    = Day;
                Systim.wHour   = Hour;
                Systim.wMinute = Minute;
                Systim.wSecond = Second;
                }
        else {
                ParGpsSecTimeSplit( Time, &Yr, &Mo, &Dy, &Hh, &Mm, &Ss, &Microsecond );
                Systim.wYear   = Yr;
                Systim.wMonth  = Mo;
                Systim.wDay    = Dy;
                Systim.wHour   = Hh;
                Systim.wMinute = Mm;
                Systim.wSecond = Ss;
                }
        
        Systim.wMilliseconds = 0;
        Systim.wDayOfWeek    = 7; // this invalid number is ignored

        if (SetSystemTime( &Systim ) == 0) {
                ParGpsLastDriverError = ParGpsOsGetLastError();
                return( 0 );
                }
        else
                return( 1 );
}

SRLOCAL long ParGpsOsLargeIntDiff( SR64BIT A, SR64BIT B ) {

        long Answer;

        Answer = (long)(A.QuadPart - B.QuadPart);

        return( Answer );
}


#elif defined( SROS_LINUX )

SRLOCAL DEVHANDLE ParGpsOsDriverOpen( char *DriverName ) {
        
        DEVHANDLE ParGpsHandle;
        char      DeviceNameBuffer[256];

        sprintf( DeviceNameBuffer, "/dev/%s", DriverName );

        
        ParGpsHandle = open( DeviceNameBuffer, O_RDWR );
        
        if (ParGpsHandle < 0)
                ParGpsHandle = BAD_DEVHANDLE;
        
        return( ParGpsHandle );
}

SRLOCAL int ParGpsOsDriverClose( DEVHANDLE *ParGpsHandle ) {

        int ret;

        if ( close( *ParGpsHandle ) < 0 )
                ret = 0;
        else
                ret = 1;
        *ParGpsHandle = BAD_DEVHANDLE;

        return( ret );
}

SRLOCAL int ParGpsOsDriverIoctl( DEVHANDLE      ParGpsHandle,
                                 unsigned long  IoCtlCode,
                                 void          *pValueIn,
                                 unsigned long  InSize,
                                 void          *pValueOut,
                                 unsigned long  OutSize,
                                 unsigned long *pBytesReturned
                               ) {
        
        int code;
        IRP IoctlData;



        // The ioctl function just takes one buffer.  On the way in,
        // it contains the input data and on the way out, it contains
        // the output data.  So, make sure we are using the largest of
        // the two and that it contains the input data.  The size of the
        // data must be the size defined for that IO control code in
        // pargpskd.h.


        IoctlData.Command       = IoCtlCode;
        IoctlData.InBuffer      = pValueIn;
        IoctlData.InSize        = InSize;
        IoctlData.OutBuffer     = pValueOut;
        IoctlData.OutSize       = OutSize;
        IoctlData.ReturnedBytes = 0;
        IoctlData.DataMethod    = METHOD_BUFFERED;
        IoctlData.ErrorCode     = 0;

        // Check in pargpskd.h for which control codes do not
        // use the more common buffered method of passing data
        
	if ((IoCtlCode == IOCTL_PARGPS_READ_PPS_DATA)     ||
	    (IoCtlCode == IOCTL_PARGPS_READ_SERIAL_DATA) ) {

                IoctlData.DataMethod = METHOD_OUT_DIRECT;
                }


        // Request the ioctl function

        code = ioctl( ParGpsHandle, IoCtlCode, &IoctlData );

        if (pBytesReturned)
                *pBytesReturned = IoctlData.ReturnedBytes;

        
        
        if (code < 0)
                return( 0 );
        else
                return( 1 );
        
}

SRLOCAL long ParGpsOsGetLastError( void ) {

        return( (long)errno );
}

SRLOCAL void ParGpsOsSerialClose( DEVHANDLE hSerial ) {

        // Close serial port.

        close( hSerial );
        
        Signal_hSerial = BAD_DEVHANDLE;
}

SRLOCAL void ParGpsOsSerialSignalHandler( int SigNum ) {

        if (Signal_hSerial != BAD_DEVHANDLE)
                ParGpsOsSerialClose( Signal_hSerial );

        exit( 1 );
}

SRLOCAL DEVHANDLE ParGpsOsSerialOpen( char *SerialPortName ) {
        
        DEVHANDLE hSerial;
        
        // Open serial port device as non-blocking.

        hSerial = open( SerialPortName, O_RDONLY | O_NONBLOCK );

        if ( hSerial < 0 ) {
                hSerial = BAD_DEVHANDLE;
                Signal_hSerial = hSerial;
                perror( SerialPortName );
                return( hSerial);
                }
        else
                Signal_hSerial = hSerial;

        return( hSerial );
}

SRLOCAL int ParGpsOsSerialInit( DEVHANDLE hSerial ) {

        struct termios NewSettings;
        int            ModemStatus;


        // Can't do init for a bad serial handle

        if (hSerial == BAD_DEVHANDLE)
                return( 0 );
        

        // Prepare new port settings.

        // For a complete discussion of these flags see the Linux info
        // system for libc:Low Level Terminal Interface:Terminal Modes

        tcgetattr( hSerial, &NewSettings );

        NewSettings.c_iflag &= ~INPCK;   // input parity checking
        NewSettings.c_iflag &= ~IGNCR;   // input not ignore CR
        NewSettings.c_iflag &= ~ICRNL;   // input not change CR to NL
        NewSettings.c_iflag &= ~INLCR;   // input not change NL to CR
        NewSettings.c_iflag &= ~IXOFF;   // input  no XON/XOFF flow control
        NewSettings.c_iflag &= ~IXON;    // output no XON/XOFF flow control
        // IGNBRK ignore break
        // BRKINT break raises SIGINT
        // IGNPAR ignore frame or parity error
        // PARMRK mark   frame or parity error
        // ISTRIP strip to 7 bits
        // IXANY  any character restarts output
        // IMAXBEL send bell when terminal input buffer full (BSD only)

        NewSettings.c_oflag &= ~OPOST;   // output processing not applied

        NewSettings.c_cflag &= ~CBAUD;   // baudrate mask
        NewSettings.c_cflag |=  B4800;   // baudrate 4800
        NewSettings.c_cflag &= ~CSIZE;   // data bit mask
        NewSettings.c_cflag |=  CS8;     // 8 data bits
        NewSettings.c_cflag &= ~CSTOPB;  // use 1 stop bits not 2
        NewSettings.c_cflag |=  CREAD;   // enable receiving characters
        NewSettings.c_cflag &= ~PARENB;  // parity disabled
        NewSettings.c_cflag &= ~HUPCL;   // no hang up on modem disconnect
        NewSettings.c_cflag |=  CLOCAL;  // local connection, no modem contol
        // PARODD odd parity

        NewSettings.c_lflag |=  ISIG;    // listen to signals
        NewSettings.c_lflag &= ~ICANON;  // disable canonical input
        NewSettings.c_lflag &= ~ECHO;    // no echo characters
        NewSettings.c_lflag &= ~ECHOE;   // no echo erase
        NewSettings.c_lflag &= ~ECHOK;   // no echo kill
        NewSettings.c_lflag &= ~ECHONL;  // no echo newline
        // NOFLSH no flush on signal
        // TOSTOP job control stop for terminal io from background

        NewSettings.c_cc[VMIN]  = 0;     // Can return with 0 chars
        NewSettings.c_cc[VTIME] = 0;     // Number of .1 secs to wait



        // Register cleanup function for Ctrl-C interrupt.

        signal( SIGINT, ParGpsOsSerialSignalHandler );



        // Activate settings.

        tcsetattr( hSerial, TCSANOW, &NewSettings );



        // Set RTS low to provide negative power for SR PARGPS board
        // The sense of the bit to the signal is somewhat confused,
        // but by experiment, it seems that clearing the bit causes
        // the signal to go to -volts which is what we need.
        
        ModemStatus = TIOCM_RTS;
        ioctl( hSerial, TIOCMBIC, &ModemStatus ); // just clr bit for active low



//      Winnt version sets up events or select here
//      EventMask = EV_RXFLAG | EV_RING;  SetCommMask( hSerial, EventMask );
        
        

        // Clear line.

        tcflush( hSerial, TCIFLUSH );

        return( 1 );
}

SRLOCAL int ParGpsOsSetPcTime( double Time,
                               int Year, int Month, int Day, 
                               int Hour, int Minute, int Second,
                               int FromYMD ) {

        struct timeval tv;
        int            Yr, Mo, Dy, Hh, Mm, Ss;
        long           Microsecond;
        double         SetTime;


        // Prepare time in timeval format Linux prefers

        if (FromYMD == 1)
                ParGpsSecTimeCombine( Year, Month, Day, Hour, Minute, Second, 0L,
                                      &SetTime );
        else
                SetTime = Time;
        
        tv.tv_sec = (long)SetTime;
        tv.tv_usec = 0;


        if (settimeofday( &tv, NULL ) == -1) {
                ParGpsLastDriverError = ParGpsOsGetLastError();
                return( 0 );
                }
        else
                return( 1 );
}

SRLOCAL long ParGpsOsLargeIntDiff( SR64BIT A, SR64BIT B ) {

        long Answer, seconds, usec;

        seconds = (long)(A.u.HighPart - B.u.HighPart);
        usec    = (long)(A.u.LowPart  - B.u.LowPart);
        Answer  = (seconds * 1000000) + usec;

        return( Answer );
}

#endif // SROS_xxx
