/* FILE: sethdrtm.c              Copyright (c), Symmetric Research, 2003-2005
 *      
 * Read in SR .DAT files and calculate the GPS time for each marked
 * data point.
 *
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>      // for exit
#include <string.h>      // for strxxx
#include <ctype.h>       // for toupper
#include <fcntl.h>       // for O_xxx defines
#include <sys/stat.h>    // for S_IREAD define (VC++ can use / instead of \)


#if defined( SROS_LINUX ) 
#define O_BINARY 0x00
#endif


#include "pargps.h"
#include "srdat.h"



// Typedefs and Prototypes

int  ProcessCommandLine( int argc, char** argv, int *SeqNum, int *Sequential );
int  CheckMarkAndCounter( long MarkChannel, long PcCounterType, long PcCounterFreq );
long SetObcChannel( long NumChannels, long MarkChannel );
int  SelectTimeMethod( long ObcChannel, long MarkChannel );
int  ReadDatRecord( void  *PC_BUFFER, unsigned int *npts, unsigned int maxpts, 
                    void *PPS_BUFFER, unsigned int *npps, unsigned int maxpps, 
                    void *SER_BUFFER, unsigned int *nser, unsigned int maxser );
int  FileExists( char *FileName );
void Usage( char *SpecificMsgText );
void PrintMsg( char *MsgArray[] );
void CleanExit( int value );



// Defines

#define XCHMODEL_UNKNOWN  0x00    // These are similar to the PARXCH_MODEL_xxx
#define XCHMODEL_PAR1CH   0x01    // defines in parxch.h but don't really need
#define XCHMODEL_PAR4CH   0x04    // to be identical
#define XCHMODEL_PAR8CH   0x08


#define PMODE                      (S_IREAD|S_IWRITE)


// MAXANALOG is defined large enough to hold data from any PARxCH from
// the PAR1CH to the PAR8CH.  The PAR1CH can have 2 channels, and the
// PAR4CH can have 5, but the PAR8CH can have up to 11 eg 8 analog + 1
// digital + 1 gps + 1 counter.

#define MAXANALOG (128*11) // 128 pts/ch * 11 channels 
#define MAXPPS        26   // This is large enough for sampling as often
#define MAXSER        26   // as about 5 samples per second.
#define MAXNAME       16   // Files names are in 8.3 format.
#define MAXTITLE     255
#define MAXFILENAME  256



// Globals (Keep large arrays in global area and not on the stack)

int        fin;   // handle of input data file 
FILE      *ftmp;  // handle of temporary file
long       PC_BUFFER[MAXANALOG];
PPSDATA    PPS_BUFFER[MAXPPS];
SERIALDATA SER_BUFFER[MAXSER];
char       TitleLine[MAXTITLE], PrefixCheck[MAXFILENAME];
char       FileIn[MAXFILENAME], FilePrefix[MAXFILENAME];


char* UsageMsg[] = {
    "Command Line Usage: sethdrtm startfile.DAT [/seq]                 ",
    "                                                                  ",
    "       examples: > sethdrtm 0000001F.DAT /seq                     ",
    "                                                                  ",
    "Sethdrtm uses the GPS information recorded in the body of the     ",
    ".DAT file to update the file start time in the header.  Sethdrtm  ",
    "takes two arguments.  The first is the name of the first .DAT     ",
    "file to process.  The second /seq argument is optional and        ",
    "indicates the file is the first in a sequence.  In this case,     ",
    "the file prefix must be numeric as sethdrtm will increment the    ",
    "prefix and update the times until the next file in the sequence   ",
    "can not be found.                                                 ",
    "                                                                  ",
    NULL
    };


char* PartialDataMsg[] = {
    "Note: Because the PPS and serial data arrive in a staggered       ",
    "manner, both types of data may not be available for the first     ",
    "few PPS events in the starting .DAT file.  Sethdrtm ignores this  ",
    "partial data and begins processing with the first complete data.  ",
    "                                                                  ",
    NULL
    };





int main( int argc, char **argv ) {

        SRDAT_HDR Hdr;
        int          hh, mm, ss, year, month, day, nval, nval1, ok;
        int          NumChannels, NumSamples, PrevUnset, TimeEntry;
        int          NotDone, SeqNum, SeqNumSave, Sequential, SerialDelay;
        int          Nvalid, TsTitle, TimeMethod;
        unsigned int PrevEventNum, PpsEvent, Npts, Npps, Nser;
        long         FirstSample, StartSample, MarkChannel;
        long         usec, ObcChannel, PcCounterType, PcCounterFreq;
        double       FirstTime, SamplingRate, PrevPpsSecTime;
        double       ObcCountsPerSample, BestCountsBetweenPps;
        TIMESTAMP    TS0, TS1, TSx;
        struct SrDatTimeData Tp;


        // Display the leader message.

        printf( "PARGPS sethdrtm: (c) SR 2004 : Rev {11/15/04}\n" );



        
        // Initialize variables

        PrevUnset      = 1;
        PrevPpsSecTime = 0.0L;
        PrevEventNum   = INVALID_PPSEVENT;
        fin            = -1;
        ftmp           = NULL;
        SerialDelay    = 0;



        // Process command line arguments

        if ( !ProcessCommandLine( argc, argv, &SeqNum, &Sequential ) )
                Usage( "Bad command line arguments" );



        printf( "\n" );



        // Open temporary output file

        ftmp = fopen( "SETHDRTM.TMP", "w+" );
        if (!ftmp) {
                printf("ERROR: Can not open temporary file SETHDRTM.TMP\n");
                CleanExit( 1 );
                }
        TsTitle = 1; // Request that we write a title the first time through



        // Make sure first file exists

        if ( !FileExists( FileIn ) ) {
                printf( "ERROR: Can not open input file %s.\n", FileIn);
                CleanExit( 1 );
                }


        // Additional message if ignore warnings are likely

        if (SeqNum != 0)
                PrintMsg( PartialDataMsg );

        SeqNumSave = SeqNum;


        // Initialize ParGps library timestamp functions

        ParGpsTsInit();



        // Loop for several files

        printf( "Pass 1:\n" );

        while ( 1 ) {

                // First check for existance of input file

                if ( !FileExists( FileIn ) ) {
                        printf( "No more files to process.  Stopping at %s.\n", FileIn);
                        break;
                        }
                
                printf( "Reading file %s ...\n", FileIn );

                
                // Open input file

                fin = open( FileIn, (O_BINARY|O_RDONLY), PMODE );
                if (fin == -1) {
                        printf("ERROR: Can not open input file %s\n", FileIn);
                        break;
                        }


                // Read header, gather useful fields and check them

                if ( !SrDatHdrRead( fin, &Hdr ) ) {
                        printf( "ERROR: Could not read .DAT header\n" );
                        CleanExit( 1 );
                        }

                if (Hdr.DatId != SRDAT_HDR_ID) {
                        printf( "ERROR: This is not a valid .DAT file." );
                        CleanExit( 1 );
                        }

                if (Hdr.DatRev != SRDAT_REV) {
                        printf( "WARNING: Header version mismatch " );
                        printf( "(%d in file, %d in program)\n",
                                                  Hdr.DatRev, SRDAT_REV );
                        }

                if ( (strlen(Hdr.DatProduct)     > SRDAT_DIM_STRING) ||
                     (strlen(Hdr.DatApplication) > SRDAT_DIM_STRING) ||
                     (strlen(Hdr.FileName)       > SRDAT_DIM_STRING) ) {
                        printf( "ERROR: Invalid strings in .DAT header." );
                        CleanExit( 1 );
                        }

                NumChannels       = (int)Hdr.Channels;
                FirstSample          = Hdr.FileStartPtNum;
                StartSample          = Hdr.FileStartPtNum;
                MarkChannel          = Hdr.ChannelMarkIndex;
                PcCounterType        = Hdr.CounterType;
                PcCounterFreq        = Hdr.CounterFreq;
                SamplingRate         = Hdr.AtodSamplingRate;
                ObcCountsPerSample   = (OBC_FREQ / SamplingRate + 0.5L);
                
                // WCT: Consider using average to improve this number.
                BestCountsBetweenPps = PcCounterFreq;

                CheckMarkAndCounter( MarkChannel, PcCounterType, PcCounterFreq );

                ObcChannel = SetObcChannel( NumChannels, MarkChannel );
                TimeMethod = SelectTimeMethod( ObcChannel, MarkChannel );

                

                // Get PC time YMD from header to start things off right
                // in the case where there is no GPS YMD info available

                if ( PrevUnset ) {
                        Tp = Hdr.TimePoint[0];
                        ParGpsSecTimeCombine( Tp.Year, Tp.Month, Tp.Day,
                                           Tp.Hour, Tp.Minute, Tp.Second,
                                           Tp.Microsecond, &PrevPpsSecTime );
                        PrevUnset = 0;
                        }


                // Loop through processing entire file

                NotDone = 1;

                while ( NotDone ) {

                        // Read next data record

                        NotDone = ReadDatRecord(  PC_BUFFER, &Npts, MAXANALOG,
                                                 PPS_BUFFER, &Npps, MAXPPS,
                                                 SER_BUFFER, &Nser, MAXSER
                                       );


                        // If it's one of the three desired types,
                        // process it into time stamps

                        if (Npts > 0)  {
                                NumSamples = Npts/NumChannels;
                                ParGpsTsProcessAnalog( PC_BUFFER, NumSamples, NumChannels,
                                                       StartSample, MarkChannel, 
                                                       ObcChannel, ObcCountsPerSample,
                                                       &PpsEvent, &Nvalid );
                                StartSample += NumSamples;
                                }

                        else if (Npps > 0)
                                ParGpsTsProcessPps( PPS_BUFFER, Npps, &PpsEvent, &Nvalid  );


                        else if (Nser > 0)
                                ParGpsTsProcessSerial( SER_BUFFER, Nser, SerialDelay,
                                                       &PrevEventNum, &PrevPpsSecTime,
                                                       &PpsEvent, &Nvalid );


                        // Read the valid time stamps from the circular buffer
                        // maintained by the PARGPS library, correct the HMS source
                        // if using OBC to compute time, and write the time stamps
                        // out to the temporary file.

                        while ( Nvalid > 0 ) {
                                ParGpsTsReadValid( &TSx );
                                if ( TimeMethod == TIME_METHOD_OBC &&
                                     TSx.HmsSource == TIME_SOURCE_GPS   )
                                        TSx.HmsSource = TIME_SOURCE_OBC;
                                ParGpsTsFileWrite( ftmp, &TSx, TsTitle );
                                TsTitle = 0;
                                Nvalid--;
                                // WCT: Consider computing best counts between pps here
                                }


                        }  // end of while NotDone


                close( fin );
                fin = -1;


                // If only a single file was requested, we're done
                
                if (!Sequential)
                        break;

                // Otherwise, compute the next sequential file name
                
                else {
                        SeqNum++;
                        sprintf( FileIn, "%08X.DAT", SeqNum );
                        }

                } // end of while 1


        if (PrevEventNum == INVALID_PPSEVENT) {
                printf( "ERROR: No valid GPS time was found in any serial NMEA string.\n" );
                printf( "Header updates skipped\n" );
                CleanExit( 1 );
                }




        // Now process the temporary and point files, updating the data files

        fseek( ftmp, 0L, SEEK_SET );             // rewind to start
        fgets( TitleLine, MAXTITLE, ftmp );      // skip title line


        // Set which time point to fill with the corrected data.  Right
        // now, just pick entry 1.  Since we know that entry 0 was
        // filled by the original acquisition program.  In the future,
        // consider searching to find the first free entry (eg one that
        // has SRDAT_TIME_TYPE_NONE)
        
        TimeEntry = 1;



        // Get first two time stamps to get things started

        nval  = ParGpsTsFileRead( ftmp, &TS0, TsTitle );
        nval1 = ParGpsTsFileRead( ftmp, &TS1, TsTitle );
        
        if ( nval == EOF || nval == 0  || nval1 == EOF || nval1 == 0 ||
             TS0.Valid != TS_VALID_ALL || TS1.Valid != TS_VALID_ALL) {
                printf("ERROR: At least two GPS marks are required to update the headers\n");
                CleanExit( 1 );
                }



        // Reset SeqNum in preparation for second pass
        
        SeqNum = SeqNumSave;

        if (Sequential)
                sprintf( FileIn, "%08X.DAT", SeqNum );



        printf( "\nPass 2:\n" );

        while ( 1 ) {

                // Open input file, read header, and rewind

                fin = open( FileIn, (O_BINARY|O_RDWR), PMODE );
                if (fin == -1) {
                        printf("Can not open input file %s\n", FileIn);
                        break;
                        }

                if ( !SrDatHdrRead( fin, &Hdr ) ) {
                        printf( "ERROR: Could not read .DAT header\n" );
                        CleanExit( 1 );
                        }

                lseek( fin, 0L, SEEK_SET );


                

                // Read lines in tmp file while lines contain earlier points
                // and there are still lines left

                FirstSample = Hdr.FileStartPtNum;

                while (TS1.Sample < FirstSample && nval != EOF) {

                        nval = ParGpsTsFileRead( ftmp, &TSx, TsTitle );
                        
                        if ( nval == EOF  ||  nval == 0 ) 
                                break;

                        // Read was successful, so update values

                        TS0 = TS1;
                        TS1 = TSx;
                        
                        } // end while TS1.Sample < FirstSample

                
                // Determine time at first point.

                if ( TimeMethod == TIME_METHOD_OBC ) {
                        ok = ParGpsTsComputeTimeObc( &TS0, &TS1,
                                                     ObcCountsPerSample,
                                                     FirstSample, &FirstTime );
                        }
                else if ( TimeMethod == TIME_METHOD_GPS ) {
                
                        ok = ParGpsTsComputeTimePps( &TS0, &TS1,
                                                     BestCountsBetweenPps, PcCounterType,
                                                     FirstSample, &FirstTime );
                        }
                else { // ( TimeMethod == TIME_METHOD_SPS )
                
                        ok = ParGpsTsComputeTimeSps( &TS0, &TS1,
                                                     SamplingRate,
                                                     FirstSample, &FirstTime );
                        }

                if (!ok)
                        printf("Error computing time\n");


                ParGpsSecTimeSplit( FirstTime, &year, &month, &day,
                                    &hh, &mm, &ss, &usec );

                printf("Updating %s: New start time is %04d-%02d-%02d %02d:%02d:%02d.%06d\n",
                       FileIn, year,month,day,hh,mm,ss,usec );




                
                // Update the time fields in the file header.


                if ( !SrDatHdrSetTimeType( &Hdr, TimeEntry,
                                           SRDAT_TIME_TYPE_FIRSTPT,
                                           FirstSample, 0, 0 ) ) {
                        printf( "ERROR: Could not set .DAT time type\n" );
                        CleanExit( 1 );
                        }

                if ( !SrDatHdrSetTimeHMS( &Hdr, TimeEntry, 
                                          hh, mm, ss, usec, TS0.HmsSource ) ) {
                        printf( "ERROR: Could not set .DAT time HMS\n" );
                        CleanExit( 1 );
                        }

                if ( !SrDatHdrSetTimeYMD( &Hdr, TimeEntry,
                                          year, month, day, TS0.YmdSource ) ) {
                        printf( "ERROR: Could not set .DAT time YMD\n" );
                        CleanExit( 1 );
                        }


                // Write out updated header and close.
                        
                if ( !SrDatHdrWrite( fin, &Hdr ) ) {
                        printf( "ERROR: Could not write .DAT header\n" );
                        CleanExit( 1 );
                        }

                close( fin );
                fin = -1;



                // If only a single file was requested, we're done
                
                if (!Sequential)
                        break;

                // Otherwise, compute the next sequential file name

                else {
                        SeqNum++;
                        sprintf( FileIn, "%08X.DAT", SeqNum );
                        }
                
                } // end while (1)




        // Close temporary time file

        fclose( ftmp );
        ftmp = NULL;


        // Done

        exit( 0 );
}

int ProcessCommandLine( int argc, char** argv, int *SeqNum, int *Sequential ) {

        int   i, iprefix, len, itmp, seq;
        char *p, *fn;
        char Msg[MAXFILENAME];


        if ( (argc < 2) || (argc > 3) )
                Usage( "Command line must have 1 or 2 arguments" );


        fn  = NULL;
        seq = 0;

        for ( i = 1 ; i < argc ; i++ ) {

                if ( strcmp( "/seq", argv[i] ) == 0 ) {
                        if (!seq)
                                seq = 1;
                        else
                                Usage( "Duplicate /seq argument" );
                        }
                else {
                        if (!fn)
                                fn = argv[i];
                        else
                                Usage( "Duplicate file name argument" );
                        }
                }



        
        // Fill file name

        strncpy( FileIn, fn, MAXFILENAME-1 );
        FileIn[MAXFILENAME-1] = '\0';



        // Find the end of the input file prefix

        len = strlen( FileIn );
        iprefix = -1;
        for ( i = len ; i > 0 ; i-- ) {
                if (FileIn[i-1] == '.') {
                        iprefix = i-1;
                        break;
                        }
                }

        if (iprefix < 0 || iprefix > len || iprefix > MAXFILENAME-4) {
                sprintf( Msg, "File name %s is invalid", FileIn );
                Usage( Msg );
                }
        


        // Extract the prefix, null terminate, and uppercase

        strncpy( FilePrefix, FileIn, iprefix );
        FilePrefix[iprefix] = '\0';


        p = FilePrefix;
        len = strlen(FilePrefix);
        for ( i = 0 ; i < len ; i++ ) {
                *p = (char)toupper( *p );
                *p++;
                }
        



        
        // If optional sequential argument was found, determine starting 
        // sequential number from prefix and verify.

        *Sequential = 0;
        *SeqNum   = -1;

        if ( seq ) {
                        
                *Sequential = 1;

                if ( sscanf(FilePrefix, "%X", &itmp) == 1 )
                        *SeqNum = itmp;

                sprintf( PrefixCheck, "%08X", *SeqNum );
                if ( strcmp( FilePrefix, PrefixCheck ) != 0 )
                        Usage( "Can't set sequential value from filename" );
                }

        return( 1 );
}

int CheckMarkAndCounter( long MarkChannel, long PcCounterType, long PcCounterFreq ) {

        // Check for valid mark channel

        if (MarkChannel == SRDAT_MARKCHANNEL_NONE) {
                printf( "ERROR: No GPS data in file %s so file header time can not be updated.\n",
                          FileIn );
                printf( "GPS must be enabled when acquiring to get PPS and serial mark data.\n" );
                CleanExit( 1 );
                }


        // Check for a valid PC counter frequency.
        // Warn and choose reasonable defaults if it isn't valid.

        if (PcCounterFreq == SRDAT_INVALID_LONG) {

                // Linux style counter, this guess probably valid
                if (PcCounterType == SRDAT_COUNTER_HIGH32LOW32)
                        PcCounterFreq = 1000000;

                // Windows style counter, this guess likely invalid
                else if (PcCounterType == SRDAT_COUNTER_INT64)
                        PcCounterFreq = 1193254;

                // Yikes, we don't even have a known CounterType
                else 
                        PcCounterFreq = 1000000;

                printf( "WARNING: Invalid GPS counter frequency.\n");
                printf( "Assuming %ld.\n", PcCounterFreq );
                }

        return( 1 );
}

long SetObcChannel( long NumChannels, long MarkChannel ) {

        // On board counter not available on PAR1CH or PAR4CH
        // (or if only analog channels were saved on the PAR8CH)

        if (NumChannels <= 8)
                return( SRDAT_MARKCHANNEL_NONE );


        // Output of O.B.C. data currently tied to output of mark channel
        
        if (MarkChannel == SRDAT_MARKCHANNEL_NONE)
                return( SRDAT_MARKCHANNEL_NONE );


        // Validity check
        
        if (MarkChannel+1 >= NumChannels)
                return( SRDAT_MARKCHANNEL_NONE );

        
        return( MarkChannel + 1 );
}

int SelectTimeMethod( long ObcChannel, long MarkChannel ) {
        
        int Method;

        
        // If there is an OBC channel, use it

        if (ObcChannel != SRDAT_MARKCHANNEL_NONE)
                Method = TIME_METHOD_OBC;

        
        // Or if there is GPS info, use it
        
        else if (MarkChannel != SRDAT_MARKCHANNEL_NONE)
                Method = TIME_METHOD_GPS;

        
        // Otherwise, just use sps

        else
                Method = TIME_METHOD_SPS;


        return( Method );
}


int ReadDatRecord( void  *PC_BUFFER, unsigned int *npts, unsigned int maxpts, 
                   void *PPS_BUFFER, unsigned int *npps, unsigned int maxpps, 
                   void *SER_BUFFER, unsigned int *nser, unsigned int maxser
                 ) {

        SRDAT_TAG Tag;


        // Initialize values

        *npts = *npps = *nser = 0;



        // Read data record tag to determine data type
        
        if ( !SrDatTagRead( fin, &Tag ) ) {
                printf( "Could not read .DAT tag\n" );
                CleanExit( 1 );
                }
                




        // Read in analog, gps, or serial data buffer


        if (Tag.TagId == SRDAT_TAGID_ANALOG24) {

                // Read in analog data buffer
                *npts = Tag.Nbytes / sizeof( long );
                if (*npts > maxpts) {
                        printf( "Size of analog data %u exceeds maximum of %u.\n",
                                  *npts, maxpts );
                        printf( "Please recompile with larger maximum.\n");
                        CleanExit( 1 );
                        }
                if ( !SrDatDataRead( fin, PC_BUFFER, Tag.Nbytes ) ) {
                        printf( "Could not read .DAT analog data\n" );
                        CleanExit( 1 );
                        }
                }


        else if (Tag.TagId == SRDAT_TAGID_GPSPPS) {

                // Read in gps data buffer
                *npps = Tag.Nbytes / sizeof( PPSDATA );
                if (*npps > maxpps) {
                        printf( "Size of PPS data %u exceeds maximum of %u.\n",
                                  *npps, maxpps );
                        printf( "Please recompile with larger maximum.\n");
                        CleanExit( 1 );
                        }
                if ( !SrDatDataRead( fin, PPS_BUFFER, Tag.Nbytes ) ) {
                        printf( "Could not read .DAT PPS data\n" );
                        CleanExit( 1 );
                        }
                }


        else if (Tag.TagId == SRDAT_TAGID_GPSSERIAL) {

                // Read in gps serial data buffer
                *nser = Tag.Nbytes / sizeof( SERIALDATA );
                if (*nser > maxser) {
                        printf( "Size of serial data %u exceeds maximum of %u.\n",
                                  *nser, maxser );
                        printf( "Please recompile with larger maximum.\n");
                        CleanExit( 1 );
                        }
                if ( !SrDatDataRead( fin, SER_BUFFER, Tag.Nbytes ) ) {
                        printf( "Could not read .DAT serial data\n" );
                        CleanExit( 1 );
                        }
                }


        else if (Tag.TagId == SRDAT_TAGID_EOF) {
                return( 0 );
                }

        else {
                printf( "Unexpected tag TagId found 0x%X\n", Tag.TagId);
                }



        return( 1 );
}


#if !defined F_OK
#define F_OK 0x00
#endif

int FileExists( char *FileName ) {

        if ( access( FileName, F_OK ) == 0 )
                return( 1 );
        else
                return( 0 );

}
        
void Usage( char *SpecificMsgText ) {

        printf( "\nERROR: %s\n", SpecificMsgText );
        PrintMsg( UsageMsg );
        exit( 1 );
}

void PrintMsg( char *MsgArray[] ) {

        int i;
        
        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );
}

void CleanExit( int value ) {
        
        // Perform orderly shutdown

        if (fin != -1)
                close( fin );

        if (ftmp)
                fclose( ftmp );

        exit( value );
}
