Attribute VB_Name = "modPARGPS"
Attribute VB_Description = "Declarations of DLL procedures for high level use of SR GPS board"

Option Explicit


' FILE: pargps.bas                  Copyright (c), Symmetric Research, 2003-2004
'
' Basic header file for PARGPS acquisition library functions.
'
' Visual Basic users should include this file in their projects to have
' access to the PARGPS library functions.
'


' In the function library prototypes (Declares), the symbols listed below
' can replace the "As Type" info for function arguments.  For example:
' (ByVal X As Long, ByRef Y As Single) equals (ByVal X&, ByRef Y!)
'
' % Integer (short)
' & Long    (32-bit int or long)
' ! Single  (float)
' # Double  (double)
' $ String  (char*)  ByVal BSTR = LPSTR, ByRef BSTR = ptr to ptr to null-term string
'
' When translating from the C declarations in pargps.h, the following
' rules were applied:
'
'   short         -> Integer
'   int           -> Long (because ints are now 32 bits, the same as longs)
'   unsigned int  -> Long
'   long          -> Long
'   unsigned long -> Long
'   float         -> Single
'   double        -> Double
'   char*         -> String
'   anything*     -> pass ByRef instead of ByVal
'   arrays        -> pass ByRef using first item in array
'
' Routines returning a value are Function, the rest are Sub.
'



' Indicator of bad handle to device

Public Const PARGPS_BAD_DEVHANDLE As Long = -1





' PARGPS DEFINES:
'
'


' GPS driver name defines ...

Public Const PARGPS_NAME                As String = "SrParGps"
Public Const PARGPS_0x378               As String = "SrParGps378"
Public Const PARGPS_0x278               As String = "SrParGps278"


' Serial port numbers

Public Const SERIAL_PORT_1           As Long = 1  ' This is Com1 or ttyS0
Public Const SERIAL_PORT_2           As Long = 2  ' This is Com2 or ttyS1


' PARGPS interrupt modes

Public Const INTERRUPT_PPS_ONLY      As Long = 0
Public Const INTERRUPT_ALTERNATING   As Long = 1


' This determines how much GPS PPS and serial data can be stored in the
' driver before it is read by a user program.  Currently it is set
' to 5 minutes of storage for data that comes in once per second.

Public Const MAX_PARGPS_STORAGE   As Long = (5 * 60)
                                  


' Error code return values returned by the PARGPS library functions.  The
' PARGPS_ERROR_MSG array, allocated in pargps.c, maps these error codes to
' short message strings for easy use with printf.

Public Const PARGPS_ERROR_NONE                           As Long = 0
Public Const PARGPS_ERROR_NOT_AVAILABLE                  As Long = 1
Public Const PARGPS_ERROR_INTERRUPT_NUMBER               As Long = 2
Public Const PARGPS_ERROR_INTERRUPT_MODE                 As Long = 3
Public Const PARGPS_ERROR_SERIAL_PORT_NUMBER             As Long = 4
Public Const PARGPS_ERROR_SERIAL_PORT_NOT_OPEN           As Long = 5
Public Const PARGPS_ERROR_DRIVER_NOT_OPEN                As Long = 6
Public Const PARGPS_ERROR_DRIVER_REQUEST_FAILED          As Long = 7
Public Const PARGPS_ERROR_KERNEL_AREA                    As Long = 8
Public Const PARGPS_ERROR_KEYPRESS                       As Long = 9
Public Const PARGPS_ERROR_BAD_PARAMETER                  As Long = 10
Public Const PARGPS_ERROR_MAX                            As Long = 11


Public Const PARGPS_ERROR_MSG_NONE                       As String = "NONE"
Public Const PARGPS_ERROR_MSG_NOT_AVAILABLE              As String = "NOT AVAILABLE"
Public Const PARGPS_ERROR_MSG_INTERRUPT_NUMBER           As String = "INTERRUPT NUM"
Public Const PARGPS_ERROR_MSG_INTERRUPT_MODE             As String = "INTERRUPT MODE"
Public Const PARGPS_ERROR_MSG_SERIAL_PORT_NUMBER         As String = "SERIAL NUM"
Public Const PARGPS_ERROR_MSG_SERIAL_PORT_NOT_OPEN       As String = "SERIAL PORT"
Public Const PARGPS_ERROR_MSG_DRIVER_NOT_OPEN            As String = "DEVICE DRIVER"
Public Const PARGPS_ERROR_MSG_DRIVER_REQUEST_FAILED      As String = "DRIVER REQUEST"
Public Const PARGPS_ERROR_MSG_KERNEL_AREA                As String = "KERNEL AREA"
Public Const PARGPS_ERROR_MSG_KEYPRESS                   As String = "KEYPRESS"
Public Const PARGPS_ERROR_MSG_BAD_PARAMETER              As String = "BAD PARAMETER"
Public Const PARGPS_ERROR_MSG_MAX                        As String = vbNullString


' PARGPS error string allocation, initialized in ParGpsInitErrorMsg() below

Public PARGPS_ERROR_MSG(PARGPS_ERROR_NONE To PARGPS_ERROR_MAX) As String


' 64 BIT SUPPORT:
'
' Support for 64-bit values varies from one operating system to
' another.  To compensate, a SR64BIT structure is defined.
' *** NOTE: LowPart should really be Unsigned Long

Public Type SR64BIT
        LowPart  As Long
        HighPart As Long
End Type





' NMEA DEFINES:
'
' NMEA, the National Marine Electonics Association, has defined a number
' of standard messages to be used with marine instruments.  See Appendix
' E of the Trimble ACE-III GPS receiver manual for a full description of
' the possible NMEA messages.  Although 7 messages are possible with the
' Trimble GPS receiver, the SR PARGPS board is set up to always use only
' the default GGA and VTG messages.  A sample GGA message is:
'
' $GPGGA,163859.0,3609.422,N,11518.713,W,1,03,1.75,00024,M,-024,M,,*62
'


Public Const NmeaStartTag  As String = "$"
Public Const NmeaTalkerId  As String = "GP"

' Note: The NMEA message IDs are listed in the order in which they come out

Public Const NMEA_MSGID_MIN  As Long = -1
Public Const NMEA_MSGID_ZDA  As Long = 0
Public Const NMEA_MSGID_GGA  As Long = 1
Public Const NMEA_MSGID_GLL  As Long = 2
Public Const NMEA_MSGID_VTG  As Long = 3
Public Const NMEA_MSGID_GSA  As Long = 4
Public Const NMEA_MSGID_GSV  As Long = 5
Public Const NMEA_MSGID_RMC  As Long = 6
Public Const NMEA_MSGID_MAX  As Long = 7

Public NmeaMsgId(NMEA_MSGID_MAX - 1) As String

Public Const MAX_NMEA_TYPE      As Long = NMEA_MSGID_MAX
Public Const MAX_NMEA_SIZE      As Long = 83
Public Const MAX_NMEA_BUFF      As Long = (MAX_NMEA_TYPE * MAX_NMEA_SIZE)

Public Const MAX_NMEA_FIELDS    As Long = 17 ' One more than truely needed
Public Const MAX_NMEA_FIELDSIZE As Long = 11

Public Type NMEAPARSE
        Nfields       As Long
        CheckSum      As Long
        Field(MAX_NMEA_FIELDS - 1, MAX_NMEA_FIELDSIZE - 1) As String
End Type


' Indexes for character array holding source of HMS or YMD.
' The first 7, from _ZDA to _RMC must match the corresponding
' NMEA_MSGID defines.

Public Const SRCSTR_MIN   As Long = -1
Public Const SRCSTR_ZDA   As Long = 0
Public Const SRCSTR_GGA   As Long = 1
Public Const SRCSTR_GLL   As Long = 2
Public Const SRCSTR_VTG   As Long = 3
Public Const SRCSTR_GSA   As Long = 4
Public Const SRCSTR_GSV   As Long = 5
Public Const SRCSTR_RMC   As Long = 6
Public Const SRCSTR_PC    As Long = 7
Public Const SRCSTR_OLD   As Long = 8
Public Const SRCSTR_UNK   As Long = 9
Public Const SRCSTR_MAX   As Long = 10


' Defines and structures for PARGPS PPS and serial data.  The PpsEventNum
' field is the big index that ties both sets of data together.

Public Const VALID_NONE       As Long = &H0
Public Const VALID_PPS        As Long = &H1
Public Const VALID_DREADY     As Long = &H2
Public Const VALID_NMEA       As Long = &H4
Public Const VALID_DATE       As Long = &H8


Public Type PPSDATA
        ValidFields   As Long
        PpsEventNum   As Long
        CountAtPps    As SR64BIT
        CountAtDready As SR64BIT
End Type

Public Type SERIALDATA
        ValidFields As Long
        PpsEventNum As Long
        NmeaCount   As Long
        NmeaMsg     As String * MAX_NMEA_BUFF
End Type

Public Type GPSTIMEDATA
        ValidFields As Long
        PpsEventNum As Long
        CountAtPps  As SR64BIT
        CountAtNow  As SR64BIT
        NmeaCount   As Long
        NmeaMsg     As String * MAX_NMEA_BUFF
End Type




' GENERAL PARGPS FUNCTIONS:
'
' The following functions provide control over the PARGPS.  When used to
' provide accurate timing for a PARxCH doing data acquisition, the
' typical pattern of usage is:
'
' 1)  Open and initialize both the PARxCH and the PARGPS
' 1A) Attach the PARGPS to the PARxCH so their drivers can communicate
' 2)  Start the PARGPS and then the PARxCH
' 3)  Call the PARxCHReadDataWithGpsMark to get analog data and the PARGPS
'     ParGpsReadPpsData and ParGpsReadSerialData to get timing data
' 4)  Loop to continue getting more data
' 5)  When done, stop both the PARGPS and PARxCH
' 5A) Release the PARGPS from the PARxCH and close both drivers
'
'
' Once the PARGPS and PARxCH drivers have been opened and started as
' described above, the PARGPS driver can be opened by a second application
' to retrieve the current GPS data.  For this case, the typical pattern
' of usage is:
'
' 1) Open the PARGPS with ParGpsOpenForTimeOnly
' 2) Call ParGpsGetCurrentTime to get current GPS like time, position,
'    and # of satellites
' 3) Close the PARGPS with the standard ParGpsClose
'
'
' The ParGpsGetRev function returns an integer indicating the library rev.
'


Public Declare Function ParGpsGetRev _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetRev@4" _
   (ByRef Rev As Long) _
    As Long


Public Declare Function ParGpsOpen _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsOpen@12" _
   (ByVal ParGpsName As String, _
    ByVal SerialPortNumber As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParGpsOpenForTimeOnly _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsOpenForTimeOnly@8" _
   (ByVal ParGpsName As String, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParGpsClose _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsClose@4" _
   (ByVal ParGpsHandle As Long) _
    As Long

Public Declare Sub ParGpsStart _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsStart@4" _
   (ByVal ParGpsHandle As Long)

Public Declare Sub ParGpsStop _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsStop@4" _
   (ByVal ParGpsHandle As Long)

Public Declare Function ParGpsGetKernelArea _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetKernelArea@8" _
   (ByVal ParGpsHandle As Long, _
    ByRef KernelArea As Long) _
    As Long

Public Declare Function ParGpsGetCounterFrequency _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetCounterFrequency@8" _
   (ByVal ParGpsHandle As Long, _
    ByRef CounterFreq As Long) _
    As Long

Public Declare Function ParGpsReadPpsData _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsReadPpsData@16" _
   (ByVal ParGpsHandle As Long, _
    ByRef Values As Long, _
    ByVal Nvalues As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParGpsReadSerialData _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsReadSerialData@16" _
   (ByVal ParGpsHandle As Long, _
    ByRef Values As Long, _
    ByVal Nvalues As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParGpsGetCurrentTime _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetCurrentTime@72" _
   (ByVal ParGpsHandle As Long, _
    ByRef Time As Double, _
    ByRef Year As Long, _
    ByRef Month As Long, _
    ByRef Day As Long, _
    ByRef Hour As Long, _
    ByRef Minute As Long, _
    ByRef Second As Long, _
    ByRef MicroSecond As Long, _
    ByRef NumSatellites As Long, _
    ByRef Latitude As Double, _
    ByRef Longitude As Double, _
    ByRef Altitude As Double, _
    ByRef SourceYMD As Long, _
    ByRef SourceHMS As Long, _
    ByRef SourceNsat As Long, _
    ByRef SourcePos As Long, _
    ByRef Error As Long) _
    As Long


' NMEA HELPER FUNCTIONS:
'
' NMEA, the National Marine Electonics Association, has defined a
' number of standard messages to be used with marine instruments.
' These helper functions work with the serial NMEA messages returned
' by the PARGPS.  Most users will want to use the first five functions
' which just take a NMEA string as input and return the specified
' quantity.
'
' Those wishing to do their own processing of the NMEA messages may
' find the ParGpsParseNmeaData function helpful since it parses the
' message into its component data fields.
'
' For the PARGPS, each NMEA message starts with a $GP and a three
' character message id.  That is followed by a comma separated list
' of data fields which may be empty and whose meaning depends on the
' message id.  Each message ends with a *, a checksum, and a carriage
' return line feed pair.  For example:
'
' $GPGGA,163859.0,3609.422,N,11518.713,W,1,03,1.75,00024,M,-024,M,,*62
'

Public Declare Function ParGpsGetNmeaSatellites _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetNmeaSatellites@12" _
   (ByVal NmeaMsg As String, _
    ByRef NumSat As Long, _
    ByRef Source As Long) _
    As Long

Public Declare Function ParGpsGetNmeaTime _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetNmeaTime@16" _
   (ByVal NmeaMsg As String, _
    ByRef Time As Double, _
    ByRef SourceYMD As Long, _
    ByRef SourceHMS As Long) _
    As Long

Public Declare Function ParGpsGetNmeaHMS _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetNmeaHMS@28" _
   (ByVal NmeaMsg As String, _
    ByRef HMS As Double, _
    ByRef Hour As Long, _
    ByRef Minute As Long, _
    ByRef Second As Long, _
    ByRef MicroSecond As Long, _
    ByRef Source As Long) _
    As Long

Public Declare Function ParGpsGetNmeaYMD _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetNmeaYMD@24" _
   (ByVal NmeaMsg As String, _
    ByRef HMS As Double, _
    ByRef Year As Long, _
    ByRef Month As Long, _
    ByRef Day As Long, _
    ByRef Source As Long) _
    As Long

Public Declare Function ParGpsGetNmeaPosition _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetNmeaPosition@20" _
   (ByVal NmeaMsg As String, _
    ByRef Latitude As Double, _
    ByRef Longitude As Double, _
    ByRef Altitude As Double, _
    ByRef Source As Long) _
    As Long

Public Declare Function ParGpsParseNmeaData _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsParseNmeaData@8" _
   (ByVal NmeaMsg As String, _
    ByRef ParseData As Long) _
    As Long



' TIME HELPER FUNCTIONS:
'
' Time can be represented in many different forms.  These helpers convert
' between some of the forms described below.
'
' 1) SECONDS SINCE 1970
'    Many of the ANSI time management functions work with time as a
'    time_t, a double that represents time as the total number of seconds
'    and fractional seconds since 1/1/1970.  These values incorporate both
'    YMD and HMS into a single number.  It is possible to make a time_t
'    number that holds just YMD or HMS, but each compensates for the local
'    time offset from UTC.  So, when adding a YMD-only time_t with an
'    HMS-only time_t you must subtract one local time offset.  Use the
'    function ParGpsCalcTimeZoneOffset to determine this offset.
'
' 2) COLLECTION OF INTS
'    Time can also be represented as a collection of individual integers:
'    Year, Month, Day, Hour, Minute, Second, Microseconds.  Microseconds
'    must be a long integer.  These separate components are similar to the
'    corresponding fields in the standard runtime library struct tm
'    structure, eg tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec.
'    Use the ParGpsSecTimeSplit and ParGpsSecTimeCombine functions to convert
'    between this collection of ints representation and the seconds since
'    1970 representation.
'
' 3) PACKED DIGITS
'    In the NMEA message strings, time is typically presented as
'    a double where three quantities are packed into pairs of digits.
'    The quantities involved can be either year, month, and day or
'    hours, minutes, and seconds.  Each quantity is allocated two
'    digits.  For example, 112307 means 23 minutes and 7 seconds past
'    11 o'clock.  Since the minutes and seconds digits can never be
'    more than 59, adding 1 second to 112359 gives 112400.  The
'    ParGpsPack3 and ParGpsUnpack3 functions convert between this packed
'    representation and a collection of individual ints.
'


Public Declare Function ParGpsPack3 _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsPack3@16" _
   (ByVal Top As Long, _
    ByVal Middle As Long, _
    ByVal Bottom As Long, _
    ByRef Xxx As Long) _
    As Long

Public Declare Function ParGpsUnpack3 _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsUnpack3@16" _
   (ByVal Xxx As Long, _
    ByRef Top As Long, _
    ByRef Middle As Long, _
    ByRef Bottom As Long) _
    As Long

Public Declare Function ParGpsGetPcTime _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsGetPcTime@28" _
   (ByRef Time As Double, _
    ByRef Year As Long, _
    ByRef Month As Long, _
    ByRef Day As Long, _
    ByRef Hour As Long, _
    ByRef Minute As Long, _
    ByRef Second As Long) _
    As Long

Public Declare Function ParGpsSetPcTime _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsSetPcTime@36" _
   (ByVal Time As Double, _
    ByVal Year As Long, _
    ByVal Month As Long, _
    ByVal Day As Long, _
    ByVal Hour As Long, _
    ByVal Minute As Long, _
    ByVal Second As Long, _
    ByVal FromYMD As Long) _
    As Long

Public Declare Function ParGpsCalcTimeZoneCorr _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsCalcTimeZoneCorr@4" _
    (ByVal StartTime As Long) _
    As Long

Public Declare Function ParGpsSecTimeSplit _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsSecTimeSplit@36" _
   (ByVal Time As Double, _
    ByRef Year As Long, _
    ByRef Month As Long, _
    ByRef Day As Long, _
    ByRef Hour As Long, _
    ByRef Minute As Long, _
    ByRef Second As Long, _
    ByRef MicroSecond As Long) _
    As Long

Public Declare Function ParGpsSecTimeCombine _
    Lib "C:\SR\PARGPS\LIB\PARGPS.DLL" _
    Alias "_ParGpsSecTimeCombine@32" _
   (ByVal Year As Long, _
    ByVal Month As Long, _
    ByVal Day As Long, _
    ByVal Hour As Long, _
    ByVal Minute As Long, _
    ByVal Second As Long, _
    ByVal MicroSecond As Long, _
    ByRef Time As Double) _
    As Long





' Initialize the error message string array

Public Sub ParGpsInitErrorMsg()

        PARGPS_ERROR_MSG(PARGPS_ERROR_NONE) = PARGPS_ERROR_MSG_NONE
        PARGPS_ERROR_MSG(PARGPS_ERROR_NOT_AVAILABLE) = PARGPS_ERROR_MSG_NOT_AVAILABLE
        PARGPS_ERROR_MSG(PARGPS_ERROR_INTERRUPT_NUMBER) = PARGPS_ERROR_MSG_INTERRUPT_NUMBER
        PARGPS_ERROR_MSG(PARGPS_ERROR_INTERRUPT_MODE) = PARGPS_ERROR_MSG_INTERRUPT_MODE
        PARGPS_ERROR_MSG(PARGPS_ERROR_SERIAL_PORT_NUMBER) = PARGPS_ERROR_MSG_SERIAL_PORT_NUMBER
        PARGPS_ERROR_MSG(PARGPS_ERROR_SERIAL_PORT_NOT_OPEN) = PARGPS_ERROR_MSG_SERIAL_PORT_NOT_OPEN
        PARGPS_ERROR_MSG(PARGPS_ERROR_DRIVER_NOT_OPEN) = PARGPS_ERROR_MSG_DRIVER_NOT_OPEN
        PARGPS_ERROR_MSG(PARGPS_ERROR_DRIVER_REQUEST_FAILED) = PARGPS_ERROR_MSG_DRIVER_REQUEST_FAILED
        PARGPS_ERROR_MSG(PARGPS_ERROR_KERNEL_AREA) = PARGPS_ERROR_MSG_KERNEL_AREA
        PARGPS_ERROR_MSG(PARGPS_ERROR_KEYPRESS) = PARGPS_ERROR_MSG_KEYPRESS
        PARGPS_ERROR_MSG(PARGPS_ERROR_BAD_PARAMETER) = PARGPS_ERROR_MSG_BAD_PARAMETER
        PARGPS_ERROR_MSG(PARGPS_ERROR_MAX) = PARGPS_ERROR_MSG_MAX

End Sub
