/* FILE: dat2asc.c                Copyright (c), Symmetric Research, 2002-2005
 *      
 * File for converting a binary .dat file to ascii.  GPS support is
 * optional, but is needed for converting the GPS PPS and serial data
 * records.  Be sure the PARGPS software is installed and define the
 * GPS_AVAILABLE macro when compiling to include GPS support.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>       // for O_xxx defines
#include <sys/stat.h>    // for S_IREAD define (VC++ can use / instead of \)
#include <time.h>

#include "srdat.h"

// If PARGPS software is available, the PPSDATA and SERIALDATA 
// structures can be converted.  Otherwise they are skipped.

#ifdef GPS_AVAILABLE
#include "pargps.h"     // for PPSDATA, SERIALDATA, TIMESTAMP, SR64BIT, etc
extern char ParGpsLastErrorMessage[];

#else // no GPS_AVAILABLE
                            // Method for computing time at an arbitrary sample
#define TIME_METHOD_NONE 0  //   Assume perfect sampling
#define TIME_METHOD_SPS  1  //   From timestamp and sampling rate
#define TIME_METHOD_GPS  2  //   From PARGPS and PC counter
#define TIME_METHOD_OBC  3  //   From PARGPS and PAR8CH 800 ns On Board Counter

typedef int PPSDATA;
typedef int SERIALDATA;
typedef struct _TIME_STAMP {  // this is a cut down version
        long   Sample;
        double SecSince1970;
        } TIMESTAMP;

int ParGpsCalcTimeZoneCorr( long StartTime );
int ParGpsSecTimeSplit( double Time, int *Year, int *Month, int *Day,
                        int *Hour, int *Minute, int *Second, long *Microsecond );
int ParGpsSecTimeCombine( int Year, int Month, int Day,
                          int Hour, int Minute, int Second, 
                          long MicroSecond, double *Time );
#endif // GPS_AVAILABLE

#if defined( SROS_LINUX )
#define O_BINARY 0x00
#endif


typedef struct SrDatTimeData TIMEPT;

   
// Prototypes
int   ProcessCommandLine( int argc, char** argv );
void  ShowHdr( FILE *fout, SRDAT_HDR *Hdr );
void  ShowTag( FILE *fout, SRDAT_TAG *Tag );
void  ShowData( FILE *fout, SRDAT_HDR *Hdr, void *Buffer,
                long DataType, unsigned long Nbytes,
                int HexFormat, int TimeFormat );
char *PickTimeTypeString( long Type );
char *PickSourceString( long Source );
char *PickDataCodingString( long DataCoding );
int   ReadDatRecord( void  *PC_BUFFER, unsigned int *npts, unsigned int maxpts, 
                     void *PPS_BUFFER, unsigned int *npps, unsigned int maxpps, 
                     void *SER_BUFFER, unsigned int *nser, unsigned int maxser
                 );
long  SetObcChannel( long NumChannels, long MarkChannel );
int   SelectTimeMethod( long ObcChannel, long MarkChannel );
void  Usage( char *msg );
void  ErrorWarning( char *msg );
int   FileExists( char *FileName );
void  CleanExit( int ret );

int InitTimeStamps( void );
int ComputeTime( SRDAT_HDR *Hdr, int TimeFormat, long Sample, char *TimeStr );
int UpdateTimeStamps( long Sample );
int FillTimeStr( int TimeFormat, double Time, char *TimeStr );





// Global defines and data

char* UsageMsg[] = {
    "USAGE: dat2asc filename.DAT [/seq] [/hex] [/tymd | /tsec] [/?]    ",
    "                                                                  ",
    "Dat2asc converts the specified .DAT file to ASCII text format.    ",
    "The output filename is the same as the input filename except the  ",
    "suffix is changed from .DAT to .ASC.  If the prefix of the input  ",
    "file contains a hex number, the optional /seq argument indicates  ",
    "the input file is the first of a sequence.  In this case, dat2asc ",
    "will continue incrementing the file prefix and converting until   ",
    "the next file in the sequence can not be found.  The optional     ",
    "/hex argument causes the analog data to be written in hex instead ",
    "of decimal format.  The /tymd and /tsec options add an extra      ",
    "column for each sample showing time as Y/M/D H:M:S or seconds     ",
    "since 1970 respectively.  Use /? to show this usage message.      ",
    "                                                                  ",
    "EXAMPLES: dat2asc SIMP4CH.DAT                                     ",
    "          dat2asc 0000001F.DAT /seq /hex /tymd                    ",
    NULL
    };




#define TIME_FORMAT_NONE  0
#define TIME_FORMAT_YMD   1
#define TIME_FORMAT_SEC   2

#define MARKCHAR_NONE  ' '
#define MARKCHAR_FOUND '<'

#define MAXDATAFMT     10
#define MAXTIMESTR     30
#define MAXFILENAME   256
#define MAXCONVERT  10000

// MAXANALOG is defined large enough to hold data from any PARxCH from
// the PAR1CH to the PAR8CH.  The PAR1CH can have 2 channels, and the
// PAR4CH can have 5, but the PAR8CH can have up to 11 eg 8 analog + 1
// digital + 1 gps + 1 counter.

#define MAXANALOG (128*11) // 128 pts/ch * 11 channels 
#define MAXPPS        26   // This is large enough for sampling as often
#define MAXSER        26   // as about 5 samples per second.


int        CurrentPt16, CurrentPt24, TimeMethod;
int        UseSeq, HexFormat, TimeFormat, FileArg, Sequential, SeqNum;
double     ObcCountsPerSample, BestCountsBetweenPps;
char       FilePrefix[MAXFILENAME], PrefixCheck[MAXFILENAME];
char       FileIn[MAXFILENAME], FileOut[MAXFILENAME];
char       Buffer[MAXCONVERT];
long       PC_BUFFER[MAXANALOG];
PPSDATA    PPS_BUFFER[MAXPPS];
SERIALDATA SER_BUFFER[MAXSER];
TIMESTAMP  TimeStamp[2];
SRDAT_HDR  Hdr;
SRDAT_TAG  Tag;
int        FHin;
FILE      *Fout, *Fts;


// Main:

int main( int argc, char** argv ) {

        unsigned int nbytes;


        // Announce program version.

        printf( "DAT2ASC [/?]: (c) SR 2004 : Rev {11/15/04}\n" );


        // Process command line arguments

        ProcessCommandLine( argc, argv );
        printf( "\n" );


        // Init file handles

        FHin = -1;
        Fout = Fts = NULL;


        // If a time option was requested, init the timestamps

        if (TimeFormat != TIME_FORMAT_NONE) {
                InitTimeStamps();
                }
        

        // Loop until no more dat files requested or found

        printf( "Converting to ASCII:\n" );

        while ( 1 ) {

                // First check for existance of input file

                if ( !FileExists( FileIn ) ) {
                        printf( "No more files to process.  Stopping at %s.\n", FileIn );
                        exit( 0 );
                        }
                
                // Open input file

                FHin = open( FileIn, (O_BINARY|O_RDONLY), (S_IREAD|S_IWRITE) );
                if (FHin == -1) {
                        printf( "Can not open input file %s\n", FileIn );
                        CleanExit( 1 );
                        }




                // Read header

                printf( "Reading file %s ...\n", FileIn );

                if ( !SrDatHdrRead( FHin, &Hdr ) )
                        ErrorWarning( "Could not read .DAT header" );
                
                if (Hdr.DatId != SRDAT_HDR_ID)
                        ErrorWarning( "This is not a valid .DAT file." );
                        
                if (Hdr.DatRev != SRDAT_REV) {
                        printf( "WARNING: Header version mismatch " );
                        printf( "(%d in file, %d in program)\n",
                                                  Hdr.DatRev, SRDAT_REV );
                        }

                if ( (strlen(Hdr.DatProduct)     > SRDAT_DIM_STRING) ||
                     (strlen(Hdr.DatApplication) > SRDAT_DIM_STRING) ||
                     (strlen(Hdr.FileName)       > SRDAT_DIM_STRING) )
                        ErrorWarning( "Invalid strings in .DAT header." );


                if ( Hdr.AtodSamplingRate <= 0.000001 ) {
                        printf( "WARNING: Bad sampling rate.  Reset to 1 sps.\n" );
                        Hdr.AtodSamplingRate = 1.0L;
                        }



                // Open output file

                Fout = fopen( FileOut, "w+" );
                if (Fout == NULL) {
                        printf( "Can not open output file %s\n", FileOut );
                        CleanExit( 1 );
                        }


                // Write ASCII header

                printf( "Writing file %s ...\n", FileOut );
                fprintf( Fout, "Converted from file %s\n\n", FileIn );

                ShowHdr( Fout, &Hdr );


                // Initialize point counts that will be updated by ShowData
                
                CurrentPt16 = CurrentPt24 = Hdr.FileStartPtNum;

                
                
                // Read and write data records until reaching the end

                if ( !SrDatTagRead( FHin, &Tag ) )
                        ErrorWarning("Could not read .DAT tag" );
                ShowTag( Fout, &Tag );
                
                while (Tag.TagId != SRDAT_TAGID_EOF) {

                        nbytes = (unsigned int)Tag.Nbytes;
                        if ( (unsigned long)nbytes != Tag.Nbytes )
                                printf( "Warning data exceeds size allowed by unsigned int.\n");

                        if (Tag.Nbytes < MAXCONVERT) {
                                if ( !SrDatDataRead( FHin, Buffer, nbytes ) )
                                        ErrorWarning("Could not read .DAT data");
                                
                                ShowData( Fout,
                                          &Hdr,
                                          Buffer,
                                          Tag.TagId,
                                          Tag.Nbytes,
                                          HexFormat,
                                          TimeFormat );
                                }
                        else {
                                SrDatDataSkip( FHin, nbytes );
                                printf("Exceeded allocated buffer size with %lu\n",
                                       Tag.Nbytes);
                                }


                        if ( !SrDatTagRead( FHin, &Tag ) )
                                ErrorWarning("Could not read .DAT tag" );
                        ShowTag( Fout, &Tag );
                        }



                // Close current files

                close( FHin );    FHin = -1;
                fclose( Fout );   Fout = NULL;


                
                // If only a single file was requested, we're done
                
                if (!Sequential)
                        break;

                // Otherwise, compute the next set of sequential file names
                
                else {
                        SeqNum++;
                        sprintf( FileIn,  "%08X.DAT", SeqNum );
                        sprintf( FileOut, "%08X.ASC", SeqNum );
                        }

                }  // end while ( 1 )

        exit( 0 );
}

int ProcessCommandLine( int argc, char** argv ) {

        int i, iprefix, len, itmp;
        char *p, Msg[MAXFILENAME];


        // Set defaults

        UseSeq     = 0;
        HexFormat  = 0;
        TimeFormat = TIME_FORMAT_NONE;
        FileArg    = 0;

        
        // Process command line arguments
        
        for ( i = 1 ; i < argc ; i++ ) {

                if (strcmp("/?",argv[i]) == 0)
                        Usage( "?" ); 
                else if (strcmp("/seq",argv[i]) == 0)
                        UseSeq = 1; 
                else if (strcmp("/hex",argv[i]) == 0)
                        HexFormat = 1; 
                else if (strcmp("/tymd",argv[i]) == 0)
                        TimeFormat = TIME_FORMAT_YMD;
                else if (strcmp("/tsec",argv[i]) == 0)
                        TimeFormat = TIME_FORMAT_SEC;
                else if (FileArg == 0)
                        FileArg = i;
                else
                        Usage( "Bad command line argument" );
                }

        if (FileArg == 0) // Did not find any file name
                Usage( "Command line must include a file name" );

        if ( (argc < 2) || (argc > 5) )
                Usage( "Wrong number of command line arguments" );


        
        // Fill file name
        
        strncpy( FileIn, argv[FileArg], MAXFILENAME-1 );
        FileIn[MAXFILENAME-1] = '\0';


        // Find the end of the input file prefix

        len = strlen( FileIn );
        iprefix = -1;
        for ( i = len ; i > 0 ; i-- ) {
                if (FileIn[i-1] == '.') {
                        iprefix = i-1;
                        break;
                        }
                }
        
        if (iprefix < 0 || iprefix > len || iprefix > MAXFILENAME-4) {
                sprintf( Msg, "File name %s is invalid", FileIn );
                Usage( Msg );
                }


        // Extract the prefix, null terminate, and uppercase

        strncpy( FilePrefix, FileIn, iprefix );
        FilePrefix[iprefix] = '\0';

        p   = FilePrefix;
        len = strlen(FilePrefix);
        for ( i = 0 ; i < len ; i++ ) {
                *p = (char)toupper( *p );
                *p++;
                }



        // If optional sequential argument was found, determine starting 
        // sequential number from prefix and verify.

        Sequential = 0;
        SeqNum     = -1;

        if ( UseSeq ) {
                        
                Sequential = 1;

                if ( sscanf(FilePrefix, "%X", &itmp) == 1 )
                        SeqNum = itmp;

                sprintf( PrefixCheck, "%08X", SeqNum );
                if ( strcmp( FilePrefix, PrefixCheck ) != 0 )
                        Usage( "Can't set sequential value from filename" );
                }



        // Make ASCII output name

        sprintf( FileOut, "%s.ASC", FilePrefix );

        
        return( 1 );
}


void ShowHdr( FILE *fout, SRDAT_HDR *Hdr ) {

        int i;
        char cDatId[5];

        
        // Check header size
        
        if (sizeof(*Hdr) != SRDAT_HDR_LAYOUTSIZE)
                ErrorWarning( "Header size not right" );

        
        // Show header info
        
        strncpy( cDatId,((char *)(&Hdr->DatId)), 4);
        cDatId[4] = '\0';
        

        fprintf(fout, "Header:\n");

        fprintf(fout, "DatId               = 0x%lX %s\n", Hdr->DatId, cDatId);
        fprintf(fout, "DatRev              = %ld\n", Hdr->DatRev);
        fprintf(fout, "DatProduct          = %s\n",  Hdr->DatProduct);
        fprintf(fout, "DatApplication      = %s\n",  Hdr->DatApplication);
        fprintf(fout, "DatUnused1          = %ld\n", Hdr->DatUnused1);
        fprintf(fout, "DatUnused2          = %ld\n", Hdr->DatUnused2);
	
        fprintf(fout, "FileName            = %s\n",  Hdr->FileName);
        fprintf(fout, "FileCreateTime      = %ld\n", Hdr->FileCreateTime);
        fprintf(fout, "FileSeqNum          = %ld\n", Hdr->FileSeqNum);
        fprintf(fout, "FileStartPtNum      = %ld\n", Hdr->FileStartPtNum);
        fprintf(fout, "FileUnused1         = %ld\n", Hdr->FileUnused1);

        for ( i = 0 ; i < SRDAT_DIM_TIME ; i++ ) {
                fprintf(fout, "TimePoint[%d]:\n", i);
                
                fprintf(fout, "   Type             = %ld",   Hdr->TimePoint[i].Type);
                fprintf(fout, PickTimeTypeString(Hdr->TimePoint[i].Type) );
                fprintf(fout, "   SampleFromRun    = %ld\n", Hdr->TimePoint[i].SampleFromRun);
                fprintf(fout, "   SampleFromFile   = %ld\n", Hdr->TimePoint[i].SampleFromFile);
                fprintf(fout, "   GpsMarkIndex     = %ld\n", Hdr->TimePoint[i].GpsMarkIndex);
                
                fprintf(fout, "   Year             = %ld\n", Hdr->TimePoint[i].Year);
                fprintf(fout, "   Month            = %ld\n", Hdr->TimePoint[i].Month);
                fprintf(fout, "   Day              = %ld\n", Hdr->TimePoint[i].Day);
                fprintf(fout, "   YMDSource        = %ld",   Hdr->TimePoint[i].YMDSource);
                fprintf(fout, PickSourceString(Hdr->TimePoint[i].YMDSource) );
                
                fprintf(fout, "   Hour             = %ld\n", Hdr->TimePoint[i].Hour);
                fprintf(fout, "   Minute           = %ld\n", Hdr->TimePoint[i].Minute);
                fprintf(fout, "   Second           = %ld\n", Hdr->TimePoint[i].Second);
                fprintf(fout, "   Microsecond      = %ld\n", Hdr->TimePoint[i].Microsecond);
                fprintf(fout, "   HMSSource        = %ld",   Hdr->TimePoint[i].HMSSource);
                fprintf(fout, PickSourceString(Hdr->TimePoint[i].HMSSource) );
                
                fprintf(fout, "   Unused1          = %ld\n", Hdr->TimePoint[i].Unused1);
                }

        fprintf(fout, "Channels            = %ld\n", Hdr->Channels);
        fprintf(fout, "ChannelMarkIndex    = %ld\n", Hdr->ChannelMarkIndex);
        fprintf(fout, "ChannelPtsPerRecord = %ld\n", Hdr->ChannelPtsPerRecord);
        fprintf(fout, "ChannelPtsPerFile   = %ld\n", Hdr->ChannelPtsPerFile);

	for ( i = 0 ; i < Hdr->Channels ; i++ ) {
		fprintf(fout, "Channel %02d          = %s\n",
                          Hdr->ChannelList[i].Channel,
                          Hdr->ChannelList[i].Title );
                }

        fprintf(fout, "AtodWordSize        = %ld\n", Hdr->AtodWordSize);
        fprintf(fout, "AtodDataCoding      = %ld",   Hdr->AtodDataCoding);
        fprintf(fout, PickDataCodingString(Hdr->AtodDataCoding) );
        fprintf(fout, "AtodSamplingRate    = %lf\n", Hdr->AtodSamplingRate);
        fprintf(fout, "AtodSpsMeasured     = %lf\n", Hdr->AtodSamplingRateMeasured);
        fprintf(fout, "AtodGain            = %lf\n", Hdr->AtodGain);
        fprintf(fout, "AtodFilterCoeff     = %lf\n", Hdr->AtodFilterCoeff);
        fprintf(fout, "AtodFilterScale     = %lf\n", Hdr->AtodFilterScale);

        fprintf(fout, "CounterType         = %ld\n", Hdr->CounterType);
        fprintf(fout, "CounterFreq         = %ld\n", Hdr->CounterFreq);
        
        fprintf(fout, "LocationLatitude    = %lf\n", Hdr->LocationLatitude);
        fprintf(fout, "LocationLongitude   = %lf\n", Hdr->LocationLongitude);
        fprintf(fout, "LocationAltitude    = %lf\n", Hdr->LocationAltitude);
        fprintf(fout, "LocationSource      = %ld",   Hdr->LocationSource);
        fprintf(fout, PickSourceString(Hdr->LocationSource) );
        
        fprintf(fout, "LocationUnused1     = %ld\n", Hdr->LocationUnused1);

}

char *PickTimeTypeString( long Type ) {
        
        if (Type == SRDAT_TIME_TYPE_NONE)
                return( " (none)\n");
        else if (Type == SRDAT_TIME_TYPE_FIRSTPT)
                return( " (first point)\n");
        else if (Type == SRDAT_TIME_TYPE_LASTPT)
                return( " (last point)\n");
        else if (Type == SRDAT_TIME_TYPE_FIRSTGPS)
                return( " (first GPS point)\n");
        else if (Type == SRDAT_TIME_TYPE_LASTGPS)
                return( " (last GPS point)\n");
        else
                return( "\n");
}

char *PickSourceString( long Source ) {
        
        if (Source == SRDAT_SOURCE_NONE)
                return( " (none)\n");
        else if (Source == SRDAT_SOURCE_CALC)
                return( " (from calculation)\n");
        else if (Source == SRDAT_SOURCE_PC)
                return( " (from PC)\n");
        else if (Source == SRDAT_SOURCE_GPS)
                return( " (from GPS)\n");
        else if (Source == SRDAT_SOURCE_USER)
                return( " (from user)\n");
        else if (Source == SRDAT_SOURCE_ACQ)
                return( " (from acq program)\n");
        else if (Source == SRDAT_SOURCE_OBC)
                return( " (from GPS + on board counter)\n");
        else
                return( "\n");
}

char *PickDataCodingString( long DataCoding ) {
        
        if (DataCoding == SRDAT_DATACODING_SIGNED)
                return( " (signed)\n");
        else if (DataCoding == SRDAT_DATACODING_OFFSET)
                return( " (offset)\n");
        else
                return( "\n");
}


void ShowTag( FILE *fout, SRDAT_TAG *Tag ) {

        char cTagType[5];
        char *TypeStr;

        
        // Check tag size
        
        if (sizeof(*Tag) != SRDAT_TAGSIZE)
                ErrorWarning( "Tag size not right" );

        strncpy( cTagType,((char *)(&Tag->TagId)), 4);
        cTagType[4] = '\0';


        // Set up tag type string

        switch (Tag->TagId) {
                case SRDAT_TAGID_ANALOG16:
                        TypeStr = "16 bit analog data";
                        break;
                case SRDAT_TAGID_ANALOG24:
                        TypeStr = "24 bit analog data";
                        break;
                case SRDAT_TAGID_GPSPPS:
                        TypeStr = "GPS pps data";
                        break;
                case SRDAT_TAGID_GPSSERIAL:
                        TypeStr = "GPS serial data";
                        break;
                case SRDAT_TAGID_EOF:
                        TypeStr = "End of File";
                        break;
                default:
                        TypeStr = "Unknown";
                        break;
                }
        

        // Show tag info

        fprintf( fout, "\nRecord Type: %s,  Tag: %s,  Bytes: %lu\n",
                                             TypeStr, cTagType, Tag->Nbytes );
        
}


void ShowData( FILE *fout, SRDAT_HDR *Hdr, void *Buffer,
               long DataType, unsigned long Nbytes,
               int HexFormat, int TimeFormat ) {

        char DataFmt16[MAXDATAFMT], DataFmt24[MAXDATAFMT];
        char TimeStr[MAXTIMESTR];
        char cTagType[5], MarkStar;
        int i, j, ToShow, Line;
        long Type, DPdiff;
        long *Analog24;
        short *Analog16;
        PPSDATA *GpsPps;
        SERIALDATA *GpsSerial;
        char *Unknown;
        

        // Show data

        Type = DataType;
        strncpy( cTagType,((char *)(&Type)), 4);
        cTagType[4] = '\0';

        if (HexFormat == 1) {
                strncpy( DataFmt16, "  0x%04X",  MAXDATAFMT-1 );
                strncpy( DataFmt24, "  0x%08lX", MAXDATAFMT-1 );
                }
        else {
                strncpy( DataFmt16, "  %+6d",   MAXDATAFMT-1 );
                strncpy( DataFmt24, "   %+9ld", MAXDATAFMT-1 );
                }
        DataFmt16[MAXDATAFMT-1] = '\0';
        DataFmt24[MAXDATAFMT-1] = '\0';
                

        
        switch (DataType) {
            case SRDAT_TAGID_ANALOG16:
                    Analog16 = (short*)Buffer;
                    ToShow = Nbytes;
                    j = 0;
                    while (ToShow > 0) {
                            if (Hdr->ChannelMarkIndex == SRDAT_MARKCHANNEL_NONE
                                || Analog16[Hdr->ChannelMarkIndex] == 0x0)
                                    MarkStar = MARKCHAR_NONE;
                            else
                                    MarkStar = MARKCHAR_FOUND;
                            fprintf( fout, "%010d  %03d", CurrentPt16, j );
                            for ( i = 0 ; i < Hdr->Channels ; i++ ) {
                                    fprintf( fout, DataFmt16, *Analog16++ );
                                    ToShow -= Hdr->AtodWordSize;
                                    }
                            fprintf( fout, " %c%c\n", MarkStar, MarkStar );
                            j++; CurrentPt16++;
                            }
                    break;
                        
            case SRDAT_TAGID_ANALOG24:
                    Analog24 = (long*)Buffer;
                    ToShow = Nbytes;
                    j = 0;
                    while (ToShow > 0) {
                            if (Hdr->ChannelMarkIndex == SRDAT_MARKCHANNEL_NONE
                                || Analog24[Hdr->ChannelMarkIndex] == 0x0)
                                    MarkStar = MARKCHAR_NONE;
                            else
                                    MarkStar = MARKCHAR_FOUND;
                            fprintf( fout, "%010d  %03d", CurrentPt24, j );
                            for ( i = 0 ; i < Hdr->Channels ; i++ ) {
                                    fprintf( fout, DataFmt24, *Analog24++ );
                                    ToShow -= Hdr->AtodWordSize;
                                    }
                            if (TimeFormat != TIME_FORMAT_NONE) {
                                    ComputeTime( Hdr, TimeFormat, CurrentPt24, TimeStr );
                                    fprintf( fout, " %s", TimeStr );
                                    }
                            fprintf( fout, " %c%c\n", MarkStar, MarkStar );
                            j++; CurrentPt24++;
                            }
                    break;
                        
#ifdef GPS_AVAILABLE
            case SRDAT_TAGID_GPSPPS:
                    GpsPps = (PPSDATA*)Buffer;
                    ToShow = Nbytes;
                    while (ToShow > 0) {
                            DPdiff = ParGpsLargeIntDiffType(
                                                    GpsPps->CountAtDready,
                                                    GpsPps->CountAtPps,
                                                    Hdr->CounterType );
                            fprintf(fout,"PpsEvent=%u, Cnt @Pps=0x%08lX%08lX @Dready=0x%08lX%08lX (D-P=%ld)\n",
                                      GpsPps->PpsEventNum,
                                      GpsPps->CountAtPps.u.HighPart,
                                      GpsPps->CountAtPps.u.LowPart,
                                      GpsPps->CountAtDready.u.HighPart,
                                      GpsPps->CountAtDready.u.LowPart,
                                      DPdiff);
                            GpsPps++;
                            ToShow -= sizeof(PPSDATA);
                            }
                    break;
                        
            case SRDAT_TAGID_GPSSERIAL:
                    GpsSerial = (SERIALDATA*)Buffer;
                    ToShow = Nbytes;
                    while (ToShow > 0) {
                            fprintf(fout, "Serial for PpsEvent = %u, NMEA string count = %d\n",
                                      GpsSerial->PpsEventNum,
                                      GpsSerial->NmeaCount);
                            for ( i = 0 ; i < GpsSerial->NmeaCount ; i++ ) {
                                    fprintf(fout,"%s",
                                       &GpsSerial->NmeaMsg[i*MAX_NMEA_SIZE]);
                                    }
                            GpsSerial++;
                            ToShow -= sizeof(SERIALDATA);
                            }
                    break;
#else                    
            case SRDAT_TAGID_GPSPPS:
            case SRDAT_TAGID_GPSSERIAL:
                    DPdiff    = 0;
                    GpsPps    = NULL;
                    GpsSerial = NULL;
                    fprintf( fout, "Compile with GPS_AVAILABLE to display PPS and serial records\n" );
                    printf("DAT2ASC must be compiled with GPS_AVAILABLE in\n" );
                    printf("order to understand GPS PPS and serial records.\n" );
                    // fall through to default case
#endif

            default:
                    fprintf(fout,"Unknown data type 0x%X\n", DataType );

                    i    = 0;
                    Line = 0;
                    Unknown = (char *)Buffer;

                    for ( i = 0 ; i < Nbytes ; i++ ) {
                            fprintf( fout, "%02X ", (Unknown[i] & 0xFF) );
                            Line++;
                            if (Line > 9) {
                                    fprintf( fout, "\n" );
                                    Line = 0;
                                    }
                            }
                    fprintf( fout, "\n" );
                        
                    break;

            } // end switch
}


int ReadDatRecord( void  *PC_BUFFER, unsigned int *npts, unsigned int maxpts, 
                   void *PPS_BUFFER, unsigned int *npps, unsigned int maxpps, 
                   void *SER_BUFFER, unsigned int *nser, unsigned int maxser
                 ) {

        SRDAT_TAG    Tag;
        unsigned int nbytes;


        // Initialize values

        *npts = *npps = *nser = 0;



        // Read data record tag to determine data type
        
        if ( !SrDatTagRead( FHin, &Tag ) ) {
                printf( "Could not read .DAT tag\n" );
                CleanExit( 1 );
                }


        // Check data size (only matters in DOS version)

        nbytes = (unsigned int)Tag.Nbytes;
        if ((unsigned long)nbytes != Tag.Nbytes) {
                printf( "Error data record exceeds size allowed by unsigned int.\n");
                CleanExit( 1 );
                }



        // Read in analog, gps, or serial data buffer

        if (Tag.TagId == SRDAT_TAGID_ANALOG24) {

                // Read in analog data buffer
                *npts = Tag.Nbytes / sizeof( long );
                if (*npts > maxpts) {
                        printf( "Size of analog data %u exceeds maximum of %u.\n",
                                  *npts, maxpts );
                        printf( "Please recompile with larger maximum.\n");
                        CleanExit( 1 );
                        }
                if ( !SrDatDataRead( FHin, PC_BUFFER, nbytes ) ) {
                        printf( "Could not read .DAT analog data\n" );
                        CleanExit( 1 );
                        }
                }


        else if (Tag.TagId == SRDAT_TAGID_GPSPPS) {

                // Read in gps data buffer
                *npps = Tag.Nbytes / sizeof( PPSDATA );
                if (*npps > maxpps) {
                        printf( "Size of PPS data %u exceeds maximum of %u.\n",
                                  *npps, maxpps );
                        printf( "Please recompile with larger maximum.\n");
                        CleanExit( 1 );
                        }
                if ( !SrDatDataRead( FHin, PPS_BUFFER, nbytes ) ) {
                        printf( "Could not read .DAT PPS data\n" );
                        CleanExit( 1 );
                        }
                }


        else if (Tag.TagId == SRDAT_TAGID_GPSSERIAL) {

                // Read in gps serial data buffer
                *nser = Tag.Nbytes / sizeof( SERIALDATA );
                if (*nser > maxser) {
                        printf( "Size of serial data %u exceeds maximum of %u.\n",
                                  *nser, maxser );
                        printf( "Please recompile with larger maximum.\n");
                        CleanExit( 1 );
                        }
                if ( !SrDatDataRead( FHin, SER_BUFFER, nbytes ) ) {
                        printf( "Could not read .DAT serial data\n" );
                        CleanExit( 1 );
                        }
                }


        else if (Tag.TagId == SRDAT_TAGID_EOF) {
                return( 0 );
                }

        else {
                printf( "Unexpected tag TagId found 0x%X\n", Tag.TagId);
                }



        return( 1 );
}

long SetObcChannel( long NumChannels, long MarkChannel ) {

        // On board counter not available on PAR1CH or PAR4CH
        // (or if only analog channels were saved on the PAR8CH)

        if (NumChannels <= 8)
                return( SRDAT_MARKCHANNEL_NONE );


        // Output of O.B.C. data currently tied to output of mark channel
        
        if (MarkChannel == SRDAT_MARKCHANNEL_NONE)
                return( SRDAT_MARKCHANNEL_NONE );


        // Validity check
        
        if (MarkChannel+1 >= NumChannels)
                return( SRDAT_MARKCHANNEL_NONE );

        
        return( MarkChannel + 1 );
}

int SelectTimeMethod( long ObcChannel, long MarkChannel ) {
        
        int Method;

        
        // If there is an OBC channel, use it

        if (ObcChannel != SRDAT_MARKCHANNEL_NONE)
                Method = TIME_METHOD_OBC;

        
        // Or if there is GPS info, use it
        
        else if (MarkChannel != SRDAT_MARKCHANNEL_NONE)
                Method = TIME_METHOD_GPS;

        
        // Otherwise, just use sps

        else
                Method = TIME_METHOD_SPS;


        return( Method );
}


int FillTimeStr( int TimeFormat, double Time, char *TimeStr ) {

        int  Year, Month, Day, Hour, Minute, Second;
        long Microsecond;

        // Error check

        if ( !TimeStr )
                return( 0 );
        
        
        // Fill time string with selected format

        if ( TimeFormat == TIME_FORMAT_YMD ) {
                
                ParGpsSecTimeSplit( Time, &Year, &Month, &Day, 
                                    &Hour, &Minute, &Second, &Microsecond );
        
                sprintf( TimeStr, "%4d/%02d/%02d %02d:%02d:%02d.%06ld",
                         Year, Month, Day, Hour, Minute, Second, Microsecond );
                }
        
        else if ( TimeFormat == TIME_FORMAT_SEC )
                sprintf( TimeStr, "%18.6lf", Time );

        else
                return( 0 );

        
        return( 1 );

}

void Usage( char *msg ) {

        int i;
        
        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );


        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s\n", UsageMsg[i] );

        exit( 1 );
}


void ErrorWarning( char *msg ) {
        printf("DAT2ASC ERROR: %s\n", msg );
        exit( 1 );
}

#if !defined F_OK
#define F_OK 0x00
#endif

int FileExists( char *FileName ) {

        if ( access( FileName, F_OK ) == 0 )
                return( 1 );
        else
                return( 0 );

}

void CleanExit( int ret ) {

        // Close any open files

        if ( FHin != -1 )
                close( FHin );

        if ( Fout != NULL )
                fclose( Fout );

        if ( Fts != NULL )
                fclose( Fts );

        // Exit

        exit( ret );
}


// The behavior of the following three functions is very different
// depending on whether or not GPS info is available.

#ifdef GPS_AVAILABLE

int InitTimeStamps( void ) {

        unsigned int PrevEventNum, PpsEvent, Npts, Npps, Nser;
        int          TsTitle, SeqNumSave, PrevUnset, ok;
        int          NotDone, Nvalid, SerialDelay;
        long         NumChannels, NumSamples;
        long         StartSample, MarkChannel, ObcChannel;
        double       PrevPpsSecTime, TpSecTime;
        TIMESTAMP    TSx;
        TIMEPT       Tp;
        SR64BIT      Zero64;


        // Go through all .dat files initializing timestamps

        
        // Open timestamp file

        Fts = fopen( "TIMESTMP.TMP", "w+" );
        if (!Fts) {
                printf( "ERROR: Can not open temporary file TIMESTMP.TMP\n" );
                CleanExit( 1 );
                }
        TsTitle = 1; // Request that we write a title the first time through
                
        
        
        // Make sure first file exists

        if ( !FileExists( FileIn ) ) {
                printf( "ERROR: Can not open input file %s.\n", FileIn );
                CleanExit( 1 );
                }

        
        SeqNumSave        = SeqNum;
        PrevEventNum      = INVALID_PPSEVENT;
        SerialDelay        = 0;
        Nvalid             = 0;
        TimeStamp[0].Valid = TS_VALID_NONE;
        TimeStamp[1].Valid = TS_VALID_NONE;
        Zero64.u.HighPart  = 0L;
        Zero64.u.LowPart   = 0L;


        // Initialize ParGps library timestamp functions

        ParGpsTsInit();
                


        // Loop for several files

        printf( "Initializing TimeStamps:\n" );

        while ( 1 ) {

                // First check for existance of input file

                if ( !FileExists( FileIn ) ) {
                        printf( "No more files to process.  Stopping at %s.\n", FileIn );
                        break;
                        }
                
                printf( "Reading file %s ...\n", FileIn );

                
                // Open input file

                FHin = open( FileIn, (O_BINARY|O_RDONLY), (S_IREAD|S_IWRITE) );
                if (FHin == -1) {
                        printf( "ERROR: Can not open input file %s\n", FileIn );
                        break;
                        }


                // Read header, gather useful fields and check them

                if ( !SrDatHdrRead( FHin, &Hdr ) ) {
                        printf( "ERROR: Could not read .DAT header\n" );
                        CleanExit( 1 );
                        }

                if (Hdr.DatId != SRDAT_HDR_ID) {
                        printf( "ERROR: This is not a valid .DAT file." );
                        CleanExit( 1 );
                        }

                if (Hdr.DatRev != SRDAT_REV) {
                        printf( "WARNING: Header version mismatch " );
                        printf( "(%d in file, %d in program)\n",
                                                  Hdr.DatRev, SRDAT_REV );
                        }

                if ( (strlen(Hdr.DatProduct)     > SRDAT_DIM_STRING) ||
                     (strlen(Hdr.DatApplication) > SRDAT_DIM_STRING) ||
                     (strlen(Hdr.FileName)       > SRDAT_DIM_STRING) ) {
                        printf( "ERROR: Invalid strings in .DAT header." );
                        CleanExit( 1 );
                        }

                NumChannels          = Hdr.Channels;
                StartSample          = Hdr.FileStartPtNum;
                MarkChannel          = Hdr.ChannelMarkIndex;
                ObcCountsPerSample   = (OBC_FREQ / Hdr.AtodSamplingRate + 0.5L);
                BestCountsBetweenPps = Hdr.CounterFreq;

                ObcChannel = SetObcChannel( NumChannels, MarkChannel );
                TimeMethod = SelectTimeMethod( ObcChannel, MarkChannel );



                // Get PC time in seconds from the header

                Tp = Hdr.TimePoint[0];
                ParGpsSecTimeCombine(
                                (int)Tp.Year, (int)Tp.Month, (int)Tp.Day,
                                (int)Tp.Hour, (int)Tp.Minute, (int)Tp.Second,
                                 Tp.Microsecond, &TpSecTime );


                // If the SPS method was selected, there won't be any
                // time info later in the file so make and save a
                // timestamp from the header info.
                        
                if ( TimeMethod == TIME_METHOD_SPS ) {
                        TSx.Valid         = TS_VALID_ALL;
                        TSx.PpsEvent      = INVALID_PPSEVENT;
                        TSx.Sample        = Hdr.FileStartPtNum;
                        TSx.CountObc      = 0;
                        TSx.CountAtPps    = Zero64;
                        TSx.CountAtDready = Zero64;
                        TSx.SecSince1970  = TpSecTime;
                        TSx.NumSat        = 0;
                        TSx.YmdSource     = Tp.YMDSource;
                        TSx.HmsSource     = Tp.HMSSource;
                        ParGpsTsFileWrite( Fts, &TSx, TsTitle );
                        TsTitle = 0;
                        }
                

                // In case there is no GPS YMD info available, this
                // PC time from header will start things off right

                if ( PrevUnset ) {
                        PrevPpsSecTime = TpSecTime;
                        PrevUnset      = 0;
                        }



                // Loop through processing entire file

                NotDone = 1;

                while ( NotDone ) {

                        // Read next data record

                        NotDone = ReadDatRecord(  PC_BUFFER, &Npts, MAXANALOG,
                                                 PPS_BUFFER, &Npps, MAXPPS,
                                                 SER_BUFFER, &Nser, MAXSER
                                       );



                        // If it's one of the three desired types,
                        // process it into time stamps

                        if (Npts > 0)  {
                                NumSamples = Npts/NumChannels;
                                ParGpsTsProcessAnalog( PC_BUFFER, NumSamples, NumChannels,
                                                       StartSample, MarkChannel, 
                                                       ObcChannel, ObcCountsPerSample,
                                                       &PpsEvent, &Nvalid );
                                StartSample += NumSamples;
                                }

                        else if (Npps > 0) {
                                ParGpsTsProcessPps( PPS_BUFFER, Npps, &PpsEvent, &Nvalid );
                                }


                        else if (Nser > 0) {
                                ParGpsTsProcessSerial( SER_BUFFER, Nser, SerialDelay,
                                                       &PrevEventNum, &PrevPpsSecTime,
                                                       &PpsEvent, &Nvalid );
                                }


                        // Read the valid time stamps from the circular buffer
                        // maintained by the PARGPS library, correct the HMS source
                        // if using OBC to compute time, and write the time stamps
                        // out to the temporary file.

                        while ( Nvalid > 0 ) {
                                ParGpsTsReadValid( &TSx );
                                if ( TimeMethod == TIME_METHOD_OBC &&
                                     TSx.HmsSource == TIME_SOURCE_GPS   )
                                        TSx.HmsSource = TIME_SOURCE_OBC;
                                ParGpsTsFileWrite( Fts, &TSx, TsTitle );
                                TsTitle = 0;
                                Nvalid--;
                                // WCT: Maybe recompute best counts between pps here
                                }


                        }  // end of while NotDone


                close( FHin );   FHin = -1;



                // If only a single file was requested, we're done
                
                if (!Sequential)
                        break;

                // Otherwise, compute the next sequential file name
                
                else {
                        SeqNum++;
                        sprintf( FileIn, "%08X.DAT", SeqNum );
                        }

                } // end of while 1


        
        // Restore orignal sequence number and input file name

        SeqNum = SeqNumSave;
        if (Sequential)
                sprintf( FileIn, "%08X.DAT", SeqNum );


        // Close the time stamp file

        fclose( Fts );  Fts = NULL;


        
        // Re-open time stamp file for reading

        Fts = fopen( "TIMESTMP.TMP", "r" );
        if (!Fts) {
                printf( "ERROR: Can not open temporary file TIMESTMP.TMP\n" );
                CleanExit( 1 );
                }

        fgets( PrefixCheck, MAXFILENAME, Fts ); // Read past title


        // Get some timestamps to start us off
        // (get two except for SPS method which only needs one).
        
        ok = ParGpsTsFileRead( Fts, &TimeStamp[0], 0 );
        if (!ok) {
                printf( "Could not read time stamp 0\n" );
                return( 0 );
                }
        
        if ( TimeMethod == TIME_METHOD_SPS ) {
                // Fill in a dummy second timestamp
                TimeStamp[1]        = TimeStamp[0];
                TimeStamp[1].Sample = 0xFFFFFFFF;
                }
        else {
                ok = ParGpsTsFileRead( Fts, &TimeStamp[1], 0 );
                if (!ok) {
                        printf( "Could not read time stamp 1\n" );
                        return( 0 );
                        }
                }
        
        return( 1 );
}


int ComputeTime( SRDAT_HDR *Hdr, int TimeFormat, long Sample, char *TimeStr ) {

        double Time;
        int    ok;


        // Uses globals: TimeStamp{}, TimeMethod,
        //               ObcCountesPerSample, BestCountsBetweenPps


        // Do extensive time computations based on timestamps prepared
        // for each PPS event.


        // Error check

        if (!Hdr || !TimeStr || TimeFormat == TIME_FORMAT_NONE )
                return( 0 );


        // Update timestamps if current sample number is
        // beyond the right-most timestamp sample number

        if ( Sample > TimeStamp[1].Sample )
                UpdateTimeStamps( Sample );



        // Compute the time using the selected method

        if ( TimeMethod == TIME_METHOD_OBC )
                ok = ParGpsTsComputeTimeObc( &TimeStamp[0], &TimeStamp[1],
                                             ObcCountsPerSample,
                                             Sample, &Time );

        else if ( TimeMethod == TIME_METHOD_GPS )
                ok = ParGpsTsComputeTimePps( &TimeStamp[0], &TimeStamp[1],
                                             BestCountsBetweenPps, Hdr->CounterType,
                                             Sample, &Time );

        else // ( TimeMethod == TIME_METHOD_SPS )
                ok = ParGpsTsComputeTimeSps( &TimeStamp[0], &TimeStamp[1],
                                             Hdr->AtodSamplingRate,
                                             Sample, &Time );

        if (!ok) {
                strcpy( TimeStr, "Time error" );
                return( 0 );
                }


        // Fill in the time string to output

        FillTimeStr( TimeFormat, Time, TimeStr );

        return( 1 );
}


int UpdateTimeStamps( long Sample ) {

        TIMESTAMP TSx;
        // Uses globals: TimeMethod, TimeStamp[], Fts


        // No update required for SPS method

        if ( TimeMethod == TIME_METHOD_SPS )
                return( 1 );
                     
        
        while ( Sample > TimeStamp[1].Sample ) {

                if ( ParGpsTsFileRead( Fts, &TSx, 0 ) ) {
                        TimeStamp[0] = TimeStamp[1];
                        TimeStamp[1] = TSx;
                        }
                else
                        return( 0 );
                
                } // end while
        
        return( 1 );
}



#else // not GPS_AVAILABLE

int InitTimeStamps( void ) {

        TIMEPT Tp;
        double TpSecTime;


        // Set global TimeMethod to NONE

        TimeMethod = TIME_METHOD_NONE;



        // Read header from first file

        printf( "Initializing TimeStamps:\n" );

        TimeStamp[0].Sample       = 0L;
        TimeStamp[0].SecSince1970 = 0.0L;




        // Open input file

        FHin = open( FileIn, (O_BINARY|O_RDONLY), (S_IREAD|S_IWRITE) );
        if (FHin == -1) {
                printf( "ERROR: Can not open input file %s\n", FileIn );
                CleanExit( 1 );
                }


        // Read header, gather useful fields and check them

        if ( !SrDatHdrRead( FHin, &Hdr ) ) {
                printf( "ERROR: Could not read .DAT header\n" );
                CleanExit( 1 );
                }

        if (Hdr.DatId != SRDAT_HDR_ID) {
                printf( "ERROR: This is not a valid .DAT file." );
                CleanExit( 1 );
                }

        if (Hdr.DatRev != SRDAT_REV) {
                printf( "WARNING: Header version mismatch " );
                printf( "(%d in file, %d in program)\n",
                                          Hdr.DatRev, SRDAT_REV );
                }

        if ( (strlen(Hdr.DatProduct)     > SRDAT_DIM_STRING) ||
             (strlen(Hdr.DatApplication) > SRDAT_DIM_STRING) ||
             (strlen(Hdr.FileName)       > SRDAT_DIM_STRING) ) {
                printf( "ERROR: Invalid strings in .DAT header." );
                CleanExit( 1 );
                }

        close( FHin );   FHin = -1;

        

        // Get PC time in seconds from the header and save time
        // as a "timestamp" (cutdown version since no pargps.h)

        Tp = Hdr.TimePoint[0];

        if (Tp.Type != SRDAT_TIME_TYPE_NONE) {
                ParGpsSecTimeCombine(
                                (int)Tp.Year, (int)Tp.Month,  (int)Tp.Day,
                                (int)Tp.Hour, (int)Tp.Minute, (int)Tp.Second,
                                 Tp.Microsecond, &TpSecTime );
        
                TimeStamp[0].Sample       = Tp.SampleFromRun;
                TimeStamp[0].SecSince1970 = TpSecTime;
                }
                

        return( 1 );
}


int ComputeTime( SRDAT_HDR *Hdr, int TimeFormat, long Sample, char *TimeStr ) {

        double  Time;


        // Use PC time and sampling rate from the header to compute time.

        // Error check

        if ( !Hdr                           ||
             !TimeStr                       ||
             TimeFormat == TIME_FORMAT_NONE ||
             TimeMethod != TIME_METHOD_NONE )
                return( 0 );



        // Get the first time point from the header and use it plus the
        // sampling rate to determine the time.  Note that we checked
        // AtodSamplingRate in main to ensure it is valid so we won't
        // divide by zero here.

        Time = TimeStamp[0].SecSince1970 +
               (double)(Sample - TimeStamp[0].Sample) /
               Hdr->AtodSamplingRate;  



        // Fill in the time string to output

        FillTimeStr( TimeFormat, Time, TimeStr );

        return( 1 );
}

int UpdateTimeStamps( long Sample ) {

        return( 1 );
}




// The following functions are duplicates of those in the
// PARGPS library, but they are so handy we're repeating them
// here if the PARGPS library is not linked in.


#define USPERSEC  1000000L // microseconds / second

int ParGpsCalcTimeZoneCorr( long StartTime ) {
        
        time_t    TargetTime, NewTime;
        struct tm tm;
        int       ZoneCorrection;

        // This function does two conversions between time stored as a structure
        // and time as seconds since the epoch (1970-01-01).  They should cancel
        // out, but the first is done using GMT time and the second is done using
        // local time.  The difference gives us the amount of the local correction.
        // Subtract this value from a local time to get a GMT time.
        
        
        // Convert epoch seconds to time structure for gmt
        
        TargetTime = (time_t)StartTime;
        tm = *gmtime( &TargetTime );

        
        // Convert time structure for local to epoch seconds (force DST unknown)
        
        tm.tm_isdst = -1;
        NewTime = mktime( &tm );


        // Difference is local correction

        ZoneCorrection = NewTime - TargetTime;
        return( ZoneCorrection );
}

int ParGpsSecTimeSplit( double Time,
                        int *Year, int *Month, int *Day,
                        int *Hour, int *Minute, int *Second, 
                        long *Microsecond ) {

        time_t    ttTime;
        struct tm tmTime;
        double    usec;


        // Split time into pieces (for UTC)

        ttTime = (time_t)Time;
        tmTime = *gmtime( &ttTime );
        usec   = Time - (double)ttTime + 0.0000005;
        
        // Save pieces
        
        if (Year)         *Year        = tmTime.tm_year + 1900;
        if (Month)        *Month       = tmTime.tm_mon + 1;
        if (Day)          *Day         = tmTime.tm_mday;
        if (Hour)         *Hour        = tmTime.tm_hour;
        if (Minute)       *Minute      = tmTime.tm_min;
        if (Second)       *Second      = tmTime.tm_sec;
        if (Microsecond)  *Microsecond = (long)(usec * USPERSEC);

        return( 1 );
}

int ParGpsSecTimeCombine( int Year, int Month, int Day,
                          int Hour, int Minute, int Second, 
                          long MicroSecond, double *Time ) {

        time_t    ttTime;
        struct tm tmTime;

        // Set default YMD values when 0's are passed in

        if (Year == 0)   Year  = 1970;
        if (Month == 0)  Month = 1;
        if (Day == 0)    Day   = 1;


        // Fill pieces
        
        tmTime.tm_year  = Year - 1900;
        tmTime.tm_mon   = Month - 1;
        tmTime.tm_mday  = Day;
        tmTime.tm_hour  = Hour;
        tmTime.tm_min   = Minute;
        tmTime.tm_sec   = Second;
        tmTime.tm_wday  = 0;
        tmTime.tm_yday  = 0;
        tmTime.tm_isdst = -1;


        // Combine time from GMT pieces to epoch seconds using local correction,
        // then use the zone offset to remove the local correction

        ttTime = mktime( &tmTime );
        if (ttTime == (time_t)-1)
                return( 0 );
        ttTime -= ParGpsCalcTimeZoneCorr( ttTime );
        if (Time)  *Time  = (double)ttTime + (double)MicroSecond / (double)USPERSEC;
        
        return( 1 );
}

#endif // GPS_AVAILABLE
