/* FILE: out2asc.c             Copyright (c), Symmetric Research 2001-2005
 *
 * This program reads in a binary .hdr and a binary .out file generated
 * by one of the SR acquisition programs and writes out the information
 * in an ASCII text version.  This file can be compiled for use with 
 * either the SR PAR24B or PARxCH A/D board.
 *
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>





//****** A/D Specific Data ******


#if defined( SRA2D_PARANY )
#include "parxch.h"

#define ANALOGCHANNELS   PARXCH_MAX_ANALOG_CHANNELS
#define DIGITALCHANNELS  PARXCH_MAX_DIGITAL_CHANNELS
#define GPSCHANNELS      PARXCH_MAX_GPS_CHANNELS
#define COUNTERCHANNELS  PARXCH_MAX_COUNTER_CHANNELS

        

#elif defined( ATOD_IS_PAR24B )

#define ANALOGCHANNELS   8
#define DIGITALCHANNELS  0
#define GPSCHANNELS      0
#define COUNTERCHANNELS  0


#else

#pragma message( "Valid A/D options are SRA2D_PARANY or ATOD_IS_PAR24B." )
#error Invalid A/D option defined.

#define ANALOGCHANNELS   8
#define DIGITALCHANNELS  0
#define GPSCHANNELS      0
#define COUNTERCHANNELS  0

#endif



#define TOTALCHANNELS    (ANALOGCHANNELS + DIGITALCHANNELS +  \
                          GPSCHANNELS + COUNTERCHANNELS)

#define MAXFILENAME      256
#define MAXFMT           20




/* Header structure
 *
 *
 *
 */

#define MAXCHANNELS             TOTALCHANNELS
#define MAXTITLE                12

#define SROUT_DATACODING_OFFSET  0
#define SROUT_DATACODING_SIGNED  1

typedef struct SrOutHdrLayout {

          char Id[32];

          long Channels;
          long RecordPts;
          long WordSize;
          long DataCoding;

        double SamplingRate;
        double Gain;
        double FilterCoeff;
        double FilterScale;

        struct channel_data { long channel; char title[ MAXTITLE ]; }

                        ChannelList[ MAXCHANNELS ];
        } SROUT_HDR;

SROUT_HDR HDR;

#define HDRSIZE         sizeof( SROUT_HDR )


// This must match the value of SRDAT_HDR_ID in srdat.h
// Dos can not handle multi-character constants so use a long

#if defined( SROS_MSDOS )
#define CHECK_HDR_ID     (0x52535253L)   // = "SRSR"
#else
#define CHECK_HDR_ID    ((long)('RSRS')) // = "SRSR"
#endif



/* Sign extension constants
 *
 * PAR4CH data can be acquired in 24-bit signed or offset format.
 * When stored as a 32-bit number, the 24-bit signed format data must
 * be sign extended if the 24th bit is set.
 *
 */

#define NEG24BIT        0x00800000
#define SIGNEXTEND      0xFF000000
#define NOSIGNEXTEND    0x00000000




char* UsageMsg[] = {
    "USAGE: out2asc file.HDR file.OUT [/seq] [/hex]                       ",
    "                                                                     ",
    "Out2asc converts the specified .HDR header and .OUT data files to    ",
    "ASCII text format.  The ASCII output filename is the same as the     ",
    "input data filename except the suffix is changed from .OUT to .ASC.  ",
    "If the prefix of the input data file contains a hex number, the      ",
    "optional /seq argument indicates the input file is the first of a    ",
    "sequence.  In this case, out2asc will continue incrementing the file ",
    "prefix and converting until the next file in the sequence can not be ",
    "found.  The optional /hex argument causes the analog data to be      ",
    "written in hex instead of decimal format.                            ",
    "                                                                     ",
    "EXAMPLES: out2asc SCOPE.HDR SCOPE.OUT                                ",
    "          out2asc SIMP.HDR 0000001F.OUT /seq /hex                    ",
    NULL
    };


void Usage( char *msg ) {

        int i;

        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );


        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( "%s\n", UsageMsg[i] );

        exit( 1 );
}


void ErrorWarning( char *msg ) {
        printf("OUT2ASC ERROR: %s\n", msg );
        exit( 1 );
}


#if !defined F_OK
#define F_OK 0x00
#endif

int FileExists( char *FileName ) {

        if ( access( FileName, F_OK ) == 0 )
                return( 1 );
        else
                return( 0 );

}

void ErrorFile( char *fmt, char *filename ) {

        /* Give an error message and quit. */

        printf( "ERROR: " );
        printf( fmt, filename );

        exit( 1 );

}



void SetSpecialIndex( int *DigitalIndex, int *GpsMarkIndex, int *GpsCtrIndex ) {

        int i;

        // Set defaults to indicate none

        *DigitalIndex = -1;
        *GpsMarkIndex = -1;
        *GpsCtrIndex  = -1;


        // Look for digital and GPS mark indicies by channel name

        for ( i = 0 ; i < HDR.Channels ; i++ ) {
                
                if (strcmp(HDR.ChannelList[i].title,"DGTL") == 0)
                        *DigitalIndex = i;

                else if (strcmp(HDR.ChannelList[i].title,"MARK") == 0)
                        *GpsMarkIndex = i;
                }

        
        // Compute counter index from GPS mark index

        if (*GpsMarkIndex != -1 && *GpsMarkIndex < HDR.Channels)
                *GpsCtrIndex = *GpsMarkIndex + 1;
}

void MakeAsciiName( char *BinaryName, char *AsciiName, int *SeqStart ) {

        int i, iprefix, len, itmp;
        char FilePrefix[MAXFILENAME];


        // Find the end of the input file prefix

        len = strlen( BinaryName );
        iprefix = -1;
        for ( i = len ; i > 0 ; i-- ) {
                if (BinaryName[i-1] == '.') {
                        iprefix = i-1;
                        break;
                        }
                }
        
        if (iprefix < 0 || iprefix > len || iprefix > MAXFILENAME-4) {
                sprintf( FilePrefix, "File name %s is invalid", BinaryName );
                Usage( FilePrefix );
                }


        // Extract the prefix and null terminate

        strncpy( FilePrefix, BinaryName, iprefix );
        FilePrefix[iprefix] = '\0';



        // Determine possible sequential number from prefix

        *SeqStart = -1;
        if ( sscanf(FilePrefix, "%X", &itmp) == 1 )
                *SeqStart = itmp;



        // Make ASCII output name

        sprintf( AsciiName, "%s.ASC", FilePrefix );

}

void ShowHdr( FILE *fout, SROUT_HDR *Hdr ) {

        int i;
        char *DataCodeName;

        if ( Hdr->DataCoding == SROUT_DATACODING_SIGNED )
                DataCodeName = "signed";
        else // Hdr->DataCoding == SROUT_DATACODING_OFFSET
                DataCodeName = "offset";
        
        
        fprintf( fout, "Header:\n" );
        fprintf( fout, "           Id = %s\n",       Hdr->Id           );
        fprintf( fout, "     Channels = %ld\n",      Hdr->Channels     );
        fprintf( fout, "    RecordPts = %ld\n",      Hdr->RecordPts    );
        fprintf( fout, "     WordSize = %ld\n",      Hdr->WordSize     );
        fprintf( fout, "   DataCoding = %ld (%s)\n", Hdr->DataCoding, DataCodeName );
        fprintf( fout, " SamplingRate = %f\n",       Hdr->SamplingRate );
        fprintf( fout, "         Gain = %f\n",       Hdr->Gain         );
        fprintf( fout, "  FilterCoeff = %f\n",       Hdr->FilterCoeff  );
        fprintf( fout, "  FilterScale = %f\n",       Hdr->FilterScale  );

        for ( i = 0 ; i < Hdr->Channels ; i++ )

                fprintf( fout, "   Channel[%1ld] = %s\n",
                                 Hdr->ChannelList[i].channel,
                                 Hdr->ChannelList[i].title
                               );
}








/* MAIN PROGRAM:
 *
 * Open the input and output files and translate to ASCII.
 *
 *
 */

int main( int argc, char *argv[] ) {

        char HeaderName[MAXFILENAME]; FILE *HeaderFile;
	char BinaryName[MAXFILENAME]; FILE *BinaryFile;
        char AsciiName[MAXFILENAME];  FILE *AsciiFile;
        char MarkStar, DataFmt[MAXFMT], DataTitle[MAXFMT];
        int  UseSeq, HexFormat, HdrFileArg, OutFileArg, SeqStart;
	int  i, ifile, DigitalIndex, GpsMarkIndex, GpsCtrIndex, Nbytes;
        long Nsample, SetTopBits, DataBuffer[TOTALCHANNELS];


        // Announce program version.

        printf( "OUT2ASC [/?]: (c) SR 2004 : Rev {09/30/04}\n" );




        // Process command line arguments

        UseSeq     = 0;
        HexFormat  = 0;
        HdrFileArg = 0;
        OutFileArg = 0;

        for ( i = 1 ; i < argc ; i++ ) {

                if (strcmp("/?",argv[i]) == 0)
                        Usage( "?" );
                else if (strcmp("/seq",argv[i]) == 0)
                        UseSeq = 1; 
                else if (strcmp("/hex",argv[i]) == 0)
                        HexFormat = 1; 
                else if (HdrFileArg == 0)
                        HdrFileArg = i;
                else if (OutFileArg == 0)
                        OutFileArg = i;
                else
                        Usage( "Bad command line argument" );
                }

        if (HdrFileArg == 0) // Did not find any hdr file name
                Usage( "Command line must include a header file name." );
        
        if (OutFileArg == 0) // Did not find any out file name
                Usage( "Command line must include a .out data file name." );

        if ( (argc < 3) || (argc > 5) )
                Usage( "Wrong number of command line arguments.  Must be 2 to 4." );

        
        // Fill some file name parameters
        
        strncpy( HeaderName, argv[HdrFileArg], MAXFILENAME-1 );
        strncpy( BinaryName, argv[OutFileArg], MAXFILENAME-1 );
        HeaderName[MAXFILENAME-1] = '\0';
        BinaryName[MAXFILENAME-1] = '\0';
        MakeAsciiName( BinaryName, AsciiName, &SeqStart );
        if (UseSeq)
                ifile = SeqStart;
        else
                ifile = -1;

        printf( "\n" );


        HeaderFile = NULL;
        BinaryFile = NULL;
        AsciiFile  = NULL;

	
	// Process header file - check existance, open, read, close.

        if ( !FileExists( HeaderName ) )
                ErrorFile( "Input header file %s does not exist", HeaderName );

	HeaderFile = fopen( HeaderName, "rb" );
	if ( !HeaderFile )
                ErrorFile( "Can't open input header file %s", HeaderName );

        printf( "Reading file %s ...\n", HeaderName );

        fread( &HDR, 1, HDRSIZE, HeaderFile );

        if ( ( strlen( HDR.Id ) > 32 )        ||
             ( HDR.Id[0] != 'S' )             ||
             ( HDR.Id[1] != 'R' )             ||
             ( HDR.Channels < 0 )             ||
             ( HDR.Channels > TOTALCHANNELS ) ) {
                ErrorFile( "Header file %s is invalid", HeaderName );
                }

        fclose( HeaderFile );



        // Set parameters based on header info ...

        // Sign extension constant for signed vs offset data

        if (HDR.DataCoding == 1)
                SetTopBits = SIGNEXTEND;
        else
                SetTopBits = NOSIGNEXTEND;


        // Format and title strings for hex vs decimal output

        if ( HexFormat == 1 ) {
                strncpy( DataTitle, "\nHex Data:\n", MAXFMT );
                strncpy( DataFmt, "%08lX ", MAXFMT);
                }
        else {
                strncpy( DataTitle, "\nData:\n", MAXFMT );
                strncpy( DataFmt, "%8ld ", MAXFMT);
                }


        // Special channel indicies for digital, GPS mark and counter channels

        SetSpecialIndex( &DigitalIndex, &GpsMarkIndex, &GpsCtrIndex );
        




        

        
        // Loop until no more out files requested or found

        while ( 1 ) {

                // First check for existance of input data file

                if ( !FileExists( BinaryName ) ) {
                        printf( "\nNo more files to process.  Stopping at %s.\n", BinaryName);
                        exit( 0 );
                        }
                
        
                // Open input and output files

                BinaryFile = fopen( BinaryName, "rb" );
                if ( !BinaryFile )
                        ErrorFile( "Can't open input data file %s", BinaryName );
                
                AsciiFile = fopen( AsciiName, "w" );
                if ( !AsciiFile )
                        ErrorFile( "Can't open output ASCII file %s", AsciiName );


                // There is no definitive test to ensure the binary file IS
                // an OUT file, but at least make sure it is NOT a DAT file.
                // Get first point, check it, and rewind.
                
                fread( DataBuffer, 1, 4, BinaryFile );
                if ( DataBuffer[0] == CHECK_HDR_ID )
                        ErrorFile( "Input file %s must be .OUT not .DAT format", BinaryName );
                fseek( BinaryFile, 0L, SEEK_SET );



                // Write header info to ASCII file

                printf( "\nWriting file %s ...\n", AsciiName );
                printf( "Converting %s header info ...\n", HeaderName );

                ShowHdr( AsciiFile, &HDR );



                // Write data column titles to ASCII file

                fprintf( AsciiFile, DataTitle );

                for ( i = 0 ; i < HDR.Channels ; i++ )
                        fprintf( AsciiFile, "%8s ", HDR.ChannelList[i].title );

                fprintf( AsciiFile, "   Index\n" );




                // Convert the binary data file to ASCII

                printf( "Converting %s binary data ...\n", BinaryName );

                Nbytes  = HDR.Channels * sizeof( long );
                Nsample = 0;

                while ( fread( DataBuffer, 1, Nbytes, BinaryFile ) == Nbytes ) {

                        // Convert the samples for each channel at this time point.

                        for ( i = 0 ; i < HDR.Channels ; i++ ) {

                                // If not a special channel, 
                                // sign extend 24 bit data to 32 bits.

                                if ( (i != DigitalIndex) &&
                                     (i != GpsMarkIndex) &&
                                     (i != GpsCtrIndex)  &&
                                     (DataBuffer[i] & NEG24BIT) )
                                        DataBuffer[i] |= SetTopBits;

                                fprintf( AsciiFile, DataFmt, DataBuffer[i] );
                                }



                        // Add final columns with the time index
                        // and GPS mark indicator.

                        if ( GpsMarkIndex == -1 || DataBuffer[GpsMarkIndex] == 0 )
                                MarkStar = ' ';
                        else
                                MarkStar = '<';

                        fprintf( AsciiFile, "%8ld  %c\n", Nsample, MarkStar );

                        Nsample++;

                        } // end while fread



                // Close current files

                fclose( BinaryFile );
                fclose( AsciiFile );


                
                // If only a single file was requested, we're done
                
                if (ifile == -1)
                        break;

                // Otherwise, compute the next set of sequential file names
                
                else {
                        ifile++;
                        sprintf( BinaryName, "%08X.OUT", ifile );
                        sprintf( AsciiName,  "%08X.ASC", ifile );
                        }


                } // end while 1 (more files to process)


        exit( 0 );

}
