// FILE: dat2sud.c               Copyright (c), Symmetric Research, 2003-2005
//
// This program converts binary .dat data file from programs like
// Symmetric Research simpxxx and scopexxx into a SUDS waveform format
// file.   Where xxx represents a64, 24b, or "" depending on whether
// this file is compiled for the SR DSPA64, PAR24B, or PARxCH A/D board.
//
// It uses the Small Systems Support SUDS support library.  When using
// a 16 bit compiler, it *MUST* be compiled large model because the
// default SUDS library support is large model.  When using a 32 bit
// compiler, it *MUST* be compiled with 2 byte packing for consistency
// with existing SUDS files and to ensure correct time fields.
//
//
//

#include <stdio.h>      // for printf, fseek, etc
#include <stdlib.h>     // for exit, _fullpath
#include <malloc.h>     // for malloc, free
#include <string.h>     // for strcpy, strcmp
#include <fcntl.h>      // for O_xxx defines
#include <sys/stat.h>   // for S_IREAD define (VC++ can use / instead of \)

#ifdef SROS_LINUX
#include <sys/types.h>  // for file status
#include <sys/time.h>   // for gettimeofday
#else
#include <sys\types.h>  // for file status
#include <sys\timeb.h>  // for ftime
#include <io.h>         // for open, lseek, close, etc
#endif

#include <time.h>       // for ftime


#include "srdat.h"      // for SR DAT file format
#include "suds.h"       // for SUDS format



#if defined( SROS_LINUX )
#define O_BINARY 0x00
#endif


//****** Prototypes ******

void DisplayUsage( void );
void OpenFiles( char *fn_dat, char *fn_sud );
void InitializeHdr( void );
void InitializeSuds( void );
void CorrectTime( void );
void AllocateArrays( void );
void WriteSudsComment( char *fn_dat, char *fn_sud );
void WriteSudsChannels( void );
void WriteSudsMuxdata( void );
void WriteSudsDescriptrace( void );
void FillAndCorrectChannel( int channel );
void CleanExit( int exit_code );




//****** Global Defines ******
                                        // Error exit codes
#define EXIT_GOOD        0
#define EXIT_BAD_CMD     1
#define EXIT_BAD_MEM     2
#define EXIT_BAD_HDR     3
#define EXIT_BAD_DATA    4
#define EXIT_BAD_SUDS    5
#define EXIT_BAD_EXIST   6
#define EXIT_BAD_DAT     7

#define MAXSTR          80              
                                        // Data range values
#define MINDATA16       -32767.0F
#define MAXDATA16       32767.0F
#define MINDATA24       -8388607.0F
#define MAXDATA24       8388607.0F
                                        // 24->32 bit sign extension values
#define NEG24BIT        0x00800000
#define SIGNEXTEND      0xFF000000

#define USECPER         1000000

#define REVSTRING       "Rev {09/30/03}"

typedef struct {
        char *name;
        int  *variable;
        int  value;
        
        } CMDSET;
        

// Correct C++ v7.0 time to 1970 (in GMT)

#ifdef SROS_LINUX

#define COMPILER_ADJUST_TIME 0L

#else

#if ( _MSC_VER == 700 )

#define COMPILER_ADJUST_TIME 2209075200L

#else

#define COMPILER_ADJUST_TIME 0L

#endif // _MSC_VER

#endif // SROS_LINUX




//****** Global Data ******
                                // Set in main
int    truncate     = 0;
int    outputmux    = 0;
                                // Set in OpenFiles
int    fsud         = 0;
int    fdat         = 0;
double begintime;
long   datfilesize;
                                // Set in InitializeHdr
long   nblock       = 0;
long   nblkbytes    = 0;
long   nchanbytes   = 0;
                                // Set in InitializeSuds
SUDS   sudsmux;
SUDS   sudstrace;
SUDS   sudschannel;
SUDS   sudscomment;

long   OutWordSize  = 0;
long   UtcToLocal   = 0;

SRDAT_HDR Hdr;
SRDAT_TAG Tag;

struct comment_info {

           char prg[MAXSTR];
           char hdr[MAXSTR];
           char scp[MAXSTR];
           char sud[MAXSTR];
           char lin[MAXSTR];
           char cnv[MAXSTR];
           
           char hdrfile[_MAX_PATH];
           char scpfile[_MAX_PATH];
           char sudfile[_MAX_PATH];
           char cmdline[_MAX_PATH];

           char *timeptr;

        } INFO = {
        
                "XXX2SUD - Rev {xx/xx/xx}",
                "    .dat header input file: ",
                "      .dat data input file: ",
                "  Suds demuxed output file: ",
                "    Converter command line: ",
                "Converted @        ",
                0,
                0,
                0,
                0,

                };

char comment[ sizeof(INFO) + 34 ];    // 34 is max size of list_mstime string           

char *UsageMsg[] = {

   "DAT2SUD : Copyright (c) Symmetric Research, 2002 : %s\n",
   "                                                                       \n",
   "  Usage: DAT2SUD [switches] datfile sudfile [switches]                 \n",
   "                                                                       \n",
   "    datfile = binary .dat data file        (input)                     \n",
   "    sudfile = demuxed SUDS data file       (output)                    \n",
   "                                                                       \n",
   NULL
   };

char *UsageNote[] = {

    "  Notes:                                                            \n",
    "    This program converts binary .dat data files from simp%3s,      \n",
    "    scope%3s, and similar programs into muxed or demuxed            \n",
    "    SUDS files.                                                     \n",
    "                                                                    \n",
    NULL
    };

        






//****** A/D Specific Data ******

// ------------------------- SRA2D_PARANY -------------------------

#if defined( SRA2D_PARANY )
#include "parxch.h"

#define APPNAME         "XCH"
#define DATAINT         long
#define WORDSIZE        4
#define ATODCHANNELS    (PARXCH_MAX_ANALOG_CHANNELS  + \
                         PARXCH_MAX_DIGITAL_CHANNELS + \
                         PARXCH_MAX_GPS_CHANNELS     + \
                         PARXCH_MAX_COUNTER_CHANNELS)
#define ZEROFFSET       0x7FFFFF

CMDSET CmdWords[] = {
        
        "/m",   &outputmux,     1,
        "/M",   &outputmux,     1,
        "/t",   &truncate,      1,
        "/T",   &truncate,      1,
        NULL,   NULL,           0
        };
        
char *UsageSwitch[] = {
        
    "  Switches:                                                         \n",
    "    /M   means output muxdata records not descriptrace records      \n",
    "    /T   means truncate data from 24 to 16 bits                     \n",
    "                                                                    \n",
    NULL
    };


        
// ------------------------- ATOD_IS_PAR24B -------------------------

#elif defined( ATOD_IS_PAR24B )

#define APPNAME         "24B"
#define DATAINT         long
#define WORDSIZE        4
#define ATODCHANNELS    8
#define ZEROFFSET       0x7FFFFF

CMDSET CmdWords[] = {
        
        "/m",   &outputmux,     1,
        "/M",   &outputmux,     1,
        "/t",   &truncate,      1,
        "/T",   &truncate,      1,
        NULL,   NULL,           0
        };
        
char *UsageSwitch[] = {
        
    "  Switches:                                                         \n",
    "    /M   means output muxdata records not descriptrace records      \n",
    "    /T   means truncate data from 24 to 16 bits                     \n",
    "                                                                    \n",
    NULL
    };


        
// ------------------------- ATOD_IS_DSPA64 -------------------------
    
#elif defined ( ATOD_IS_DSPA64 )

#define APPHDRID        "SR DSPA64"
#define DATAINT         short
#define WORDSIZE        2
#define ATODCHANNELS    64
#define ZEROFFSET       0x7FFF

CMDSET CmdWords[] = {  

        "/m",   &outputmux,     1,
        "/M",   &outputmux,     1,
        NULL,   NULL,           0
        };

char *UsageSwitch[] = {

    "  Switches:                                                         \n",
    "    /M   means output muxdata records not descriptrace records      \n",
    "                                                                    \n",
    NULL
    };


// ------------------------- ATOD_IS_?????? -------------------------
                
#else

#error Invalid A/D option defined.
#error Use SRA2D_PARANY, ATOD_IS_PAR24B, or ATOD_IS_DSPA64.

#define APPNAME         "???"
#define DATAINT         short
#define WORDSIZE        2
#define ATODCHANNELS    16
#define ZEROFFSET       0x7FFF

CMDSET CmdWords[]   = { NULL, NULL, 0 };
char *UsageSwitch[] = { NULL };

#endif





//****** A/D Dependent Data ******

                                // Set in AllocateArrays
DATAINT *data_array  = 0;
DATAINT *demuxed     = 0;







//****** OUT2SUD Program ******


#ifdef SROS_LINUX

char* _fullpath( char *abspath, char *relpath, int maxpath ) {

//        getcwd( abspath, maxpath );

        // For now, just pass on the relative path

        strncpy( abspath, relpath, maxpath );

        return( abspath );
}
        
#endif // SROS_LINUX

int main( int argc, char **argv ) {

        int    i;
        int    isopt, idatfile, isudfile;
        CMDSET *cmd;




        // Check command line and display usage message if needed.

        if (argc <= 2)
        
                DisplayUsage();


        truncate  = 0;
        outputmux = 0;
        idatfile  = 0;
        isudfile  = 0;

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for valid switches
                
                cmd   = CmdWords;
                isopt = 0;

                while ( (cmd->name != NULL) ) {
                        if ( !strncmp(argv[i],cmd->name,strlen(cmd->name)) ) {
                                *cmd->variable = cmd->value;
                                isopt = 1;
                                break;
                                }
                        cmd++;
                        }
                        

                // Non-switch options must be file names

                if (!isopt) {

                        if (idatfile == 0)
                                idatfile = i;
                        else if (isudfile == 0)
                                isudfile = i;
                        else
                                DisplayUsage();
                        }
                }



                                


        // Do initializations.

        OpenFiles( argv[idatfile], argv[isudfile] );

        InitializeHdr( );

        InitializeSuds( );

        AllocateArrays();



        
        // Output suds information
        
        WriteSudsComment( argv[idatfile], argv[isudfile] );

        WriteSudsChannels();

        if (outputmux)
                WriteSudsMuxdata();
        else
                WriteSudsDescriptrace();




        CleanExit( EXIT_GOOD );

        return( 0 );
}


void DisplayUsage( void ) {

        int i;

        for ( i = 0 ; UsageMsg[i] ; i++ )
                printf( UsageMsg[i], REVSTRING );

        for ( i = 0 ; UsageSwitch[i] ; i++ )
                printf( UsageSwitch[i] );

        for ( i = 0 ; UsageNote[i] ; i++ )
                printf( UsageNote[i], APPNAME );


        CleanExit( EXIT_BAD_CMD );

}


void OpenFiles( char *fn_dat, char *fn_sud ) {


        // Open input binary dat file.

        fdat = open( fn_dat, (O_BINARY|O_RDONLY), (S_IREAD|S_IWRITE) );
        if (fdat == -1) {
                printf("Can not open input file %s\n", fn_dat);
                CleanExit( EXIT_BAD_DAT );
                }

        
        // Open output SUDS file.

        if ( !(fsud = suds_open( fn_sud, SUDS_CREATE )) ) {
                printf( "Could not create output suds file %s.\n", fn_sud );
                CleanExit( EXIT_BAD_SUDS );
                }


        return;
}

void InitializeHdr( void ) {


        // Read header contents.

        if ( !SrDatHdrRead( fdat, &Hdr ) ) {
                printf( "Could not read .DAT header\n" );
                CleanExit( EXIT_BAD_DAT );
                }
        if (Hdr.DatRev != SRDAT_REV) {
                printf( "WARNING: Header version mismatch " );
                printf( "(%d in file, %d in program)\n", Hdr.DatRev, SRDAT_REV );
                }


        // Do some error checking.

        if ( Hdr.DatId != SRDAT_HDR_ID ) {

                printf( "Error in header tag 0x%lX. (Id = %s).\n",
                          Hdr.DatId, Hdr.DatProduct );
                printf( "DAT2SUD only works with headers from SR products.\n");
                CleanExit( EXIT_BAD_HDR );

                }


        if ( Hdr.AtodWordSize != WORDSIZE ) {

                printf( "Error in header WordSize (%d).\n", Hdr.AtodWordSize );
                printf( "DAT2SUD only works with %d byte data.\n", WORDSIZE );
                CleanExit( EXIT_BAD_HDR );

                }

        if ( Hdr.ChannelPtsPerFile <= 0 ) {

                printf( "Error in header RecordPts (%ld).\n", Hdr.ChannelPtsPerFile );
                printf( "DAT2SUD only works where each channel has " );
                printf( "more than 0 data points.\n");
                CleanExit( EXIT_BAD_HDR );

                }

        if ( Hdr.AtodSamplingRate <= 0.000001 ) {

                printf( "Error in header SampleRate (%f).\n", Hdr.AtodSamplingRate );
                printf( "DAT2SUD only works where the sample rate is " );
                printf( "more than 0.000001 .\n");
                CleanExit( EXIT_BAD_HDR );

                }

        if ( Hdr.Channels == 0 ) {

                printf( "Error in header Channels (%d).\n", Hdr.Channels );
                printf( "DAT2SUD only works with 1 or more channels.\n" );
                CleanExit( EXIT_BAD_HDR );

                }



        // Prepare helper variables
        //
        // All data is read in from the .DAT file in records.
        // In each record, there are Hdr.OneRecordPts per channel.
        //
        // Variable                 Is number of ...
        // --------                 ----------------
        // Hdr.ChannelPtsPerFile    pts/channel in DAT file
        // Hdr.ChannelPtsPerRecord  pts/channel in a data block
        // nchanbytes               bytes/channel in a data block
        // nblkbytes                bytes in a data block
        // nblock                   data blocks in DAT file


        nchanbytes = Hdr.ChannelPtsPerRecord * Hdr.AtodWordSize;
        nblkbytes  = Hdr.ChannelPtsPerRecord * Hdr.AtodWordSize * Hdr.Channels;
        nblock     = Hdr.ChannelPtsPerFile / Hdr.ChannelPtsPerRecord;


        return;
}


void InitializeSuds( void ) {

        char          component, data_type, *netname;
        float         minval, maxval;
        short         timezone;

#ifdef SROS_LINUX
        struct tm *tmbuffer;
        time_t dummy;
#else        
        struct _timeb timebuffer;
#endif



        // Channel identification must match between SUDS structures.

        netname   = APPNAME;
        component = 'V';                        // vertical


        if ((WORDSIZE == 4) && (truncate == 0)) {
                data_type   = '2';              // 24 bit signed stored as long
                minval      = MINDATA24;
                maxval      = MAXDATA24;
                OutWordSize = 4;
                }
                
        else {  // WORDSIZE == 2 || truncate == 1
                data_type   = 'i';              // 16 bit signed
                minval      = MINDATA16;
                maxval      = MAXDATA16;
                OutWordSize = 2;
                }



        // Get timezone information

#ifdef SROS_LINUX
        dummy      = 0; // ignore daylight savings to be like Windows
        tmbuffer   = localtime( &dummy );
        timezone   = (-tmbuffer->tm_gmtoff)/60;  // from sec east to min west
        UtcToLocal = -tmbuffer->tm_gmtoff;
#else        
        _ftime( &timebuffer );
        timezone   = timebuffer.timezone;
        UtcToLocal = 60L * timebuffer.timezone;
#endif



        // Correct begintime

        CorrectTime();



        // SUDS Station structure defaults.

        sudschannel.type = STATIONCOMP;
        suds_init( &sudschannel );

        sprintf( sudschannel.un.sc.sc_name.network, "%-.4s", netname );
        sudschannel.un.sc.sc_name.component = component;

        sudschannel.un.sc.data_type  = data_type;
        sudschannel.un.sc.data_units = 'd';                // digital counts
        sudschannel.un.sc.polarity   = 'n';                // normal polarity
        sudschannel.un.sc.atod_gain  = (short)Hdr.AtodGain;




        // SUDS MuxData structure defaults.

        sudsmux.type = MUXDATA;
        suds_init( &sudsmux );

        sprintf( sudsmux.un.mu.netname, "%-.4s", netname );
        
        sudsmux.un.mu.begintime = begintime;
        sudsmux.un.mu.loctime   = timezone;
        sudsmux.un.mu.numchans  = Hdr.Channels;
        sudsmux.un.mu.dig_rate  = (float)Hdr.AtodSamplingRate;
        sudsmux.un.mu.typedata  = data_type;
        sudsmux.un.mu.descript  = 'g';                // good trace
        sudsmux.un.mu.blocksize = Hdr.ChannelPtsPerRecord;
        sudsmux.un.mu.numsamps  = nblock;             // num sample "sweeps"

        sudsmux.data_len = Hdr.ChannelPtsPerRecord * Hdr.Channels * OutWordSize;
                                  
                                  



        


        // SUDS Demuxed trace structure defaults.

        sudstrace.type = DESCRIPTRACE;
        suds_init( &sudstrace );

        sprintf( sudstrace.un.dt.dt_name.network, "%-.4s", netname );
        sudstrace.un.dt.dt_name.component = component;

        sudstrace.un.dt.begintime  = begintime;
        sudstrace.un.dt.localtime  = timezone;
        sudstrace.un.dt.descriptor = 'g';                  // good trace
        sudstrace.un.dt.datatype   = data_type;
        sudstrace.un.dt.length     = Hdr.ChannelPtsPerFile;
        sudstrace.un.dt.rate       = (float)Hdr.AtodSamplingRate;
        sudstrace.un.dt.mindata    = minval;
        sudstrace.un.dt.maxdata    = maxval;

        sudstrace.data_len = Hdr.ChannelPtsPerFile * OutWordSize;




        // SUDS Comment structure defaults.

        sudscomment.type = COMMENT;
        suds_init( &sudscomment );


        return;
}

void CorrectTime( void ) {

        int         i, ipt;
        time_t      sec_time;
        struct tm   utc_time;
        struct stat filestat;

        
        // Perform standard time correction by one of several methods.

        // Get start time from .DAT header

        ipt = -1;
        for ( i = 0 ; i < SRDAT_DIM_TIME ; i++ ) {
                if (Hdr.TimePoint[i].Type == SRDAT_TIME_TYPE_FIRSTPT) {
                        ipt = i;
                        break;
                        }
                }
                        
        if ( (ipt != -1)                                            &&
             (Hdr.TimePoint[ipt].Year        != SRDAT_INVALID_LONG) &&
             (Hdr.TimePoint[ipt].Month       != SRDAT_INVALID_LONG) &&
             (Hdr.TimePoint[ipt].Day         != SRDAT_INVALID_LONG) &&
             (Hdr.TimePoint[ipt].Hour        != SRDAT_INVALID_LONG) &&
             (Hdr.TimePoint[ipt].Minute      != SRDAT_INVALID_LONG) &&
             (Hdr.TimePoint[ipt].Second      != SRDAT_INVALID_LONG) &&
             (Hdr.TimePoint[ipt].Microsecond != SRDAT_INVALID_LONG) ) {

                utc_time.tm_year  = Hdr.TimePoint[ipt].Year - 1900;
                utc_time.tm_mon   = Hdr.TimePoint[ipt].Month - 1;
                utc_time.tm_mday  = Hdr.TimePoint[ipt].Day;
                utc_time.tm_hour  = Hdr.TimePoint[ipt].Hour;
                utc_time.tm_min   = Hdr.TimePoint[ipt].Minute;
                utc_time.tm_sec   = Hdr.TimePoint[ipt].Second;
                utc_time.tm_isdst = 0;

                sec_time = mktime( &utc_time );
                
                begintime = (double)sec_time +
                            (double)Hdr.TimePoint[ipt].Microsecond / USECPER;

                // Now correct begintime seconds from UTC back to local
                // since the mktime utility expected local time as input

                begintime -= UtcToLocal;
                }


        
        // Use file create time from .DAT header
        
        else if (Hdr.FileCreateTime != SRDAT_INVALID_LONG) {
                begintime = (double)Hdr.FileCreateTime;
                }
        
        
        // Get the modify time of the file from the file stats.  Then
        // update it to contain the time of the FIRST sample instead of
        // the LAST.

        else {
                if (fstat( fdat, &filestat ) == -1) {
                        printf( "Can't get status info on .DAT file.\n" );
                        CleanExit( EXIT_BAD_DATA );
                        }
        
                filestat.st_mtime -= COMPILER_ADJUST_TIME;

                begintime = (double)( filestat.st_mtime ); // really end time
                begintime -= (Hdr.ChannelPtsPerFile / Hdr.AtodSamplingRate);
                }

        return;
}

void AllocateArrays( void ) {

        unsigned int nublk, nuchan;


        // Must downcast for malloc, verify that it proceeded ok.
        
        nublk  = (unsigned int) nblkbytes;
        nuchan = (unsigned int) nchanbytes;

        if ( (nblkbytes != (long)nublk) || (nchanbytes != (long)nuchan) ) {
        
                printf( "Memory allocations denied.\n" );
                printf( "Request for data block of %ld ", nblkbytes );
                printf( "or demuxed of %ld is too large.\n", nchanbytes );
                CleanExit( EXIT_BAD_MEM );
                }


        // Allocate array for muxed data.

        data_array = (DATAINT *) malloc( nublk );

        if (!data_array) {
                printf( "Could not allocate memory for raw data.\n" );
                printf( "Requested space for %u bytes\n", nublk );
                CleanExit( EXIT_BAD_MEM );
                }


        // Allocate array for demuxed data.

        demuxed = (DATAINT *) malloc( nuchan );

        if (!demuxed) {
                printf( "Could not allocate memory for demuxed data.\n" );
                printf( "Requested space for %u bytes\n", nuchan );
                CleanExit( EXIT_BAD_MEM );
                }

                
        return;
}


void WriteSudsComment( char *fn_dat, char *fn_sud ) {

        int i;


        // Capture execution time in a string.
        
        INFO.timeptr = list_mstime( get_mstime(), 4 );


        // Correct program name and revision

        sprintf( INFO.prg, "DAT2SUD - %s", REVSTRING );

        
        
        // Get the full path names for each file
                
        _fullpath( INFO.hdrfile, fn_dat, _MAX_PATH );
        _fullpath( INFO.scpfile, fn_dat, _MAX_PATH );
        _fullpath( INFO.sudfile, fn_sud, _MAX_PATH );


        
        // Prepare converter command line info

        strncpy( INFO.cmdline, "         ", _MAX_PATH-1 );

        i = 0;
                
        if (outputmux) {INFO.cmdline[i++] = '/'; INFO.cmdline[i++] = 'm'; i++;}
        if (truncate)  {INFO.cmdline[i++] = '/'; INFO.cmdline[i++] = 't'; i++;}
                


         
         
        // Fill in comment text.

        sudscomment.un.co.length = sprintf( comment,
        
                                "%s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n",
                                
                                INFO.prg,
                                INFO.hdr, INFO.hdrfile,
                                INFO.scp, INFO.scpfile,
                                INFO.sud, INFO.sudfile,
                                INFO.lin, INFO.cmdline,
                                INFO.cnv, INFO.timeptr );

        sudscomment.data_len = sudscomment.un.co.length;

        



        // Output comment record header and data.

        if ( !suds_write( fsud, &sudscomment ) ) {
                printf( "Could not write comment header.\n" );
                CleanExit( EXIT_BAD_SUDS );
                }

        if ( !suds_write_data( fsud, comment, sudscomment.data_len ) ) {
                printf("Could not write suds data for comment.\n" );
                CleanExit( EXIT_BAD_SUDS );
                }


                
        return;
}


void WriteSudsChannels() {

        int  ichan;
        char *stname;

        
        // Output station records for all channels
        //
        // For each channel
        //   Set channel specific suds station information
        //   Output station information record

        
        for ( ichan = 0 ; ichan < Hdr.Channels ; ichan++ ) {

                sudschannel.un.sc.channel = (short)Hdr.ChannelList[ichan].Channel;
                stname                    = Hdr.ChannelList[ichan].Title;
                sprintf( sudschannel.un.sc.sc_name.st_name, "%-.4s", stname );


                if ( !suds_write( fsud, &sudschannel ) ) {
                        printf( "Could not write channel %-.4s (#%02d)\n",
                                sudschannel.un.sc.sc_name.st_name,
                                sudschannel.un.sc.channel );
                        CleanExit( EXIT_BAD_SUDS );
                        }

                }
}

                
void WriteSudsMuxdata() {

        int    ichan;
        long   chan_len;
        double blocktime;


        // Output muxdata header and data in several blocks
        //
        // Initialize variables
        // Position input file past header and read first tag
        //
        // For each analog tag (skip others)
        //   Output header part of muxdata record
        //   Check size and read in the analog data record
        //
        //   For each channel
        //     Fill data for 1 channel and correct for offset/truncation/etc
        //     Output partial data part of record
        //
        //   Update SUDS record time
        //   Read next tag

        
        blocktime = (double)Hdr.ChannelPtsPerRecord / Hdr.AtodSamplingRate;
        chan_len  = Hdr.ChannelPtsPerRecord * OutWordSize;


        lseek( fdat, 0L, SEEK_SET );
        
        if ( !SrDatHdrSkip( fdat ) ) {
                printf( "Could not skip .DAT header\n" );
                CleanExit( EXIT_BAD_DAT );
                }
        if ( !SrDatTagRead( fdat, &Tag ) ) {
                printf( "Could not read .DAT tag\n" );
                CleanExit( EXIT_BAD_DAT );
                }


        while (Tag.TagId != SRDAT_TAGID_EOF) {

            if ( (Tag.TagId == SRDAT_TAGID_ANALOG16) ||
                 (Tag.TagId == SRDAT_TAGID_ANALOG24) ) {

                if ( !suds_write( fsud, &sudsmux ) ) {
                      printf("Could not write header part of muxdata\n");
                      CleanExit( EXIT_BAD_SUDS );
                      }


                if (nblkbytes != Tag.Nbytes) {
                        printf("Analog data record not expected size\n");
                        CleanExit( EXIT_BAD_DATA );
                        }
                if ( !SrDatDataRead(fdat,data_array,(unsigned int)nblkbytes) ){
                        printf( "Could not read .DAT data\n" );
                        CleanExit( EXIT_BAD_DAT );
                        }


                        
                for ( ichan = 0 ; ichan < Hdr.Channels ; ichan++ ) {

                        FillAndCorrectChannel( ichan );

                        if ( !suds_write_data(fsud, demuxed, chan_len) ) {
                            printf("Could not write suds data for channel %d.\n",ichan);
                            CleanExit( EXIT_BAD_SUDS );
                            }

                        } // end for ichan

                        
                sudsmux.un.mu.begintime += blocktime;
                        
                } // end if analog16/24


            else
                if ( !SrDatDataSkip( fdat, (unsigned int)Tag.Nbytes ) ) {
                        printf( "Could not skip .DAT data\n" );
                        CleanExit( EXIT_BAD_DAT );
                        }



            if ( !SrDatTagRead( fdat, &Tag ) ) {
                    printf( "Could not read .DAT tag\n" );
                    CleanExit( EXIT_BAD_DAT );
                    }

            } // end of while tag != none

}

void WriteSudsDescriptrace() {

        int  ichan;
        long chan_len;
        char *stname;


        // Output descriptrace header and data
        //
        // Initialize variable
        //
        // For each channel
        //   Set channel specific station info
        //   Write header part of record
        //   Position input file past header and read first tag
        //
        //   For each analog tag (skip others)
        //     Check size and read in the analog data record
        //     Fill data for 1 channel and correct for offset/truncation/etc
        //     Write data part of record
        //     Read next tag

        
        chan_len = Hdr.ChannelPtsPerRecord * OutWordSize;
        
        for ( ichan = 0 ; ichan < Hdr.Channels ; ichan++ ) {

            stname = Hdr.ChannelList[ichan].Title;
            sprintf( sudstrace.un.dt.dt_name.st_name,   "%-.4s", stname );

                
            if ( !suds_write( fsud, &sudstrace ) ) {
                printf("Could not write header part of descriptrace ");
                printf("for channel %-.4s\n",sudstrace.un.dt.dt_name.st_name);
                CleanExit( EXIT_BAD_SUDS );
                }


            lseek( fdat, 0L, SEEK_SET );
            if ( !SrDatHdrSkip( fdat ) ) {
                    printf( "Could not skip .DAT header\n" );
                    CleanExit( EXIT_BAD_DAT );
                    }
            if ( !SrDatTagRead( fdat, &Tag ) ) {
                    printf( "Could not read .DAT tag\n" );
                    CleanExit( EXIT_BAD_DAT );
                    }

            
            while (Tag.TagId != SRDAT_TAGID_EOF) {

                if ( (Tag.TagId == SRDAT_TAGID_ANALOG16) ||
                     (Tag.TagId == SRDAT_TAGID_ANALOG24) ) {

                    if (nblkbytes != Tag.Nbytes) {
                            printf("Analog data record not expected size\n");
                            CleanExit( EXIT_BAD_DATA );
                            }
                    if ( !SrDatDataRead( fdat,
                                         data_array,
                                         (unsigned int)nblkbytes ) ) {
                            printf( "Could not read .DAT data\n" );
                            CleanExit( EXIT_BAD_DAT );
                            }

                    FillAndCorrectChannel( ichan );

                    if ( !suds_write_data(fsud, demuxed, chan_len) ) {
                            printf("Could not write suds data for channel %d.\n",ichan);
                            CleanExit( EXIT_BAD_SUDS );
                            }
                                
                    } // end if analog16/24

                else
                    if ( !SrDatDataSkip( fdat, (unsigned int)Tag.Nbytes ) ) {
                            printf( "Could not skip .DAT data\n" );
                            CleanExit( EXIT_BAD_DAT );
                            }


                if ( !SrDatTagRead( fdat, &Tag ) ) {
                        printf( "Could not read .DAT tag\n" );
                        CleanExit( EXIT_BAD_DAT );
                        }
                        
                } // end while tag != none
                        
                        
            } // end for ichan

}


void FillAndCorrectChannel( int ichannel ) {

        int      i, startbyte;
        long     pointcount, nchpts;
        char    *bytptr;
        short   *trnptr;  
        DATAINT *datptr;
        DATAINT *dmxptr;

        
        // Collect up data for one channel from one analog data record
        // and correct it for various factors.
        //
        // *** Note that records must NOT be greater than 64k in size!


        // Transfer data for desired channel from data_array to demuxed

        datptr     = data_array + ichannel;
        dmxptr     = demuxed;
        pointcount = Hdr.ChannelPtsPerRecord;  // = pts per channel per record


        while ( pointcount-- ) {

                *dmxptr++ = *datptr;
                datptr   += Hdr.Channels;

                }


        
        
        // Otherwise apply various corrections to the analog data 

        nchpts = Hdr.ChannelPtsPerRecord;
        

// A/D specific corrections     

#if defined( ATOD_IS_PAR24B )
        
        // Correct for 24->32 bit sign extension, if data coding is signed
        // and this is not the mark channel

        if ( (Hdr.AtodDataCoding == SRDAT_DATACODING_SIGNED) &&
             (Hdr.ChannelMarkIndex != ichannel)              ){

                dmxptr = demuxed;

                for ( i = 0 ; i < nchpts ; i++ ) {
                
                        if (*dmxptr & NEG24BIT)
                                *dmxptr |= SIGNEXTEND;

                        dmxptr++;
                                                
                        }
                }
#endif
        
#if defined( ATOD_IS_DSPA64 )
        // None currently 
#endif  

#if defined( SRA2D_PARANY )
        // None currently 
#endif
        


        // Correct for offset, if data coding is offset and not signed
        // and this is not the mark channel

        if ( (Hdr.AtodDataCoding == SRDAT_DATACODING_OFFSET) &&
             (Hdr.ChannelMarkIndex != ichannel)              ){

                dmxptr = demuxed;

                for ( i = 0 ; i < nchpts ; i++ )

                        *dmxptr++ -= ZEROFFSET;

                }


        
                
        // Correct for truncation, if requested on command line
        //
        // Analog channels take middle two bytes, while the Mark channel
        // takes the low two bytes, effectively doing a left shift before 
        // the truncation occurs.

        if (truncate == 1) {

                if (Hdr.ChannelMarkIndex != ichannel)
                        startbyte = 1;
                else
                        startbyte = 0;


                trnptr = (short*)demuxed;
                bytptr = ((char*)demuxed)+startbyte;

                
                for ( i = 0 ; i < nchpts ; i++ ) {

                        *trnptr++ = *((short*)bytptr);
                        bytptr   += 4;
                        
                        }
                }

}


void CleanExit( int exit_code ) {

        if (data_array)  free(       data_array );
        if (demuxed)     free(       demuxed );
        if (fdat)        close(      fdat );
        if (fsud)        suds_close( fsud );


        exit( exit_code );
}
