// FILE: out2sud.c               Copyright (c), Symmetric Research, 2003
//
// This program converts binary .out data file from programs like
// Symmetric Research simpxxx and scopexxx into a SUDS waveform format
// file.   Where xxx represents a64, 24b, or "" depending on whether
// this file is compiled for the SR DSPA64, PAR24B, or PARxCH A/D board.
//
// It uses the Small Systems Support SUDS support library.  When using
// a 16 bit compiler, it *MUST* be compiled large model because the
// default SUDS library support is large model.  When using a 32 bit
// compiler, it *MUST* be compiled with 2 byte packing for consistency
// with existing SUDS files and to ensure correct time fields.
//
//
//

#include <stdio.h>      // for printf, fseek, etc
#include <stdlib.h>     // for exit, _fullpath
#include <malloc.h>     // for malloc, free
#include <string.h>     // for strcpy, strcmp

#ifdef SROS_LINUX
#include <sys/types.h>  // for file status
#include <sys/stat.h>   // for file status
#include <sys/time.h>   // for gettimeofday
#else
#include <sys\types.h>  // for file status
#include <sys\stat.h>   // for file status
#include <sys\timeb.h>  // for ftime
#endif

#include <time.h>       // for ftime

#include "suds.h"




//****** Prototypes ******

void display_usage( void );
void open_files( char *fn_hdr, char *fn_scp, char *fn_sud );
void initialize_hdr( void );
void initialize_suds( void );
void correct_time( void );
int  gps_adjust_time( void );
void allocate_arrays( void );
void write_suds_comment( char *fn_hdr, char *fn_scp, char *fn_sud );
void write_suds_channels( void );
void write_suds_muxdata( void );
void write_suds_descriptrace( void );
void read_channel_block( int channel, int block );
void correct_demuxed( long nchpts, long ndfpts );
void clean_exit( int exit_code );




//****** Global Defines ******
                                        // Error exit codes
#define EXIT_GOOD        0
#define EXIT_BAD_CMD     1
#define EXIT_BAD_MEM     2
#define EXIT_BAD_HDR     3
#define EXIT_BAD_DATA    4
#define EXIT_BAD_SUDS    5
#define EXIT_BAD_EXIST   6
                                        // Selected so buffer for 64 ch
#define MAXCHANPTS      256             // is < 64k for 16 bit compile
                                        
#define MAXSTR          80              
                                        // Data range values
#define MINDATA16       -32767.0F
#define MAXDATA16       32767.0F
#define MINDATA24       -8388607.0F
#define MAXDATA24       8388607.0F
                                        // 24->32 bit sign extension values
#define NEG24BIT        0x00800000
#define SIGNEXTEND      0xFF000000

#define GPS_CORR_BADPPS 1000
#define GPS_CORR_LIMIT  350

#define REVSTRING       "Rev {09/30/03}"

typedef struct {
        char *name;
        int  *variable;
        int  value;
        
        } CMDSET;
        

// Correct C++ v7.0 time to 1970 (in GMT)

#ifdef SROS_LINUX

#define COMPILER_ADJUST_TIME 0L

#else

#if ( _MSC_VER == 700 )

#define COMPILER_ADJUST_TIME 2209075200L

#else

#define COMPILER_ADJUST_TIME 0L

#endif // _MSC_VER

#endif // SROS_LINUX




//****** Global Data ******
                                // Set in main
int    truncate     = 0;
int    outputmux    = 0;
int    gpstime      = 0;
                                // Set in open_files
int    fsud         = 0;
FILE   *fhdr        = 0;
FILE   *fscp        = 0;
double begintime;
long   outfilesize;
                                // Set in initialize_hdr
long   nblock          = 0;
long   nchanblkpts     = 0;
long   nchanblklastpts = 0;
long   nzeropts        = 0;
long   nblkbytes       = 0;
long   nblklastbytes   = 0;
long   nmaxchanbytes   = 0;
                                // Set in initialize_suds
SUDS   sudsmux;
SUDS   sudstrace;
SUDS   sudschannel;
SUDS   sudscomment;

long   OutWordSize     = 0;

struct gps_data {

        unsigned short secs_hi;
        unsigned short secs_lo;
        unsigned short msecs;
        unsigned short corr;
        
        } GPS;

        

struct comment_info {

           char prg[MAXSTR];
           char hdr[MAXSTR];
           char scp[MAXSTR];
           char sud[MAXSTR];
           char lin[MAXSTR];
           char cnv[MAXSTR];
           
           char hdrfile[_MAX_PATH];
           char scpfile[_MAX_PATH];
           char sudfile[_MAX_PATH];
           char cmdline[_MAX_PATH];

           char *timeptr;

        } INFO = {
        
                "OUT2SUD - Rev {xx/xx/xx}",
                "    .hdr header input file: ",
                "      .out data input file: ",
                "  Suds demuxed output file: ",
                "    Converter command line: ",
                "Converted @        ",
                0,
                0,
                0,
                0,

                };

char comment[ sizeof(INFO) + 34 ];    // 34 is max size of list_mstime string           

char *usage_msg[] = {

   "OUT2SUD : Copyright (c) Symmetric Research, 1997-2003 : %s\n",
   "                                                                       \n",
   "  Usage: OUT2SUD [switches] hdrfile scpfile sudfile [switches]         \n",
   "                                                                       \n",
   "    hdrfile = binary .hdr header file      (input)                     \n",
   "    scpfile = binary .out data file        (input)                     \n",
   "    sudfile = demuxed SUDS data file       (output)                    \n",
   "                                                                       \n",
   NULL
   };

char *usage_note[] = {

    "  Notes:                                                            \n",
    "    This program converts binary .out data files from simp%3s,      \n",
    "    scope%3s, and similar programs into muxed or demuxed            \n",
    "    SUDS files.                                                     \n",
    "                                                                    \n",
    NULL
    };

        






//****** A/D Specific Data ******

// ------------------------- SRA2D_PARANY -------------------------

#if defined( SRA2D_PARANY )

#include "parxch.h"

#define APPHDRID        "SR PAR"
#define APPNAME         "XCH"
#define DATAINT         long
#define WORDSIZE        4
#define ATODCHANNELS    (PARXCH_MAX_ANALOG_CHANNELS  + \
                         PARXCH_MAX_DIGITAL_CHANNELS + \
                         PARXCH_MAX_GPS_CHANNELS     + \
                         PARXCH_MAX_COUNTER_CHANNELS)
#define ZEROFFSET       0x7FFFFF

CMDSET cmd_words[] = {
        
        "/g",   &gpstime,       1,
        "/G",   &gpstime,       1,
        "/m",   &outputmux,     1,
        "/M",   &outputmux,     1,
        "/t",   &truncate,      1,
        "/T",   &truncate,      1,
        NULL,   NULL,           0
        };
        
char *usage_switch[] = {
        
    "  Switches:                                                         \n",
    "    /G   means read GPS time from first 4 points of channel 0       \n",
    "    /M   means output muxdata records not descriptrace records      \n",
    "    /T   means truncate data from 24 to 16 bits                     \n",
    "                                                                    \n",
    NULL
    };


        
// ------------------------- ATOD_IS_PAR24B -------------------------

#elif defined( ATOD_IS_PAR24B )

#define APPHDRID        "SR PAR24B"
#define APPNAME         "24B"
#define DATAINT         long
#define WORDSIZE        4
#define ATODCHANNELS    8
#define ZEROFFSET       0x7FFFFF

CMDSET cmd_words[] = {
        
        "/g",   &gpstime,       1,
        "/G",   &gpstime,       1,
        "/m",   &outputmux,     1,
        "/M",   &outputmux,     1,
        "/t",   &truncate,      1,
        "/T",   &truncate,      1,
        NULL,   NULL,           0
        };
        
char *usage_switch[] = {
        
    "  Switches:                                                         \n",
    "    /G   means read GPS time from first 4 points of channel 0       \n",
    "    /M   means output muxdata records not descriptrace records      \n",
    "    /T   means truncate data from 24 to 16 bits                     \n",
    "                                                                    \n",
    NULL
    };


        
// ------------------------- ATOD_IS_DSPA64 -------------------------
    
#elif defined ( ATOD_IS_DSPA64 )

#define APPHDRID        "SR DSPA64"
#define APPNAME         "A64"
#define DATAINT         short
#define WORDSIZE        2
#define ATODCHANNELS    64
#define ZEROFFSET       0x7FFF

CMDSET cmd_words[] = {  

        "/g",   &gpstime,       1,
        "/G",   &gpstime,       1,
        "/m",   &outputmux,     1,
        "/M",   &outputmux,     1,
        NULL,   NULL,           0
        };

char *usage_switch[] = {

    "  Switches:                                                         \n",
    "    /G   means read GPS time from first 4 points of channel 0       \n",
    "    /M   means output muxdata records not descriptrace records      \n",
    "                                                                    \n",
    NULL
    };


// ------------------------- ATOD_IS_?????? -------------------------
                
#else

#error Invalid A/D option defined.
#error Use SRA2D_PARANY, ATOD_IS_PAR24B, or ATOD_IS_DSPA64.

#define APPHDRID        "SR ??????"
#define APPHDRID2       "SR ?????? GPS"
#define APPNAME         "???"
#define DATAINT         short
#define WORDSIZE        2
#define ATODCHANNELS    16
#define ZEROFFSET       0x7FFF

CMDSET cmd_words[]   = { NULL, NULL, 0 };
char *usage_switch[] = { NULL };

#endif





//****** A/D Dependent Data ******

                                // Set in allocate_arrays
DATAINT *data_array  = 0;
DATAINT *demuxed     = 0;


// HDR FILE DATA LAYOUT
//
// This structure and its associated defines must exactly match the
// corresponding values found in the program generating the .out data
// file.
//

#define MAXCHANNELS             ATODCHANNELS
#define MAXTITLE                12

struct HDR_DATA {

          char Id[32];

          long Channels;
          long RecordPts;
          long WordSize;
          long DataCoding;

        double SampleRate;
        double Gain;
        double FilterCoeff;
        double FilterScale;

        struct channel_data { long channel; char title[ MAXTITLE]; }

                        ChannelList[MAXCHANNELS];
        }

        HDR;








//****** OUT2SUD Program ******


#ifdef SROS_LINUX

char* _fullpath(char *abspath, char *relpath, int maxpath) {

//        getcwd( abspath, maxpath );

        // For now, just pass on the relative path

        strncpy(abspath, relpath, maxpath );

        return( abspath );
}
        
#endif // SROS_LINUX



int main( int argc, char **argv ) {

        int    i;
        int    isopt, ihdrfile, ioutfile, isudfile;
        CMDSET *cmd;




        // Check command line and display usage message if needed.

        if (argc <= 3)
        
                display_usage();


        truncate  = 0;
        outputmux = 0;
        gpstime   = 0;
        ihdrfile  = 0;
        ioutfile  = 0;
        isudfile  = 0;

        for ( i = 1 ; i < argc ; i++ ) {


                // Check for valid switches
                
                cmd   = cmd_words;
                isopt = 0;

                while ( (cmd->name != NULL) ) {
                        if ( !strncmp(argv[i],cmd->name,strlen(cmd->name)) ) {
                                *cmd->variable = cmd->value;
                                isopt = 1;
                                break;
                                }
                        cmd++;
                        }
                        

                // Non-switch options must be file names

                if (!isopt) {

                        if (ihdrfile == 0)
                                ihdrfile = i;
                        else if (ioutfile == 0)
                                ioutfile = i;
                        else if (isudfile == 0)
                                isudfile = i;
                        else
                                display_usage();
                        }
                }



                                


        // Do initializations.

        open_files( argv[ihdrfile], argv[ioutfile], argv[isudfile] );

        initialize_hdr( );

        initialize_suds( );

        allocate_arrays();



        
        // Output suds information
        
        write_suds_comment( argv[ihdrfile], argv[ioutfile], argv[isudfile] );

        write_suds_channels();

        if (outputmux)
                write_suds_muxdata();
        else
                write_suds_descriptrace();




        clean_exit( EXIT_GOOD );

        return( 0 );
}


void display_usage( void ) {

        int i;

        for ( i = 0 ; usage_msg[i] ; i++ )
                printf( usage_msg[i], REVSTRING );

        for ( i = 0 ; usage_switch[i] ; i++ )
                printf( usage_switch[i], APPNAME );

        for ( i = 0 ; usage_note[i] ; i++ )
                printf( usage_note[i], APPNAME );


        clean_exit( EXIT_BAD_CMD );

}


void open_files( char *fn_hdr, char *fn_scp, char *fn_sud ) {

        struct stat filestat;



        // Open input hdr file.

        fhdr = fopen( fn_hdr, "rb" );

        if (!fhdr) {
                printf( "Can not open hdr file %s.\n", fn_hdr );
                clean_exit( EXIT_BAD_HDR );
                }





        // Open input scope binary multiplexed data file.

        fscp = fopen( fn_scp, "rb" );

        if (!fscp) {
                printf( "Can not read data file %s.\n", fn_scp );
                clean_exit( EXIT_BAD_DATA );
                }


        // Determine time and size from file status info.

        if (stat( fn_scp, &filestat ) == -1) {
                printf( "Can't get status info on file %s.\n", fn_scp );
                clean_exit( EXIT_BAD_DATA );
                }

        
#if ( _MSC_VER == 700 )
        // C++ v7.0 used a different offset, correct to 1970 (in GMT)
        filestat.st_mtime -= 2209075200L;
#endif

        
        begintime = (double)( filestat.st_mtime ); // actually is end time

        outfilesize = filestat.st_size;


        

        // Open output SUDS file.

        if ( !(fsud = suds_open( fn_sud, SUDS_CREATE )) ) {
                printf( "Could not create output suds file %s.\n", fn_sud );
                clean_exit( EXIT_BAD_SUDS );
                }



        return;
}


void initialize_hdr( void ) {

        int  nread, nhdrsize, lenid;
        long npadpts, noutsize, newrecordpts;



        nhdrsize = sizeof( HDR );
        

        // Read header in binary mode.

        nread = fread( &HDR, 1, nhdrsize, fhdr );

        if ( fclose( fhdr ) ) {
                printf( "Unable to close scope header file.\n");
                clean_exit( EXIT_BAD_HDR );
                }
        else
                fhdr = 0;




        // Do some error checking.

        lenid = strlen( APPHDRID );
        if ( strncmp( HDR.Id, APPHDRID, lenid )  != 0 ) {

                printf( "Error in header id (%s).\n", HDR.Id );
                printf( "OUT2SUD only works with headers starting with %s.\n",
                                                                   APPHDRID );
                clean_exit( EXIT_BAD_HDR );

                }


        if ( HDR.WordSize != WORDSIZE ) {

                printf( "Error in header WordSize (%d).\n", HDR.WordSize );
                printf( "OUT2SUD only works with %d byte data.\n", WORDSIZE );
                                                        
                clean_exit( EXIT_BAD_HDR );

                }

        if ( HDR.RecordPts <= 0 ) {

                printf( "Error in header RecordPts (%ld).\n", HDR.RecordPts );
                printf( "OUT2SUD only works where each channel has " );
                printf( "more than 0 data points.\n");
                clean_exit( EXIT_BAD_HDR );

                }

        if ( HDR.SampleRate <= 0.000001 ) {

                printf( "Error in header SampleRate (%f).\n", HDR.SampleRate );
                printf( "OUTSUD only works where the sample rate is ");
                printf( "more than 0.000001 .\n");
                clean_exit( EXIT_BAD_HDR );

                }

        if ( HDR.Channels == 0 ) {

                printf( "Error in header Channels (%d).\n", HDR.Channels );
                printf( "OUT2SUD only works with 1 or more channels.\n");
                clean_exit( EXIT_BAD_HDR );

                }



        // Verify binary data file is size expected by header.

        noutsize = HDR.RecordPts * HDR.Channels * HDR.WordSize;

        if (noutsize != outfilesize) {
                printf( "Error in data file size.\n" );
                printf( "OUT2SUD expected %ld but found %ld\n",
                                                noutsize, outfilesize);

                newrecordpts = outfilesize / (HDR.Channels * HDR.WordSize);
                noutsize     = newrecordpts * HDR.Channels * HDR.WordSize;
                if (noutsize != outfilesize) {
                        printf( "File size is not a multiple of %ld channels\n",
                                                HDR.Channels);
                        clean_exit( EXIT_BAD_DATA );
                        }

                printf( "Changing number of record points from %ld to %ld ",
                                              HDR.RecordPts, newrecordpts);
                printf( "and continuing ...\n");
                HDR.RecordPts = newrecordpts;
                }

                


        // Prepare helper variables
        //
        // All data is read in from the .OUT file in blocks.
        // In each block, there are MAXCHANPTS per channel
        // except for the last block which may contain fewer.
        //
        // Since muxdata records are output in blocks of similar
        // size, the existing data may need to be padded with
        // zeros to fill up the final block.
        //
        // Variable         Is number of ...
        // --------         ----------------
        // HDR.RecordPts    pts/channel in OUT file
        // MAXCHANPTS       pts/channel in a data block
        // nblock           data blocks in OUT file
        // nchanblkpts      pts/channel in a complete data block
        // nchanblklastpts  pts/channel in the last partial data block
        // npadpts          pts/channel in OUT file rounded up to whole blocks
        // nzeropts         additional pts/channel to fill last partial block
        // nblkbytes        bytes in a complete data block
        // nblklastbytes    bytes in the last partial data block
        // nmaxchanbytes    bytes/channel in a complete data block


        npadpts = MAXCHANPTS;
        
        while (npadpts < HDR.RecordPts)
                npadpts += MAXCHANPTS;

                
        nzeropts = npadpts - HDR.RecordPts;


        nchanblkpts = HDR.RecordPts;
        if (nchanblkpts > MAXCHANPTS)
                nchanblkpts = MAXCHANPTS;

                
        nblock          = HDR.RecordPts / nchanblkpts;
        nchanblklastpts = HDR.RecordPts - (nblock*nchanblkpts);

        if (nchanblklastpts != 0)
                nblock++;
        else
                nchanblklastpts = nchanblkpts;

        
        nblkbytes     = nchanblkpts     * HDR.Channels * HDR.WordSize;
        nblklastbytes = nchanblklastpts * HDR.Channels * HDR.WordSize;
        nmaxchanbytes = MAXCHANPTS                     * HDR.WordSize;


        return;
}


void initialize_suds( void ) {

        char          component, data_type, *netname;
        float         minval, maxval;
        short         timezone;

#ifdef SROS_LINUX
        struct tm *tmbuffer;
        time_t dummy;
#else        
        struct _timeb timebuffer;
#endif



        // Channel identification must match between SUDS structures.

        netname   = APPNAME;
        component = 'V';                        // vertical


        if ((WORDSIZE == 4) && (truncate == 0)) {
                data_type   = '2';              // 24 bit signed stored as long
                minval      = MINDATA24;
                maxval      = MAXDATA24;
                OutWordSize = 4;
                }
                
        else {  // WORDSIZE == 2 || truncate == 1
                data_type   = 'i';              // 16 bit signed
                minval      = MINDATA16;
                maxval      = MAXDATA16;
                OutWordSize = 2;
                }



        // Get timezone information

#ifdef SROS_LINUX
        dummy    = 0; // ignore daylight savings to be like Windows
        tmbuffer = localtime( &dummy );
        timezone = (-tmbuffer->tm_gmtoff)/60;  // from sec east to min west
#else        
        _ftime( &timebuffer );
        timezone = timebuffer.timezone;
#endif
        


        // Correct begintime

        correct_time();



        // SUDS Station structure defaults.

        sudschannel.type = STATIONCOMP;
        suds_init( &sudschannel );

        sprintf( sudschannel.un.sc.sc_name.network, "%-.4s", netname );
        sudschannel.un.sc.sc_name.component = component;

        sudschannel.un.sc.data_type  = data_type;
        sudschannel.un.sc.data_units = 'd';                // digital counts
        sudschannel.un.sc.polarity   = 'n';                // normal polarity
        sudschannel.un.sc.atod_gain  = (short)HDR.Gain;




        // SUDS MuxData structure defaults.

        sudsmux.type = MUXDATA;
        suds_init( &sudsmux );

        sprintf( sudsmux.un.mu.netname, "%-.4s", netname );
        
        sudsmux.un.mu.begintime    = begintime;
        sudsmux.un.mu.loctime      = timezone;
        sudsmux.un.mu.numchans     = HDR.Channels;
        sudsmux.un.mu.dig_rate     = (float)HDR.SampleRate;
        sudsmux.un.mu.typedata     = data_type;
        sudsmux.un.mu.descript     = 'g';                // good trace
        sudsmux.un.mu.blocksize    = MAXCHANPTS;
        sudsmux.un.mu.numsamps     = nblock;             // num sample "sweeps"

        sudsmux.data_len        = MAXCHANPTS * HDR.Channels * OutWordSize;



        


        // SUDS Demuxed trace structure defaults.

        sudstrace.type = DESCRIPTRACE;
        suds_init( &sudstrace );

        sprintf( sudstrace.un.dt.dt_name.network, "%-.4s", netname );
        sudstrace.un.dt.dt_name.component = component;

        sudstrace.un.dt.begintime  = begintime;
        sudstrace.un.dt.localtime  = timezone;
        sudstrace.un.dt.descriptor = 'g';                  // good trace
        sudstrace.un.dt.datatype   = data_type;
        sudstrace.un.dt.length     = HDR.RecordPts;
        sudstrace.un.dt.rate       = (float)HDR.SampleRate;
        sudstrace.un.dt.mindata    = minval;
        sudstrace.un.dt.maxdata    = maxval;

        sudstrace.data_len      = HDR.RecordPts * OutWordSize;




        // SUDS Comment structure defaults.

        sudscomment.type = COMMENT;
        suds_init( &sudscomment );


        return;
}

void correct_time( void ) {

        // Perform GPS time correction if requested on the command line.
        // If this fails for any reason, perform the standard correction
        // time correction instead.
        
        if (gpstime) {

                if ( gps_adjust_time() )
                        return;
                
                else {
                        printf( "Error in GPS time correction, " );
                        printf( "using standard time correction instead.\n" );
                        }
                }
                

        // Perform standard time correction.  The modify time of the file
        // actually contains the time of the last sample, not the first.
        // So the following statement corrects back to the true begin time.

        begintime -= (HDR.RecordPts / HDR.SampleRate);

        return;
}

int gps_adjust_time( void ) {

        // Correct time using GPS data from the first 4 points of channel 0.
        
        int           i, ierr, nread, next;
        unsigned long gpsseconds, correction;
        double        dseconds;
        DATAINT       ch0array[4];
        struct tm    *goodtime;


        // Read in GPS data.
        
        ierr = 0;
        next = HDR.WordSize * (HDR.Channels-1);

        for ( i = 0 ; i < 4 ; i++ ) {

                // Get GPS data from channel 0.

                nread = fread( &ch0array[i], 1, (int)HDR.WordSize, fscp );

                if (nread != HDR.WordSize)
                        ierr = 1;

                
                // Skip remaining channels.

                ierr += fseek( fscp, next, SEEK_CUR );

                }


        // Transfer to a structure.

        GPS.secs_hi = (unsigned short)(ch0array[0]);
        GPS.secs_lo = (unsigned short)(ch0array[1]);
        GPS.msecs   = (unsigned short)(ch0array[2]);
        GPS.corr    = (unsigned short)(ch0array[3]);


        // Rewind file so all data can be read later.

        rewind( fscp );



        // Check for errors.
        // The secs_hi check is roughly for years between 1999 and 2022.

        if ( (ierr != 0)                                    ||
             (GPS.secs_hi < 14000) || (GPS.secs_hi > 25000) ||
             (GPS.msecs > 1000)                             )

                return( 0 );

        if (GPS.corr > GPS_CORR_BADPPS) {
                printf( "Warning: GPS correction of %d indicates ", GPS.corr );
                printf( "invalid PPS signal.  No correction used.\n" );
                correction = 0;
                }
        else if (GPS.corr > GPS_CORR_LIMIT) {
                printf("Warning: GPS correction of %d exceeds ", GPS.corr );
                printf("limit of %d.  Subtracting 1 second.\n",GPS_CORR_LIMIT);
                correction = 1;
                }
        else
                correction = 0;



        // Set begin time based on GPS timing info
        //
        // 1) recombine hi and low words for seconds since 1970
        // 2) convert to a time structure
        // 3) include milliseconds
        // 4) combine all date/time parts into a MS time double


        gpsseconds = (GPS.secs_hi << 16) + GPS.secs_lo - correction;
        
        goodtime   = gmtime( &gpsseconds );
        
        dseconds   = (double)(goodtime->tm_sec) + (double)GPS.msecs*0.001;

        begintime  = make_mstime( goodtime->tm_year+1900,
                                  goodtime->tm_mon+1,
                                  goodtime->tm_mday,
                                  goodtime->tm_hour,
                                  goodtime->tm_min,
                                  dseconds );
        
        return( 1 );
}

void allocate_arrays( void ) {

        unsigned int nublk, numax;


        // Must downcast for malloc, verify that it proceeded ok.
        
        nublk = (unsigned int) nblkbytes;
        numax = (unsigned int) nmaxchanbytes;

        if ( (nblkbytes != (long)nublk) || (nmaxchanbytes != (long)numax) ) {
        
                printf( "Memory allocations denied.\n" );
                printf( "Request for data block of %ld ", nblkbytes );
                printf( "or demuxed of %ld is too large.\n", nmaxchanbytes );
                clean_exit( EXIT_BAD_MEM );
                }


        // Allocate array for muxed data.

        data_array = (DATAINT *) malloc( nublk );

        if (!data_array) {
                printf( "Could not allocate memory for raw data.\n" );
                printf( "Requested space for %u bytes\n", nublk );
                clean_exit( EXIT_BAD_MEM );
                }


        // Allocate array for demuxed data.

        demuxed = (DATAINT *) malloc( numax );

        if (!demuxed) {
                printf( "Could not allocate memory for demuxed data.\n" );
                printf( "Requested space for %u bytes\n", numax );
                clean_exit( EXIT_BAD_MEM );
                }

                
        return;
}


void write_suds_comment( char *fn_hdr, char *fn_scp, char *fn_sud ) {

        int i;


        // Capture execution time in a string.
        
        INFO.timeptr = list_mstime( get_mstime(), 4 );


        // Correct program name and revision

        sprintf( INFO.prg, "OUT2SUD - %s", REVSTRING );

        
        
        // Get the full path names for each file
                
        _fullpath( INFO.hdrfile, fn_hdr, _MAX_PATH );
        _fullpath( INFO.scpfile, fn_scp, _MAX_PATH );
        _fullpath( INFO.sudfile, fn_sud, _MAX_PATH );


        
        // Prepare converter command line info

        strncpy( INFO.cmdline, "         ", _MAX_PATH-1 );

        i = 0;
                
        if (gpstime)   {INFO.cmdline[i++] = '/'; INFO.cmdline[i++] = 'g'; i++;}
        if (outputmux) {INFO.cmdline[i++] = '/'; INFO.cmdline[i++] = 'm'; i++;}
        if (truncate)  {INFO.cmdline[i++] = '/'; INFO.cmdline[i++] = 't'; i++;}
                


         
         
        // Fill in comment text.

        sudscomment.un.co.length = sprintf( comment,
        
                                "%s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n",
                                
                                INFO.prg,
                                INFO.hdr, INFO.hdrfile,
                                INFO.scp, INFO.scpfile,
                                INFO.sud, INFO.sudfile,
                                INFO.lin, INFO.cmdline,
                                INFO.cnv, INFO.timeptr );

        sudscomment.data_len = sudscomment.un.co.length;

        



        // Output comment record header and data.

        if ( !suds_write( fsud, &sudscomment ) ) {
                printf( "Could not write comment header.\n" );
                clean_exit( EXIT_BAD_SUDS );
                }

        if ( !suds_write_data( fsud, comment, sudscomment.data_len ) ) {
                printf("Could not write suds data for comment.\n" );
                clean_exit( EXIT_BAD_SUDS );
                }


                
        return;
}


void write_suds_channels() {

        int  ichan;
        char *stname;

        
        // Output station records for all channels
        //
        // For each channel
        //   Set channel specific suds station information
        //   Output station information record

        
        for ( ichan = 0 ; ichan < HDR.Channels ; ichan++ ) {

                sudschannel.un.sc.channel = (short)HDR.ChannelList[ichan].channel;
                stname                    = HDR.ChannelList[ichan].title;
                sprintf( sudschannel.un.sc.sc_name.st_name, "%-.4s", stname );


                if ( !suds_write( fsud, &sudschannel ) ) {
                        printf( "Could not write channel %-.4s (#%02d)\n",
                                sudschannel.un.sc.sc_name.st_name,
                                sudschannel.un.sc.channel );
                        clean_exit( EXIT_BAD_SUDS );
                        }

                }
}

                
void write_suds_muxdata() {

        int    ichan, iblock;
        long   seekstart, passchanpts, passzeropts, chan_len;
        double blocktime;


        seekstart = 0L;
        blocktime = (double)nchanblkpts / HDR.SampleRate;
        

        passzeropts = 0;
        passchanpts = nchanblkpts;
        chan_len    = MAXCHANPTS * OutWordSize;

        
        
        // Output muxdata header and data in several blocks
        //
        // For each block
        //   Before last pass, update read sizes
        //   Output header part of muxdata record
        //
        //   For each channel
        //     Position input file
        //     Collect data for one channel
        //     Correct data for offset/padding/truncation
        //     Output partial data part of record
        //
        //   Update file position and SUDS record time


        for ( iblock = 0 ; iblock < nblock ; iblock++ ) {
        
                if (iblock == nblock-1) {
                        passzeropts = nzeropts;
                        passchanpts = nchanblklastpts;
                        }

                if ( !suds_write( fsud, &sudsmux ) ) {
                      printf("Could not write header part of muxdata\n");
                      clean_exit( EXIT_BAD_SUDS );
                      }


                for ( ichan = 0 ; ichan < HDR.Channels ; ichan++ ) {

                        fseek( fscp, seekstart, SEEK_SET );

                        read_channel_block( ichan, iblock );

                        correct_demuxed( passchanpts, passzeropts );

                        if ( !suds_write_data(fsud, demuxed, chan_len) ) {
                          printf("Could not write suds data for channel %d.\n",
                                                                          ichan);
                          clean_exit( EXIT_BAD_SUDS );
                          }

                        }


                seekstart               += nblkbytes;
                sudsmux.un.mu.begintime += blocktime;
                
                }

}

void write_suds_descriptrace() {

        int  ichan, iblock;
        long chan_len, passchanpts;
        char *stname;



        // Output descriptrace header and data
        //
        // For each channel
        //   Set channel specific station info
        //   Write header part of record
        //   Position input file
        //
        //   For each block
        //     Before last pass, update read sizes
        //     Collect data for 1 channel
        //     Correct for offset/padding/truncation,
        //     Write data part of record
        
        for ( ichan = 0 ; ichan < HDR.Channels ; ichan++ ) {

                passchanpts = nchanblkpts;
                chan_len    = nchanblkpts * OutWordSize;
                
                stname      = HDR.ChannelList[ichan].title;
                sprintf( sudstrace.un.dt.dt_name.st_name,   "%-.4s", stname );

                
                if ( !suds_write( fsud, &sudstrace ) ) {
                    printf("Could not write header part of descriptrace ");
                    printf("for channel %-.4s\n",sudstrace.un.dt.dt_name.st_name);
                    clean_exit( EXIT_BAD_SUDS );
                    }

                
                fseek( fscp, 0L, SEEK_SET );


                for ( iblock = 0 ; iblock < nblock ; iblock++ ) {

                        if (iblock == nblock-1) {
                                passchanpts = nchanblklastpts;
                                chan_len    = nchanblklastpts * OutWordSize;
                                }
                        
                        read_channel_block( ichan, iblock );

                        correct_demuxed( passchanpts, 0L );

                        if ( !suds_write_data(fsud, demuxed, chan_len) ) {
                           printf("Could not write suds data for channel %d.\n",
                                                                   ichan);
                           clean_exit( EXIT_BAD_SUDS );
                           }

                        }
                        
                }

}


void read_channel_block( int ichannel, int iblock ) {

        unsigned int bytes_read, bytes_to_read;
        long         pointcount;
        DATAINT     *datptr;
        DATAINT     *dmxptr;

        
        // Collect up data for one channel from one block.
        // Last block may contain less data.
        //
        // Bytes_to_read contains the total bytes per block.
        // Pointcount contains the points per channel per block.
        // The values for these numbers were computed in initialize_hdr.
        //
        // *** Note that blocks must NOT be greater than 64k in size!

        if (iblock < nblock-1) {
                
                bytes_to_read = (unsigned int) nblkbytes;
                pointcount    = nchanblkpts;
                }
        else {
                bytes_to_read = (unsigned int) nblklastbytes;
                pointcount    = nchanblklastpts;
                }



        // Read in one block of data

        bytes_read = fread( data_array, 1, bytes_to_read, fscp );

        if ( bytes_read != bytes_to_read ) {
                printf( "Error reading binary data\nRead %u, requested %u\n",
                                                bytes_read, bytes_to_read);
                clean_exit( EXIT_BAD_DATA );
                }



        // Transfer data for desired channel

        datptr = data_array + ichannel;
        dmxptr = demuxed;
        
        while ( pointcount-- ) {

                *dmxptr++ = *datptr;
                datptr   += HDR.Channels;

                }
}


void correct_demuxed( long nchpts, long nzpts ) {

        int      i;
        char    *bytptr;
        short   *trnptr;  
        DATAINT *dmxptr;


// A/D specific corrections     

#if defined( ATOD_IS_PAR24B )
        
        // Correct for 24->32 bit sign extension, if data coding is signed

        if (HDR.DataCoding == 1) {

                dmxptr = demuxed;

                for ( i = 0 ; i < nchpts ; i++ ) {
                
                        if (*dmxptr & NEG24BIT)
                                *dmxptr |= SIGNEXTEND;

                        dmxptr++;
                                                
                        }
                }
#endif
        
#if defined( ATOD_IS_DSPA64 )
#endif  



        // Correct for offset, if data coding is offset and not signed

        if (HDR.DataCoding == 0) {

                dmxptr = demuxed;

                for ( i = 0 ; i < nchpts ; i++ )

                        *dmxptr++ -= ZEROFFSET;

                }


        
        // Correct mux data for padding

        if (outputmux == 1) {
                
                dmxptr = &demuxed[nchpts];
                
                for ( i = 0 ; i < nzpts ; i++ )
                        
                        *dmxptr++ = 0;

                }


                
        // Correct for truncation, if requested on command line

        if (truncate == 1) {
                
                trnptr = (short*)demuxed;
                bytptr = ((char*)demuxed)+1;

                
                for ( i = 0 ; i < nchpts+nzpts ; i++ ) {
                
                        *trnptr++ = *((short*)bytptr);
                        bytptr   += 4;
                        
                        }
                }

}


void clean_exit( int exit_code ) {

        if (data_array)  free(       data_array );
        if (demuxed)     free(       demuxed );
        if (fhdr)        fclose(     fhdr );
        if (fscp)        fclose(     fscp );
        if (fsud)        suds_close( fsud );


        exit( exit_code );
}
