/* FILE: diag.c                 Copyright (c), Symmetric Research, 2003-2005
 *
 * This diagnostic program checks for proper installation and functioning
 * of the SR PARxCH board.  It contains a collection of many different
 * tests that can be selected with command line options.  Run the program
 * without any options to get a help screen listing the availble tests.
 *
 *
 *
 */

#include <stdio.h>
#include <time.h>



// OS dependent includes ...

#if defined( SROS_WIN2K ) || defined( SROS_WINNT )
#include <windows.h>
#elif defined( SROS_WIN95 ) || defined( SROS_MSDOS )
#elif defined( SROS_LINUX )
#endif  // SROS_xxxxx



// PARxCH specific includes ...

#include "srhelper.h"    // SR helper utility functions
#include "parxch.h"      // PARxCH library functions
#include "parxchkd.h"    // PARxCH driver defines


// Os dependent prototypes from parxch.c

DEVHANDLE ParXchOsDriverOpen( char *DriverName );
int ParXchOsDriverIoctl( DEVHANDLE PassedParXchHandle,
                         unsigned long IoCtlCode,
                         void *pValueIn,
                         unsigned long InSize,
                         void *pValueOut,
                         unsigned long OutSize,
                         unsigned long *ReturnSize );
long ParXchOsGetLastError( void );




/* Defines
 *
 *
 */

#define   LED_COUNT   10
#define  BEEP_COUNT    1
#define VGOOD_COUNT   10
#define      MAXSTR  256
#define      MAXPTS   10        



/* Global variables
 *
 *
 */

char      DriverName[MAXSTR], Msg[MAXSTR];
int       XchModel, PortMode, OptBeep, NeedHelp, ready, full, ntest;
DEVHANDLE ParXchHandle;



/* DIAGNOSTIC MESSAGES:
 *
 *
 *
 *
 */




char *NOT_FOUND[] = {

    "Failed to find PARXCH at any Port address or mode ...                ",
    "                                                                     ",
    "Meanings for failure messages:                                       ",
    "                                                                     ",
    "   DEVICE DRIVER -> Can't open device driver                         ",
    "   NO RESPONSE   -> Can't read/write over parallel port              ",
    "   A/D READBACK  -> Didn't get expected response from A/D            ",
    "                                                                     ",
    " * Is a board powered up and connected to the right cable/port ?     ",
    " * Did the green port LED flash ?  If not, recheck cable and port.   ",
    " * Is a device driver with that name installed ?                     ",
    " * Is driver in use by another process ?                             ",
    " * Does the CMOS parallel port mode match those tried (PS/2 = BPP) ? ",
    " * Check Win 95 Plug and Play.  Reserve desired port address.        ",
    " * Remove/disable a competing printer driver (HP, lpd etc)           ",

    NULL,

    };




/* TEST REQUEST GLOBAL VARIABLES:

The following variables are used to turn on various tests from the
command line at run time.  Include the test request option on the
command line for any test you wish to perform.  Some of the options
such as "all" or "fifo" turn on a popular combination of other tests.

*/

int test_std        = 0;
int test_all        = 0;
int test_fifo       = 0;
int test_turnon     = 0;
int test_turnoff    = 0;
int test_port       = 0;
int test_model      = 0;
int test_default    = 0;
int test_led        = 0;
int test_acquire    = 0;
int test_fifo_empty = 0;
int test_fifo_full  = 0;
int test_fifo_slam  = 0;
int test_fifo_slide = 0;
int test_userio     = 0;
int test_vgood      = 0;

struct TestOption {
        int *testflag;
        char *name;
        char *text;
        } *t, table[] = {

   &test_all,        "all",        NULL,
   &test_fifo,       "fifo",       NULL,
   &test_std,        "std",        "Standard diag, checks port + led (REQUIRES KEYPRESS)",
   &test_port,       "port",       "Check parallel port address and mode",
   &test_model,      "model",      "Check Xch model PAR1CH, PAR4CH, or PAR8CH",
   &test_default,    "default",    "Check default Xch model",
   &test_led,        "led",        "Toggle the yellow LED",
   &test_acquire,    "acquire",    "Acquire data",
   &test_fifo_empty, "empty",      "Check FIFO empty flag",
   &test_fifo_full,  "full",       "Check FIFO full flag              (TAKES ~5 MINUTES)",
   &test_fifo_slam,  "slam",       "Run FIFO slamming ones            (TAKES ~5 MINUTES)",
   &test_userio,     "userio",     "Use digital I/O                  (REQUIRES TEST JIG)",
   &test_vgood,      "vgood",      "Check voltage good signal",
   &test_turnon,     "turnon",     "Power up (parallel port led stays on after exit)",
   &test_turnoff,    "turnoff",    "Power down",
   NULL,             NULL,
   };

#define CMD_DONE   -1
#define CMD_HELP    0
#define CMD_DRIVER  1
#define CMD_MODEL   2
#define CMD_PORT    3
#define CMD_BEEP    4
#define CMD_TESTS   5
#define CMD_MAX     6

int CmdFilled[CMD_MAX];

struct CommandLineOption {
        int   Category;
        char *Name;
        char *ParmChar;
        int  *ParmInt;
        int   Default;
        int   Value;
        } *Opt, OptionsTable[] = {

  CMD_HELP,   "/?",          NULL,       &NeedHelp, 0, 1,

  CMD_DRIVER, "SrParXch378", DriverName, NULL,      1, 0,                        
  CMD_DRIVER, "SrParXch278", DriverName, NULL,      0, 0,                        
  CMD_DRIVER, "SrParXch3BC", DriverName, NULL,      0, 0,
  
  CMD_MODEL,  "PAR1CH",      NULL,       &XchModel, 0, PARXCH_MODEL_PAR1CH,
  CMD_MODEL,  "PAR4CH",      NULL,       &XchModel, 0, PARXCH_MODEL_PAR4CH,
  CMD_MODEL,  "PAR8CH",      NULL,       &XchModel, 0, PARXCH_MODEL_PAR8CH,
  
  CMD_PORT,   "bpp",         NULL,       &PortMode, 0, PARXCH_PORT_MODE_BPP,     
  CMD_PORT,   "epp",         NULL,       &PortMode, 1, PARXCH_PORT_MODE_EPP,     
  CMD_PORT,   "ecp/bpp",     NULL,       &PortMode, 0, PARXCH_PORT_MODE_ECP_BPP, 
  CMD_PORT,   "ecp/epp",     NULL,       &PortMode, 0, PARXCH_PORT_MODE_ECP_EPP,
  
  CMD_BEEP,   "beep",        NULL,       &OptBeep,  0, 1,
  
  CMD_DONE,    NULL,         NULL,       NULL,      0, 0,                        

        };



/* COMMAND LINE PROCESSING FUNCTIONS:

These helper functions set the test options based on the values given
on the command line.

*/

void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );

}


// Usage message arrays:

char *UsageMsg[] = {
  "USAGE: diag [DriverName] [XchModel] [PortMode] [beep] test [test ...] [/?] ",
  "                                                                           ",
  "  DriverName = SrParXch378, SrParXch278, SrParXch3BC  default SrParXch378  ",
  "  XchModel   = PAR1CH, PAR4CH, PAR8CH                 default from setmodel",
  "  PortMode   = bpp, epp, ecp/bpp, ecp/epp             default epp          ",
  "  beep       = makes the program beep when done                            ",
  "  /?         = show this usage message                                     ",
  NULL
  };

char *ExampleMsg[] = {
  "EXAMPLES: diag std                                                         ",
  "          diag SrParXch378 PAR4CH epp led acquire beep                     ",
  NULL
  };

void Usage( char *msg ) {

        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );



        // Print usage message top part.

        PrintMsg( UsageMsg );


        // Wait for keypress to continue

        printf( "\nPress ENTER to see list of tests:" );
        SrGetKeyWait( );
        printf( "\n" );
        

        
        // Print out tests

        t = table;
        while ( t->name ) {
                if (t->text)
                        printf( "      %-10s %s\n", t->name, t->text );
                t++;
                }
        printf("\n");



        // Print usage message examples part.
        
        PrintMsg( ExampleMsg );
        
        exit( 1 );
}


void CountTests( void ) {

        // Count up total number of tests
        
        ntest = 0;
        t     = table;
        while ( t->name ) {
                if (*t->testflag == 1)
                        ntest++;
                t++;
                }
}
        
int ProcessOptions( char **optionname ) {
        

        // Search for matching option.

        Opt = OptionsTable;
        while ( Opt->Name && strcmp( *optionname, Opt->Name ) )
                        Opt++;


        // Argument does not match a valid option, so return to
        // calling function which will check for a valid test.

        if ( !Opt->Name )
                return( 0 );


        // Check for duplicate argument.

        if ( CmdFilled[ Opt->Category ] == 1 )
                Usage( "more than one option per category" );
        else
                 CmdFilled[ Opt->Category ] = 1;



        // Fill in parameter value for known option.

        if ( Opt->ParmInt )
                *Opt->ParmInt = Opt->Value;

        else if ( Opt->ParmChar )
                SrStringCopy( Opt->ParmChar, Opt->Name, MAXSTR );

        else
                Usage( "option of unknown type" );

        return( 1 );
}

int ProcessTests( char **testname ) {

        // Search for matching test.

        t = table;
        while ( t->name && strcmp(*testname, t->name) )
                t++;


        // Matching test option found.

        if ( t->name ) {

                // Turn its test on.

                *t->testflag = 1;


                // If it is a multiple option (all, fifo),
                // also turn on the tests for its related options.

                if (t->testflag == &test_fifo ||
                    t->testflag == &test_all ) {
                        test_fifo_empty = 1;
                        test_fifo_full  = 1;
                        test_fifo_slam  = 1;
                        test_fifo_slide = 0; // not available
                        }

                if (t->testflag == &test_all) {
                        test_turnon  = 1;
                        test_turnoff = 1;
                        test_port    = 1;
                        test_led     = 1;
                        test_acquire = 1;
                        test_userio  = 1;
                        test_vgood   = 1;
                        }

                return( 1 );
                }


        // Test not recognized.

        else {
                sprintf( Msg, "unknown test %s", *testname );
                Usage( Msg );
                }

        return( 0 );        
}

void ProcessCommandLine( int argc, char *argv[] ) {

        int i;

        // Get the command line arguments, there must be at least one.
        
        if ( argc < 2 )
                Usage( "wrong number of arguments" );


        
        // Set default options

        OptBeep  = 0;
        NeedHelp = 0;
        PortMode = PARXCH_PORT_MODE_EPP;
        XchModel = PARXCH_MODEL_UNKNOWN;
        SrStringCopy( DriverName, "SrParXch378", MAXSTR );

        for ( i = 0 ; i < CMD_MAX ; i++ )
                CmdFilled[i] = 0;
        


        // Loop through processing all arguments as either options or tests

        while ( --argc ) {

                argv++;

                // Did not match a valid option or test.

                if ( !ProcessOptions( argv ) && !ProcessTests( argv ) )
                        Usage( "invalid option or test" );

                } // end while


        // Check for help request

        if ( NeedHelp == 1 )
                Usage( "?" );



        // Finalize selection of specific PARxCH model

        if (XchModel == PARXCH_MODEL_UNKNOWN) // Not given by user on command line
                XchModel = ParXchGetDefaultModel( DriverName );

        

        // Count up total number of selected tests
        
        CountTests( );

        return;
}



/* DISPLAY HELPER FUNCTIONS:
 *
 *
 */

#define PS_NONE     0x00
#define PS_DATE     0x01
#define PS_FLUSH    0x02
#define PS_DISPLAY  0x04

#define PrintStr( msg )           PrintString( msg, PS_DISPLAY )
#define PrintStrDate( msg )       PrintString( msg, PS_DISPLAY | PS_DATE )
#define PrintStrFlush( msg )      PrintString( msg, PS_DISPLAY | PS_FLUSH )
#define PrintStrDateFlush( msg )  PrintString( msg, PS_DISPLAY | PS_DATE | PS_FLUSH )
#define PrintStrNoDisplay( msg )  PrintString( msg, PS_NONE )

void PrintString( char *msg, int mode ) {

        FILE *RptFp;
        char *datestr, dbuffer[16];
        char *timestr, tbuffer[16];
        time_t current_time;
        struct tm *local_time;


        // All output messages should be printed via this function
        // which sends the output to both a report file and the screen,
        // optionally adding a time and date string when writing to the
        // report file.



        // Get the date and time if requested.

        if ( mode & PS_DATE ) {

                time( &current_time );
                local_time = localtime( &current_time );

                strftime( dbuffer, 16, "%m/%d/%y", local_time );
                strftime( tbuffer, 16, "%H:%M:%S", local_time );

                datestr = dbuffer;
                timestr = tbuffer;
                
                }




        // Write the message to the report file, optionally including the date.

        if ( RptFp = fopen( "DIAG.RPT", "a" ) ) {
                if ( mode & PS_DATE )
                        fprintf( RptFp, "[%s] [%s] : ", datestr, timestr );
                fprintf( RptFp, "%s", msg );
                fclose( RptFp );
                }

        
        
        
        // Display the message to the screen, optionally flushing
        // the line even if the message did not end with a new line
        // character.
        
        if ( mode & PS_DISPLAY )
                printf( "%s", msg );

        if ( mode & PS_FLUSH )
                SrFlushPrint();

}

char *StrPortMode( int value ) {


        // Convert the PortMode flag into a string.

        if ( value == PARXCH_PORT_MODE_BPP )
                return( "BPP" );

        else if ( value == PARXCH_PORT_MODE_EPP )
                return( "EPP" );

        else if ( value == PARXCH_PORT_MODE_ECP_BPP )
                return( "ECP/BPP" );

        else if ( value == PARXCH_PORT_MODE_ECP_EPP )
                return( "ECP/EPP" );

        else
                return( "UNKNOWN" );

}

char *StrXchModel( int PassedXchModel ) {


        // Convert the XchModel flag into a string.

        if ( PassedXchModel == PARXCH_MODEL_PAR1CH )
                return( "PAR1CH" );

        else if ( PassedXchModel == PARXCH_MODEL_PAR4CH )
                return( "PAR4CH" );

        else if ( PassedXchModel == PARXCH_MODEL_PAR8CH )
                return( "PAR8CH" );

        else
                return( "UNKNOWN" );

}

int GetAnalogChannels( int PassedXchModel ) {

        // Get number of analog channels for the Xch model type

        if ( PassedXchModel == PARXCH_MODEL_PAR1CH )
                return( PAR1CH_ANALOG_CHANNELS );

        else if ( PassedXchModel == PARXCH_MODEL_PAR4CH )
                return( PAR4CH_ANALOG_CHANNELS );

        else if ( PassedXchModel == PARXCH_MODEL_PAR8CH)
                return( PAR8CH_ANALOG_CHANNELS );

        else
                return( 0 );
        
}






/* REV MESSAGE:
 *
 *
 *
 */

void Rev( void ) {

        int LibRev;
        
        // Print the rev info.

        ParXchGetRev( &LibRev );
        sprintf( Msg,
                 "PARxCH diag [/?]: (c) SR 2004 : Rev {01/01/04} : ParXch.lib rev %d\n",
                 LibRev );
        PrintStrDate( Msg );
}






/* DRIVER ACCESS HELPER FUNCTIONS:

These helper functions access some OS and device driver functions
that are usually off limits because they are local to the parxch.c
and parxchkd.c library and driver files.  These functions should be
used with care and only by those EXTREMELY familiar with the device
driver code.

*/

DEVHANDLE DrvxchOpen( char *PassedDriverName ) {

        unsigned long LastDriverError;

        ParXchHandle = ParXchOsDriverOpen( PassedDriverName );

        if (ParXchHandle == BAD_DEVHANDLE)
                LastDriverError = ParXchOsGetLastError();

        return( ParXchHandle );
}

int DrvxchInit( int PassedXchModel,
                int PassedPortMode,
                long *Expected_CrValue,
                long *ADS1210_CrValue,
                int *Error ) {

        unsigned long LastDriverError;
        INITXCHPARMS  Parm;

        // Set up parameters to be passed to driver.

        Parm.XchModel        = PassedXchModel;
        Parm.ParPortMode     = PassedPortMode;
        Parm.Df              = PARXCH_DF_SIGNED;
        Parm.GainLog         = PARXCH_GAIN_1;
        Parm.TurboLog        = PARXCH_TURBO_16;
        Parm.Decimation      = PARXCH_DECIMATION_100Hz_T16;
        Parm.ExtraDecimation = PARXCH_EXTRADECIMATION_MIN;
        Parm.ADS1210_CrValue = 0;
        Parm.Error           = PARXCH_ERROR_NONE;


        // Note: This value was computed by hand based on the
        // PARXCH_ settings given above and the ADS1210_ defines
        // located in the parxchkd.c driver source.

        if (Expected_CrValue)
                *Expected_CrValue = 0x83208C34;
        

        // Initialize the A/D device.

        if ( !ParXchOsDriverIoctl(
                                  ParXchHandle,
                                  IOCTL_PARXCH_INIT,
                                  &Parm,
                                  sizeof(Parm), 
                                  &Parm,
                                  sizeof(Parm),
                                  NULL
                                 ) )
              Parm.Error = PARXCH_ERROR_DRIVER_REQUEST_FAILED;


        // Save CrValue set by kernel init for optional output.
        
        if (ADS1210_CrValue)
                *ADS1210_CrValue = Parm.ADS1210_CrValue;


        // Trap IoCtrl failure and IoCtrl success but Kernel Init failure.
        // Despite error, don't close driver so further tests can be tried.
        
        if ( Parm.Error != PARXCH_ERROR_NONE )
                LastDriverError = ParXchOsGetLastError();


        // Finish.

        if ( Error )
                *Error = Parm.Error;


        if ( Parm.Error == PARXCH_ERROR_NONE )
                return( 1 );
        else
                return( 0 );
}

void DrvxchRecoverFlags( int flags ) {

        // Set the ready and full variables based on the data
        // packed into the flags variable by the driver.  The
        // two DIAG_FLAG defines are located in parxchkd.h
        // since they must be shared with the driver code.

        if (flags & DIAG_FLAG_READY)
                ready = 1;
        else
                ready = 0;

        if (flags & DIAG_FLAG_FULL)
                full = 1;
        else
                full = 0;

        return;
}

// These defines must match those in parxchkd.c
#define BDCTRL_MODE_HALT_INIT  (0x0 << 0)
#define BDCTRL_MODE_HALT_UDIO  (0x1 << 0)

void DrvxchBdCtrlMode( int Value ) {

        // Set the PARXCH board control register mode bits.

        ParXchOsDriverIoctl(
                            ParXchHandle,
                            IOCTL_PARXCH_DIAG_BDCTRL_MODE,
                            &Value,
                            sizeof(int),
                            NULL,
                            0,
                            NULL
                           );
}
int DrvxchFifoResetFlags( void ) {

        int flags, result;

        // Reset the PARXCH FIFO.  Note, this function will fail
        // unless the board control mode is set to HALT_INIT.

        result =  ParXchOsDriverIoctl(
                                      ParXchHandle,
                                      IOCTL_PARXCH_DIAG_FIFO_RESET,
                                      NULL,
                                      0,
                                      &flags,
                                      sizeof(int),
                                      NULL
                                     );

        DrvxchRecoverFlags( flags );

        return( result );
}
void DrvxchFifoReadFlags( unsigned char *Value ) {

        int flags;

        // Read one nibble from the PARXCH Fifo.

        ParXchOsDriverIoctl(
                            ParXchHandle,
                            IOCTL_PARXCH_DIAG_FIFO_READ,
                            NULL,
                            0,
                            &flags,
                            sizeof(int),
                            NULL
                           );

        *Value = (unsigned char)(flags & DIAG_FLAG_VALUE);

        DrvxchRecoverFlags( flags );
}
void DrvxchFifoWriteFlags( unsigned char Value ) {

        int flags;

        // Write one nibble to the PARXCH Fifo.

        flags = (int)(Value);

        ParXchOsDriverIoctl(
                            ParXchHandle,
                            IOCTL_PARXCH_DIAG_FIFO_WRITE,
                            &flags,
                            sizeof(int),
                            &flags,
                            sizeof(int),
                            NULL
                           );

        DrvxchRecoverFlags( flags );
}



/* Open, Init, and Close Helper Functions
 *
 *
 *
 */


DEVHANDLE OpenXch( void ) {

        // Open A/D driver

        PrintStrDate( "Opening PARxCH A/D board with:\n" );
        sprintf( Msg, "   DriverName = %s\n", DriverName );                PrintStr( Msg );
        sprintf( Msg, "   XchModel   = %s\n", StrXchModel( XchModel ) );   PrintStr( Msg );
        sprintf( Msg, "   PortMode   = %s\n\n", StrPortMode( PortMode ) ); PrintStr( Msg );

        PrintStrDate( "Performing low level open ...\n" );


        ParXchHandle = DrvxchOpen( DriverName );

        if (ParXchHandle == BAD_DEVHANDLE)
                PrintStr( " ERROR: Could not open driver\n" );
        else
                PrintStr( "\n");

        return( ParXchHandle );
}

int InitXch( void ) {

        int ErrorCode, Result;
        long ADS1210_CrValue, Expected_CrValue;

        
        // Initialize and already open PARXCH with a reasonable sampling
        // rate using the specified port address/mode.

        PrintStrDate( "Performing low level initialization ...\n" );

        ADS1210_CrValue = 0x0L;

        Result = DrvxchInit(
                            XchModel,
                            PortMode,
                            &Expected_CrValue,
                            &ADS1210_CrValue,
                            &ErrorCode
                           );

        if ( !Result || (ErrorCode != PARXCH_ERROR_NONE) ) {
                sprintf( Msg,
                         " ERROR: failure %s (Readback = 0x%lX, Expected = 0x%lX)\n",
                         PARXCH_ERROR_MSG[ErrorCode],
                         ADS1210_CrValue,
                         Expected_CrValue);
                PrintStr( Msg );
                return( 0 );
                }

        PrintStr( "\n" );
        return( 1 );
}

DEVHANDLE OpenInitXch( char *PassedDriverName,
                       int   PassedXchModel,
                       int   PassedPortMode,
                       long *ADS1210_CrValue,
                       int  *ErrorCode
                     ) {

        int  Error;
        long CrValue;

        ParXchHandle = ParXchFullOpen(
                                      PassedDriverName,
                                      PassedXchModel,
                                      PassedPortMode,

                                      PARXCH_DF_SIGNED,
                                      PARXCH_GAIN_1,
                                      PARXCH_TURBO_16,
                                      PARXCH_DECIMATION_100Hz_T16,
                                      PARXCH_EXTRADECIMATION_MIN,
                                      PARXCH_UNUSED,

                                      &CrValue,
                                      &Error
                                     );

        if (ADS1210_CrValue)
                *ADS1210_CrValue = CrValue;
        if (ErrorCode)
                *ErrorCode = Error;
        
        return( ParXchHandle );
}

void CloseXch( void ) {

        // Power down and close driver

        ParXchClose( ParXchHandle );
        ParXchHandle = BAD_DEVHANDLE;
}










/* TEST DATA:
 *
 *
 *
 *
 */

#define SUCCESS         1
#define FAILURE         0


struct test_port_info

   { char *DriverName; int PortMode; int Status; }

        TestPortInfo[] = {

                PARXCH_0x378, PARXCH_PORT_MODE_EPP,     FAILURE,
                PARXCH_0x378, PARXCH_PORT_MODE_BPP,     FAILURE,
                PARXCH_0x378, PARXCH_PORT_MODE_ECP_EPP, FAILURE,
                PARXCH_0x378, PARXCH_PORT_MODE_ECP_BPP, FAILURE,

                PARXCH_0x278, PARXCH_PORT_MODE_EPP,     FAILURE,
                PARXCH_0x278, PARXCH_PORT_MODE_BPP,     FAILURE,
                PARXCH_0x278, PARXCH_PORT_MODE_ECP_EPP, FAILURE,
                PARXCH_0x278, PARXCH_PORT_MODE_ECP_BPP, FAILURE,

                0x000,

                };

struct test_model_info

   { int Id; char *Name; int Channels; }

        TestModelInfo[] = {

                PARXCH_MODEL_PAR8CH,  PAR8CH_MODELNAME, PAR8CH_ANALOG_CHANNELS,
                PARXCH_MODEL_PAR4CH,  PAR4CH_MODELNAME, PAR4CH_ANALOG_CHANNELS,
                PARXCH_MODEL_PAR1CH,  PAR1CH_MODELNAME, PAR1CH_ANALOG_CHANNELS,
                PARXCH_MODEL_UNKNOWN, NULL,             PARXCH_MAX_ANALOG_CHANNELS,
        };


struct test_default_info

   { char *DriverName; }

        TestDefaultInfo[] = {

                PARXCH_0x378,
                PARXCH_0x278,
                0x000,

                };



/* TEST FUNCTIONS:
 *
 *
 *
 *
 *
 *
 *
 */

int TestPort( void ) {

        int i, ibest, ErrorCode;
        long ADS1210_CrValue;



        // Test for the PARXCH at various port addresses and modes.
        // This is done by writing and reading to the various addresses.
        // (assume PAR1CH since it has the least rigorous FIFO check)



        ibest = -1;

        PrintStr( "          Driver      Port Mode    Result\n" );
        PrintStr( "          ------      ---------    ------\n" );

        for ( i = 0 ; TestPortInfo[i].DriverName ; i++ ) {


                // Display the Port being tested ...

                sprintf( Msg, "  Testing %s  %7s ... ",
                                                TestPortInfo[i].DriverName,
                                   StrPortMode( TestPortInfo[i].PortMode ) );
                PrintStrFlush( Msg );



                // Do a presence test.

                ParXchHandle = OpenInitXch(
                                           TestPortInfo[i].DriverName,
                                           PARXCH_MODEL_PAR1CH,
                                           TestPortInfo[i].PortMode,

                                           &ADS1210_CrValue,
                                           &ErrorCode
                                          );

                if ( ParXchHandle == BAD_DEVHANDLE ) {

                        TestPortInfo[i].Status = FAILURE;
                        sprintf( Msg,
                                 "failure %s\n",
                                 PARXCH_ERROR_MSG[ErrorCode]
                               );
                        PrintStr( Msg );
                        }

                else {
                        TestPortInfo[i].Status = SUCCESS;
                        PrintStr( "SUCCESS\n" );
                        CloseXch( );
                        }





                // Save the index of the first successful scheme.

                if ( TestPortInfo[i].Status == SUCCESS && ibest == -1)

                        ibest = i;

                } // end for i


        printf( "\n" );



        // If no (address,mode) succeeded, give help.

        if ( ibest == -1 ) {
                PrintStrDate( "\n" );
                PrintMsg( NOT_FOUND );
                CloseXch( );
                }


        return( ibest );
}

int TestModel( char *TestDriverName, int TestPortMode ) {

        int  i, ibest, npts, ErrorCode;
        long ADS1210_CrValue, Values[PARXCH_MAX_ANALOG_CHANNELS*MAXPTS];


        // Test for the specific A/D model

        ibest = -1;

        PrintStr( "              Model      Result\n" );
        PrintStr( "              -----      ------\n" );

        for ( i = 0 ; TestModelInfo[i].Name ; i++ ) {
                
                sprintf( Msg, "  Testing for %s ... ", TestModelInfo[i].Name );
                PrintStrFlush( Msg );


                // Open device for full FIFO test

                ParXchHandle = OpenInitXch(
                                           TestDriverName,
                                           TestModelInfo[i].Id,
                                           TestPortMode,

                                           &ADS1210_CrValue,
                                           &ErrorCode
                                          );
                
                if ( ParXchHandle == BAD_DEVHANDLE )
                        PrintStr( "failure (Fifo Check)\n" );

                else {
                        // Try to acquire some samples ...
                        // This will test the sanity bits and interval
                        // which differ from product to product
                        
                        
                        // Start and wait for some samples to arrive.
                        
                        ParXchStart( ParXchHandle );
                        
                        SrWait( 300 ); // ms

                        
                        if ( !ParXchReady( ParXchHandle ) )
                                PrintStr( "failure (Not Ready)\n" );
                        else { 
                                
                                // Read them.

                                npts = ParXchReadData( ParXchHandle,
                                                       Values,
                                                       TestModelInfo[i].Channels*MAXPTS,
                                                       &ErrorCode
                                                     );


                                if  (ErrorCode != PARXCH_ERROR_NONE) 
                                        PrintStr( "failure (Bad Read)\n" );
                                
                                else if (npts <= 0)
                                                PrintStr( "failure (No Points)\n" );
                                
                                else if ( TestModelInfo[i].Id == PARXCH_MODEL_PAR4CH &&
                                          Values[1] == 0xFFFFFFFF  &&
                                          Values[2] == 0xFFFFFFFF  &&
                                          Values[3] == 0xFFFFFFFF
                                          )
                                        PrintStr( "failure (Bad Channels)\n" );

                                else {
                                        if (ibest == -1)
                                                ibest = i;
                                        PrintStr( "SUCCESS\n" );
                                        }

                                } // end else ready





                        // Stop and close device

                        ParXchStop( ParXchHandle );
                        CloseXch( );
                        }

                } // end for i TestModel

        PrintStr( "\n" );

        if (ibest == -1) {
                ibest = sizeof(TestModelInfo) / sizeof( struct test_model_info) - 1;
                PrintStr( "  Device not recognized\n\n" );
                }
        else {
                sprintf( Msg, "  Xch model appears to be a %s\n\n",
                         TestModelInfo[ibest].Name );
                PrintStr( Msg );
                }


        return( ibest );
}

int TestDefault( void ) {

        int i, DefaultXchModel;

        PrintStr( "          Driver          Default\n" );
        PrintStr( "          ------          -------\n" );

        for ( i = 0 ; TestDefaultInfo[i].DriverName ; i++ ) {

                sprintf( Msg, "  Testing %s ... ", TestDefaultInfo[i].DriverName );
                PrintStrFlush( Msg );
                
                DefaultXchModel = ParXchGetDefaultModel( TestDefaultInfo[i].DriverName );

                sprintf( Msg, "%s\n", StrXchModel( DefaultXchModel ) );
                PrintStrFlush( Msg );
                }
                
        return( 1 );
}

int TestTurnon( void ) {

        int PowerOff, dummy;

        // Turn on the parallel port

        dummy = 0;
        ParXchOsDriverIoctl(
                            ParXchHandle,
                            IOCTL_PARXCH_DIAG_TURN_ON,
                            &dummy,
                            sizeof(int),
                            NULL,
                            0,
                            NULL
                           );



        // Set flag controlling power down behavior at close
        // so the port led will stay on.

        PowerOff = 0;
        ParXchOsDriverIoctl(
                            ParXchHandle,
                            IOCTL_PARXCH_DIAG_SET_POWER_DOWN,
                            &PowerOff,
                            sizeof(int),
                            NULL,
                            0,
                            NULL
                           );


        return( 1 );
}

int TestTurnoff( void ) {

        int PowerOff, dummy;


        // Set flag controlling power down behavior at close
        // so the port led will go off.

        PowerOff = 1;
        ParXchOsDriverIoctl(
                            ParXchHandle,
                            IOCTL_PARXCH_DIAG_SET_POWER_DOWN,
                            &PowerOff,
                            sizeof(int),
                            NULL,
                            0,
                            NULL
                           );


        // Turn off the parallel port

        dummy = 0;
        ParXchOsDriverIoctl(
                            ParXchHandle,
                            IOCTL_PARXCH_DIAG_TURN_OFF,
                            &dummy,
                            sizeof(int),
                            NULL,
                            0,
                            NULL
                           );


        return( 1 );
}

int TestVgood( void ) {

        int vg, Count, Delay, PowerGood;

        // Test the voltage good signal on the PARXCH several times
        // in a row.  If the user turns the power to the PARXCH off
        // and back on during this test, the output message should
        // indicate bad voltage while the power is off.

        Delay = 200;

        sprintf(  Msg, "  Testing voltage every %d ms\n", Delay );
        PrintStr( Msg );
        printf( "\n" );

        Count     = 0;
        PowerGood = 1;
        vg        = -1;

        while ( Count < VGOOD_COUNT ) {

                vg = ParXchVoltageGood( ParXchHandle );
                if (vg == 1)
                        PrintStr( "  Voltage Good\n" );
                else if (vg == 0)
                        PrintStr( "  Voltage Bad\n" );
                else {
                        PrintStr( "  Voltage Unknown\n" );
                        PowerGood = 0;
                        }

                vg = -1;

                Count++;

                SrWait( Delay );
                }

        printf( "\n" );

        
        return( PowerGood );
}

void TestLed( int Interactive ) {

        int LedValue, Count;

        // Test the yellow LED by toggling it on and off several
        // times in a row.  The printed message indicates whether
        // the LED is currently on (1) or off (0).
        

        LedValue = 0;
        Count    = 0;

        printf( "\n" );

        while ( Count < LED_COUNT ) {

                ParXchUserLed( ParXchHandle, LedValue );

                sprintf( Msg, "  Yellow LED = %1d", LedValue );
                PrintStrFlush( Msg );
                if (Interactive)
                        PrintStrNoDisplay( "\n" );
                else
                        PrintStr( "\n" );
                        
                    
                LedValue ^= 1;


                if (Interactive) {
                        if ( SrGetKeyWait( ) == 'q' )
                                break;
                        }
                else {
                        SrWait( 100 ); // ms
                        Count++;
                        }
                }

        printf( "\n" );

        return;
}


void TestUserIO( void ) {

        int i, IoValue;


        // The User Digital I/O is tested by writing sixteen hex values
        // out the digital output lines and immediately reading those
        // values back in on the digital input lines.  In order to run
        // correctly, a simple test jig is required.  See the readme.txt
        // file for a description of how to make this jig from a TRM15E
        // board and a few wires.
        

        PrintStr( "  NOTE: A test jig is required to test the digital I/O.\n" );
        printf( "\n" );

        for ( i = 0 ; i < 16 ; i++ ) {
                sprintf( Msg, "  Testing value 0x%1X ... ", i );
                PrintStrFlush( Msg );
                ParXchUserIoWr( ParXchHandle, i );
                ParXchUserIoRd( ParXchHandle, &IoValue );
                if (IoValue == i)
                        PrintStr( "OK\n" );
                else {
                        sprintf( Msg, "BAD   read = 0x%1X\n", IoValue & 0xF );
                        PrintStr( Msg );
                        }
                }

        printf( "\n" );

        return;
}

void TestAcquire( void ) {

        int i, Npts, Count, Err, ErrorCode, Timeout, AnalogChannels;
        long Values[PARXCH_MAX_ANALOG_CHANNELS];


        // This function tests acquiring a few samples of data.  It is
        // not necessary to actually have any analog input signals
        // connected for this test.  It starts acquisition at the 100 SPS
        // rate set in OpenXch, waits for a sample to become ready, then
        // reads and displays the data.  After several samples have been
        // read, it stops acquisition.

        Err   = 0;
        Count = 0;
        Timeout = 0;
        AnalogChannels = GetAnalogChannels( XchModel );

        ParXchStart( ParXchHandle );


        // Always get two samples, but just print the last one 
        // since the first one might contain some startup effects.

        while ( Count < 2 ) {

                // Wait until the next data value is ready, but include
                // a timeout counter so the test does not wait forever
                // in case of trouble.

                while ( !ParXchReady( ParXchHandle ) ) {
                        SrWait( 100 ); // ms
                        Timeout++;

                        if (Timeout > 30) {
                                PrintStrDate( "Error: Ready flag never came true. \n" );
                                ParXchStop( ParXchHandle );
                                return;
                                }
                        }


                // Read it.

                Npts = ParXchReadData( ParXchHandle,
                                       Values,
                                       AnalogChannels,
                                       &ErrorCode
                                     );


                // Check for errors.

                if (ErrorCode != PARXCH_ERROR_NONE) {
                        sprintf( Msg,
                                 "ReadData error %s\n",
                                 PARXCH_ERROR_MSG[ErrorCode]
                               );
                        PrintStrDate( Msg );
                        Err = 1;
                        break;
                        }

                if (Npts == 0) {
                        PrintStrDate( "ReadData returned 0 points\n" );
                        Err = 1;
                        break;
                        }


                Count++;
                }

        ParXchStop( ParXchHandle );

        if (Err == 1)
                printf( "Error occurred, but showing channel values anyway ...\n" );

        for ( i = 0 ; i < AnalogChannels ; i++ ) {
                sprintf( Msg, "  Channel %d: 0x%08X\n", i, Values[i] );
                PrintStr( Msg );
                }
        
        printf( "\n" );

        return;
}





#define OS_SUCCESS   0
#define OS_FAILURE   1

#define FIFOMAX   ((long)4194303)    // 2^^22 - 1 max LSFR fifo count

#define STR_YR_YF           "YES ready, YES full"
#define STR_YR_NF           "YES ready, NOT full"
#define STR_NR_YF           "NOT ready, YES full"
#define STR_NR_NF           "NOT ready, NOT full"

#define FMT_CHECK_SLIDE     "  Sliding write then read with value 0x%X    ... "
#define FMT_CHECK_SLAM      "  Slamming write then read %5s then %5s ... "
#define FMT_CHECK_RESET     "  Reset fifo, expecting %s ... "
#define FMT_CHECK_WRITE     "  Write 0x%X,  expecting %s ... "
#define FMT_CHECK_READ      "  Read  0x%X,  expecting %s ... "
#define FMT_AT_POINT        "  At %7ld, expecting %s ... "

#define FMT_MANY_POINTS     "  %7s %7ld points.  Please wait      ... "


#define FMT_OK              "OK \n"
#define FMT_BAD             "BAD\n"
#define FMT_BAD_YES_READY   "BAD\n    Found YES ready, expected NOT ready         "
#define FMT_BAD_NOT_READY   "BAD\n    Found NOT ready, expected YES ready         "
#define FMT_BAD_YES_FULL    "BAD\n    Found YES full,  expected NOT full          "
#define FMT_BAD_NOT_FULL    "BAD\n    Found NOT full,  expected YES full          "
#define FMT_BAD_VALUE       "BAD\n    Found 0x%d,       expected 0x%d             "
#define FMT_BAD_RESET_MODE  "BAD\n    Tried to reset FIFO when PC not in control  "
#define FMT_BAD_SLAMMING    "BAD\n    At pt %7ld found %5s = 0x%X %5s = 0x%X "
#define FMT_BAD_SLIDING     "BAD\n    At pt %7ld, found 0x%X,0x%X wanted 0x%X,0x%X "

#define FMT_MSG_POINT_ZERO  "   \n    First point gets overwritten on full        "



void FifoWriteRead( long count ) {

        unsigned char value;
        unsigned long pt;

        // This helper function moves the "empty" position
        // in the FIFO ahead "count" nibbles by first writing
        // and then reading a value.  No checks of the values
        // or the flags are done.

        sprintf( Msg, FMT_MANY_POINTS, "Moving ", count );
        PrintStrFlush( Msg );

        for ( pt = 0L ; pt < (unsigned long)count ; pt++ ) {
                DrvxchFifoWriteFlags( (unsigned char)pt );
                DrvxchFifoReadFlags( &value );
                }

        PrintStr( FMT_OK );
}

void TestFifoReset( void ) {

        // Reset the FIFO so it is empty and the "empty" position
        // is located at address 1.  Check that the reset succeeds
        // and that the resulting flag settings are correct.


        // Announce test.
        
        sprintf( Msg, FMT_CHECK_RESET, STR_NR_NF );
        PrintStrFlush( Msg );

        
        // Do the reset.
        
        if ( !DrvxchFifoResetFlags( ) ) {
                PrintStr( FMT_BAD_RESET_MODE );
                exit( OS_FAILURE );
                }

        
        // Check flag settings.
        
        if ( !ready && !full )
                PrintStr( FMT_OK );
        else {
                if (ready)
                        PrintStr( FMT_BAD_YES_READY );
                if (full)
                        PrintStr( FMT_BAD_YES_FULL );
                exit( OS_FAILURE );
                }
}

void TestFifoEmpty( void ) {

        unsigned char value, input, expected;


        // Test the behavior of the FIFO empty flag by writing
        // two points and reading them back in.  After each step,
        // the state of the empty and full flags and the value of
        // any data read in are checked.  The FIFO is reset at the 
        // beginning and end of this test to put it in a known state.


        
        // Reset FIFO.

        TestFifoReset();
        

        
        // Write 0x3 as point 1.

        input = 0x3;
        DrvxchFifoWriteFlags( input );
        sprintf( Msg, FMT_CHECK_WRITE, input, STR_YR_NF );
        PrintStrFlush( Msg );
        if ( ready && !full )
                PrintStr( FMT_OK );
        else {
                if (!ready)
                        PrintStr( FMT_BAD_NOT_READY );
                if (full)
                        PrintStr( FMT_BAD_YES_FULL );
                exit( OS_FAILURE );
                }


        // Write 0x5 as point 2.

        input = 0x5;
        DrvxchFifoWriteFlags( input );
        sprintf( Msg, FMT_CHECK_WRITE, input, STR_YR_NF );
        PrintStrFlush( Msg );
        if ( ready && !full )
                PrintStr( FMT_OK );
        else {
                if (!ready)
                        PrintStr( FMT_BAD_NOT_READY );
                if (full)
                        PrintStr( FMT_BAD_YES_FULL );
                exit( OS_FAILURE );
                }


        // Read 0x3 from point 1.

        expected = 0x3;
        DrvxchFifoReadFlags( &value );
        sprintf( Msg, FMT_CHECK_READ, expected, STR_YR_NF );
        PrintStrFlush( Msg );
        if ( ready && !full && value == expected )
                PrintStr( FMT_OK );
        else {
                if (!ready)
                        PrintStr( FMT_BAD_NOT_READY );
                if (full)
                        PrintStr( FMT_BAD_YES_FULL );
                if (value != expected) {
                        sprintf( Msg, FMT_BAD_VALUE, value, expected );
                        PrintStr( Msg );
                        }
                exit( OS_FAILURE );
                }


        // Read 0x5 from point 2.

        expected = 0x5;
        DrvxchFifoReadFlags( &value );
        sprintf( Msg, FMT_CHECK_READ, expected, STR_NR_NF );
        PrintStrFlush( Msg );
        if ( !ready && !full && value == expected )
                PrintStr( FMT_OK );
        else {
                if (ready)
                        PrintStr( FMT_BAD_YES_READY );
                if (full)
                        PrintStr( FMT_BAD_YES_FULL );
                if (value != expected) {
                        sprintf( Msg, FMT_BAD_VALUE, value, expected );
                        PrintStr( Msg );
                        }
                exit( OS_FAILURE );
                }


        // Try to read a non-existing point (should get last valid value).

        expected = 0x5;
        DrvxchFifoReadFlags( &value );
        sprintf( Msg, FMT_CHECK_READ, expected, STR_NR_NF );
        PrintStrFlush( Msg );
        if ( !ready && !full && value == expected )
                PrintStr( FMT_OK );
        else {
                if (ready)
                        PrintStr( FMT_BAD_YES_READY );
                if (full)
                        PrintStr( FMT_BAD_YES_FULL );
                if (value != expected) {
                        sprintf( Msg, FMT_BAD_VALUE, value, expected );
                        PrintStr( Msg );
                        }
                exit( OS_FAILURE );
                }


        // Reset FIFO.

        TestFifoReset();
        

        printf( "\n" );
        
        return;
}

void TestFifoFull( long count ) {

        int nbad;
        unsigned char value, input, expected;
        unsigned long pt;


        // Test the behavior of the FIFO full flag.  Start by
        // resetting the FIFO and moving the "empty" postion to a
        // new address.  Next, write data until the FIFO is almost
        // full and write a few more points so it is completely full.
        // Then begin reading the data back in.  After each step, the
        // state of the empty and full flags and the value of any data
        // read in are checked.  Once the full flag has been set, it
        // should remain in effect until the FIFO is reset.


        
        // Reset FIFO

        TestFifoReset();
        

        
        // Move empty position ahead by count.

        FifoWriteRead( count );


        // Fill FIFO until almost full.

        sprintf( Msg, FMT_MANY_POINTS, "Filling", (FIFOMAX-1) );
        PrintStr( Msg );

        nbad = 0;
        for ( pt = 0 ; pt < FIFOMAX-1 ; pt++ ) {
                DrvxchFifoWriteFlags( (unsigned char)pt );
                if ( ready && !full )
                        ; // good, continue
                else if (nbad < 5) {
                        sprintf( Msg, FMT_AT_POINT, pt, STR_YR_NF );
                        PrintStr( Msg );
                        if (!ready)
                                PrintStr( FMT_BAD_NOT_READY );
                        if (full)
                                PrintStr( FMT_BAD_YES_FULL );
                        nbad++;
                        }
                else
                        exit( OS_FAILURE );
                }
                if (nbad == 0)
                        PrintStr( FMT_OK );



        // Write last point so FIFO is completely full.

        input = 0x5;
        DrvxchFifoWriteFlags( input );
        sprintf( Msg, FMT_CHECK_WRITE, input, STR_YR_YF );
        PrintStrFlush( Msg );
        if ( ready && full )
                PrintStr( FMT_OK );
        else {
                if (!ready)
                        PrintStr( FMT_BAD_NOT_READY );
                if (!full)
                        PrintStr( FMT_BAD_NOT_FULL );
                exit( OS_FAILURE );
                }


        // Write another point so FIFO is over full (should be discarded).

        input = 0x3;
        DrvxchFifoWriteFlags( input );
        sprintf( Msg, FMT_CHECK_WRITE, input, STR_YR_YF );
        PrintStrFlush( Msg );
        if ( ready && full )
                PrintStr( FMT_OK );
        else {
                if (!ready)
                        PrintStr( FMT_BAD_NOT_READY );
                if (!full)
                        PrintStr( FMT_BAD_NOT_FULL );
                exit( OS_FAILURE );
                }

        ParXchUserLed( ParXchHandle, 1 );


        // Read FIFO until almost empty (full should stay on).

        sprintf( Msg, FMT_MANY_POINTS, "Reading", (FIFOMAX-1) );
        PrintStrFlush( Msg );
        nbad = 0;
        for ( pt = 0L ; pt < FIFOMAX-1 ; pt++ ) {
                expected = (0xF & (unsigned char)pt);
                DrvxchFifoReadFlags(&value );
                if ( ready && full && value == expected )
                        ; // good, continue
                else if (pt == 0L)
                        PrintStr( FMT_MSG_POINT_ZERO );
                else if (nbad < 5) {
                        sprintf( Msg, FMT_AT_POINT, pt, STR_YR_YF );
                        PrintStr( Msg );
                        if (!ready)
                                PrintStr( FMT_BAD_NOT_READY );
                        if (!full)
                                PrintStr( FMT_BAD_NOT_FULL );
                        if (value != expected) {
                                sprintf( Msg, FMT_BAD_VALUE, value, expected );
                                PrintStr( Msg );
                                }
                        nbad++;
                        }
                else
                        exit( OS_FAILURE );
                }
                if (nbad == 0)
                        PrintStr( FMT_OK );




        // Read last point (should be last full pt, not discarded one).

        expected = 0x5;
        DrvxchFifoReadFlags(&value );
        sprintf( Msg, FMT_CHECK_READ, expected, STR_NR_YF );
        PrintStrFlush( Msg );
        if ( !ready && full && value == expected )
                PrintStr( FMT_OK );
        else {
                if (ready)
                        PrintStr( FMT_BAD_YES_READY );
                if (!full)
                        PrintStr( FMT_BAD_NOT_FULL );
                if (value != expected) {
                        sprintf( Msg, FMT_BAD_VALUE, value, expected );
                        PrintStr( Msg );
                        }
                exit( OS_FAILURE );
                }

        ParXchUserLed( ParXchHandle, 0 );



        // Write one new point (should not happen due to sticky full).

        input = 0x4;
        DrvxchFifoWriteFlags( input );
        sprintf( Msg, FMT_CHECK_WRITE, input, STR_NR_YF );
        PrintStrFlush( Msg );
        if ( !ready && full )
                PrintStr( FMT_OK );
        else {
                if (ready)
                        PrintStr( FMT_BAD_YES_READY );
                if (!full)
                        PrintStr( FMT_BAD_NOT_FULL );
                exit( OS_FAILURE );
                }

        
        // Reset FIFO.

        TestFifoReset();
        

        printf( "\n" );
        

        return;
}

void TestFifoSlam( void ) {

#define SLAMONES   0xF
#define SLAMZEROS  0x0
#define SLAMMAX    FIFOMAX-1

        int nbad;
        long count;
        unsigned char value0, value1;


        // Test all the bits in the FIFO by writing alternating
        // ones and zeros and then reading them back in.  After 
        // each pair of reads and the incoming data is checked.
        // The test is repeated using zeros first and ones
        // second, so all FIFO addresses are checked.  The FIFO
        // is reset at the beginning and end of this test to 
        // put it in a known state.

        
        // Reset FIFO.

        TestFifoReset();
        nbad = 0;

        
        

        // Announce test.
        
        sprintf( Msg, FMT_CHECK_SLAM, " ones", "zeros" );
        PrintStrFlush( Msg );

        
        // Write alternating ones, then zeros.

        for ( count = 0 ; count < SLAMMAX ; count += 2 ) {
                
                DrvxchFifoWriteFlags( SLAMONES );
                DrvxchFifoWriteFlags( SLAMZEROS );
                }
        
        
        // Read them back, checking the values.

        for ( count = 0 ; count < SLAMMAX ; count += 2 ) {
                
                DrvxchFifoReadFlags( &value1 );
                DrvxchFifoReadFlags( &value0 );
                
                if ( (value1 != SLAMONES) || (value0 != SLAMZEROS) ) {
                        sprintf( Msg, FMT_BAD_SLAMMING,
                                 count, " ones", value1, "zeros", value0 );
                        PrintStr( Msg );
                        if (nbad < 5)
                                nbad++;
                        else
                                break;
		        }
                }

        if (nbad == 0)
                PrintStr( FMT_OK );
        else
                exit( OS_FAILURE );

        

        // Previous section didn't write to the last point to
        // avoid hitting the sticky full condition.  Now test
        // "last" point with alternating ones, then zeros.

        DrvxchFifoWriteFlags( SLAMONES );
        DrvxchFifoWriteFlags( SLAMZEROS );
                
        DrvxchFifoReadFlags( &value1 );
        DrvxchFifoReadFlags( &value0 );

        if ( (value1 != SLAMONES) || (value0 != SLAMZEROS) ) {
                sprintf( Msg, FMT_BAD_SLAMMING,
                         count, " ones", value1, "zeros", value0 );
                PrintStr( Msg );
                exit( OS_FAILURE );
                }


        
        // Reset FIFO.

        TestFifoReset();

        
        // Now switch the order and write alternating zeros, then ones.

        sprintf( Msg, FMT_CHECK_SLAM, "zeros", " ones" );
        PrintStrFlush( Msg );
        
        for ( count = 0 ; count < SLAMMAX ; count += 2 ) {
                
                DrvxchFifoWriteFlags( SLAMZEROS );
                DrvxchFifoWriteFlags( SLAMONES );
                }

        
        // Read them back, checking the values.
        
        for ( count = 0 ; count < SLAMMAX ; count += 2 ) {
                
                DrvxchFifoReadFlags( &value0 );
                DrvxchFifoReadFlags( &value1 );
                
                if ( (value0 != SLAMZEROS) || (value1 != SLAMONES) ) {
                        sprintf( Msg, FMT_BAD_SLAMMING,
                                 count, "zeros", value0, " ones", value1 );
                        PrintStr( Msg );
                        if (nbad < 5)
                                nbad++;
                        else
                                break;
		        }
                }
        
        if (nbad == 0)
                PrintStr( FMT_OK );
        else
                exit( OS_FAILURE );
        

        
        // Previous section didn't write to the last point to
        // avoid hitting the sticky full condition.  Now test
        // "last" point with alternating zeros, then ones.

        DrvxchFifoWriteFlags( SLAMZEROS );
        DrvxchFifoWriteFlags( SLAMONES );
                
        DrvxchFifoReadFlags( &value0 );
        DrvxchFifoReadFlags( &value1 );

        if ( (value0 != SLAMZEROS) || (value1 != SLAMONES) ) {
                sprintf( Msg, FMT_BAD_SLAMMING,
                         count, " ones", value1, "zeros", value0 );
                PrintStr( Msg );
                exit( OS_FAILURE );
                }

        
        // Reset FIFO.

        TestFifoReset();

        printf( "\n" );
        
        return;
}

void TestFifoSlide( void ) {

#define SLIDEMAX FIFOMAX-1

        int nbad, ival, istart;
        long count;
        unsigned char value, reverse, newvalue, newreverse;



        // THIS TEST IS VERY SLOW, SO IT IS CURRENTLY NOT ENABLED.
        
        // Test all the bits in the FIFO by writing and then 
        // reading data with one bit turned on.  Cycle through
        // nibbles so that each bit is tested.  A second value
        // which is the reverse of the first is also written and
        // read to ensure that the correct data is truely being read
        // and is not just remaining on the bus or in memory.
        // The test is repeated so all FIFO addresses are checked.  
        // The FIFO is reset at the beginning and end of this test to 
        // put it in a known state.

        
        
        // Reset FIFO

        TestFifoReset();
        
        nbad = 0;


        // Run sliding ones test twice, offsetting the start between runs.

        for ( istart = 0 ; istart < 2 ; istart++ ) {
        
                // Test for values 1,2,4,8

                value = 0x1;

                for ( ival = 0 ; ival < 4 ; ival++ ) {

                        sprintf( Msg, FMT_CHECK_SLIDE, value );
                        PrintStrFlush( Msg );

                        reverse = (~value & 0xF);

                        for ( count = 0 ; count < SLIDEMAX ; count++ ) {

                                DrvxchFifoWriteFlags( value );
                                DrvxchFifoWriteFlags( reverse );
                                DrvxchFifoReadFlags( &newvalue );
                                DrvxchFifoReadFlags( &newreverse );


                                if ( (newvalue != value) || (newreverse != reverse) ) {
                                        sprintf( Msg, FMT_BAD_SLIDING, count, 
                                                 newvalue, newreverse, value, reverse );
                                        PrintStr( Msg );
                                        if (nbad < 5)
                                                nbad++;
                                        else
                                                break;
                                        }

                                }

                                if (nbad == 0)
                                        PrintStr( FMT_OK );
                                else
                                        exit( OS_FAILURE );

                                value = value * 2;
                        }

                
                // Reset and offset start position by 1.

                if (istart == 0) {
                        TestFifoReset();
                        FifoWriteRead( 1 );
                        }

                }

        
        // Reset FIFO.

        TestFifoReset();

        printf( "\n" );
        
        return;
}


int TestStandard( void ) {

        int  iport, imodel, ErrorCode;
        int  BestPortMode, BestXchModel, DefaultXchModel;
        long ADS1210_CrValue;
        char *DefaultXchModelName;
        char *BestDriverName, *BestXchModelName, *BestPortModeName;


        
        // Test for the PARXCH at various port addresses and modes and
        // use the selected address/mode for all the additional tests
        // requested.

        PrintStr( "\n  Test 1: Check port and mode\n\n" );

        iport = TestPort( );
        if ( iport == -1 )
                return( 0 );

        BestDriverName   = TestPortInfo[iport].DriverName;
        BestPortMode     = TestPortInfo[iport].PortMode;
        BestPortModeName = StrPortMode( BestPortMode );

        




        // Now test for the specific A/D model

        PrintStr( "\n  Test 2: Check for specific PARxCH A/D model (x = 1, 4, or 8)\n\n" );

        imodel = TestModel( BestDriverName, BestPortMode );

        BestXchModel     = TestModelInfo[imodel].Id;
        BestXchModelName = StrXchModel( BestXchModel );




        // Now test for the default A/D model

        PrintStr( "\n  Test 3: Check for default PARxCH model\n\n" );
        
        DefaultXchModel     = ParXchGetDefaultModel( BestDriverName );
        DefaultXchModelName = StrXchModel( DefaultXchModel );
                    
        sprintf( Msg, "  For driver %s, the registry/environment default model is %s\n\n",
                 BestDriverName, DefaultXchModelName );
        PrintStr( Msg );

        if (DefaultXchModel != BestXchModel) {
                if (DefaultXchModel == PARXCH_MODEL_PAR1CH &&
                    BestXchModel    == PARXCH_MODEL_PAR4CH) {
                        ;
                        }
                else {
                        sprintf( Msg,
                                 "  %s %s,\n  %s %s.\n  %s\n\n",
                                 "WARNING: Registry/environment default model is set to",
                                 DefaultXchModelName,
                                 "         but PARxCH appears to be a",
                                 BestXchModelName,
                                 "         Use driver\\setmodel to change the default."
                               );
                        PrintStr( Msg );
                        }
                }




        
        // Let them twiddle the yellow LED after reinitializing.

        PrintStr( "\n  Test 4: Check yellow LED\n\n" );


        PrintStr( "  Initializing PARxCH with:\n" );
        sprintf( Msg, "    DriverName = %s\n",  BestDriverName );    PrintStr( Msg );
        sprintf( Msg, "    XchModel   = %s\n",  BestXchModelName );  PrintStr( Msg );
        sprintf( Msg, "    PortMode   = %-s\n", BestPortModeName );  PrintStr( Msg );
        
        ParXchHandle = OpenInitXch(
                                   BestDriverName,
                                   BestXchModel,
                                   BestPortMode,

                                   &ADS1210_CrValue,
                                   &ErrorCode
                                  );

        if ( ParXchHandle == BAD_DEVHANDLE ) {
                sprintf( Msg, "\n  FAILURE: %s\n", PARXCH_ERROR_MSG[ErrorCode] );
                PrintStr( Msg );
                }

        printf( "\n" );

        
        PrintStr( "  Hit ENTER to twiddle the yellow LED, q+ENTER continues ... \n");
        
        TestLed( 1 );
        
        printf( "\n");
        CloseXch( );




        
        return( 1 );
}





/* MAIN:
 *
 * Organize and perform all the PARXCH diagnostic tests requested on
 * the command line by the user.
 *
 *
 */

int main( int argc, char** argv ) {

        int i, Result;


        // Announce program version.

        Rev( );



        // Select desired tests from command line.

        ProcessCommandLine( argc, argv );
        printf( "\n" );
        





        /* MULTIPLE INIT TESTS:
         *
         * The standard and port tests initialize the device several
         * times in different modes, so run them as the first tests.
         *
         */

        // Run standard diag

        if ( test_std ) {
                PrintStrDate( "Beginning 4 Standard tests ... \n" );
                printf( "\n" );
                TestStandard();
                if (ntest == 1)
                        goto done;
                ntest--;
                }


        // Check ports

        if ( test_port ) {
                PrintStrDate( "Beginning port mode and address test ... \n" );
                printf( "\n" );
                Result = TestPort( );
                if ( (Result == -1) || (ntest == 1) )
                        goto done;
                ntest--;
                }


        // Check model

        if ( test_model ) {
                PrintStrDate( "Beginning model test ... \n" );
                printf( "\n" );
                Result = TestModel( DriverName, PortMode );
                if ( (Result == -1) || (ntest == 1) )
                        goto done;
                ntest--;
                }


        // Test default model setting

        if ( test_default ) {
                PrintStrDate( "Testing default model settings ... \n" );
                printf( "\n" );
                TestDefault( );
                if (ntest == 1)
                        goto done;
                ntest--;
                }

        



        /* NO INIT TESTS:
         *
         * Next, we open the driver and run the turnon and turnoff
         * "tests" which are so basic they don't require the device
         * to be initialized.
         *
         */

        // Open the A/D driver (with global driver name from command line)

        ParXchHandle = OpenXch( );



        // Turn device power on (and stay on at exit)

        if ( test_turnon ) {
                PrintStrDate( "Powering up device ... \n" );
                TestTurnon( );
                if (ntest == 1)
                        goto done;
                ntest--;
                }


        // Turn port off (and go off at exit)

        if ( test_turnoff ) {
                PrintStrDate( "Powering down device ... \n" );
                TestTurnoff( );
                if (ntest == 1)
                        goto done;
                ntest--;
                }





        /* SINGLE INIT TESTS:
         *
         * Now, initialize the device to the requested or default mode.
         * All remaining tests are performed before the driver is
         * closed.
         *
         * Even if the init fails, we still continue.  This is
         * especially useful for the led test since that can prove we
         * are writing to the device OK even though we can not read back
         * (which suggests the port is not truely bi-directional).
         *
         */


        // Init the A/D driver (with global model and port from command line)

        Result = InitXch( );
        if (!Result)
                PrintStr(" Continuing anyway\n" );

        

        
        // Test the power good signal.

        if ( test_vgood ) {
                PrintStrDate( "Beginning Vgood test ... \n" );
                printf( "\n" );
                Result = TestVgood( );
                }



        // Test the yellow User LED.

        if ( test_led ) {
                PrintStrDate( "Beginning LED test ... \n" );
                TestLed( 0 );
                }


        

        // Test the User Digital I/O.
        //
        // This test requires that you have a test jig hooked up
        // that connects input 0 to output 0, etc.

        if ( test_userio ) {
                PrintStrDate( "Beginning User Digital I/O test ... \n" );
                printf( "\n" );
                DrvxchBdCtrlMode( BDCTRL_MODE_HALT_UDIO );
                TestUserIO( );
                }


        // Test the FIFO.

        if ( test_fifo_empty ) {
                PrintStrDate( "Beginning FIFO empty test ... \n" );
                printf( "\n" );
                DrvxchBdCtrlMode( BDCTRL_MODE_HALT_INIT );
                TestFifoEmpty( );
                }


        if ( test_fifo_full ) {
                PrintStrDate( "Please be patient, the FIFO full test may take several minutes.\n" );
                PrintStrDate( "Beginning FIFO full test ... \n" );
                printf( "\n" );
                DrvxchBdCtrlMode( BDCTRL_MODE_HALT_INIT );
                TestFifoFull( 20L );
                }


        if ( test_fifo_slam ) {
                PrintStrDate( "Please be patient, the FIFO slam test may take several minutes.\n" );
                PrintStrDate( "Beginning FIFO slam test ... \n" );
                printf( "\n" );
                DrvxchBdCtrlMode( BDCTRL_MODE_HALT_INIT );
                TestFifoSlam( );
                }


        if ( test_fifo_slide ) {
                PrintStrDate( "Please be patient, the FIFO slide test may take several minutes.\n" );
                PrintStrDate( "Beginning FIFO slide test ... \n" );
                printf( "\n" );
                DrvxchBdCtrlMode( BDCTRL_MODE_HALT_INIT );
                TestFifoSlide( );
                }



        // Test acquisition.

        if ( test_acquire ) {
                PrintStrDate( "Acquiring one value from each channel ... \n" );
                printf( "\n" );
                TestAcquire();
                }



        // Power down and close driver.

        CloseXch( );




done:        
        // Ring bell to signal completion.

        if (OptBeep) {
                for ( i = 0 ; i < BEEP_COUNT ; i++ )
                        printf( "\a" );
                }

        return( 0 );
}
