/* FILE: digio.c             Copyright (c), Symmetric Research, 2003-2005

This program demonstrates writing and reading the digital I/O bits
that are available on the DB15 connector on the right hand side of the
front panel of the PARxCH.

Digital output bits on the PARxCH are latched by the hardware.  Write
various output values with this program and test them with a voltmeter
after the write.  Because they are latched they will remain stable.
Output values are standard full scale 0v and +5v TTL.

Another easy way to see the digital output is to wire up the TRM15E
terminal board with 4 LEDs connected between the output bits and
ground.  No other parts are required since current limiting resistors
on the PARxCH board make it possible to power a T1 LED directly off
the digital output signals.

For example, to see output bit 0, connect the negative LED lead to
ground at TRM15E position 8 and the positive LED lead to position 9.
This LED should light whenever output bit 0 is set (1).

An easy way to perform a loopback read test with this program is to
prepare a test jig using the TRM15E terminal board and 4 pieces of
wire.  Use the wire to connect each of the digital output bits to the
corresponding digital input bit by connecting the following pairs of
TRM15E euro block positions together: 1 + 9, 3 + 11, 5 + 13, 7 + 15

With these connections, any value that is written to the output bits
will be read back on the input bits and displayed by digio.exe.

The following table shows the relationship between the digital I/O
bits, the DB15 connector pins, and the TRM15E terminal board euro
block positions.  For example, access output bit 2 is available by
connecting to either pin #7 on the DB15 connector or euro block
position 13 on the TRM15E.

  INPUT bit    0   1   2   3
  OUTPUT bit                   0   1   2   3 
  Ground                                       G   G   G   G   G   G   G
  DB pin #     1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
  TRM15E pos   1   3   5   7   9  11  13  15   2   4   6   8  10  12  14

*/



#include <stdio.h>
#include <stdlib.h>  // for exit
#include <string.h>  // for strcmp

#include "srhelper.h"
#include "parxch.h"


// Global variables

#define MAXNAME 20

char DriverName[MAXNAME], *XchModelName, *PortModeName;
int  NeedHelp, XchModel, PortMode;


// Message text

char *UsageMsg[] = {

 "USAGE: digio [DriverName] [XchModel] [PortMode] [/?]                        ",
 "                                                                            ",
 "   DriverName = SrParXch378, SrParXch278, SrParXch3BC  default SrParXch378  ",
 "     XchModel = PAR1CH, PAR4CH, PAR8CH                 default from setmodel",
 "     PortMode = bpp, epp, ecp/bpp, ecp/epp             default epp          ",
 "           /? = show this help message                                      ",
 "                                                                            ",
 "EXAMPLE: digio SrParXch378 PAR4CH epp                                       ",
 NULL,
 };


char *ExplainMsg[] = {
 "                                                                     ",
 "Digio prompts for a four bit value to write to the PARxCH DB15       ",
 "digital I/O connector.  Output values are latched.  After the output,",
 "the digital input bits are read and also displayed.  See digio.c     ",
 "for comments on making a simple loopback test jig with the TRM15E.   ",
 "                                                                     ",
 NULL,
 };

char *ShowBinary[] = {
        "0000", "0001", "0010", "0011", "0100", "0101", "0110", "0111",
        "1000", "1001", "1010", "1011", "1100", "1101", "1110", "1111"
        };



// Function Prototypes

void ProcessCommandLine( int argc, char **argv );
char *StrXchModel( int PassedXchModel );
char *StrPortMode( int PassedPortMode );
void Usage( char *msg );
void PrintMsg( char *MsgArray[] );




/* Main Digital I/O Program Code
 *
 *
 */ 
        
int main( int argc, char *argv[] ) {

        int       Error, Output, Input, Nscan;
        char      Response[15];
        DEVHANDLE ParXchHandle;


        // Announce program version

        printf( "PARxCH digio [/?] : (c) SR 2004 : Rev {01/01/04}\n" );
        
        
        // Process the command line arguments, if any.

        ProcessCommandLine( argc, argv );


        // Give info message (so they can read it while waiting for driver open)

        PrintMsg( ExplainMsg );



        // Open the PARxCH A/D driver.

        printf( "Opening PARxCH A/D board with:\n" );
        printf( "   DriverName = %s\n", DriverName );
        printf( "   XchModel   = %s\n",   XchModelName );
        printf( "   PortMode   = %s\n\n", PortModeName );

        ParXchHandle = ParXchOpen( DriverName,
                                   XchModel,
                                   PortMode,
                                   100.0,
                                   NULL,
                                   &Error
                                 );

        if ( Error != PARXCH_ERROR_NONE ) {
                printf( "\nERROR: Could not open PARxCH device driver. (Err=%s)\n",
                          PARXCH_ERROR_MSG[Error] );
                exit( 1 );
                }

        

        // Loop until the user enters q.  Prompt for a value, output
        // it, read the digital I/O, and show the results.

        while ( 1 ) {

                // Prompt user for a value and process response.

                printf( "\nEnter hex output value or q to quit: " );
                SrFlushPrint();

                scanf( "%s", Response );
                if ( (Response[0] == 'q') || (Response[0] == 'Q') )
                        break;

                Nscan = sscanf( Response, "%X", &Output );
                if ( Nscan != 1 || Output < 0x0 || Output > 0xF ) {
                        printf( "Value must be between 0x0 and 0xF.  Please try again.\n\n" );
                        continue;
                        }

        
                // Write the digital output.

                ParXchUserIoWr( ParXchHandle, Output );



                // Read the digital input.

                ParXchUserIoRd( ParXchHandle, &Input );



                // Display the results.

                printf( "Digital output = 0x%X ( binary %s )\n", Output, 
                                                           ShowBinary[Output] );
                printf( "Digital input  = 0x%X ( binary %s )\n", Input, 
                                                           ShowBinary[Input] );

                } // end while

        

        // Close the PARxCH.

        ParXchClose( ParXchHandle );
        exit( 0 );
}


/* Command Line Processing function
 *
 *
 */ 
        
// Command line option data

#define CMD_DONE   -1
#define CMD_HELP    0
#define CMD_DRIVER  1
#define CMD_MODEL   2
#define CMD_PORT    3
#define CMD_MAX     4

struct CommandLineOption {
        int   Category;
        char *Name;
        char *ParmChar;
        int  *ParmInt;
        int   Value;
        } *Opt, OptionsTable[] = {

  CMD_HELP,   "/?",          NULL,       &NeedHelp, 1,

  CMD_DRIVER, "SrParXch378", DriverName, NULL,      0,
  CMD_DRIVER, "SrParXch278", DriverName, NULL,      0,
  CMD_DRIVER, "SrParXch3BC", DriverName, NULL,      0,

  CMD_MODEL,  "PAR1CH",      NULL,       &XchModel, PARXCH_MODEL_PAR1CH,
  CMD_MODEL,  "PAR4CH",      NULL,       &XchModel, PARXCH_MODEL_PAR4CH,
  CMD_MODEL,  "PAR8CH",      NULL,       &XchModel, PARXCH_MODEL_PAR8CH,

  CMD_PORT,   "bpp",         NULL,       &PortMode, PARXCH_PORT_MODE_BPP,
  CMD_PORT,   "epp",         NULL,       &PortMode, PARXCH_PORT_MODE_EPP,
  CMD_PORT,   "ecp/bpp",     NULL,       &PortMode, PARXCH_PORT_MODE_ECP_BPP,
  CMD_PORT,   "ecp/epp",     NULL,       &PortMode, PARXCH_PORT_MODE_ECP_EPP,

  CMD_DONE,    NULL,         NULL,       NULL,      0,

        };


void ProcessCommandLine( int argc, char **argv ) {

        int  i, CmdFilled[CMD_MAX];
        char Msg[256];


        // Set defaults

        SrStringCopy( DriverName, "SrParXch378", MAXNAME );
        XchModel = PARXCH_MODEL_UNKNOWN;
        PortMode = PARXCH_PORT_MODE_EPP;
        NeedHelp = 0;

        for ( i = 0 ; i < CMD_MAX ; i++ )
                CmdFilled[i] = 0;
        

        // Verify number of arguments (none to 2)

        if ( argc < 1 || argc > 3 )
                Usage( "wrong number of arguments" );


        // Process arguments

        while ( --argc ) {

                argv++; // Move to next argument (1st time skips program name)
                
                // Check all names in the OptionsTable for a match.

                Opt = OptionsTable;
                while ( Opt->Name && strcmp( *argv, Opt->Name ) )
                                Opt++;


                // Argument did not match a valid option.

                if ( !Opt->Name ) {
                        sprintf( Msg, "unknown option %s", *argv );
                        Usage( Msg );
                        }


                // Check for duplicate argument.

                if ( CmdFilled[ Opt->Category ] == 1 )
                        Usage( "more than one option per category" );
                else
                         CmdFilled[ Opt->Category ] = 1;

                
                // Fill in known option.

                if ( Opt->ParmInt )
                        *Opt->ParmInt = Opt->Value;
                
                else if ( Opt->ParmChar )
                        SrStringCopy( Opt->ParmChar, Opt->Name, MAXNAME );

                else
                        Usage( "option of unknown type" );

                } // end while argc


        // Check for help request

        if ( NeedHelp == 1 )
                Usage( "?" );



        // Finalize selection of specific PARxCH model

        if (XchModel == PARXCH_MODEL_UNKNOWN) // Not given by user on command line
                XchModel = ParXchGetDefaultModel( DriverName );

        XchModelName = StrXchModel( XchModel );
        PortModeName = StrPortMode( PortMode );

        
        return;
}


/* Helper functions
 *
 *
 */ 
        
char *StrXchModel( int PassedXchModel ) {

        if (PassedXchModel == PARXCH_MODEL_PAR1CH)
                return( PAR1CH_MODELNAME );
        else if (PassedXchModel == PARXCH_MODEL_PAR4CH)
                return( PAR4CH_MODELNAME );
        else if (PassedXchModel == PARXCH_MODEL_PAR8CH)
                return( PAR8CH_MODELNAME );
        else 
                return( "unknown" );
}

char *StrPortMode( int PassedPortMode ) {

        if (PassedPortMode == PARXCH_PORT_MODE_BPP)
                return( "BPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_EPP)
                return( "EPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_ECP_BPP)
                return( "ECP/BPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_ECP_EPP)
                return( "ECP/EPP" );
        else 
                return( "unknown" );
}


void Usage( char *msg ) {

        // Print a command line usage message and quit.

        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );

        PrintMsg( UsageMsg );
        exit( 1 );
}

void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );
}

