// FILE: indriver.c              Copyright (c), Symmetric Research, 2003
//
// Program to install PARxCH kernel mode driver.
//
// This is done by starting the Service Control Manager, creating the
// driver service, adding driver information to the registry, and then
// starting the service.  Additional registry entries are also added to
// allow PARxCH driver messages to be properly interpreted by the
// system log EventViewer.
//
// Refer to the readme.txt file for information about how to install
// the device driver manually.
//
//


#include <windows.h>
#include <stdio.h>

#include "parxchkd.h"   // for REG_PARM constants




// Include the DriverInfo[] array with allowed driver names:

#include "indriver.h"
DRIVER_INFO_TABLE *Driver;




// Helper function protos:

void PrintMsg( char *MsgArray[] );
void Usage( void );
void Error( char *Msg, int Fatal );



// Global variables shared by main and helper functions:

SC_HANDLE SCManager, SCService;




// Usage and Error message arrays:

char *Usage1Msg[] = {
  "                                                                         ",
  "ERROR: Command Line Usage: indriver ParXchDriverName XchModel            ",
  "                                                                         ",
  "This utility installs the PARxCH device driver.   This one driver        ",
  "supports members of the PARxCH family of boards for x = 1, 4, or 8.      ",
  "                                                                         ",
  "ParXchDriverName identifies the driver and should be used when running   ",
  "rmdriver, with the ParXchOpen() function, and with application programs  ",
  "like simp and scope.  ParXchDriverName is case sensitive and must be     ",
  "from the following list:                                                 ",
  "                                                                         ",
  NULL
  };

char *Usage2Msg[] = {
  "                                                                         ",
  "XchModel indicates which model is the default for application programs.  ",
  "This choice can be changed later with the setmodel utility without       ",
  "reinstalling the driver.  It can also be overridden within applications  ",
  "like simp and scope as needed.  XchModel is case sensitive and must be   ",
  "one of the following: PAR1CH, PAR4CH, PAR8CH                             ",
  "                                                                         ",
  "INDRIVER EXAMPLE: > indriver SrParXch378 PAR4CH                          ",
  NULL
  };

char *SuccessMsg[] = {
  "                                                                       ",
  "SUCCESS: The device has been installed.                                ",
  "                                                                       ",
  "Now run the PARxCH diagnostic program diag to verify the PARxCH is     ",
  "working correctly.  If using the PARGPS, please install its driver     ",
  "using the indriver utility located in the pargps\\driver directory.    ",
  "                                                                       ",
  "Note that the exact driver name is needed when accessing the PARxCH in:",
  "      calls to ParXchOpen                                              ",
  "      the scope hardware dialog                                        ",
  "      the simp ini file                                                ",
  "      the rmdriver utility to remove the driver                        ",
  "                                                                       ",
  "If you forget the name, you can find by clicking the Windows NT Devices",
  "Applet from the Control Panel and looking for names that start with Sr.",
  "                                                                       ",
  "The name of the installed PARxCH device driver is:                     ",
  "                                                                       ",
  NULL
  };

char *ErrorServiceExistsMsg[] = {

  "  Note: Use rmdriver.exe to clear out the previous version and        ",
  "  try again.                                                          ",

  NULL
  };

char *ErrorAccessDeniedMsg[] = {

  "The PARxCH device driver could not start because no port was found at ",
  "the requested port address.                                           ",
  "                                                                      ",
  "Possible causes include:                                              ",
  "                                                                      ",
  "   1) The port is already in use by a printer driver or the default   ",
  "      Printer Port and Parallel Class Enumerator drivers.             ",
  "   2) The parallel port is using a different address.  eg you asked   "
  "      for 0x378, but it is using 0x278.                               "
  "                                                                      ",
  "Run rmdriver to remove the failed PARxCH installation and see the     ",
  "troubleshooting section of readme.txt.  BIOS CMOS setup, the Device   ",
  "Manager and the Printers folder may also be useful in determining and ",
  "correcting the problem.  Then try the PARxCH installation again.      ",

  NULL
  };

char *ErrorNoSystemResourcesMsg[] = {

  "This error most likely indicates that the PARxCH device driver could  ",
  "not start because another device was using the requested port         ",
  "addresses.                                                            ",
  "                                                                      ",
  "Once the conflict is resolved, you can start the PARxCH device driver ",
  "by typing: net start DriverName where DriverName is the actual name   ",
  "of the PARxCH driver.                                                 ",
  "                                                                      ",
  "Please examine the System Log in the EventViewer for more details on  ",
  "the resource conflict.  To access the EventViewer, type eventvwr at   ",
  "the CommandPrompt or select Programs > Administrative Tools >         ",
  "EventViewer from the Start menu.                                      ",
  "                                                                      ",
  "The current PARxCH driver name is:                                    ",

  NULL
  };




// Character string areas:

#define MAXSTR  256

char
        XchModelName[MAXSTR],
        ParameterKeyName[MAXSTR],
        EventLogKeyName[MAXSTR],
        DriverPath[MAXSTR],
        EventMessageFile[MAXSTR],
        DriverDescBuffer[MAXSTR];





// Main:

void main( int argc, char *argv[] ) {

        HKEY   ParXchKey;
        BOOL   Ret;
        LONG   Code, Code1, Code2, Code3, Code4;
        DWORD  TagId, Disposition, TypesSupported;
        char  *DriverDesc, *LoadGroup;



        
        // Banner message

        printf( "PARxCH Kernel Mode Driver Installation Program : Rev {01/01/04}\n");



        // Process command line arguments for driver name and Xch model name
        // (port address and irq derived from driver name)

        if ( argc != 3 ) 
                Usage();

        Driver = DriverInfo;

        while ( Driver->Name )  {
                if (strcmp(Driver->Name, argv[1]) == 0)
                        break;
                Driver++;
                }
        
        if (Driver->Name == NULL)
                Usage();

        strncpy( XchModelName, argv[2], MAXSTR );
        if ( (strcmp(XchModelName,"PAR1CH") != 0) &&
             (strcmp(XchModelName,"PAR4CH") != 0) &&
             (strcmp(XchModelName,"PAR8CH") != 0) )
                Usage();




        // Open Service Control Manager

        SCManager = NULL;
        SCService = NULL;
        
        SCManager = OpenSCManager(
                        NULL,                  // use local machine
                        NULL,                  // use SERVICES_ACTIVE_DATABASE
                        SC_MANAGER_ALL_ACCESS  // full access permission
                        );

        if (!SCManager)
                Error("Could not create service manager", 1);
        



        // Create the device driver service.
        //
        // This involves creating a registry key with several 
        // values and adding the service to the services database.  
        // The registry entries can be made by hand.  But in that
        // case, you must reboot the machine before the service 
        // will be included in the services database.  Once the 
        // service is available, it can be started.  The pertinent 
        // registry entries are as follows (notes in parenthesis are
        // just information and are not included in the registry):
        //
        // HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\SrParXchXXX
        //  DisplayName  = SrParXchXXX A/D Device Driver (DriverDesc)
        //  ErrorControl = 0x00000001                     (SERVICE_ERROR_NORMAL)
        //  Group        = Data Acquisition Devices       (LoadGroup)
        //  ImagePath    = \??\C:\sr\parxch\driver\XXX\parxchkd.sys  (\??\C:DriverPath)
        //  Start        = 0x00000002                     (SERVICE_AUTO_START)
        //  Tag          = 0x00000001                     (TagId)
        //  Type         = 0x00000001                     (SERVICE_KERNEL_DRIVER)
        //

        GetCurrentDirectory( MAXSTR, DriverPath );
        lstrcat( DriverPath,  "\\parxchkd.sys");
        
        sprintf( DriverDescBuffer, "%s A/D Device Driver", Driver->Name );
        DriverDesc  = DriverDescBuffer;
        LoadGroup   = "Data Acquisition Devices";
        TagId       = 0x1L;
        
        SCService = CreateService(
                        SCManager,             // Service Control Manager
                        Driver->Name,          // name of device driver
                        DriverDesc,            // description of driver
                        SERVICE_ALL_ACCESS,    // full access permission
                        SERVICE_KERNEL_DRIVER, // type is a kernel driver
                        SERVICE_AUTO_START,    // starts at end of booting
                        SERVICE_ERROR_NORMAL,  // give warning, but still boot
                        DriverPath,            // location of .sys file
                        LoadGroup,             // loads in custom A/D group
                        &TagId,                // load order within group
                        NULL,                  // depends on no other services
                        NULL,                  // use default ServiceStartName
                        NULL                   // no password
                       );


        if (!SCService)
                Error("Could not create device driver service", 1);




        // Add registry entries for device specific parameters
        
        lstrcpy( ParameterKeyName, "System\\CurrentControlSet\\Services\\");
        lstrcat( ParameterKeyName, Driver->Name );
        lstrcat( ParameterKeyName, "\\Parameters" );

        Code = RegCreateKeyEx(
                       HKEY_LOCAL_MACHINE,      // put under this key
                       ParameterKeyName,        // at this location
                       0,                       // reserved
                       NULL,                    // object type of key
                       REG_OPTION_NON_VOLATILE, // keep value when reboot
                       KEY_ALL_ACCESS,          // full access permission
                       NULL,                    // key handle not inheritable
                       &ParXchKey,              // handle to key
                       &Disposition             // indicates key new or existing
                      );

        if (Code != ERROR_SUCCESS)
                Error("Failed to create parameter key", 1);


        Code1 = RegSetValueEx(
                    ParXchKey,                      // use this key
                    REGENVSTR_MODELNAME,            // value name
                    0,                              // reserved
                    REG_SZ,                         // value data type
                    &(CONST BYTE)XchModelName[0],   // ptr to value contents
                    strlen(XchModelName)+1          // contents size
                   );
        
        Code2 = RegSetValueEx(
                    ParXchKey,                      // use this key
                    REG_PARM_DRIVERNAME,            // value name
                    0,                              // reserved
                    REG_SZ,                         // value data type
                    &(CONST BYTE)Driver->Name[0],   // ptr to value contents
                    strlen(Driver->Name)+1          // contents size
                   );
        
        Code3 = RegSetValueEx(
                    ParXchKey,                      // use this key
                    REG_PARM_PORTADDR,              // value name
                    0,                              // reserved
                    REG_DWORD,                      // value data type
                    &(CONST BYTE)Driver->PortAddr,  // ptr to value contents
                    sizeof(ULONG)                   // contents size
                   );
        
        Code4 = RegSetValueEx(
                    ParXchKey,                      // use this key
                    REG_PARM_PORTIRQ,               // value name
                    0,                              // reserved
                    REG_DWORD,                      // value data type
                    &(CONST BYTE)Driver->PortIrq,   // ptr to value contents
                    sizeof(ULONG)                   // contents size
                   );
        
        if ( (Code1 != ERROR_SUCCESS) ||
             (Code2 != ERROR_SUCCESS) ||
             (Code3 != ERROR_SUCCESS) ||
             (Code4 != ERROR_SUCCESS) )
                printf("Failed writing parameter key values to registry\n");
        
                       
        Code = RegCloseKey( ParXchKey );
        
        if (Code != ERROR_SUCCESS) 
                Error("Failed to close parameter key", 1);
        

        

        // Add registry entries for interpreting EventViewer entries
        //
        // Note that these entries are NOT deleted when rmdriver.exe
        // is run.  This is so that System Log entries relating to the
        // PARxCH driver can still be interpreted even when the driver
        // is no longer running.  Except for making the System Log
        // entries less informative, deleting these registry entries
        // by hand produces no bad effects.
        //
        
        lstrcpy( EventLogKeyName,
                 "System\\CurrentControlSet\\Services\\EventLog\\System\\");
        lstrcat( EventLogKeyName, Driver->Name );


        TypesSupported = 0x7L;
        lstrcpy( EventMessageFile, DriverPath );
        lstrcat( EventMessageFile, ";%SystemRoot%\\System32\\IoLogMsg.dll");
                         
        Code = RegCreateKeyEx(
                       HKEY_LOCAL_MACHINE,      // put under this key
                       EventLogKeyName,         // at this location
                       0,                       // reserved
                       NULL,                    // object type of key
                       REG_OPTION_NON_VOLATILE, // keep value when reboot
                       KEY_ALL_ACCESS,          // full access permission
                       NULL,                    // key handle not inheritable
                       &ParXchKey,              // handle to key 
                       &Disposition             // indicates key new or existing
                      );

        if (Code != ERROR_SUCCESS) 
                Error("Failed creating Event Log Key", 1);


        Code1 = RegSetValueEx(
                    ParXchKey,                      // use this key
                    "EventMessageFile",             // value name
                    0,                              // reserved
                    REG_EXPAND_SZ,                  // value data type
                    (CONST BYTE*)EventMessageFile,  // ptr to value contents
                    strlen(EventMessageFile)        // contents size
                    );
        
                              
        Code2 = RegSetValueEx(
                    ParXchKey,                      // use this key
                    "TypesSupported",               // value name
                    0,                              // reserved
                    REG_DWORD,                      // value data type
                    &(CONST BYTE)TypesSupported,    // ptr to value contents
                    sizeof(DWORD)                   // contents size
                   );
        
        if ( (Code1 != ERROR_SUCCESS) ||
             (Code2 != ERROR_SUCCESS) )
                printf("Failed writing event log key values to registry\n");
        

        Code = RegCloseKey( ParXchKey );
        if (Code != ERROR_SUCCESS)
                Error("Failed closing Event Log Key", 1);




        // Indicate install successful
        
        printf( "\n%s driver has been successfully installed.\n", Driver->Name );

        
        
        // Start Service

        Ret = StartService(

                   SCService,     // service identifier
                   0,             // number of arguments
                   NULL           // pointer to arguments

                   );


        if (!Ret)
                Error("Failed to start PARxCH device driver service", 1);
        else
                printf( "%s driver was started using port address 0x%X and IRQ %d.\n",

                        Driver->Name,
                        Driver->PortAddr,
                        Driver->PortIrq

                        );
        

        // Close handles (leaving service loaded and running)

        CloseServiceHandle( SCService );
        CloseServiceHandle( SCManager );


        PrintMsg( SuccessMsg );
        printf( "\t\t%s   (at 0x%03lX, irq %d)\n",
                  Driver->Name, Driver->PortAddr, Driver->PortIrq );
        
}



// Helper functions:

void PrintMsg( char *MsgArray[] ) {

        int i;

        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );

}

void Usage( void ) {

        int i;
        
        PrintMsg( Usage1Msg );
                
        for ( i = 0 ; DriverInfo[i].Name ; i++ )

                printf( "     %s   << uses parallel port 0x%X, irq %d\n",

                          DriverInfo[i].Name,
                          DriverInfo[i].PortAddr,
                          DriverInfo[i].PortIrq

                        );

        PrintMsg( Usage2Msg );

        exit( 1 );
}

void Error( char *Msg, int Fatal ) {

        DWORD  Err;
        LPVOID lpMessageBuffer;


        // Display local message, error number, and system message
        
        Err = GetLastError();

        FormatMessage(
                      FORMAT_MESSAGE_ALLOCATE_BUFFER |
                      FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      Err,
                      MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                      (LPTSTR) &lpMessageBuffer,
                      0,
                      NULL
                     );

        printf( "\n%s.  Error #%d\n\n%s\n", Msg, Err, lpMessageBuffer );

        LocalFree( lpMessageBuffer );


        
        // Give additional explanation for common errors
        
        if (Err == ERROR_SERVICE_EXISTS)  // #1073
                PrintMsg( ErrorServiceExistsMsg );
        
        else if (Err == ERROR_ACCESS_DENIED) // #5
                PrintMsg( ErrorAccessDeniedMsg );
        
        else if (Err == ERROR_NO_SYSTEM_RESOURCES) { // #1450
                PrintMsg( ErrorNoSystemResourcesMsg );
                printf( "     %s\n", Driver->Name );
                }

        else
                printf( "Sorry, no extra help available for this error.\n" );


        // Clean up and exit if fatal.
        
        if ( Fatal ) {

                if (SCService)  CloseServiceHandle( SCService );
                if (SCManager)  CloseServiceHandle( SCManager );

                exit( 1 );

                }
}
