/* FILE: setmodel.c             Copyright (c), Symmetric Research, 2004

This utility program queries and sets which PARxCH model will be used
by default in the various application programs if no model is explicitly
given by the user.

For WinNT and Win2K/XP, the default model information is stored in the
registry under the driver's service key.

*/


#include <windows.h>
#include <stdio.h>

#include "parxchkd.h" // for device name registry parameter

// Defines and globals

#define NOT_FATAL   0
#define FATAL       1
#define MAXSTR      256


char *DriverOptions[] = {
        "SrParXch378",
        "SrParXch278",
        "SrParXch3BC",
        NULL
        };
char *XchModelOptions[] = {
        "PAR1CH",
        "PAR4CH",
        "PAR8CH",
        NULL
        };
        
char DriverName[MAXSTR], XchModelName[MAXSTR], msg[MAXSTR];


char *UsageMsg[] = {
  "USAGE: setmodel [DriverName] [XchModel]                                ",
  "                                                                       ",
  "  where,                                                               ",
  "          DriverName = SrParXch378, SrParXch278, SrParXch3BC           ",
  "          XchModel = PAR1CH, PAR4CH, PAR8CH                            ",
  "                                                                       ",
  "This utility can be used to set which of the PARxCH models is to be    ",
  "used as the default when running application programs such as diag,    ",
  "meter, simp, and scope.  DriverName and XchModel are case sensitive.   ",
  "                                                                       ",
  "If DriverName is not given, the default SrParXch378 will be used.  If  ",
  "the second argument, XchModel is not given, setmodel will report the   ",
  "current default model value for the specified driver.                  ",
  "                                                                       ",
  "EXAMPLES: setmodel SrParXch378 PAR4CH                                  ",
  "          setmodel SrParXch278                                         ",
  "          setmodel PAR8CH                                              ",
  "          setmodel                                                     ",
   NULL
   };

//Protos

void ProcessCommandLine( int argc, char *argv[] );
void Usage( char *msg );
void PrintMsg( char *MsgArray[] );
void Error( char *msg, int IsFatal );


// Main program

void main( int argc, char** argv ) {
        
        char  ServiceParmKeyName[MAXSTR], CurrentModel[MAXSTR];
        HKEY  ServiceParmKey;
        DWORD Code, BufferSize, RegType;

        // Print banner

        printf( "setmodel [DriverName] [XchModel] : (c) SR 2004 : Rev {01/01/04}\n" );


        // Get DriverName and XchModel

        ProcessCommandLine( argc, argv );


        // Open service parameters registry key

        sprintf( ServiceParmKeyName,
                 "System\\CurrentControlSet\\Services\\%s\\Parameters",
                 DriverName );

        Code = RegOpenKeyEx(
                       HKEY_LOCAL_MACHINE,      // already open key
                       ServiceParmKeyName,      // subkey to open
                       0,                       // reserved
                       KEY_ALL_ACCESS,          // full access permission
                       &ServiceParmKey          // handle to key
                      );

        if (Code != ERROR_SUCCESS) {
                sprintf( msg, "Could not open service parameter registry key.\nDriver %s may not be installed yet.", DriverName);
                Error( msg, FATAL );
                }
        


        // Get current value of XchModelName

        BufferSize = sizeof( CurrentModel );
        Code = RegQueryValueEx(
                               ServiceParmKey,
                               REGENVSTR_MODELNAME,
                               NULL,
                               &RegType,
                               CurrentModel,
                               &BufferSize
                              );
        if (Code != ERROR_SUCCESS)
                sprintf( CurrentModel, "Unavailable (err=%d)", Code );



        // If requested, set new value of default PARxCH model

        if ( XchModelName[0] != '\0') {

                BufferSize = strlen( XchModelName );
                Code = RegSetValueEx(
                                     ServiceParmKey,        // handle to key
                                     REGENVSTR_MODELNAME,   // value name
                                     0,                     // reserved
                                     REG_SZ,                // value type
                                     XchModelName,          // value data
                                     BufferSize             // size of value data
                            );
                        
                if (Code != ERROR_SUCCESS) {
                        sprintf( msg, "Could not set default model to %s\n",XchModelName);
                        Error( msg, NOT_FATAL );
                        }
                else
                        sprintf( CurrentModel, "%s", XchModelName );


                } // end if XchModelName
        

        // Announce current model

        printf( "\n  The current XchModel for driver %s is: %s\n",
                  DriverName, CurrentModel );



        // Close service parm key

        Code = RegCloseKey( ServiceParmKey );
        if (Code != ERROR_SUCCESS)
                Error( "Could not close service parm key", FATAL );
        
}

void ProcessCommandLine( int argc, char *argv[] ) {

        int i, narg, DriverSet, ModelSet, TotalSet;

        // Set defaults

        DriverSet = ModelSet = 0;
        strncpy( DriverName, "SrParXch378", MAXSTR );  
        strncpy( XchModelName, "", MAXSTR );


        // Get the command line arguments.

        if ( argc < 1 || argc > 3 )
                Usage( "Must not have more than 2 arguments" );


        if ( argc == 1 )
                printf( "Type setmodel /? for a usage message.\n" );


        narg = 0;
        while ( --argc ) {

                argv++;

                // Check for help request

                if ( strcmp( *argv, "/?" ) == 0 )
                        Usage( "?" );

                
                // Check for driver name

                i = 0;
                while ( DriverOptions[i] ) {
                        if ( strcmp( *argv, DriverOptions[i] ) == 0 ) {
                                if (DriverSet)
                                        Usage( "Driver name appears twice" );
                                else {
                                        DriverSet = 1;
                                        strncpy( DriverName,
                                                DriverOptions[i], MAXSTR );
                                        } // end else
                                } // end if strcmp
                        i++;
                        } // end while DriverOptions

                
                // Check for model name

                i = 0;
                while ( XchModelOptions[i] ) {
                        if ( strcmp( *argv, XchModelOptions[i] ) == 0 ) {
                                if (ModelSet)
                                        Usage( "Model name appears twice" );
                                else {
                                        ModelSet = 1;
                                        strncpy( XchModelName,
                                                 XchModelOptions[i], MAXSTR );
                                        } // end else
                                } // end if strcmp
                        i++;
                        } // end while XchModelOptions



                // Argument did not match a valid option.

                narg++;
                TotalSet = DriverSet + ModelSet;

                if ( narg != TotalSet)
                        Usage( "Unknown argument" );

                
                } // end while argc

        return;
}

void Usage( char *msg ) {

        // Print a command line usage message and quit.

        // Select error line depending on 
        // 1) no message, 2) help requested, or 3) message given

        if ( !msg )
                printf( "\nERROR: Command line usage.\n\n" );

        else if ( strcmp( msg, "?" ) == 0 )
                printf( "\n" );

        else
                printf( "\nERROR: Command line usage (%s).\n\n", msg );

        PrintMsg( UsageMsg );
        exit( 1 );
}

void PrintMsg( char *MsgArray[] ) {

        int i;
        
        for ( i = 0 ; MsgArray[i] ; i++ )
                printf( "%s\n", MsgArray[i] );
}

void Error( char *msg, int IsFatal ) {
        
        printf( "ERROR: %s\n", msg );
        if ( IsFatal )
                exit( 0 );
}
