/* FILE: sample.c               Copyright (c), Symmetric Research, 2003-2005
 *
 * This is a simple bare bones program to demonstrate running the PARxCH
 * board with the PARxCH function library.  It has the minimum code required
 * to run the board.
 *
 * If you want to change any of the sample.c initialization parameters,
 * you must change its source code and recompile.  See the makefile for
 * how to compile.
 *
 * Simp.c in the simple directory provides a more flexible interface 
 * for changing the init parameters.  Use this program for experimenting
 * with programming the PARxCH in C.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "parxch.h"
#include "srhelper.h"



// Function prototypes

int  GetAnalogChannels( int PassedXchModel );
char *StrXchModel( int PassedXchModel );
char *StrPortMode( int PassedPortMode );
void DisplayValues( long *Values, unsigned int Nvalues, int AnalogChannels );
void Error( char *msg, int errorcode );




/* MAIN:
 *
 *
 *
 *
 */

int main( void ) {

        int  ErrorCode, XchModel, PortMode, AnalogChannels;
        char *DriverName;
        long *PC_BUFFER;
        double RequestedSps, ActualSps;
        unsigned int Nvalues, Nremain, NextPt, Nread;
        DEVHANDLE ParXchHandle;


        // Announce program version
        
        printf( "PARxCH sample : (c) SR 2004 : Rev {01/01/04}\n" );

          
        // Set initialization parameters

        DriverName     = PARXCH_0x378;
        XchModel       = ParXchGetDefaultModel( DriverName );
        PortMode       = PARXCH_PORT_MODE_EPP;
        RequestedSps   = 100.0;
        AnalogChannels = GetAnalogChannels( XchModel );



        // Allow getchar to return even if a key has not been pressed.

        if ( SrKeyboardNonBlock() == -1 )
               Error( "Can't set non-blocking read ...", -1 );





        /* Initialize the PARxCH.
         *
         * The defined constants used with ParXchOpen come from parxch.h.
         *
         *
         */

        printf( "\nOpening PARxCH A/D board with:\n" );
        printf( "   DriverName = %s\n",   DriverName );
        printf( "   XchModel   = %s\n",   StrXchModel( XchModel ) );
        printf( "   PortMode   = %s\n\n", StrPortMode( PortMode ) );


        ParXchHandle = ParXchOpen(
                                  DriverName,
                                  XchModel,
                                  PortMode,
                                  RequestedSps,

                                  &ActualSps,
                                  &ErrorCode
                                 );

        if ( ParXchHandle == BAD_DEVHANDLE )
                Error( "ParXchOpen failed ...", ErrorCode );





        /* PC DATA BUFFER:
         *
         * Values from the PARxCH board are saved in this array.  The
         * size can be any multiple of the number of analog channels.
         * 
         * Each 24 bit value is saved as a 32 bit long integer.  Values 
         * from the analog channels are saved in multiplexed fashion.
         *
         *
         */

        Nvalues   = 128 * AnalogChannels;
        PC_BUFFER = (long *)malloc( Nvalues*sizeof( long ) );






        /* Start executing and wait for data to become ready.
         *
         *
         */

        printf( "Press ENTER to start waiting for data, ENTER again quits ...\n" );
        SrGetKeyWait();

        ParXchStart( ParXchHandle );


        // Initialize data counters.

        Nremain = Nvalues;
        NextPt  = 0;

        
polling_loop:

        
        // Read available data into array.

        Nread = ParXchReadData(
                               ParXchHandle,
                               &PC_BUFFER[NextPt],
                               Nremain,
                               &ErrorCode
                              );
        Nremain -= Nread;
        NextPt  += Nread;



        // When all requested points have been read,
        // reset the data counters and display the data.
                
        if (Nremain == 0) {

                Nremain = Nvalues;
                NextPt  = 0;

                DisplayValues( PC_BUFFER, Nvalues, AnalogChannels );
                
                }


        
        // Check for an error or a kbhit to quit.

        if (ErrorCode != PARXCH_ERROR_NONE) {
                
                ParXchStop( ParXchHandle );
                ParXchClose( ParXchHandle );
                
                Error( "ParXchReadData failed ...", ErrorCode );
                
                }
        
        else if ( SrGetKeyCheck() != EOF ) {

                printf( "Kbhit ... quitting\n" );

                ParXchStop( ParXchHandle );
                ParXchClose( ParXchHandle );

                exit( 0 );

                }




        // Continue polling.

        goto polling_loop;

        return( 0 );
}



/* HELPER FUNCTIONS:
 *
 *
 *
 */

int GetAnalogChannels( int PassedXchModel ) {

        if (PassedXchModel == PARXCH_MODEL_PAR1CH)
                return( PAR1CH_ANALOG_CHANNELS );
        else if (PassedXchModel == PARXCH_MODEL_PAR4CH)
                return( PAR4CH_ANALOG_CHANNELS );
        else if (PassedXchModel == PARXCH_MODEL_PAR8CH)
                return( PAR8CH_ANALOG_CHANNELS );
        else 
                return( PARXCH_MAX_ANALOG_CHANNELS );
}

char *StrXchModel( int PassedXchModel ) {

        if (PassedXchModel == PARXCH_MODEL_PAR1CH)
                return( PAR1CH_MODELNAME );
        else if (PassedXchModel == PARXCH_MODEL_PAR4CH)
                return( PAR4CH_MODELNAME );
        else if (PassedXchModel == PARXCH_MODEL_PAR8CH)
                return( PAR8CH_MODELNAME );
        else 
                return( "unknown" );
}

char *StrPortMode( int PassedPortMode ) {

        if (PassedPortMode == PARXCH_PORT_MODE_BPP)
                return( "BPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_EPP)
                return( "EPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_ECP_BPP)
                return( "ECP/BPP" );
        else if (PassedPortMode == PARXCH_PORT_MODE_ECP_EPP)
                return( "ECP/EPP" );
        else 
                return( "unknown" );
}

void DisplayValues( long *Values, unsigned int Nvalues, int AnalogChannels ) {

        unsigned int i;
        int          j;

        for ( i = 0 ; i < Nvalues ; i += AnalogChannels ) {

                printf( "   " );

                for ( j = 0 ; j < AnalogChannels ; j++ ) {

                        printf( " %08lX", Values[i+j] );

                        } // end for j

                printf( "\n" );

                } // end for i
}

void Error( char *msg, int errorcode ) {

        // Print an error message and quit.

        if (errorcode < 0)
                printf( "ERROR: %s\n", msg );
        else
                printf( "ERROR: %s (%s)\n", msg, PARXCH_ERROR_MSG[errorcode] );

        exit( 1);
}

