/* FILE: srdat.h              Copyright (c), Symmetric Research, 2003-2004

SR DAT file defines, structures, and helper function protos.


The new DAT file format has been introduced to better support
recording GPS timing information along with analog data.  SR will
continue to support the older OUT data file format for applications
not using GPS or digital recording and better suited to a simple
analog data only file format.


The DAT file format is comprised of a header followed by records.

The header size is fixed at 4096 bytes and is comprised of a C
structure, struct SrDatHdrLayout, followed by zero padding.  The
fields in the header structure include items such as sampling rate,
number of channels and etc.

Records following the header are comprised of a record tag structure
followed by the record information.  Records can have information
about items such as analog data or GPS timing information, etc.  The
record tag structure has an identifier indicating the type of record
information following the tag structure.  See the comments below for
more information.



HEADER FIELD DESCRIPTIONS:

Following are short descriptions of the fields in the header struct
SrDatHdrLayout.  They are listed by functional groups.


GENERAL HEADER PARAMETERS:

* DatId

A long integer for identifying a DAT file easily.  Must always equate
to the string "SRSR".

* DatRev

The DAT file format revision number.

* DatProduct

An ascii string indicating which SR A/D hardware produced the acquired
data stored in the DAT file.  Typical examples would be:

        "SR PAR4CH with GPS"
        "SR PAR4CH"
        "SR PAR24B"
        "SR DSPA64"

* DatApplication

An ascii string indicating the application program that generated the
DAT file.  Typical examples would be:

        "SR SIMP4CH"
        "SR SCOPE4CH"
        "SR SCOPEA64"

* DatUnused1, DatUnused2

Fields for possible future use and 8 byte alignment.  For Windows and
Linux compatibility it is important to have doubles aligned on 8 byte
boundaries.  Unused fields are used as needed in the header structure
to force such alignment.




FILE RELATED PARAMETERS:

* FileName

The original name of the DAT file when it was created by the
application.

* FileCreateTime

The local time when the DAT file was created, stored as seconds since
1/1/1970.

* FileSeqNum

A sequential number indicating this is the n-th file written during
the current run of the application.

* FileStartPtNum

The number of the first point in this file, where point 0 was the
first point in the first file during the current acquisition run.

* FileUnused1

Future use and 8 byte boundary alignment.




DATE AND TIME PARAMETERS:

* TimePoint[]

A structure with YMD date and HMS time information for up to four time
points.  The various fields in the structure are detailed below.

* Meaning

An indicator of what the time data represents.  Possible values are
the SRDAT_TIME_TYPE_ defines below.  Typically the meaning will
indicate whether the time is for the first or last point in the file
or the first or last point with a GPS mark.

* SampleFromFile, SampleFromRun

Index of the analog data sample the time information is for.  Index 0
is the first point in the file for SampleFromFile and the first point
in the current run of the application for SampleFromRun.

* Year, Month, Day, Hour, Minute, Second, Microsecond

UTC date (YMD) and time (HMS) for the identified sample point.  Year
is a standard 4 digit value like 2002, month goes from 1 to 12 and day
goes from 1 to 31.  Hour goes from 0 to 23, minute and second go from
0 to 59, and microsecond goes from 0 to 999999.

* YMDQuality, HMSQuality

Indicators of how the YMD and HMS values were determined.  Possible
values are the SRDAT_SOURCE_ defines below.  Typically the quality
will indicate whether the source was the PC clock or GPS system.






CHANNEL RELATED PARAMETERS:

* Channels

Total number of channels recorded.  Most of the channels acquired are
analog data, but one may be a special "GPS mark" channel as described
below.  The maximum number of channels allowed is 65.

* ChannelMarkIndex

The index of a special "GPS mark" channel.  An example would be a SR
PAR4CH system with GPS attached.  Such a system has 4 analog channels
(0,1,2,3) and one "GPS mark" channel (4).  For this system, Channels = 5,
and ChannelMarkIndex = 4.

If only analog data has been recorded and there is no GPS mark
channel, then ChannelMarkIndex = SRDAT_MARKCHANNEL_NONE.

* ChannelPtsPerRecord

The number of analog samples per channel in one ANALOG16 or ANALOG24
data record.  Multiply by Channels to get the total number of samples
per record.

* ChannelPtsPerFile

The number of analog samples per channel in the entire file.  These
samples may be spread across many ANALOG16 or ANALOG24 data records.
Note that DAT files are assumed to have only ANALOG16 or ANALOG24
records, but not both, and all ANALOG records are also assumed to
have the same number of points.

* ChannelList

An array of structures containing an scan list index and short 12
character title string for each channel.  This array is dimensioned
to the maximum number of channels, but only contains entries for the
recorded channels counted in the Channels field.



A/D ANALOG CONVERTER PARAMETERS:

* AtodWordSize

Number of bytes per sample for the acquired analog data.  SR 24 bit
products store their data in 32 bit longs and have AtodWordSize = 4.
SR 16 bit products have AtodWordSize = 2.

* AtodDataCoding

Whether the data uses signed or offset format.  Possible values are
the SRDAT_DATACODING_ defines below.

* AtodSamplingRate

Rate at which the analog data was acquired in samples per second.

* AtodSamplingRateMeasured

Actual measured sampling rate.  This may differ slightly from the
theoretical AtodSamplingRate above, and is determined by counting the
actual number of sampling points per GPS PPS mark on installations
with a GPS attached.

* AtodGain

Amplification factor applied by the circuitry on the A/D board.

* AtodFilterScale, AtodFilterCoeff

Filter scale and smoothing coefficients for the SR DSPA64/HLF
product.  Will be set to 1.0 and 0.0 respectively for no filtering.
See the DSPA64/HLF documentation for more information.




64-BIT COUNTER FIELDS:

Most PCs have a high speed 64 bit counter on their motherboard which
the SR GPS system uses as part of its time recording system.  The
folowing two fields characterize this counter.  See the description
of the DAT record types below for more information on how the counter
is used and recorded.


* CounterType

CounterType indicates the OS specific counter format for the system
on which the DAT file was recorded.  Use the SRDAT_COUNTER_ defines
below.

Under Windows, the counter is a simple 64 bit unsigned integer.

Under Linux, the counter is a small structure occupying 64 bits.  The
high 32 bits hold seconds, while the low 32 bits hold microseconds.
The low 32 bits advance from 0 to 999999 microseconds and then roll
back to 0, incrementing the high 32 bit second part of the count.


* CounterFreq

CounterFreq indicates the rate in counts/second at which the 64 bit
counter advances.

Under Windows, this value is usually 1193254 counts/second, which
corresponds to about .8 microseconds per count.  Be aware however,
the exact value may vary with processor and motherboard.  Always
parameterize your programs with respect to CounterFreq rather than a
hardcoded value.

Under Linux, CounterFreq is always 1000000 counts/second, indicating
the microsecond part of the 64 bit count structure advances at one
microsecond per count.



LOCATION PARAMETERS:

* LocationLatitude
* LocationLongitude
* LocationAltitude

If a GPS system is attached to the SR equipment, then these are the
corresponding location parameters.  See your GPS documentation for
more information.

* LocationQuality

An indicator of how the locations was determined.  Possible values
are the SRDAT_SOURCE_ defines below.  Normally the location will
have been determined by the GPS system, but other methods are
possible.

*/






// HEADER STRUCTURE AND DEFINES:
//
// Note: If revising the header structure, be sure to
//
// 1) Change the Rev field.
// 2) Keep all data in 4 bytes multiples so Linux stays aligned (ie no
//    shorts).
// 3) Keep an even number of 4 byte items (ie 8 byte alignment) so the
//    doubles don't need alignment padding in Windows.
// 4) Do not let the total size exceed 4096.
//
//

#define SRDAT_HDR_ID                    ((long)('RSRS')) // = "SRSR"
#define SRDAT_REV                       105

#define SRDAT_INVALID_LONG              -12345
#define SRDAT_INVALID_DOUBLE            -12345.0L

#define SRDAT_SOURCE_NONE               0       // none
#define SRDAT_SOURCE_CALC               1       // calculated
#define SRDAT_SOURCE_PC                 2       // from the PC clock
#define SRDAT_SOURCE_GPS                3       // from GPS
#define SRDAT_SOURCE_USER               4       // set by the user
#define SRDAT_SOURCE_ACQ                5       // set by acquisition program
#define SRDAT_SOURCE_OBC                6       // from GPS + on board counter

#define SRDAT_TIME_TYPE_NONE            0       // none
#define SRDAT_TIME_TYPE_FIRSTPT         1       // first point in file
#define SRDAT_TIME_TYPE_LASTPT          2       // last  point in file
#define SRDAT_TIME_TYPE_FIRSTGPS        3       // first GPS marked point
#define SRDAT_TIME_TYPE_LASTGPS         4       // last  GPS marked point
#define SRDAT_TIME_TYPE_FIRSTEVENT      5       // first event occurred
#define SRDAT_TIME_TYPE_LASTEVENT       6       // last  event occurred
#define SRDAT_TIME_TYPE_USER1           7       // first user defined meaning
#define SRDAT_TIME_TYPE_USER2           8       // next  user defined meaning

#define SRDAT_MARKCHANNEL_NONE         -1

#define SRDAT_DATACODING_OFFSET         0
#define SRDAT_DATACODING_SIGNED         1

#define SRDAT_COUNTER_INT64             0
#define SRDAT_COUNTER_HIGH32LOW32       1


// Define array dimensions

#define SRDAT_DIM_CHANNELS             65   // Allow for 64ch + GPS mark
#define SRDAT_DIM_TITLE                12
#define SRDAT_DIM_STRING               32
#define SRDAT_DIM_TIME                  4



struct SrDatChannelData {
        long Channel;
        char Title[ SRDAT_DIM_TITLE ];
        };


struct SrDatTimeData {
        long Type;
        long SampleFromRun;
        long SampleFromFile;
        long GpsMarkIndex;
        long Year;
        long Month;
        long Day;
        long YMDSource;
        long Hour;
        long Minute;
        long Second;
        long Microsecond;
        long HMSSource;
        long Unused1;
        };


typedef struct SrDatHdrLayout {


        // DAT Id, rev, source
        
        long DatId;             // = "SRSR"
        long DatRev;
        char DatProduct[ SRDAT_DIM_STRING ];
        char DatApplication[ SRDAT_DIM_STRING ];

        long DatUnused1;        // align to 8 byte boundary ...
        long DatUnused2;

        
        // File name parameters
        
        char FileName[ SRDAT_DIM_STRING ];
        long FileCreateTime;
        long FileSeqNum;
        long FileStartPtNum;
        long FileUnused1;       // align to 8 byte boundary ...


        // Date and Time

        struct SrDatTimeData TimePoint[ SRDAT_DIM_TIME ];


        // Channel layout
        
        long Channels;
        long ChannelMarkIndex;
        long ChannelPtsPerRecord;
        long ChannelPtsPerFile;

        struct SrDatChannelData ChannelList[ SRDAT_DIM_CHANNELS ];
        
        
        // Analog converter parameters
        
        long AtodWordSize;
        long AtodDataCoding;
        double AtodSamplingRate;
        double AtodSamplingRateMeasured;
        double AtodGain;
        double AtodFilterCoeff;
        double AtodFilterScale;

        
        // 64-bit counter
        
        long CounterType;
        long CounterFreq;

        
        // GPS location
        
        double LocationLatitude;
        double LocationLongitude;
        double LocationAltitude;
        long LocationSource;
        long LocationUnused1;   // align to 8 byte boundary ...


        } SRDAT_HDR;





// HEADER PADDING ARRAY:
//
// The DAT file header area is defined to be 4096 bytes long.  Since the
// header layout structure above may not fill the entire 4096 bytes, a
// padding array is used to make up the difference.  The padding array
// itself is defined in srdat.c.
//
//

#define SRDAT_HDR_TOTALSIZE    4096
#define SRDAT_HDR_LAYOUTSIZE   sizeof(SRDAT_HDR)
#define SRDAT_HDR_PADDINGSIZE  (signed int)(SRDAT_HDR_TOTALSIZE-SRDAT_HDR_LAYOUTSIZE)

extern char SrDatHdrPadding[ SRDAT_HDR_PADDINGSIZE ];







/* RECORD TAG STRUCTURE AND DEFINES:

A record tag structure includes TagId and Nbytes indentifing the type
and size of record information that follows the tag structure.

Besides analog data, many users will be interested in the format of
the GPS timing information.  The layout of the GPS data in the
records reflects the functioning of the underlying hardware.  And is
as follows:

GPS systems generally produce two data streams.  One stream is a PPS
(pulse per second) tick that is very accurately aligned with the GPS
second.  The second is an RS232 data stream with coarse hours, minutes,
seconds and location etc.  After starting, the SR GPS system begins
counting off PPS ticks, 1,2,3,4 ...

When recording, an additional "GPS mark" channel is added on to the
analog data streams.  This channel is normally zero except when a PPS
tick occurs.  It then records the PPS count index of the tick, 1,2,3,4
...

To obtain accurate time information for the analog values at a GPS
mark index, you must examine both the GPSPPS and GPSSERIAL records.
Those records will have the accurate high speed 64 bit counts and
coarse serial information for the corresponding index.  Use the
CounterType and CounterFrequency values from the DAT header to
accurately interpolate the sample time with respect to the PPS count.
Usual accuracy will be better than 100 microseconds and often 1
microsecond.



RECORD TAG ID DESCRIPTIONS:

These tag types indicate the meaning of the data in the record
information that follows the tag structure.


* SRDAT_TAGID_ANALOG16

16 bit time multiplexed analog data.  The multiplexed data layout is:

        channel 0, sample 0
        channel 1, sample 0
        channel 2, sample 0 ...
        channel N, sample 0

        channel 0, sample 1
        channel 1, sample 1
        channel 2, sample 1 ...
        channel N, sample 1

        channel 0, sample 2
        channel 1, sample 2 ...

where channel N may be the "GPS mark" channel if a GPS system is
attached.


* SRDAT_TAGID_ANALOG24

24 bit time multiplexed analog data.  The samples are packed into 32
bit longs in the same multiplexed fashion as above.



* SRDAT_TAGID_GPSPPS

A GPS mark index and two high speed 64 bit counter values.  The first
counter value is the 64 bit count at the PPS tick associated with the
respective index.  The second is the 64 bit count when the analog data
was recorded.

* SRDAT_TAGID_GPSSERIAL

A GPS mark index and the NMEA serial messages that correspond to that
index.  Typically, two NMEA messages will be received.  The GGA
message contains time and location information while the VTG message
is for velocity track good info.  See the documentation that came
with the GPS system for more information.


* SRDAT_TAGID_EOF

Used to indicate the end of a DAT file.

*/


#define SRDAT_TAGID_ANALOG16    ((long)('61AT'))   // = "TA16"
#define SRDAT_TAGID_ANALOG24    ((long)('42AT'))   // = "TA24"
#define SRDAT_TAGID_GPSPPS      ((long)('SPPT'))   // = "TPPS"
#define SRDAT_TAGID_GPSSERIAL   ((long)('REST'))   // = "TSER"
#define SRDAT_TAGID_EOF         ((long)('FOET'))   // = "TEOF"


typedef struct RecordTag {

        unsigned long TagId;
        unsigned long Nbytes;
        unsigned long Unused1;
        unsigned long Unused2;

        } SRDAT_TAG;


#define SRDAT_TAGSIZE   sizeof( SRDAT_TAG )






// Helper function prototypes:
//
// These functions help programming with SR DAT files.  See srdat.c for
// more information.
//
//

// Use these defines to select which group of header time variables
// to set in the SrDatHdrSetDate and SrDatHdrSetTime functions


int SrDatHdrInit( SRDAT_HDR *Hdr );

int SrDatHdrSetFileParm(

                    SRDAT_HDR *Hdr,
                    char *FileName,
                    long FileCreateTime,
                    long FileSeqNum,
                    long FileStartPtNum

                   );

int SrDatHdrSetTimeType(

                    SRDAT_HDR *Hdr,
                    int WhichTime,
                    int Type,
                    int SampleFromRun,
                    int SampleFromFile,
                    int GpsMarkIndex
                           
                    );

int SrDatHdrSetTimeYMD(

                    SRDAT_HDR *Hdr,
                    int WhichTime,
                    int Year,
                    int Month,
                    int Day,
                    int Source

                   );

int SrDatHdrSetTimeHMS(

                    SRDAT_HDR *Hdr,
                    int WhichTime,
                    int hh,
                    int mm,
                    int ss,
                    long usec,
                    int Source

                   );

int SrDatHdrSetLocation(

                    SRDAT_HDR *Hdr,
                    long Latitude,
                    long Longitude,
                    long Altitude,
                    int Source

                   );

int SrDatHdrWrite( int fh, SRDAT_HDR *Hdr );
int SrDatWrite( int fh, long DataType, void *Buffer, unsigned int Nbytes );

int SrDatHdrRead( int fh, SRDAT_HDR *Hdr );
int SrDatTagRead( int fh, SRDAT_TAG *Tag );
int SrDatDataRead( int fh, void *Buffer, unsigned int Nbytes );

int SrDatHdrSkip( int fh );
int SrDatDataSkip( int fh, unsigned int Nbytes );
