/* FILE: srhelper.c             Copyright (c), Symmetric Research, 2002-2004

These are the SR user library helper functions.  They are mostly
wrappers around some utility functions that differ between various OS's.

*/

#include <stdio.h>
#include <sys/stat.h>    // fstat function  (VC++ can use / in path)


// OS dependent includes ...

#if defined( SROS_WIN2K ) || defined( SROS_WINNT )
#include <windows.h>     // windef.h -> winnt.h -> defines GENERIC_READ etc
#include <io.h>          // access function
#include <conio.h>       // kbhit, getch functions


#elif defined( SROS_WIN95 )
#include <stdlib.h>      // sleep function
#include <io.h>          // access, commit functions
#include <conio.h>       // kbhit, getch functions


#elif defined( SROS_MSDOS )
#include <io.h>          // access, commit functions
#include <conio.h>       // kbhit, getch functions
#include <sys\timeb.h>   // ftime function


#elif defined( SROS_LINUX )
#include <unistd.h>      // close, access, fsync fns and STDIN_FILENO define
#include <fcntl.h>       // open  function


#endif  // SROS_xxxxx


#if !defined( SRLOCAL )
#define SRLOCAL static
#endif


#include "srhelper.h"


/* RELEASE REV INFO:

This function allows the user to determine the library rev they are
working with.

Short rev history:

  SRHELPER_REV  101  ( 10/15/2003 )  < Changed SrFileExist return, added Rev fn

*/

#define SRHELPER_REV      101

FUNCTYPE( int )  SrHelperRev( int* Rev ) {
        
        *Rev = SRHELPER_REV;
        return( *Rev );
}


/* OS DEPENDENT HELPER FUNCTIONS:
 *
 * These functions help hide OS dependencies from the rest of
 * the library by providing a common interface.  The functions
 * are repeated once for each supported OS, but the SROS_xxxxx
 * defines select only one set to be compiled in to the code.
 * 
 * Supported OS's include
 *
 *     SROS_WIN2K  Windows 2000
 *     SROS_WINNT  Windows NT 4.0
 *     SROS_WIN95  Windows 95, 98, and ME
 *     SROS_MSDOS  MS Dos
 *     SROS_LINUX  Linux 2.2.12 kernel (RedHat 6.0,6.1,6.2)
 *
 *
 */


#if defined( SROS_WIN2K ) || defined( SROS_WINNT )
SRLOCAL int OsSetNonBlock( int value ) {

        // Winnt is always in non-blocking mode (value=1)
        //   value = 0 blocks reads until a character is ready
        //   value = 1 allow reads to return immediately
        //   returns -1 for failure, 0 for success

        if (value == 1)
                return( 0 );
        else
                return( -1 );
}

SRLOCAL int OsGetKey( int wait ) {

        int c;

        // Get a ready keypress or wait until one is ready if requested.

        c = EOF;
        if ( kbhit() || wait ) {
                c = getch();
                if (c == 0xD) printf("\n");  // add newline if char was ENTER
                if (c == 0x00 || c == 0xE0 ) // deal with extended key (fn,->)
                        c = getch();         // c has scan code
                }
        
        return( c );
}

SRLOCAL void OsFlushPrint( void ) {
        
        // WinNT automatically flushes printf.
}

SRLOCAL int OsSleep( int ms, int hardwait ) {

        // Sleep for specifed number of milliseconds.
        // Returns 0 for success, -1 for error.
        

        if ( ms <= 0 )
                return( 0 );
        
        _sleep( ms );
        return( 0 );
}

#define OsCommit _commit



#elif defined( SROS_WIN95 ) || defined( SROS_MSDOS )

SRLOCAL int OsSetNonBlock( int value ) {

        // Win95/MsDOS are always in non-blocking mode (value=1)
        //   value = 0 blocks reads until a character is ready
        //   value = 1 allow reads to return immediately
        //   returns -1 for failure, 0 for success

        if (value == 1)
                return( 0 );
        else
                return( -1 );
}

SRLOCAL int OsGetKey( int wait ) {

        int c;

        // Get a ready keypress or wait until one is ready if requested.

        c = EOF;
        if ( kbhit() || wait ) {
                c = getch();
                if (c == 0xD) printf("\n");  // add newline if char was ENTER
                if (c == 0x00 || c == 0xE0 ) // deal with extended key (fn,->)
                        c = getch();         // c has scan code
                }
        
        return( c );
}

SRLOCAL void OsFlushPrint( void ) {
        
        // Win95/MsDOS automatically flushes printf.
}

#if defined ( SROS_WIN95 )
SRLOCAL int OsSleep( int ms, int hardwait ) {

        // Sleep for specifed number of milliseconds.
        // Returns 0 for success, -1 for error.
        
        if ( ms <= 0 )
                return( 0 );

        _sleep( ms );

        return( 0 );
}
#elif defined( SROS_MSDOS )
SRLOCAL int OsSleep( int ms, int hardwait ) {

        struct _timeb timebuffer;
        unsigned short endms;
        time_t endsec;

        // Sleep for specifed number of milliseconds.
        // Returns 0 for success, -1 for error.
        
        if ( ms <= 0 )
                return( 0 );

        // No sleep function is available for MSDOS, so just
        // return.  Unless the user has requested that they
        // really want to wait, even if it means spinning in
        // a loop wasting CPU cycles.
        
        if ( hardwait ) {
                
                _ftime( &timebuffer );
                endsec = timebuffer.time;
                endms  = timebuffer.millitm + ms;
                
                while (endms > 1000) {
                        endms  -= 1000;
                        endsec += 1;
                        }
                
                while (timebuffer.time < endsec)
                        _ftime( &timebuffer );
                
                while ( (timebuffer.time == endsec) &&
                        (timebuffer.millitm < endms) )
                        _ftime( &timebuffer );
                }

                
//                ms *= 1000; // Convert ms to us
//                while ( ms-- )
//                        _asm {  mov  dx,0x379  // Read from status port
//                                in   al,dx     // takes about 1 microsecond
//                                }

        return( 0 );
}
#endif // SROS_WIN95 or SROS_MSDOS for sleep function difference


#define OsCommit _commit




#elif defined( SROS_LINUX )
SRLOCAL int OsSetNonBlock( int value ) {

        // Set non-blocking mode for keyboard reads
        //   value = 0 blocks reads until a character is ready
        //   value = 1 allow reads to return immediately
        //   returns -1 for failure, 0 for success

        int oldflags, result;

        oldflags = fcntl( STDIN_FILENO, F_GETFL, 0 );

        if (oldflags == -1)
                return( -1 );

        if (value != 0)
                oldflags |= O_NONBLOCK;
        else
                oldflags &= ~O_NONBLOCK;

       result = fcntl( STDIN_FILENO, F_SETFL, oldflags );

       return( result );
}

SRLOCAL int OsGetKey( int wait ) {

        int c;

        // Get a ready keypress or wait until one is ready if requested.

        // Notes:
        //
        // This version assumes that non-blocking reads have been
        // enabled.  Otherwise, the first getchar will always wait until
        // a character is ready.
        //
        // Getchar requires an ENTER after the character before it will
        // start processing the character.


        c = getchar();
        while ( (c == EOF) && wait )
                c = getchar();

        return( c );
}

SRLOCAL void OsFlushPrint( void ) {
        
        fflush( STDIN_FILENO );
}

SRLOCAL int OsSleep( int ms, int hardwait ) {

        // Sleep for specifed number of milliseconds.
        // Returns 0 for success, -1 for error.

        struct timespec twait, tleft;
        
        if ( ms <= 0 )
                return( 0 );

        
        // Linux nanosleep takes structure with seconds and nanoseconds.
        
        tleft.tv_sec  = tleft.tv_nsec = 0;
        twait.tv_sec  = (long)(ms / 1000);
        twait.tv_nsec = (ms - (twait.tv_sec*1000)) * 1000000;
        
        return( nanosleep( &twait, &tleft ) );
}

#define OsCommit fsync

#endif // SROS_xxxxx   End of OS dependent helper functions





/* GENERIC HELPER FUNCTIONS - compensate for differences between OS's

Most of these functions are just wrappers around their OS specific
counterparts defined in the OS section above.

SrKeyboardNonBlock ensures non-blocking mode is enabled, allowing
SrGetKeyCheck to return even if a key has not been pressed.

SrGetKeyWait and SrGetKeyCheck are handy shortcuts for the two
keyboard behaviors OsGetKey provides.  If wait is true, OsGetKey
waits until a character is ready and then return it.  Otherwise, it
checks for an available character.  If one is there, it is read and
returned.  Otherwise, EOF is returned.

SrFlushPrint ensures that the last printf string is displayed, even
if it didn't end with a new line (\n) character.

SrSleep puts the calling program to sleep for the specified number of
milliseconds.

SrFileLength returns the length of the specified file in bytes.

*/


FUNCTYPE( int )  SrKeyboardNonBlock( void ) { return( OsSetNonBlock( 1 ) ); }
FUNCTYPE( int )  SrGetKeyCheck( void ) { return( OsGetKey( 0 ) ); }
FUNCTYPE( int )  SrGetKeyWait(  void ) { return( OsGetKey( 1 ) ); }
FUNCTYPE( void ) SrFlushPrint( void ) { OsFlushPrint(); }
FUNCTYPE( int )  SrSleep( int ms ) { return( OsSleep( ms, 0 ) ); }
FUNCTYPE( int )  SrWait(  int ms ) { return( OsSleep( ms, 1 ) ); }

FUNCTYPE( int )  SrCommit( int fh ) {
        if (fh == FH_STDOUT) // Don't try to commit stdout
                return( 0 ); // 0 means success, like underlying commit
        else
                return( OsCommit( fh ) );
}


// NOTE: The return value changed as of 10/15/2003.  
//       Now it is 1 when the file exists and 0 otherwise.

#if !defined F_OK
#define F_OK 0x00
#endif

FUNCTYPE( int )  SrFileExist( char *FileName ) {
        if ( access( FileName, F_OK ) == 0 )
                return( 1 );
        else
                return( 0 );
}

FUNCTYPE( int )  SrFileLength( int fh ) {

        struct stat buffer;

        if ( fstat( fh, &buffer ) == -1 ) {
                printf( "Could not get file stats\n" );
                return( -1 );
                }

        return( buffer.st_size );
}

// This function copies up to MaxChar-1 characters from SrcStr to DestStr,
// ensuring that the last character in the destination string is a null.
// It returns the total number of characters in the destination string,
// including the terminating null.
//
//  *** Be sure MaxChar is <= the dimension of DestStr. ***

FUNCTYPE( int  ) SrStringCopy( char *DestStr, const char *SrcStr, int MaxChar ) {

        int i;

        // Copy characters from SrcStr to DestStr, stopping after
        // MaxChar-1 characters or a null at the end of the source
        // string is reached.

        i = 1;
        while ( (i < MaxChar) && (*DestStr++ = *SrcStr++) )
                i++;


        // If the copying stopped because of hitting the character
        // limit, fill the last character with null (if the copy stopped
        // because of reaching the end of the source string, the last
        // character is already null).

        if ( i == MaxChar )
                *DestStr = '\0';

        return( i );
}
