Attribute VB_Name = "modPARXCH"
Attribute VB_Description = "Declarations of DLL procedures for high level use of SR PARxCH board"

Option Explicit


' FILE: parxch.bas                  Copyright (c), Symmetric Research, 2003
'
' Basic header file for PARxCH acquisition library functions.
'
' Visual Basic users should include this file in their projects to have
' access to the PARxCH library functions.
'


' In the function library prototypes (Declares), the symbols listed below
' can replace the "As Type" info for function arguments.  For example:
' (ByVal X As Long, ByRef Y As Single) equals (ByVal X&, ByRef Y!)
'
' % Integer (short)
' & Long    (32-bit int or long)
' ! Single  (float)
' # Double  (double)
' $ String  (char*)  ByVal BSTR = LPSTR, ByRef BSTR = ptr to ptr to null-term string
'
' When translating from the C declarations in parxch.h, the following
' rules were applied:
'
'   short         -> Integer
'   int           -> Long (because ints are now 32 bits, the same as longs)
'   unsigned int  -> Long
'   long          -> Long
'   unsigned long -> Long
'   float         -> Single
'   double        -> Double
'   char*         -> String
'   anything*     -> pass ByRef instead of ByVal
'   arrays        -> pass ByRef using first item in array
'
' Routines returning a value are Function, the rest are Sub.
'




' RELEASE REV:
'
' This function allows the user to determine the library rev at
' runtime.  The rev is an integer.  A number like 123 would be referred
' to as rev "1.23".
'

Public Declare Function ParXchGetRev _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchGetRev@4" _
   (ByRef Rev As Long) _
    As Long




' DEFAULT XCH MODEL:
'
' The PARxCH family of 24 bit A/D boards includes the 1, 4, and 8
' channel PAR1CH, PAR4CH, PAR8CH models.  While the software is
' designed to work with any of the PARxCH models, the application
' programs often need to know which model is being used.  Typically,
' the specific model is given as a command line argument.  But for
' convenience, a default choice can be saved (in either the registry
' or an environment variable depending on OS) by using the setmodel
' utility.  This function retrieves this default model information.
'


Public Declare Function ParXchGetDefaultModel _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchGetDefaultModel@4" _
   (ByVal DriverName As String) _
    As Long




' SIMPLE PARxCH CONTROL/EXECUTION DEFINES AND FUNCTIONS:
'
' The following "easy to use" defines and functions provide a basic
' level of control over the PARxCH.  Many applications will not need
' anything else.  The basic pattern of usage is:
'
' 1) Open and initialize the PARxCH, setting the sampling rate
' 2) Start conversion
' 3) Call ReadData to copy ready data from the PARxCH to the PC 
' 4) Loop to continue getting more data
' 5) When done, stop conversion and close the PARxCH
' 
' When using the PARxCH with a PARGPS unit for accurate timing, the basic
' pattern of usage is similar but with a few additional steps:
' 
' 1)  Open and initialize both the PARxCH and the PARGPS
' 1A) Attach the PARGPS to the PARxCH so their drivers can communicate
' 2)  Start the PARGPS and then the PARxCH
' 3)  Call the PARxCH ReadDataWithGpsMark to get analog data and the PARGPS
'     ReadPpsData and ReadSerialData to get timing data
' 4)  Loop to continue getting more data
' 5)  When done, stop both the PARGPS and PARxCH
' 5A) Release the PARGPS from the PARxCH and close both drivers
' 
' See pargps.h for more information about opening, closing, and reading
' timing data from the PARGPS device.
' 
' 
' For more detailed control of the PARxCH see the ADVANCED defines and
' functions below.
'
' Some of the abbreviations are:
'
'        EPP    enhanced parallel port
'        BPP    bidirectional parallel port
'        ECP    extended capabilities port
'        SPS    samples per second
'
'
' Note that the function ParXchOpen returns a 1 or 0 for general
' success or failure.  It also fills in an integer error code, which on
' return has more information about what may have gone right or wrong.
' The PARXCH_ERROR ... defines provide a mapping from these error codes
' to strings.
'


' The PARxCH family of A/D boards includes the several members.  The
' following ID defines are used to identify the specific A/D.

Public Const PARXCH_MODEL_UNKNOWN  As Long = &H00
Public Const PARXCH_MODEL_PAR1CH   As Long = &H01
Public Const PARXCH_MODEL_PAR4CH   As Long = &H04
Public Const PARXCH_MODEL_PAR8CH   As Long = &H08



' PARxCH number of channels ...

' Several different types of information can be acquired from the
' PARxCH boards.  These include analog, digital, GPS and counter
' channels.  The number of channels for each type of PARxCH, defined
' below, control what data is saved to the output files.  An additional
' set of defines indicate the channel counts for plotting purposes as
' this may be different than for saving.  For example, the PAR8CH
' digital information is stored as 1 channel, but is plotted as 4
' traces.


' Channel counts for the PAR8CH

Public Const PAR8CH_ANALOG_CHANNELS    As Long = 8
Public Const PAR8CH_DIGITAL_CHANNELS   As Long = 1
Public Const PAR8CH_GPS_CHANNELS       As Long = 1
Public Const PAR8CH_COUNTER_CHANNELS   As Long = 1

Public Const PAR8CH_ANAPLOT_CHANNELS   As Long = 8
Public Const PAR8CH_DIGPLOT_CHANNELS   As Long = 4
Public Const PAR8CH_GPSPLOT_CHANNELS   As Long = 1
Public Const PAR8CH_CNTPLOT_CHANNELS   As Long = 0

Public Const PAR8CH_MODELNAME          As String = "PAR8CH"



' Channel counts for the PAR4CH

Public Const PAR4CH_ANALOG_CHANNELS    As Long = 4
Public Const PAR4CH_DIGITAL_CHANNELS   As Long = 0
Public Const PAR4CH_GPS_CHANNELS       As Long = 1
Public Const PAR4CH_COUNTER_CHANNELS   As Long = 0

Public Const PAR4CH_ANAPLOT_CHANNELS   As Long = 4
Public Const PAR4CH_DIGPLOT_CHANNELS   As Long = 0
Public Const PAR4CH_GPSPLOT_CHANNELS   As Long = 1
Public Const PAR4CH_CNTPLOT_CHANNELS   As Long = 0

Public Const PAR4CH_MODELNAME          As String = "PAR4CH"



' Channel counts for the PAR1CH

Public Const PAR1CH_ANALOG_CHANNELS    As Long = 1
Public Const PAR1CH_DIGITAL_CHANNELS   As Long = 0
Public Const PAR1CH_GPS_CHANNELS       As Long = 1
Public Const PAR1CH_COUNTER_CHANNELS   As Long = 0

Public Const PAR1CH_ANAPLOT_CHANNELS   As Long = 1
Public Const PAR1CH_DIGPLOT_CHANNELS   As Long = 0
Public Const PAR1CH_GPSPLOT_CHANNELS   As Long = 1
Public Const PAR1CH_CNTPLOT_CHANNELS   As Long = 0

Public Const PAR1CH_MODELNAME          As String = "PAR1CH"



' Maximum channel counts for all PARxCH

Public Const PARXCH_MAX_ANALOG_CHANNELS    As Long = 8
Public Const PARXCH_MAX_DIGITAL_CHANNELS   As Long = 1
Public Const PARXCH_MAX_GPS_CHANNELS       As Long = 1
Public Const PARXCH_MAX_COUNTER_CHANNELS   As Long = 1

Public Const PARXCH_MAX_ANAPLOT_CHANNELS   As Long = 8
Public Const PARXCH_MAX_DIGPLOT_CHANNELS   As Long = 4
Public Const PARXCH_MAX_GPSPLOT_CHANNELS   As Long = 1
Public Const PARXCH_MAX_CNTPLOT_CHANNELS   As Long = 0



' Set the general channel count defines and an Id define to match the
' specific PARxCH A/D requested at compile time.  The requested A/D is
' identified by the SRA2D_PARxxx macro.  The allowed values are:
'
' SRA2D_PAR8CH  for the 8 channel A/D
' SRA2D_PAR4CH  for the 4 channel A/D
' SRA2D_PAR1CH  for the 1 channel A/D
' SRA2D_PARANY  for any board in the PARxCH family (used for common driver)
'


#If SRA2D_PAR8CH Then

Public Const PARXCH_MODEL               As Long = PARXCH_MODEL_PAR8CH
Public Const PARXCH_MODELNAME           As String = PAR8CH_MODELNAME

Public Const PARXCH_ANALOG_CHANNELS     As Long = PAR8CH_ANALOG_CHANNELS
Public Const PARXCH_DIGITAL_CHANNELS    As Long = PAR8CH_DIGITAL_CHANNELS
Public Const PARXCH_GPS_CHANNELS        As Long = PAR8CH_GPS_CHANNELS
Public Const PARXCH_COUNTER_CHANNELS    As Long = PAR8CH_COUNTER_CHANNELS

Public Const PARXCH_ANAPLOT_CHANNELS    As Long = PAR8CH_ANAPLOT_CHANNELS
Public Const PARXCH_DIGPLOT_CHANNELS    As Long = PAR8CH_DIGPLOT_CHANNELS
Public Const PARXCH_GPSPLOT_CHANNELS    As Long = PAR8CH_GPSPLOT_CHANNELS
Public Const PARXCH_CNTPLOT_CHANNELS    As Long = PAR8CH_CNTPLOT_CHANNELS



#ElseIf SRA2D_PAR4CH Then

Public Const PARXCH_MODEL               As Long = PARXCH_MODEL_PAR4CH
Public Const PARXCH_MODELNAME           As String = PAR4CH_MODELNAME

Public Const PARXCH_ANALOG_CHANNELS     As Long = PAR4CH_ANALOG_CHANNELS
Public Const PARXCH_DIGITAL_CHANNELS    As Long = PAR4CH_DIGITAL_CHANNELS
Public Const PARXCH_GPS_CHANNELS        As Long = PAR4CH_GPS_CHANNELS
Public Const PARXCH_COUNTER_CHANNELS    As Long = PAR4CH_COUNTER_CHANNELS

Public Const PARXCH_ANAPLOT_CHANNELS    As Long = PAR4CH_ANAPLOT_CHANNELS
Public Const PARXCH_DIGPLOT_CHANNELS    As Long = PAR4CH_DIGPLOT_CHANNELS
Public Const PARXCH_GPSPLOT_CHANNELS    As Long = PAR4CH_GPSPLOT_CHANNELS
Public Const PARXCH_CNTPLOT_CHANNELS    As Long = PAR4CH_CNTPLOT_CHANNELS


#ElseIf SRA2D_PAR1CH Then

Public Const PARXCH_MODEL               As Long   = PARXCH_MODEL_PAR1CH
Public Const PARXCH_MODELNAME           As String = PAR1CH_MODELNAME

Public Const PARXCH_ANALOG_CHANNELS     As Long = PAR1CH_ANALOG_CHANNELS
Public Const PARXCH_DIGITAL_CHANNELS    As Long = PAR1CH_DIGITAL_CHANNELS
Public Const PARXCH_GPS_CHANNELS        As Long = PAR1CH_GPS_CHANNELS
Public Const PARXCH_COUNTER_CHANNELS    As Long = PAR1CH_COUNTER_CHANNELS

Public Const PARXCH_ANAPLOT_CHANNELS    As Long = PAR1CH_ANAPLOT_CHANNELS
Public Const PARXCH_DIGPLOT_CHANNELS    As Long = PAR1CH_DIGPLOT_CHANNELS
Public Const PARXCH_GPSPLOT_CHANNELS    As Long = PAR1CH_GPSPLOT_CHANNELS
Public Const PARXCH_CNTPLOT_CHANNELS    As Long = PAR1CH_CNTPLOT_CHANNELS


' Works for any PARxCH (SRA2D_PARANY) or none

#Else

Public Const PARXCH_MODEL               As Long = PARXCH_MODEL_UNKNOWN
Public Const PARXCH_MODELNAME           As String = "PARxCH"

#End If
' SRA2D_PARxxx





' Driver name and parallel port mode defines ...

Public Const PARXCH_0x378              As String = "SrParXch378"
Public Const PARXCH_IRQ_0x378          As Long   = 7

Public Const PARXCH_0x278              As String = "SrParXch278"
Public Const PARXCH_IRQ_0x278          As Long = 5

Public Const PARXCH_DRIVER_NAME_1      As String = PARXCH_0x378
Public Const PARXCH_DRIVER_NAME_2      As String = PARXCH_0x278

Public Const PARXCH_DRIVER_IRQ_1       As Long = PARXCH_IRQ_0x378
Public Const PARXCH_DRIVER_IRQ_2       As Long = PARXCH_IRQ_0x278

Public Const PARXCH_PORT_MODE_BPP      As Long = 0
Public Const PARXCH_PORT_MODE_EPP      As Long = 1
Public Const PARXCH_PORT_MODE_ECP_BPP  As Long = 2
Public Const PARXCH_PORT_MODE_ECP_EPP  As Long = 3


' Minimum and maximum Sps values ...

Public Const PARXCH_SPS_MIN            As Single = 0.00001     ' Hz
Public Const PARXCH_SPS_MAX            As Single = 15625#      ' Hz
Public Const PARXCH_SPS_DEFAULT        As Single = 100#        ' Hz


' Error code return values ...

Public Const PARXCH_ERROR_NONE                  As Long = 0
Public Const PARXCH_ERROR_PORT_ADDRESS          As Long = 1
Public Const PARXCH_ERROR_PORT_MODE             As Long = 2
Public Const PARXCH_ERROR_DATA_FORMAT           As Long = 3
Public Const PARXCH_ERROR_CAL_MODE              As Long = 4
Public Const PARXCH_ERROR_GAIN                  As Long = 5
Public Const PARXCH_ERROR_TURBO                 As Long = 6
Public Const PARXCH_ERROR_GAIN_TURBO_PRODUCT    As Long = 7
Public Const PARXCH_ERROR_DECIMATION            As Long = 8
Public Const PARXCH_ERROR_DRIVER_NOT_OPEN       As Long = 9
Public Const PARXCH_ERROR_DRIVER_REQUEST_FAILED As Long = 10
Public Const PARXCH_ERROR_VOLTAGE_BAD_ON        As Long = 11
Public Const PARXCH_ERROR_VOLTAGE_BAD_OFF       As Long = 12
Public Const PARXCH_ERROR_NOT_UNLOCKED          As Long = 13
Public Const PARXCH_ERROR_ADS1210_CR_READBACK   As Long = 14
Public Const PARXCH_ERROR_FIFO_SANITY           As Long = 15
Public Const PARXCH_ERROR_OVERFLOW              As Long = 16
Public Const PARXCH_ERROR_KEYPRESS              As Long = 17
Public Const PARXCH_ERROR_GPS_NOT_AVAILABLE     As Long = 18
Public Const PARXCH_ERROR_GPS_NOT_ATTACHED      As Long = 19
Public Const PARXCH_ERROR_GPS_VERSION_MISMATCH  As Long = 20
Public Const PARXCH_ERROR_DIGITAL_NOT_AVAILABLE As Long = 21
Public Const PARXCH_ERROR_ATOD_UNKNOWN          As Long = 22
Public Const PARXCH_ERROR_MAX                   As Long = 23


Public Const PARXCH_ERROR_MSG_NONE                  As String = "NONE"
Public Const PARXCH_ERROR_MSG_PORT_ADDRESS          As String = "PORT ADDRESS"
Public Const PARXCH_ERROR_MSG_PORT_MODE             As String = "PORT MODE"
Public Const PARXCH_ERROR_MSG_DATA_FORMAT           As String = "DATA FORMAT"
Public Const PARXCH_ERROR_MSG_CAL_MODE              As String = "CALIBRATION"
Public Const PARXCH_ERROR_MSG_GAIN                  As String = "GAIN"
Public Const PARXCH_ERROR_MSG_TURBO                 As String = "TURBO"
Public Const PARXCH_ERROR_MSG_GAIN_TURBO_PRODUCT    As String = "GAIN*TURBO"
Public Const PARXCH_ERROR_MSG_DECIMATION            As String = "DECIMATION"
Public Const PARXCH_ERROR_MSG_DRIVER_NOT_OPEN       As String = "DEVICE DRIVER"
Public Const PARXCH_ERROR_MSG_DRIVER_REQUEST_FAILED As String = "DRIVER REQUEST"
Public Const PARXCH_ERROR_MSG_VOLTAGE_BAD_ON        As String = "VOLTAGE STILL ON"
Public Const PARXCH_ERROR_MSG_VOLTAGE_BAD_OFF       As String = "VOLTAGE BAD/OFF"
Public Const PARXCH_ERROR_MSG_NOT_UNLOCKED          As String = "NO RESPONSE"
Public Const PARXCH_ERROR_MSG_ADS1210_CR_READBACK   As String = "A/D READBACK"
Public Const PARXCH_ERROR_MSG_FIFO_SANITY           As String = "FIFO SANITY"
Public Const PARXCH_ERROR_MSG_OVERFLOW              As String = "OVERFLOW"
Public Const PARXCH_ERROR_MSG_KEYPRESS              As String = "KEYPRESS"
Public Const PARXCH_ERROR_MSG_GPS_NOT_AVAILABLE     As String = "PARGPS NOT AVAILABLE"
Public Const PARXCH_ERROR_MSG_GPS_NOT_ATTACHED      As String = "PARGPS AREA NOT ATTACHED"
Public Const PARXCH_ERROR_MSG_GPS_VERSION_MISMATCH  As String = "PARGPS VERSION MISMATCH"
Public Const PARXCH_ERROR_MSG_DIGITAL_NOT_AVAILABLE As String = "DIGITAL NOT AVAILABLE"
Public Const PARXCH_ERROR_MSG_ATOD_UNKNOWN          As String = "ATOD NOT PAR1CH,4CH,8CH"
Public Const PARXCH_ERROR_MSG_MAX                   As String = vbNullString


' PARXCH error string allocation, initialized in ParXchInitErrorMsg() below

Public PARXCH_ERROR_MSG(PARXCH_ERROR_NONE To PARXCH_ERROR_MAX) As String



Public Const PARXCH_BAD_DEVHANDLE As Long = -1




' Simple PARxCH functions ...

Public Declare Function Par8chOpen _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_Par8chOpen@28" _
   (ByVal DriverName As String, _
    ByVal PortMode As Long, _
    ByVal Sps As Double, _
    ByRef ActualSps As Double, _
    ByRef Error As Long) _
    As Long

Public Declare Function Par4chOpen _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_Par4chOpen@28" _
   (ByVal DriverName As String, _
    ByVal PortMode As Long, _
    ByVal Sps As Double, _
    ByRef ActualSps As Double, _
    ByRef Error As Long) _
    As Long

Public Declare Function Par1chOpen _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_Par1chOpen@28" _
   (ByVal DriverName As String, _
    ByVal PortMode As Long, _
    ByVal Sps As Double, _
    ByRef ActualSps As Double, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchOpen _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchOpen@32" _
   (ByVal DriverName As String, _
    ByVal XchModel As Long, _
    ByVal PortMode As Long, _
    ByVal Sps As Double, _
    ByRef ActualSps As Double, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchClose _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchClose@4" _
   (ByVal ParXchHandle As Long) _
    As Long

Public Declare Sub ParXchStart _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchStart@4" _
   (ByVal ParXchHandle As Long)

Public Declare Sub ParXchStop _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchStop@4" _
   (ByVal ParXchHandle As Long)

Public Declare Function ParXchReady _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchReady@4" _
   (ByVal ParXchHandle As Long) _
    As Long

Public Declare Function ParXchOverflow _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchOverflow@4" _
   (ByVal ParXchHandle As Long) _
    As Long

Public Declare Function ParXchReadData _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchReadData@16" _
   (ByVal ParXchHandle As Long, _
    ByRef Values As Long, _
    ByVal Nvalues As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchReadDataWithDigital _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchReadDataWithDigital@16" _
   (ByVal ParXchHandle As Long, _
    ByRef Values As Long, _
    ByVal Nvalues As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchReadDataWithGpsMark _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchReadDataWithGpsMark@16" _
   (ByVal ParXchHandle As Long, _
    ByRef Values As Long, _
    ByVal Nvalues As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchReadDataAll _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchReadDataAll@16" _
   (ByVal ParXchHandle As Long, _
    ByRef Values As Long, _
    ByVal Nvalues As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchAttachGps _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchAttachGps@12" _
   (ByVal ParXchHandle As Long, _
    ByVal GpsHandle As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchReleaseGps _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchReleaseGps@4" _
   (ByVal ParXchHandle As Long) _
    As Long




' PARxCH USER DIGITAL IO FUNCTIONS:
'
' These functions allow programs to communicate with the DB15 digital
' IO connector on the right hand side of the PARxCH front panel.
'

Public Declare Sub ParXchUserIoRd _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchUserIoRd@8" _
   (ByVal ParXchHandle As Long, _
    ByRef Value As Long)

Public Declare Sub ParXchUserIoWr _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchUserIoWr@8" _
   (ByVal ParXchHandle As Long, _
    ByVal Value As Long)

Public Declare Sub ParXchUserLed _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchUserLed@8" _
   (ByVal ParXchHandle As Long, _
    ByVal State As Long)










' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY
' EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY EXPERTS ONLY


' ADVANCED PARxCH CONTROL/EXECUTION DEFINES AND FUNCTIONS:
'
' These defines and functions allow a more detailed level of control
' over PARxCH execution.  They are intended for users who are familiar
' with the ADS1210 command register bit fields and wish to program them
' directly.
'
'
' Use these defines with ParXchFullOpen to initialize the PARxCH.
'
' Several of these constants map directly into ADS1210 command register
' bit setings.  Don't change their values !!
'
' Some of the abbreviations are:
'
'        DF               ADS1210 data format
'        CAL_MODE         ADS1210 calibration mode
'        GAIN             ADS1210 preamp gain
'        TURBO            ADS1210 modulator turbo setting
'        DECIMATION       ADS1210 modulator decimation ratio
'
'
' These functions control the operation of the PARxCH.
'
' Use the defines above and helper functions for setting ParXchFullOpen
' parameters.
'
' Note that PARxCH user digital IO will not work until the board has
' been initialized with ParXchOpen.
'



' PARxCH ADS1210 command register defines ...

Public Const PARXCH_DF_SIGNED              As Long = 0
Public Const PARXCH_DF_OFFSET              As Long = 1

Public Const PARXCH_CAL_NORMAL             As Long = 0
Public Const PARXCH_CAL_SELF               As Long = 1
Public Const PARXCH_CAL_SYSTEM_OFFSET      As Long = 2
Public Const PARXCH_CAL_SYSTEM_FS          As Long = 3
Public Const PARXCH_CAL_PSEUDO             As Long = 4
Public Const PARXCH_CAL_BACKGROUND         As Long = 5

Public Const PARXCH_CAL_MIN                As Long = 0
Public Const PARXCH_CAL_MAX                As Long = 5

Public Const PARXCH_GAIN_1                 As Long = 0
Public Const PARXCH_GAIN_2                 As Long = 1
Public Const PARXCH_GAIN_4                 As Long = 2
Public Const PARXCH_GAIN_8                 As Long = 3
Public Const PARXCH_GAIN_16                As Long = 4

Public Const PARXCH_GAIN_MIN               As Long = 0
Public Const PARXCH_GAIN_MAX               As Long = 4

Public Const PARXCH_TURBO_1                As Long = 0
Public Const PARXCH_TURBO_2                As Long = 1
Public Const PARXCH_TURBO_4                As Long = 2
Public Const PARXCH_TURBO_8                As Long = 3
Public Const PARXCH_TURBO_16               As Long = 4

Public Const PARXCH_TURBO_MIN              As Long = 0
Public Const PARXCH_TURBO_MAX              As Long = 4

Public Const PARXCH_GAINTURBO_MIN          As Long = 0
Public Const PARXCH_GAINTURBO_MAX          As Long = 4

Public Const PARXCH_DECIMATION_MIN         As Long = 19
Public Const PARXCH_DECIMATION_100Hz_T1    As Long = 194
Public Const PARXCH_DECIMATION_100Hz_T16   As Long = 3124
Public Const PARXCH_DECIMATION_50Hz_T1     As Long = 390
Public Const PARXCH_DECIMATION_50Hz_T16    As Long = 6249
Public Const PARXCH_DECIMATION_MAX         As Long = 8000

Public Const PARXCH_EXTRADECIMATION_MIN    As Long = 1

Public Const PARXCH_UNUSED                 As Long = -1



' Driver read request defines ...

Public Const PARXCH_REQUEST_ANALOG   As Long = 0
Public Const PARXCH_REQUEST_GPS      As Long = 1
Public Const PARXCH_REQUEST_DIGITAL  As Long = 2
Public Const PARXCH_REQUEST_COUNTER  As Long = 4
Public Const PARXCH_REQUEST_ALL      As Long = (PARXCH_REQUEST_ANALOG  + _
                                                PARXCH_REQUEST_GPS     + _
                                                PARXCH_REQUEST_DIGITAL + _
                                                PARXCH_REQUEST_COUNTER)



' Advanced PARxCH functions ...


' Note: Two parameters have been put on some lines to avoid exceeding
'       the maximum number of line continuations allowed in VB 4.0

Public Declare Function ParXchFullOpen _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchFullOpen@44" _
   (ByVal DriverName As String, _
    ByVal ParXchId As Long, ByVal PortMode As Long, _
    ByVal Df As Long, ByVal GainLog As Long, _
    ByVal TurboLog As Long, ByVal Decimation As Long, _
    ByVal ExtraDecimation As Long, ByVal Unused As Long, _
    ByRef ADS1210_CrValue As Long, _
    ByRef Error As Long) As Long

Public Declare Function ParXchFullReadData _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchFullReadData@20" _
   (ByVal ParXchHandle As Long, _
    ByRef Values As Long, _
    ByVal Nvalues As Long, _
    ByVal RequestedData As Long, _
    ByRef Error As Long) _
    As Long

Public Declare Function ParXchSpsGainToTde _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchSpsGainToTde@24" _
   (ByVal Sps As Double, _
    ByVal GainLog As Long, _
    ByRef TurboLog As Long, _
    ByRef Decimation As Long, _
    ByRef ExtraDecimation As Long) _
    As Long

Public Declare Function ParXchTdeToSpsGain _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchTdeToSpsGain@20" _
   (ByVal TurboLog As Long, _
    ByVal Decimation As Long, _
    ByVal ExtraDecimation As Long, _
    ByRef Sps As Double, _
    ByRef GainLog As Long) _
    As Long





' PARxCH PC INTERRUPT FUNCTIONS:
'
' These functions allow programs to enable and disable the PC interrupt.
'

Public Declare Sub ParXchInterruptEnable _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchInterruptEnable@4" _
   (ByVal ParXchHandle As Long)

Public Declare Sub ParXchInterruptDisable _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchInterruptDisable@4" _
   (ByVal ParXchHandle As Long)






' PARxCH ARM DREADY INTERRUPT FUNCTION:
' 
' This function arms the PARxCH board to trigger a parallel port
' interrupt when the next acquired data point is ready.  This is a one
' shot interrupt.  After it has been sent, the PARxCH will return to
' passing on interrupts from the pin 3 of the digital I/O header.
'

Public Declare Sub ParXchArmDreadyInterrupt _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchArmDreadyInterrupt@4" _
   (ByVal ParXchHandle As Long)




' PARxCH VOLTAGE CHECK FUNCTION:
'
' This function checks the PARxCH board voltage level.  If the voltage is
' good, this function returns 1.  Otherwise, 0.
'

Public Declare Function ParXchVoltageGood _
    Lib "C:\SR\PARXCH\LIB\PARXCH.DLL" _
    Alias "_ParXchVoltageGood@4" _
   (ByVal ParXchHandle As Long) _
    As Long






' Initialize the error message string array

Public Sub ParXchInitErrorMsg()

        PARXCH_ERROR_MSG(PARXCH_ERROR_NONE) = PARXCH_ERROR_MSG_NONE
        PARXCH_ERROR_MSG(PARXCH_ERROR_PORT_ADDRESS) = PARXCH_ERROR_MSG_PORT_ADDRESS
        PARXCH_ERROR_MSG(PARXCH_ERROR_PORT_MODE) = PARXCH_ERROR_MSG_PORT_MODE
        PARXCH_ERROR_MSG(PARXCH_ERROR_DATA_FORMAT) = PARXCH_ERROR_MSG_DATA_FORMAT
        PARXCH_ERROR_MSG(PARXCH_ERROR_CAL_MODE) = PARXCH_ERROR_MSG_CAL_MODE
        PARXCH_ERROR_MSG(PARXCH_ERROR_GAIN) = PARXCH_ERROR_MSG_GAIN
        PARXCH_ERROR_MSG(PARXCH_ERROR_TURBO) = PARXCH_ERROR_MSG_TURBO
        PARXCH_ERROR_MSG(PARXCH_ERROR_GAIN_TURBO_PRODUCT) = PARXCH_ERROR_MSG_GAIN_TURBO_PRODUCT
        PARXCH_ERROR_MSG(PARXCH_ERROR_DECIMATION) = PARXCH_ERROR_MSG_DECIMATION
        PARXCH_ERROR_MSG(PARXCH_ERROR_DRIVER_NOT_OPEN) = PARXCH_ERROR_MSG_DRIVER_NOT_OPEN
        PARXCH_ERROR_MSG(PARXCH_ERROR_DRIVER_REQUEST_FAILED) = PARXCH_ERROR_MSG_DRIVER_REQUEST_FAILED
        PARXCH_ERROR_MSG(PARXCH_ERROR_VOLTAGE_BAD_ON) = PARXCH_ERROR_MSG_VOLTAGE_BAD_ON
        PARXCH_ERROR_MSG(PARXCH_ERROR_VOLTAGE_BAD_OFF) = PARXCH_ERROR_MSG_VOLTAGE_BAD_OFF
        PARXCH_ERROR_MSG(PARXCH_ERROR_NOT_UNLOCKED) = PARXCH_ERROR_MSG_NOT_UNLOCKED
        PARXCH_ERROR_MSG(PARXCH_ERROR_ADS1210_CR_READBACK) = PARXCH_ERROR_MSG_ADS1210_CR_READBACK
        PARXCH_ERROR_MSG(PARXCH_ERROR_FIFO_SANITY) = PARXCH_ERROR_MSG_FIFO_SANITY
        PARXCH_ERROR_MSG(PARXCH_ERROR_OVERFLOW) = PARXCH_ERROR_MSG_OVERFLOW
        PARXCH_ERROR_MSG(PARXCH_ERROR_KEYPRESS) = PARXCH_ERROR_MSG_KEYPRESS
        PARXCH_ERROR_MSG(PARXCH_ERROR_GPS_NOT_AVAILABLE) = PARXCH_ERROR_MSG_GPS_NOT_AVAILABLE
        PARXCH_ERROR_MSG(PARXCH_ERROR_GPS_NOT_ATTACHED) = PARXCH_ERROR_MSG_GPS_NOT_ATTACHED
        PARXCH_ERROR_MSG(PARXCH_ERROR_GPS_VERSION_MISMATCH) = PARXCH_ERROR_MSG_GPS_VERSION_MISMATCH
        PARXCH_ERROR_MSG(PARXCH_ERROR_DIGITAL_NOT_AVAILABLE) = PARXCH_ERROR_MSG_DIGITAL_NOT_AVAILABLE
        PARXCH_ERROR_MSG(PARXCH_ERROR_ATOD_UNKNOWN) = PARXCH_ERROR_MSG_ATOD_UNKNOWN
        PARXCH_ERROR_MSG(PARXCH_ERROR_MAX) = PARXCH_ERROR_MSG_MAX

End Sub
